/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "http_protocol.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"
#include "util_script.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include <math.h>
#include <alloca.h>
#include "vm.h"
#include "config-info.h"
#include "eni.h"

/* essembly  Ǵ instruction 鿡     
   ⼭  д.  */
enum insn_code
  {
#define DEF_CODE(ENUM, NAME, NUMBER, FORMAT, CLASS, OPLEN)   ENUM ,
#include "vm_code.def"
#undef DEF_CODE
    MAX_INSN_CODE
  };

/*  Opcode   ̿  迭.  Opcode  Ǿ ,
   run_virtual_machine ()   PC (Program Counter)   PC 
     ȴ.  */
const int insn_oplen[MAX_INSN_CODE] = {
#define DEF_CODE(ENUM, NAME, NUMBER, FORMAT, CLASS, OPLEN)   OPLEN,
#include "vm_code.def"
#undef DEF_CODE
};

const char *insn_name[MAX_INSN_CODE] = {
#define DEF_CODE(ENUM, NAME, NUMBER, FORMAT, CLASS, OPLEN)   NAME,
#include "vm_code.def"
#undef DEF_CODE
};

#define getopcode()             ((int) code[pc])
#define getpc(n)		(code[pc+1+(n)])
#define getcode(n)		(code[(n)])
#define	putpc(n, v)		code[pc+(n)] = (v)

#define	class_object()		(crinfo)
#define current_environment()	(METHOD_ENV (m))

#define	get_class_info(NODE)					\
  crinfo = get_envstruct_with_idx (NODE, METHOD_ENV (m));	\
  if (crinfo == 0)						\
    fprintf (stderr, "get_envstruct_with_idx: Error.\n");

#define	get_field_info(NODE)					\
  if (get_field (NODE, METHOD_ENV (m), NON_STATIC_FIELD_FLAG, &finfo) == -1)	\
    fprintf (stderr, "get_field: Error.\n");

#define	get_static_field_info(NODE)				\
  if (get_field (NODE, METHOD_ENV (m), STATIC_FIELD_FLAG, &finfo) == -1)	\
    fprintf (stderr, "get_field: Error.\n");

#define get_method_info(NODE)			\
  if (get_method_type_environment (NODE, METHOD_ENV (m), 1, 0, &cinfo) == -1)	\
    fprintf (stderr, "get_method_type_environment: Error.\n");

#define get_special_method_info(NODE)		\
  if (get_method_type_environment (NODE, METHOD_ENV (m), 1, 1, &cinfo) == -1) \
    fprintf (stderr, "get_method_type_environment: Error.\n");

#define get_interface_method_info(NODE)		\
  if (get_method_type_environment (NODE, METHOD_ENV (m), 1, 2, &cinfo) == -1) \
    fprintf (stderr, "get_method_type_environment: Error.\n");

#define field_env()		(finfo.e)
#define field_field()		(finfo.field)
#define field_name()		(finfo.name)
#define field_sig()		(finfo.type)

#define	method_name()		(cinfo.name)
#define	method_type()		(cinfo.type)
#define method_method()		(cinfo.method)
#define method_idx()		(METHOD_INDEX (cinfo.method))
#define method_class()		(cinfo.e)
#define	method_nargs()		(MTYPE_ARGC (METHOD_TYPE (cinfo.method)))
#define	method_returntype()	(STRCONST_BUFFER_IDX (MTYPE_NAME (METHOD_TYPE (cinfo.method)), MTYPE_RETUEN (METHOD_TYPE (cinfo.method))))

#define	method_dtable_offset	EOBJECTREF_DTABLE_OFFSET

#define	object_array_offset	(EARRAYREF_DATA_OFFSET)
#define	object_array_length	(EARRAYREF_SIZE_OFFSET)

#define	switchpair_size		8
#define	switchpair_addr		4
#define	switchtable_shift	2

#define	PUSH(N)			sp += (N)
#define	POP(N)			sp -= (N)

#define	stack(N)		(&sp[-(N)])
#define	stack_long(N)		(&sp[-(N)-1])
#define	stack_ref		stack
#define	stack_float		stack
#define	stack_double		stack_long

#define	rstack			stack
#define	wstack			stack
#define	rstack_ref		stack_ref
#define	wstack_ref		stack_ref
#define	rstack_long		stack_long
#define	wstack_long		stack_long
#define	rstack_float		stack_float
#define	wstack_float		stack_float
#define	rstack_double		stack_double
#define	wstack_double		stack_double

#define	local(N)		(&local[(N)])
#define local_ref		local
#define local_long		local
#define local_float		local
#define local_double		local_long

#define	METHOD_RETURN_VALUE()			\
do						\
  {						\
    /* Store the return type (if necessary) */	\
    switch (low)				\
      {						\
      case 'V':					\
	break;					\
      case 'L':					\
      case '[':					\
	PUSH (1);				\
       	return_ref (stack (0));			\
       	break;					\
      case 'I':					\
      case 'Z':					\
      case 'S':					\
      case 'B':					\
      case 'C':					\
	PUSH (1);				\
	return_int (stack (0));			\
	break;					\
      case 'F':					\
	PUSH (1);				\
	return_float (stack (0));		\
        break;					\
      case 'J':					\
	PUSH (2);				\
	return_long (stack_long (0));		\
	break;					\
      case 'D':					\
	PUSH (2);				\
	return_double (stack_double (0));	\
	break;					\
      default:					\
	ENV_ABORT ();				\
      }					        \
  }						\
while (0)

#define EXPLICIT_CHECK_NULL(_i, _s, _n)                       \
      cbranch_ref_const_ne ((_s), 0, reference_label (_i, _n)); \
      fprintf (stderr, "Exception: Null Pointer Exception.\n");	\
      set_label (_i, _n)

#define explicit_check_null(x,obj,y)	EXPLICIT_CHECK_NULL (x,obj,y)

/*
 * This macro initialises the static data associated with a class.
 */
#define	INIT_STATIC_CLASS(e)			\
  if (e != 0					\
      && ENV_STATE (e) != ENV_STATE_COMPLETE)	\
    process_env (e, 1);

static int call_method_core (struct emethod_callinfo *);

int run_vm (emethod_t *, evalue_t *, evalue_t *);
void CODE_DEBUG (const char *, ...);
int run_vm_core (emethod_t *, evalue_t *, evalue_t *, evalue_t *, int, struct VMexception *);
int call_method_v (emethod_t *, void *, void *, evalue_t *, va_list);
eobjectref_t * call_constructor_v (const char *, env_t, const char *, va_list);
int setup_exception_handling (struct VMexception *, emethod_t *, eobjectref_t *);
int cleanup_exception_handling (struct VMexception *);

void
CODE_DEBUG (const char *fmt, ...)
{
  va_list args;

  va_start (args, fmt);
  env_log_error_core (ENVLOG_MARK, ENVLOG_INFO, fmt, args);
  va_end (args);
}

int
run_vm_core (m, ret, local, sp, npc, mjbuf)
     emethod_t *m;
     evalue_t *local, *sp, *ret;
     int npc;
     struct VMexception *mjbuf;
{
  /* Variables used directly in the machine */
  int idx, low, high;
  env_t crinfo;
  evalue_t tmp[1], tmp2[1], mtable[1];
  long long lcc;
  struct efield_info finfo;
  struct emethod_info cinfo;
  bytecode *code = (bytecode *) METHOD_BCODE (m);

  while (1)
    {
      register int pc = npc;
      
      VMEXCEPTION_PC (mjbuf) = pc;
      npc = pc + insn_oplen[(code[(pc)])];

      /* Stack   0   , ֱ ϵ stack  Ų.  
	 stack (1)  stack (0)   ϵ stack  Ų.  */
      switch (code[pc])
	{
	case _NOP:		/* 0 */
	  CODE_DEBUG ("nop");
	  break;
	case _ACONST_NULL:	/* 1 */
	  CODE_DEBUG ("aconst 0");
	  PUSH (1);
	  move_ref_const (stack (0), 0);
	  break;
	case _ICONST_M1:	/* 2 */
	case _ICONST_0:		/* 3 */
	case _ICONST_1:		/* 4 */
	case _ICONST_2:		/* 5 */
	case _ICONST_3:		/* 6 */
	case _ICONST_4:		/* 7 */
	case _ICONST_5:		/* 8 */
	  CODE_DEBUG ("iconst\t%d", getopcode () - _ICONST_0);
	  PUSH (1);
	  move_int_const (stack (0), getopcode () - _ICONST_0);
	  break;
	case _LCONST_0:		/* 9 */
	case _LCONST_1:		/* 10 */
	  CODE_DEBUG ("lconst\t%d", getopcode () - _LCONST_0);
	  PUSH (2);
	  move_long_const (stack_long (0), getopcode () - _LCONST_0);
	  break;
	case _FCONST_0:		/* 11 */
	case _FCONST_1:		/* 12 */
	case _FCONST_2:		/* 13 */
	  CODE_DEBUG ("fconst %f", (double) (float)(getopcode () - _FCONST_0));
	  PUSH (1);
	  move_float_const (stack (0), (float)(getopcode () - _FCONST_0));
	  break;
	case _DCONST_0:		/* 14 */
	case _DCONST_1:		/* 15 */
	  CODE_DEBUG ("dconst %f", (double) (double) (getopcode () - _DCONST_0));
	  PUSH (2);
	  move_double_const (stack_double (0), (double) (getopcode () - _DCONST_0));
	  break;
	case _BIPUSH:		/* 16 */
	  low = (int8) getpc (0);
	  CODE_DEBUG ("bipush\t%d", low);
	  PUSH (1);
	  move_int_const (stack (0), low);
	  break;
	case _SIPUSH:		/* 17 */
	  low = (int16) ((getpc (0) << 8) | getpc (1));
	  CODE_DEBUG ("sipush\t%d", low);
	  PUSH (1);
	  move_int_const (stack (0), low);
	  break;
	case _LDC:		/* 18 */
	  {
	    idx = (uint8) getpc (0);
	    CODE_DEBUG ("ldc\t%d", idx);
	    PUSH (1);
	    low = CONSTPOOL_TAGS_IDX (ENV_CONSTANT_POOL (current_environment ()), idx);
	    switch (low) 
	      {
	      case CONSTANT_String:
		{
		  void *s;

		  s = resolve_string (current_environment (), idx);
		  if (!s)
		    throw_error (THROW_NullPointerException);

		  move_string_const (stack (0), s);
		  break;
		}
	      case CONSTANT_EnvedString:
		move_string_const (stack (0), (void *) CONSTPOOL_DATA_IDX (ENV_CONSTANT_POOL (current_environment ()), idx));
		break;
	      case CONSTANT_Float:
		move_float_const (stack (0), *(float*)&CONSTPOOL_DATA_IDX (ENV_CONSTANT_POOL (current_environment ()), idx));
		break;
	      case CONSTANT_Integer:
		move_int_const (stack (0), CONSTPOOL_DATA_IDX (ENV_CONSTANT_POOL (current_environment ()), idx));
		break;
	      default:
		ENV_ABORT ();
		break;
	      }
	    break;
	  }
	case _LDC_W: /* 19 */
	  {
	    idx = (uint16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("ldc_w\t%d", idx);
	    PUSH (1);
	    low = CONSTPOOL_TAGS_IDX (ENV_CONSTANT_POOL (current_environment ()), idx);
	    switch (low) 
	      {
	      case CONSTANT_String:
		{
		  void *s = resolve_string (current_environment (), idx);
		  if (s == 0)
		    throw_error (THROW_NullPointerException);
		  move_string_const (stack (0), s);
		  break;
		}
	      case CONSTANT_EnvedString:
		move_string_const (stack (0), (void *) CONSTPOOL_DATA_IDX (ENV_CONSTANT_POOL (current_environment ()), idx));
		break;
	      case CONSTANT_Float:
		move_float_const (stack (0), *(float*)&CONSTPOOL_DATA_IDX (ENV_CONSTANT_POOL (current_environment ()), idx));
		break;
	      case CONSTANT_Integer:
		move_int_const (stack (0), CONSTPOOL_DATA_IDX (ENV_CONSTANT_POOL (current_environment ()), idx));
		break;
	      default:
		ENV_ABORT ();
		break;
	      }
	    break;
	  }
	case _LDC2_W:
	  {
	    idx = (uint16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("ldc_w\t%d", idx);
	    PUSH (2);
	    low = CONSTPOOL_TAGS_IDX (ENV_CONSTANT_POOL (current_environment ()), idx);
	    switch (low) 
	      {
	      case CONSTANT_Long:
		tmp[0].v.elong = ENV_CONSTPOOL_LONG (current_environment (), idx);
		move_long_const (stack_long (0), tmp[0].v.elong);
		break;
	      case CONSTANT_Double:
		tmp[0].v.elong = ENV_CONSTPOOL_DOUBLE (current_environment (), idx);
		move_double_const (stack_double (0), tmp[0].v.edouble);
		break;
	      default:
		ENV_ABORT ();
		break;		
	      }
	    break;
	  }
	case _ILOAD: /* 21 */
	  idx = (uint8) getpc (0);
	  CODE_DEBUG ("iload\t%d", idx);
	  PUSH (1);
	  move_int (stack (0), local (idx));
	  break;
	case _LLOAD: /* 22 */
	  idx = (uint8) getpc (0);
	  CODE_DEBUG ("lload\t%d", idx);
	  PUSH (2);
	  move_long (stack_long (0), local_long (idx));
	  break;
	case _FLOAD: /* 23 */
	  idx = (uint8) getpc (0);
	  CODE_DEBUG ("fload\t%d", idx);
	  PUSH (1);
	  move_float (stack (0), local_float (idx));
	  break;
	case _DLOAD: /* 24 */
	  idx = (uint8) getpc (0);
	  CODE_DEBUG ("dload\t%d", idx);
	  PUSH (2);
	  move_double (stack_double (0), local_double (idx));
	  break;
	case _ALOAD: /* 25 */
	  idx = (uint8) getpc (0);
	  CODE_DEBUG ("aload\t%d", idx);
	  PUSH (1);
#if 0
	  CODE_DEBUG ("aload pointer [stack] = %p %p", stack (0), (stack (0))[0].v.eaddr);
	  CODE_DEBUG ("aload pointer [local] = %p %p", local (idx), (local (idx))[0].v.eaddr);
#endif
	  move_ref (stack (0), local (idx));
	  break;
	case _ILOAD_0: /* 26 */
	case _ILOAD_1: /* 27 */
	case _ILOAD_2: /* 28 */
	case _ILOAD_3: /* 29 */
	  CODE_DEBUG ("iload\t%d", getopcode () - _ILOAD_0);
	  PUSH (1);
	  move_int (stack (0), local (getopcode () - _ILOAD_0));
	  break;
	case _LLOAD_0: /* 30 */
	case _LLOAD_1: /* 31 */
	case _LLOAD_2: /* 32 */
	case _LLOAD_3: /* 33 */
	  CODE_DEBUG ("lload\t%d", getopcode () - _LLOAD_0);
	  PUSH (2);
	  move_long (stack_long (0), local_long (getopcode () - _LLOAD_0));
	  break;
	case _FLOAD_0: /* 34 */
	case _FLOAD_1: /* 35 */
	case _FLOAD_2: /* 36 */
	case _FLOAD_3: /* 37 */
	  CODE_DEBUG ("fload\t%d", getopcode () - _FLOAD_0);
	  PUSH (1);
	  move_float (stack_float (0), local_float (getopcode () - _FLOAD_0));
	  break;
	case _DLOAD_0: /* 38 */
	case _DLOAD_1: /* 39 */
	case _DLOAD_2: /* 40 */
	case _DLOAD_3: /* 41 */
	  CODE_DEBUG ("dload\t%d", getopcode () - _DLOAD_0);
	  PUSH (2);
	  move_double (stack_double (0), local_double (getopcode () - _DLOAD_0));
	  break;
	case _ALOAD_0: /* 42 */
	case _ALOAD_1: /* 43 */
	case _ALOAD_2: /* 44 */
	case _ALOAD_3: /* 45 */
	  CODE_DEBUG ("aload\t%d", getopcode () - _ALOAD_0);
	  PUSH (1);
	  move_ref (stack (0), local (getopcode () - _ALOAD_0));
	  break;
	case _IALOAD: /* 46 */
	  CODE_DEBUG ("iaload");
	  load_offset_scaled_int (stack (1), rstack (1), rstack (0), object_array_offset);
	  POP (1);
	  break;
	case _LALOAD: /* 47 */
	  CODE_DEBUG ("laload");
	  load_offset_scaled_long (stack_long (0), stack (1), stack (0), object_array_offset);
	  break;
	case _FALOAD: /* 48 */
	  CODE_DEBUG ("faload");
	  load_offset_scaled_float (stack_float (1), stack (1), stack (0), object_array_offset);
	  POP (1);
	  break;
	case _DALOAD: /* 49 */
	  {
	    CODE_DEBUG ("daload");
	    load_offset_scaled_double (stack_double (0), stack (1), stack (0), object_array_offset);
	    break;
	  }
	case _AALOAD: /* 50 */
	  {
	    CODE_DEBUG ("aaload");
	    load_offset_scaled_ref (stack (1), rstack (1), rstack (0), object_array_offset);
	    POP (1);
	    break;
	  }
	case _BALOAD: /* 51 */
	  {
	    CODE_DEBUG ("baload");
	    load_offset_scaled_byte (stack (1), rstack (1), rstack (0), object_array_offset);
	    POP (1);
	    break;
	  }
	case _CALOAD: /* 52 */
	  {
	    CODE_DEBUG ("caload");
	    load_offset_scaled_char (stack (1), rstack (1), rstack (0), object_array_offset);
	    POP (1);
	    break;
	  }
	case _SALOAD: /* 53 */
	  {
	    CODE_DEBUG ("saload");
	    load_offset_scaled_short (stack (1), rstack (1), rstack (0), object_array_offset);
	    POP (1);
	    break;
	  }
	case _ISTORE: /* 54 */
	  {
	    idx = (uint8) getpc (0);
	    CODE_DEBUG ("istore\t%d", idx);
	    move_int (local (idx), stack (0));
	    POP (1);
	    break;
	  }
	case _LSTORE: /* 55 */
	  {
	    idx = (uint8) getpc (0);
	    CODE_DEBUG ("lstore\t%d", idx);
	    move_long (local_long (idx), stack_long (0));
	    POP (2);
	    break;
	  }
	case _FSTORE: /* 56 */
	  {
	    idx = (uint8) getpc (0);
	    CODE_DEBUG ("fstore\t%d", idx);
	    move_float (local_float (idx), stack_float (0));
	    POP (1);
	    break;
	  }
	case _DSTORE: /* 57 */
	  {
	    idx = (uint8) getpc (0);
	    CODE_DEBUG ("dstore\t%d", idx);
	    move_double (local_double (idx), stack_double (0));
	    POP (2);
	    break;
	  }
	case _ASTORE: /* 58 */
	  {
	    idx = (uint8) getpc (0);
	    CODE_DEBUG ("astore\t%d", idx);
	    move_ref (local (idx), stack (0));
	    POP (1);
	    break;
	  }
	case _ISTORE_0: /* 59 */
	case _ISTORE_1: /* 60 */
	case _ISTORE_2: /* 61 */
	case _ISTORE_3: /* 62 */
	  {
	    CODE_DEBUG ("istore\t%d", getopcode () - _ISTORE_0);
	    move_int (local (getopcode () - _ISTORE_0), stack (0));
	    POP (1);
	    break;
	  }
	case _LSTORE_0: /* 63 */
	case _LSTORE_1: /* 64 */
	case _LSTORE_2: /* 65 */
	case _LSTORE_3: /* 66 */
	  {
	    CODE_DEBUG ("lstore\t%d", getopcode () - _LSTORE_0);
	    move_long (local_long (getopcode () - _LSTORE_0), stack_long (0));
	    POP (2);
	    break;
	  }
	case _FSTORE_0: /* 67 */
	case _FSTORE_1: /* 68 */
	case _FSTORE_2: /* 69 */
	case _FSTORE_3: /* 70 */
	  {
	    CODE_DEBUG ("lstore\t%d", getopcode () - _FSTORE_0);
	    move_float (local_float (getopcode () - _FSTORE_0), stack (0));
	    POP (1);
	    break;
	  }
	case _DSTORE_0: /* 71 */
	case _DSTORE_1: /* 72 */
	case _DSTORE_2: /* 73 */
	case _DSTORE_3: /* 74 */
	  {
	    CODE_DEBUG ("dstore\t%d", getopcode () - _DSTORE_0);
	    move_double (local_double (getopcode () - _DSTORE_0), stack_double (0));
	    POP (2);
	    break;
	  }
	case _ASTORE_0: /* 75 */
	case _ASTORE_1: /* 76 */
	case _ASTORE_2: /* 77 */
	case _ASTORE_3: /* 78 */
	  {
	    CODE_DEBUG ("astore\t%d", getopcode () - _ASTORE_0);
	    move_ref (local (getopcode () - _ASTORE_0), stack (0));
	    POP (1);
	    break;
	  }
	case _IASTORE: /* 79 */
	  {
	    CODE_DEBUG ("iastore");
	    store_offset_scaled_int (rstack (2), rstack (1), object_array_offset, rstack (0));
	    POP (3);
	    break;
	  }
	case _LASTORE: /* 80 */
	  {
	    CODE_DEBUG ("lastore");
	    store_offset_scaled_long (stack (3), stack (2), object_array_offset, stack_long (0));
	    POP (4);
	    break;
	  }
	case _FASTORE: /* 81 */
	  {
	    CODE_DEBUG ("fastore");
	    store_offset_scaled_float (stack (2), stack (1), object_array_offset, stack (0));
	    POP (3);
	    break;
	  }
	case _DASTORE: /* 82 */
	  {
	    CODE_DEBUG ("dastore");
	    store_offset_scaled_double (stack (3), stack (2), object_array_offset, stack_double (0));
	    POP (4);
	    break;
	  }
	case _AASTORE: /* 83 */
	  {
	    CODE_DEBUG ("aastore");
	    store_offset_scaled_ref (rstack (2), rstack (1), object_array_offset, rstack (0));
	    POP (3);
	    break;
	  }
	case _BASTORE: /* 84 */
	  {
	    CODE_DEBUG ("bastore");
#if 0
	    CODE_DEBUG ("bastore arrayref = %p int = %d offset = %d, pos = %p, pos2 = %p", 
			 (rstack (2))[0].v.eaddr, (rstack (1))[0].v.eint, object_array_offset,
			 ((char *)(((unsigned char *)((rstack (2))[0].v.eaddr)) + (object_array_offset))),
			 &(((char *)(((unsigned char *)((rstack (2))[0].v.eaddr)) + (object_array_offset)))[(rstack (1))[0].v.eint]));
#endif
	    store_offset_scaled_byte (rstack (2), rstack (1), object_array_offset, rstack (0));
#if 0
	    dump_buffer ((char *) (rstack (2))[0].v.eaddr, 32);
#endif
	    POP (3);
	    break;
	  }
	case _CASTORE: /* 85 */
	  {
	    CODE_DEBUG ("castore");
	    store_offset_scaled_char (rstack (2), rstack (1), object_array_offset, rstack (0));
	    POP (3);
	    break;
	  }
	case _SASTORE: /* 86 */
	  {
	    CODE_DEBUG ("sastore");
	    store_offset_scaled_short (rstack (2), rstack (1), object_array_offset, rstack (0));
	    POP (3);
	    break;
	  }
	case _POP: /* 87 */
	  {
	    CODE_DEBUG ("pop");
	    POP (1);
	    break;
	  }
	case _POP2: /* 88 */
	  {
	    CODE_DEBUG ("pop");
	    POP (2);
	    break;
	  }
	case _DUP: /* 89 */
	  {
	    CODE_DEBUG ("dup");
	    PUSH (1);
	    move_any (stack (0), stack (1));
	    break;
	  }
	case _DUP_X1: /* 90 */
	  {
	    CODE_DEBUG ("dup_x1");
	    PUSH (1);
	    move_any (stack (0), stack (1));
	    move_any (stack (1), stack (2));
	    move_any (stack (2), stack (0));
	    break;
	  }
	case _DUP_X2: /* 91 */
	  {
	    CODE_DEBUG ("dup_x2");
	    PUSH (1);
	    move_any (stack (0), stack (1));
	    move_any (stack (1), stack (2));
	    move_any (stack (2), stack (3));
	    move_any (stack (3), stack (0));
	    break;
	  }
	case _DUP2: /* 92 */
	  {
	    CODE_DEBUG ("dup2");
	    PUSH (2);
	    move_any (stack (0), stack (2));
	    move_any (stack (1), stack (3));
	    break;
	  }
	case _DUP2_X1: /* 93 */
	  {
	    CODE_DEBUG ("dup2_x1");
	    PUSH (2);
	    move_any (stack (0), stack (2));
	    move_any (stack (1), stack (3));
	    move_any (stack (2), stack (4));
	    move_any (stack (3), stack (0));
	    move_any (stack (4), stack (1));
	    break;
	  }
	case _DUP2_X2: /* 94 */
	  {
	    CODE_DEBUG ("dup2_x2");
	    PUSH (2);
	    move_any (stack (0), stack (2));
	    move_any (stack (1), stack (3));
	    move_any (stack (2), stack (4));
	    move_any (stack (3), stack (5));
	    move_any (stack (4), stack (0));
	    move_any (stack (5), stack (1));
	    break;
	  }
	case _SWAP: /* 95 */
	  {
	    CODE_DEBUG ("swap");
	    swap_any (stack (0), stack (1));
	    break;
	  }
	case _IADD: /* 96 */
	  {
	    CODE_DEBUG ("iadd");
	    add_int (stack (1), rstack (1), rstack (0));
	    POP (1);
	    break;
	  }
	case _LADD: /* 97 */
	  {
	    CODE_DEBUG ("ladd");
	    add_long (stack_long (2), rstack_long (2), rstack_long (0));
	    POP (2);
	    break;
	  }
	case _FADD: /* 98 */
	  {
	    CODE_DEBUG ("fadd");
	    add_float (stack_float (1), rstack_float (1), rstack_float (0));
	    POP (1);
	    break;
	  }
	case _DADD: /* 99 */
	  {
	    CODE_DEBUG ("dadd");
	    add_double (stack_double (2), rstack_double (2), rstack_double (0));
	    POP (2);
	    break;
	  }
	case _ISUB: /* 100 */
	  {
	    CODE_DEBUG ("isub");
	    sub_int (stack (1), rstack (1), rstack (0));
	    POP (1);
	    break;
	  }
	case _LSUB: /* 101 */
	  {
	    CODE_DEBUG ("lsub");
	    sub_long (stack_long (2), rstack_long (2), rstack_long (0));
	    POP (2);
	    break;
	  }
	case _FSUB: /* 102 */
	  {
	    CODE_DEBUG ("fsub");
	    sub_float (stack_float (1), rstack_float (1), rstack_float (0));
	    POP (1);
	    break;
	  }
	case _DSUB: /* 103 */
	  {
	    CODE_DEBUG ("dsub");
	    sub_double (stack_double (2), rstack_double (2), rstack_double (0));
	    POP (2);
	    break;
	  }
	case _IMUL: /* 104 */
	  {
	    CODE_DEBUG ("imul");
	    mul_int (stack (1), rstack (1), rstack (0));
	    POP (1);
	    break;
	  }
	case _LMUL: /* 105 */
	  {
	    CODE_DEBUG ("lmul");
	    mul_long (stack_long (2), rstack_long (2), rstack_long (0));
	    POP (2);
	    break;
	  }
	case _FMUL: /* 106 */
	  {
	    CODE_DEBUG ("fmul");
	    mul_float (stack_float (1), rstack_float (1), rstack_float (0));
	    POP (1);
	    break;
	  }
	case _DMUL: /* 107 */
	  {
	    CODE_DEBUG ("dmul");
	    mul_double (stack_double (2), rstack_double (2), rstack_double (0));
	    POP (2);
	    break;
	  }
	case _IDIV: /* 108 */
	  {
	    CODE_DEBUG ("idiv");
	    div_int (stack (1), rstack (1), rstack (0));
	    POP (1);
	    break;
	  }
	case _LDIV: /* 109 */
	  {
	    CODE_DEBUG ("ldiv");
	    div_long (stack_long (2), rstack_long (2), rstack_long (0));
	    POP (2);
	    break;
	  }
	case _FDIV: /* 110 */
	  {
	    CODE_DEBUG ("fdiv");
	    div_float (stack_float (1), rstack_float (1), rstack_float (0));
	    POP (1);
	    break;
	  }
	case _DDIV: /* 111 */
	  {
	    CODE_DEBUG ("ddiv");
	    div_double (stack_double (2), rstack_double (2), rstack_double (0));
	    POP (2);
	    break;
	  }
	case _IREM: /* 112 */
	  {
	    CODE_DEBUG ("irem");
	    rem_int (stack (1), rstack (1), rstack (0));
	    POP (1);
	    break;
	  }
	case _LREM: /* 113 */
	  {
	    CODE_DEBUG ("lrem");
	    rem_long (stack_long (2), rstack_long (2), rstack_long (0));
	    POP (2);
	    break;
	  }
	case _FREM: /* 114 */
	  {
	    CODE_DEBUG ("frem");
	    rem_float (stack_float (1), rstack_float (1), rstack_float (0));
	    POP (1);
	    break;
	  }
	case _DREM: /* 115 */
	  {
	    CODE_DEBUG ("drem");
	    rem_double (stack_double (2), rstack_double (2), rstack_double (0));
	    POP (2);
	    break;
	  }
	case _INEG: /* 116 */
	  {
	    CODE_DEBUG ("ineg");
	    negate_int (stack (0), rstack (0));
	    break;
	  }
	case _LNEG: /* 117 */
	  {
	    CODE_DEBUG ("lneg");
	    negate_long (stack_long (0), rstack_long (0));
	    break;
	  }
	case _FNEG: /* 118 */
	  {
	    CODE_DEBUG ("");
	    negate_float (stack_float (0), rstack_float (0));
	    break;
	  }
	case _DNEG: /* 119 */
	  {
	    CODE_DEBUG ("");
	    negate_double (stack_double (0), rstack_double (0));
	    break;
	  }
	case _ISHL: /* 120 */
	  {
	    CODE_DEBUG ("ishl");
	    lshl_int (stack (1), rstack (1), rstack (0));
	    POP (1);
	    break;
	  }
	case _LSHL: /* 121 */
	  {
	    CODE_DEBUG ("lshl");
	    lshl_long (stack_long (1), stack_long (1), stack (0));
	    POP (1);
	    break;
	  }
	case _ISHR: /* 122 */
	  {
	    CODE_DEBUG ("ishr");
	    ashr_int (stack (1), rstack (1), rstack (0));
	    POP (1);
	    break;
	  }
	case _LSHR: /* 123 */
	  {
	    CODE_DEBUG ("lshr");
	    ashr_long (stack_long (1), stack_long (1), stack (0));
	    POP (1);
	    break;
	  }
	case _IUSHR: /* 124 */
	  {
	    CODE_DEBUG ("iushr");
	    lshr_int (stack (1), rstack (1), rstack (0));
	    POP (1);
	    break;
	  }
	case _LUSHR: /* 125 */
	  {
	    CODE_DEBUG ("lushr");
	    lshr_long (stack_long (1), stack_long (1), stack (0));
	    POP (1);
	    break;
	  }
	case _IAND: /* 126 */
	  {
	    CODE_DEBUG ("iand");
	    and_int (stack (1), rstack (1), rstack (0));
	    POP (1);
	    break;
	  }
	case _LAND: /* 127 */
	  {
	    CODE_DEBUG ("land");
	    and_long (stack_long (2), rstack_long (2), rstack_long (0));
	    POP (2);
	    break;
	  }
	case _IOR: /* 128 */
	  {
	    CODE_DEBUG ("ior");
	    or_int (stack (1), rstack (1), rstack (0));
	    POP (1);
	    break;
	  }
	case _LOR: /* 129 */
	  {
	    CODE_DEBUG ("lor");
	    or_long (stack_long (2), rstack_long (2), rstack_long (0));
	    POP (2);
	    break;
	  }
	case _IXOR: /* 130 */
	  {
	    CODE_DEBUG ("ixor");
	    xor_int (stack (1), rstack (1), rstack (0));
	    POP (1);
	    break;
	  }
	case _LXOR: /* 131 */
	  {
	    CODE_DEBUG ("lxor");
	    xor_long (stack_long (2), rstack_long (2), rstack_long (0));
	    POP (2);
	    break;
	  }
	case _IINC: /* 132 */
	  {
	    idx = (uint8) getpc (0);
	    low = (int8) getpc (1);
	    CODE_DEBUG ("iinc\t%d %d", idx, low);
	    add_int_const (local (idx), local (idx), low);
	    break;
	  }
	case _I2L: /* 133 */
	  {
	    CODE_DEBUG ("i2l");
	    PUSH (1);
	    cvt_int_long (stack_long (0), stack (1));
	    break;
	  }
	case _I2F: /* 134 */
	  {
	    CODE_DEBUG ("i2f");
	    cvt_int_float (stack (0), stack (0));
	    break;
	  }
	case _I2D: /* 135 */
	  {
	    CODE_DEBUG ("i2d");
	    PUSH (1);
	    cvt_int_double (stack_double (0), stack (1));
	    break;
	  }
	case _L2I: /* 136 */
	  {
	    CODE_DEBUG ("l2i");
	    cvt_long_int (stack (1), stack_long (0));
	    POP (1);
	    break;
	  }
	case _L2F: /* 137 */
	  {
	    CODE_DEBUG ("l2f");
	    cvt_long_float (stack (1), stack_long (0));
	    POP (1);
	    break;
	  }
	case _L2D: /* 138 */
	  {
	    CODE_DEBUG ("l2d");
	    cvt_long_double (stack_double (0), stack_long (0));
	    break;
	  }
	case _F2I: /* 139 */
	  {
	    CODE_DEBUG ("f2i");
	    cvt_float_int (stack (0), stack (0));
	    break;
	  }
	case _F2L: /* 140 */
	  {
	    CODE_DEBUG ("f2l");
	    PUSH (1);
	    cvt_float_long (stack_long (0), stack (1));
	    break;
	  }
	case _F2D: /* 141 */
	  {
	    CODE_DEBUG ("f2d");
	    PUSH (1);
	    cvt_float_double (stack_double (0), stack (1));
	    break;
	  }
	case _D2I: /* 142 */
	  {
	    CODE_DEBUG ("d2i");
	    cvt_double_int (stack (1), stack_double (0));
	    POP (1);
	    break;
	  }
	case _D2L: /* 143 */
	  {
	    CODE_DEBUG ("d2l");
	    cvt_double_long (stack_long (0), stack_double (0));
	    break;
	  }
	case _D2F: /* 144 */
	  {
	    CODE_DEBUG ("d2f");
	    cvt_double_float (stack (1), stack_double (0));
	    POP (1);
	    break;
	  }
	case _I2B: /* 145 */
	  {
	    CODE_DEBUG ("i2b");
	    cvt_int_byte (stack (0), rstack (0));
	    break;
	  }
	case _I2C: /* 146 */
	  {
	    CODE_DEBUG ("i2c");
	    cvt_int_char (stack (0), rstack (0));
	    break;
	  }
	case _I2S: /* 147 */
	  {
	    CODE_DEBUG ("i2s");
	    cvt_int_short (stack (0), rstack (0));
	    break;
	  }
	case _LCMP: /* 148 */
	  {
	    CODE_DEBUG ("lcmp");
	    lcmp (stack (3), rstack_long (0), rstack_long (2));
	    POP (3);
	    break;
	  }
	case _FCMPL: /* 149 */
	  {
	    CODE_DEBUG ("fcmpl");
	    cmpl_float (stack (1), stack (1), stack (0));
	    POP (1);
	    break;
	  }
	case _FCMPG: /* 150 */
	  {
	    CODE_DEBUG ("fcmpg");
	    cmpg_float (stack (1), stack (1), stack (0));
	    POP (1);
	    break;
	  }
	case _DCMPL: /* 151 */
	  {
	    CODE_DEBUG ("dcmpl");
	    cmpl_double (stack (3), stack_double (2), stack_double (0));
	    POP (3);
	    break;
	  }
	case _DCMPG: /* 152 */
	  {
	    CODE_DEBUG ("dcmpg");
	    cmpg_double (stack (3), stack_double (2), stack_double (0));
	    POP (3);
	    break;
	  }
	case _IFEQ: /* 153 */
	  {
	    idx = (int16) (((getpc (0) << 8) | getpc (1)));
	    CODE_DEBUG ("ifeq\t%d", pc + idx);
	    cbranch_int_const_eq (rstack (0), 0, reference_code_label (pc+idx));
	    POP (1);
	    break;
	  }
	case _IFNE: /* 154 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("ifne\t%d", pc + idx);
	    cbranch_int_const_ne (rstack (0), 0, reference_code_label (pc+idx));
	    POP (1);	
	    break;
	  }
	case _IFLT: /* 155 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("iflt\t%d", pc + idx);
	    cbranch_int_const_lt (rstack (0), 0, reference_code_label (pc+idx));
	    POP (1);
	    break;
	  }
	case _IFGE: /* 156 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("ifge\t%d", pc + idx);
	    cbranch_int_const_ge (rstack (0), 0, reference_code_label (pc+idx));
	    POP (1);
	    break;
	  }
	case _IFGT: /* 157 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("ifgt\t%d", pc + idx);
	    cbranch_int_const_gt (rstack (0), 0, reference_code_label (pc+idx));
	    POP (1);
	    break;
	  }
	case _IFLE: /* 158 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("ifle\t%d", pc + idx);
	    cbranch_int_const_le (rstack (0), 0, reference_code_label (pc+idx));
	    POP (1);
	    break;
	  }
	case _IF_ICMPEQ: /* 159 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("if_icmpeq\t%d", pc + idx);
	    cbranch_int_eq (rstack (1), rstack (0), reference_code_label (pc+idx));
	    POP (2);	
	    break;
	  }
	case _IF_ICMPNE: /* 160 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("if_icmpne\t%d", pc + idx);
	    cbranch_int_ne (rstack (1), rstack (0), reference_code_label (pc+idx));
	    POP (2);
	    break;
	  }
	case _IF_ICMPLT: /* 161 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("if_icmplt\t%d", pc + idx);
	    cbranch_int_lt (rstack (1), rstack (0), reference_code_label (pc+idx));
	    POP (2);
	    break;
	  }
	case _IF_ICMPGE: /* 162 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("if_icmpge\t%d", pc + idx);
	    cbranch_int_ge (rstack (1), rstack (0), reference_code_label (pc+idx));
	    POP (2);
	    break;
	  }
	case _IF_ICMPGT: /* 163 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("if_icmpgt\t%d", pc + idx);
	    cbranch_int_gt (rstack (1), rstack (0), reference_code_label (pc + idx));
	    POP (2);
	    break;
	  }
	case _IF_ICMPLE: /* 164 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("if_icmple\t%d", pc + idx);
	    cbranch_int_le (rstack (1), rstack (0), reference_code_label (pc+idx));
	    POP (2);
	    break;
	  }
	case _IF_ACMPEQ: /* 165 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("if_acmpeq\t%d", pc + idx);
	    cbranch_ref_eq (rstack (1), rstack (0), reference_code_label (pc+idx));
	    POP (2);
	    break;
	  }
	case _IF_ACMPNE: /* 166 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("if_acmpne\t%d", pc + idx);
	    cbranch_ref_ne (rstack (1), rstack (0), reference_code_label (pc+idx));
	    POP (2);
	    break;
	  }
	case _GOTO: /* 167 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("goto\t%d", pc + idx);
	    branch_a (reference_code_label (pc + idx));
	    break;
	  }
	case _JSR: /* 168 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("jsr\t%d", pc + idx);
	    PUSH (1);
	    move_label_const (stack (0), reference_code_label (npc));
	    branch_a (reference_code_label (pc+idx));
	    break;
	  }
	case _RET: /* 169 */
	  {
	    idx = (uint8) getpc (0);
	    CODE_DEBUG ("ret %d", idx);
	    branch_indirect (stored_code_label (local (idx)));
	    break;
	  }
	case _TABLESWITCH: /* 170 */
	  {
	    npc = (pc + 1 + 3) & -4;
	    low = (int32) ((getcode (npc+4) << 24) | (getcode (npc+5) << 16) |
			   (getcode (npc+6) << 8) | getcode (npc+7));
	    high = (int32) ((getcode (npc+8) << 24) | (getcode (npc+9) << 16) |
			    (getcode (npc+10) << 8) | getcode (npc+11));
	    npc = npc + 12;
	    CODE_DEBUG ("tableswitch %d %d", low, high);
	    cbranch_int_const_lt (stack (0), low, reference_label (_TABLESWITCH, 8));
	    cbranch_int_const_le (stack (0), high, reference_label (_TABLESWITCH, 7));
	    
	    set_label (_TABLESWITCH, 8);
	    move_int_const (stack (0), low-3); /* Position at default entry */

	    set_label (_TABLESWITCH, 7);
	    if (low != 0) 
	      {
		sub_int_const (stack (0), stack (0), low);
	      }
	    lshl_int_const (stack (0), stack (0), switchtable_shift);
	    move_label_const (tmp2, reference_table_label (9));
	    add_ref (stack (0), tmp2, stack (0));
	    load_code_ref (stack (0), stack (0));
	    branch_indirect (table_code_label (stack (0)));
	    POP (1);
	    break;
	  }
	case _LOOKUPSWITCH:
	  {
	    npc = (pc + 1 + 3) & -4;
	    idx = (int32) ((getcode (npc+4) << 24) | (getcode (npc+5) << 16) |
			   (getcode (npc+6) << 8) | getcode (npc+7));
	    
	    CODE_DEBUG ("lookupswitch %d", idx);
	    
	    move_label_const (tmp, reference_table_label (7));
	    if (idx != 0) 
	      {
		move_ref (tmp2, tmp);
		add_ref_const (tmp, tmp, idx * switchpair_size);
		
		set_label (LOOKUPSWITCH, 5);
		load_key (mtable, tmp);
		cbranch_int_eq (mtable, stack (0), reference_label (LOOKUPSWITCH, 6));
		
		add_ref_const (tmp, tmp, -switchpair_size);
		cbranch_ref_ne (tmp, tmp2, reference_label (LOOKUPSWITCH, 5));
		
		add_ref_const (tmp, tmp, -switchpair_addr);

		set_label (LOOKUPSWITCH, 6);
		add_ref_const (tmp, tmp, switchpair_addr);
	      }
	    load_code_ref (tmp, tmp);
	    branch_indirect (table_code_label (tmp));
	    POP (1);
	    break;
	  }
	case _IRETURN: /* 172 */
	  {
	    CODE_DEBUG ("ireturn");
	    returnarg_int (rstack (0));
	    ret ();
	    break;
	  }
	case _LRETURN: /* 173 */
	  {
	    CODE_DEBUG ("lreturn");
	    returnarg_long (rstack_long (0));
	    ret ();
	    break;
	  }
	case _FRETURN: /* 174 */
	  {
	    CODE_DEBUG ("freturn");
	    returnarg_float (rstack_float (0));
	    ret ();
	    break;
	  }
	case _DRETURN: /* 175 */
	  {
	    CODE_DEBUG ("dreturn");
	    returnarg_double (rstack_double (0));
	    ret ();
	    break;
	  }
	case _ARETURN: /* 176 */
	  {
	    CODE_DEBUG ("areturn");
	    returnarg_ref (rstack_ref (0));
	    ret ();
	    break;
	  }
	case _RETURN: /* 177 */
	  {
	    CODE_DEBUG ("return");
	    ret ();
	    break;
	  }	  
	case _GETSTATIC: /* 178 */
	  {
	    idx = (uint16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("getstatic\t%d", idx);
	    get_static_field_info (idx);
	    INIT_STATIC_CLASS (field_env ());
	    if (field_field () == 0) 
	      {
		throw_error (THROW_NoSuchFieldError);

		switch (STRCONST_BUFFER_IDX (field_sig (), 0))
		  {
		  case 'J':
		  case 'D':
		    POP (2);
		    break;
		  default:
		    POP (1);
		    break;
		  }
	      }
	    else if (FIELD_IS_OBJECTREF_OR_ARRAYREF_TYPE (field_field ()))
	      {
		PUSH (1);
		load_static_ref (stack (0), ENV_STATIC_FIELDS_DATA_IDX (field_env (), FIELD_STATIC_DATA_INDEX (field_field ())));
	      }
	    else
	      {
		switch (ETYPE_KIND (FIELD_TYPE (field_field ())))
		  {
		  case ETYPE_INT:
		    PUSH (1);
		    load_static_int (stack (0), ENV_STATIC_FIELDS_DATA_IDX (field_env (), FIELD_STATIC_DATA_INDEX (field_field ())));
		    break;
		  case ETYPE_SHORT:
		    PUSH (1);
		    load_static_short (stack (0), ENV_STATIC_FIELDS_DATA_IDX (field_env (), FIELD_STATIC_DATA_INDEX (field_field ())));
		    break;
		  case ETYPE_BYTE:
		  case ETYPE_BOOLEAN:
		    PUSH (1);
		    load_static_byte (stack (0), ENV_STATIC_FIELDS_DATA_IDX (field_env (), FIELD_STATIC_DATA_INDEX (field_field ())));
		    break;
		  case ETYPE_CHAR:
		    PUSH (1);
		    load_static_char (stack (0), ENV_STATIC_FIELDS_DATA_IDX (field_env (), FIELD_STATIC_DATA_INDEX (field_field ())));
		    break;
		  case ETYPE_FLOAT:
		    PUSH (1);
		    load_static_float (stack (0), ENV_STATIC_FIELDS_DATA_IDX (field_env (), FIELD_STATIC_DATA_INDEX (field_field ())));
		    break;
		  case ETYPE_LONG:
		    PUSH (2);
		    load_static_long (stack_long (0), ENV_STATIC_FIELDS_DATA_IDX (field_env (), FIELD_STATIC_DATA_INDEX (field_field ())));
		    break;
		  case ETYPE_DOUBLE:
		    PUSH (2);
		    load_static_double (stack_double (0), ENV_STATIC_FIELDS_DATA_IDX (field_env (), FIELD_STATIC_DATA_INDEX (field_field ())));
		    break;
		  default:
		    ENV_ABORT ();
		    break;
		  }
	      }
	    break;
	  }
	case _PUTSTATIC:	/* 179 */
	  {
	    idx = (uint16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("putstatic\t%d", idx);
	    get_static_field_info (idx);
	    INIT_STATIC_CLASS (field_env ());
	  
	    if (field_field () == 0) 
	      {
		throw_error (THROW_NoSuchFieldError);

		switch (STRCONST_BUFFER_IDX (field_sig (), 0))
		  {
		  case 'J':
		  case 'D':
		    POP (2);
		    break;
		  default:
		    POP (1);
		    break;
		  }
	      }
	    else if (FIELD_IS_OBJECTREF_OR_ARRAYREF_TYPE (field_field ()))
	      {
		store_static_ref (ENV_STATIC_FIELDS_DATA_IDX (field_env (), FIELD_STATIC_DATA_INDEX (field_field ())), stack (0));
		POP (1);
	      }
	    else
	      {
		switch (ETYPE_KIND (FIELD_TYPE (field_field ())))
		  {
		  case ETYPE_INT:
		    store_static_int (ENV_STATIC_FIELDS_DATA_IDX (field_env (), FIELD_STATIC_DATA_INDEX (field_field ())), stack (0));
		    POP (1);
		    break;
		  case ETYPE_SHORT:
		    store_static_short (ENV_STATIC_FIELDS_DATA_IDX (field_env (), FIELD_STATIC_DATA_INDEX (field_field ())), stack (0));
		    POP (1);
		    break;
		  case ETYPE_BYTE:
		  case ETYPE_BOOLEAN:
		    store_static_byte (ENV_STATIC_FIELDS_DATA_IDX (field_env (), FIELD_STATIC_DATA_INDEX (field_field ())), stack (0));
		    POP (1);
		    break;
		  case ETYPE_CHAR:
		    store_static_char (ENV_STATIC_FIELDS_DATA_IDX (field_env (), FIELD_STATIC_DATA_INDEX (field_field ())), stack (0));
		    POP (1);
		    break;
		  case ETYPE_FLOAT:
		    store_static_float (ENV_STATIC_FIELDS_DATA_IDX (field_env (), FIELD_STATIC_DATA_INDEX (field_field ())), stack (0));
		    POP (1);
		    break;
		  case ETYPE_LONG:
		    store_static_long (ENV_STATIC_FIELDS_DATA_IDX (field_env (), FIELD_STATIC_DATA_INDEX (field_field ())), stack_long (0));
		    POP (2);
		    break;
		  case ETYPE_DOUBLE:
		    store_static_double (ENV_STATIC_FIELDS_DATA_IDX (field_env (), FIELD_STATIC_DATA_INDEX (field_field ())), stack_double (0));
		    POP (2);
		    break;
		  default:
		    ENV_ABORT ();
		    break;
		  }
	      }
	    break;
	  }
	case _GETFIELD:		/* 180 */
	  {
	    idx = (uint16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("getfield\t#%d", idx);
	    get_field_info (idx);	  

	    if (field_field () == 0)
	      {
		throw_error (THROW_NoSuchFieldError);

		switch (STRCONST_BUFFER_IDX (field_sig (), 0))
		  {
		  case 'J':
		  case 'D':
		    POP (1);
		    break;
		  default:
		    break;
		  }
	      }
	    else if (FIELD_IS_OBJECTREF_OR_ARRAYREF_TYPE (field_field ()))
	      {
		load_field_ref (stack (0), rstack (0), FIELD_OBJECT_DATA_OFFSET (field_field ()));
	      }
	    else
	      {
		switch (ETYPE_KIND (FIELD_TYPE (field_field ())))
		  {
		  case ETYPE_INT:
		    {
		      load_field_int (stack (0), rstack (0), FIELD_OBJECT_DATA_OFFSET (field_field ()));
		      break;
		    }
		  case ETYPE_SHORT:
		    {
		      load_field_short (stack (0), rstack (0), FIELD_OBJECT_DATA_OFFSET (field_field ()));
		      break;
		    }
		  case ETYPE_BYTE:
		  case ETYPE_BOOLEAN:
		    {
		      load_field_byte (stack (0), rstack (0), FIELD_OBJECT_DATA_OFFSET (field_field ()));
		      break;
		    }
		  case ETYPE_CHAR:
		    {
		      load_field_char (stack (0), rstack (0), FIELD_OBJECT_DATA_OFFSET (field_field ()));
		      break;
		    }
		  case ETYPE_FLOAT:
		    {
		      load_field_float (stack (0), rstack (0), FIELD_OBJECT_DATA_OFFSET (field_field ()));
		      break;
		    }
		  case ETYPE_LONG:
		    {
		      load_field_long (stack (0), rstack (0), FIELD_OBJECT_DATA_OFFSET (field_field ()));
		      PUSH (1);
		      break;
		    }
		  case ETYPE_DOUBLE:
		    {
		      load_field_double (stack (0), rstack (0), FIELD_OBJECT_DATA_OFFSET (field_field ()));
		      PUSH (1);
		      break;
		    }
		  default:
		    ENV_ABORT ();
		    break;
		  }
	      }
	    break;
	  }
	case _PUTFIELD: /* 181 */
	  {
	    /* ..., obj-ref, value  -> ...  */

	    idx = (uint16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("putfield\t#%d", idx);
	    get_field_info (idx);	  
	    if (field_field () == 0)
	      {
		throw_error (THROW_NoSuchFieldError);

		switch (STRCONST_BUFFER_IDX (field_sig (), 0))
		  {
		  case 'J':
		  case 'D':
		    POP (3);
		    break;
		  default:
		    POP (2);
		    break;
		  }
	      }
	    else if (FIELD_IS_OBJECTREF_OR_ARRAYREF_TYPE (field_field ()))
	      {
		/*  FIELD  '[' Ȥ 'L'  , ̸ ˸° óؾ Ѵ.  */
		store_field_ref (stack (1), FIELD_OBJECT_DATA_OFFSET (field_field ()), stack (0));
		POP (2);
	      }
	    else
	      {
		/* Ȯϰ ؼϱ ؼ  obj-ref   ׿  offset 
		     ϴ     ,  get_field_info () 
		    ش field       , Offset  
		   ʿ䰡 ν .  */
		switch (ETYPE_KIND (FIELD_TYPE (field_field ())))
		  {
		  case ETYPE_INT:
		    {
		      store_field_int (stack (1), FIELD_OBJECT_DATA_OFFSET (field_field ()), stack (0));
		      POP (2);
		      break;
		    }
		  case ETYPE_SHORT:
		    {
		      store_field_short (stack (1), FIELD_OBJECT_DATA_OFFSET (field_field ()), stack (0));
		      POP (2);
		      break;
		    }
		  case ETYPE_BYTE:
		  case ETYPE_BOOLEAN:
		    {
		      store_field_byte (stack (1), FIELD_OBJECT_DATA_OFFSET (field_field ()), stack (0));
		      POP (2);
		      break;
		    }
		  case ETYPE_CHAR:
		    {
		      store_field_char (stack (1), FIELD_OBJECT_DATA_OFFSET (field_field ()), stack (0));
		      POP (2);
		      break;
		    }
		  case ETYPE_FLOAT:
		    {
		      store_field_float (stack (1), FIELD_OBJECT_DATA_OFFSET (field_field ()), stack (0));
		      POP (2);
		      break;
		    }
		  case ETYPE_LONG:
		    {
		      store_field_long (stack (1), FIELD_OBJECT_DATA_OFFSET (field_field ()), stack (0));
		      POP (3);
		      break;
		    }
		  case ETYPE_DOUBLE:
		    {
		      store_field_double (stack (1), FIELD_OBJECT_DATA_OFFSET (field_field ()), stack (0));
		      POP (3);
		      break;
		    }
		  default:
		    ENV_ABORT ();
		    break;
		  }
	      }
	    break;
	  }
	case _INVOKEVIRTUAL: /* 182 */
	  {
	    idx = (uint16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("invokevirtual\t%d", idx);
	    get_method_info (idx);
	    if (method_class () == 0)
	      throw_error (THROW_NoClassDefFoundError);
	    else if (!method_method ())
	      throw_error (THROW_NoSuchMethodError);
	    else
	      {
		int is_virtual = (METHOD_INDEX (method_method ()) >= 0);

		idx = method_nargs ();
		if (is_virtual)
		  {
		    /* Find dispatch table in object */
		    load_offset_ref (tmp, stack (idx), method_dtable_offset);
		    
		    /* Check method table for cached entry */
		    load_offset_ref (tmp, tmp, DISPATCH_METHODOFFSET 
				     + method_idx () * DISPATCH_METHODSIZE);
		  }
		else {
		  explicit_check_null (_INVOKEVIRTUAL, stack (idx), 35);
		}
		
		idx++;
		POP (idx);
		
		/* Call it */
		low = method_returntype ();
		
		/* Ʒ IF { ~ }  ϰ ϱ ؼ, { Ȥ }  ؼ ȵȴ.
		   call  call_indirect_method     ũε̱ ̴.  */
		if (is_virtual) {
		  call (tmp);
		} else {
		  /* ȣϰ ϴ Method  AF_STATIC Ȥ AF_PRIVATE  Ǿ
		     ִٸ, is_virtual  ݵ -1  ̴.    method 
		     ȣ caller    𸣰 ȣ ɼ ϴ.  ׷
		     Ƹ object ref  ־  ̱ , ̸ ݵ  ְ
		     ȣ ؾ Ѵ.  */
		  call_indirect_method_skip_objectref (method_method ());
		}
		
		METHOD_RETURN_VALUE ();
	      }
	    break;
	  }
	case _INVOKESPECIAL: /* 183 */
	  {
	    idx = (uint16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("invokespecial\t%d", idx);
	    get_special_method_info (idx);
	    if (method_class () == 0)
	      throw_error (THROW_NoClassDefFoundError);
	    else if (method_method () == 0)
	      throw_error (THROW_NoSuchMethodError);
	    else
	      {
		idx = method_nargs ();
		idx++;
		POP (idx);
		/* Call it */
		low = method_returntype ();
		
		call_indirect_method (method_method ());
		METHOD_RETURN_VALUE ();
	      }
	    break;
	  }
	case _INVOKESTATIC: /* 184 */
	  {
	    idx = (uint16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("invokestatic\t%d", idx);
	    get_method_info (idx);
	    if (method_class () == 0)
	      throw_error (THROW_NoClassDefFoundError);
	    else if (method_method () == 0)
	      throw_error (THROW_NoSuchMethodError);
	    else
	      {
		idx = method_nargs ();
		POP (idx);
		/* Call it */
		low = method_returntype ();
		
		call_indirect_method (method_method ());
		METHOD_RETURN_VALUE ();
	      }
	    break;
	  }
	case _INVOKEINTERFACE:	/* 185 */
	  {
	    idx = (uint16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("invokeinterface\t%d", idx);
	    get_interface_method_info (idx);
	    if (method_class () == 0)
	      throw_error (THROW_NoClassDefFoundError);
	    else if (method_method () == 0)
	      throw_error (THROW_NoSuchMethodError);
	    else
	      {
		idx = ((unsigned int) getpc (2)) - 1;

		softcall_lookupinterfacemethod (tmp, method_method (), stack (idx));

		/* Push arguments & object */
		idx++;
		POP (idx);

		/* Call it */
		low = method_returntype ();
		call (tmp);

		METHOD_RETURN_VALUE();
	      }
	    break;
	  }
	case _NEW: /* 187 */
	  {
	    idx = (uint16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("new\t%d", idx);
	    get_class_info (idx);
	    PUSH (1);
	    if (class_object () == 0)
	      throw_error (THROW_NoClassDefFoundError);
	    else
	      softcall_new (stack (0), class_object ());
	    break;
	  }
	case _NEWARRAY: /* 188 */
	  {
	    /* Arrayref  ,  쿡  ٸ, Ʒ low  TYPE   object.h Ͽ
	       ִ  ü õǾ .  arrayref   մܿ struct eobject  ڸϸ
	        ڿ length,  ڿ  data  TYPE  ũ⸸ŭ ڸϰ ȴ.  */
	    low = (uint8) getpc (0);
	    CODE_DEBUG ("newarray	%d", low);
	    softcall_newarray (stack (0), stack (0), low);
	    break;
	  }
	case _ANEWARRAY:	/* 189 */
	  {
	    idx = (uint16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("anewarray\t%d", idx);
	    get_class_info (idx);
	    if (class_object () == 0)
	      throw_error (THROW_NoClassDefFoundError);
	    else
	      {
		softcall_anewarray (stack (0), stack (0), class_object ());
	      }
	    break;
	  }
	case _ARRAYLENGTH:	/* 190 */
	  {
	    CODE_DEBUG ("arraylength");
	    load_offset_int (stack (0), rstack (0), object_array_length);
	    break;
	  }
	case _ATHROW:		/* 191 */
	  {
	    CODE_DEBUG ("athrow");
	    softcall_athrow (rstack (0));
	    break;
	  }
	case _CHECKCAST:	/* 192 */
	  {
	    idx = (uint16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("checkcast\t%d", idx);
	    get_class_info (idx);
	    if (class_object() == 0)
	      throw_error (THROW_NoClassDefFoundError);
	    else
	      {
		softcall_checkcast (stack (0), rstack (0), class_object ());
	      }
	    break;
	  }
	case _INSTANCEOF:	/* 193 */
	  {
	    idx = (uint16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("instanceof\t%d", idx);
	    get_class_info (idx);
	    if (class_object() == 0)
	      throw_error (THROW_NoClassDefFoundError);
	    else
	      {
		softcall_instanceof (stack (0), rstack (0), class_object ());
	      }	    
	    break;
	  }
	case _MONITORENTER:	/* 194 */
	  {
	    CODE_DEBUG ("monitorenter");
	    softcall_monitorenter (rstack (0));
	    POP (1);
	    break;
	  }
	case _MONITOREXIT:	/* 195 */
	  {
	    CODE_DEBUG ("monitorexit");
	    softcall_monitorexit (rstack (0));
	    POP (1);
	    break;
	  }
	case _WIDE:
	  {
	    /* Move PC to widened instruction */
	    pc = npc;
	    npc = pc + insn_oplen [getopcode ()] + 1;

	    switch(getopcode ()) 
	      {
	      case _ILOAD:
		{
		  idx = (uint16) ((getpc (0) << 8) | getpc (1));
		  CODE_DEBUG ("wide iload %d\n", idx);
		  PUSH (1);
		  move_int (stack (0), local (idx));
		  break;
		}
	      case _FLOAD:
		{
		  idx = (uint16) ((getpc (0) << 8) | getpc (1));
		  CODE_DEBUG ("wide fload %d\n", idx);
		  PUSH (1);
		  move_float (stack (0), local_float (idx));
		  break;
		}
	      case _ALOAD:
		{
		  idx = (uint16) ((getpc (0) << 8) | getpc (1));
		  CODE_DEBUG ("wide aload %d\n", idx);
		  PUSH (1);
		  move_ref (stack (0), local (idx));
		  break;
		}
	      case _LLOAD:
		{
		  idx = (uint16) ((getpc (0) << 8) | getpc (1));
		  CODE_DEBUG ("wide lload %d\n", idx);
		  PUSH (2);
		  move_long (stack_long (0), local_long (idx));
		  break;
		}
	      case _DLOAD:
		{
		  idx = (uint16) ((getpc (0) << 8) | getpc (1));
		  CODE_DEBUG ("wide dload %d\n", idx);
		  PUSH (2);
		  move_double (stack_double (0), local_double (idx));
		  break;
		}
	      case _ISTORE:
		{
		  idx = (uint16) ((getpc (0) << 8) | getpc (1));
		  CODE_DEBUG ("wide istore %d\n", idx);
		  move_int (local (idx), stack (0));
		  POP (1); 
		  break;
		}
	      case _FSTORE:
		{
		  idx = (uint16) ((getpc (0) << 8) | getpc (1));
		  CODE_DEBUG ("wide fstore %d\n", idx);
		  move_float (local_float (idx), stack (0));
		  POP (2); 
		  break;
		}
	      case _ASTORE:
		{
		  idx = (uint16) ((getpc (0) << 8) | getpc (1));
		  CODE_DEBUG ("wide astore %d\n", idx);
		  move_ref (local (idx), stack (0));
		  POP (1);
		  break;
		}
	      case _LSTORE:
		{
		  idx = (uint16) ((getpc (0) << 8) | getpc (1));
		  CODE_DEBUG ("wide lstore %d\n", idx);
		  move_long (local_long (idx), stack_long (0));
		  POP (2);
		  break;
		}
	      case _DSTORE:
		{
		  idx = (uint16) ((getpc (0) << 8) | getpc (1));
		  CODE_DEBUG ("wide dstore %d\n", idx);
		  move_double (local_double (idx), stack_double (0));
		  POP (2);
		  break;
		}
	      case _IINC:
		{
		  idx = (uint16) ((getpc (0) << 8) | getpc (1));
		  low = (int16) ((getpc (2) << 8) | getpc (3));
		  CODE_DEBUG ("wide iinc %d %d\n", idx, low);
		  add_int_const (local (idx), local (idx), low);
		  npc += 1;	/* Extra one */
		  break;
		}
	      default:
		ENV_ABORT ();
	      }
	    break;
	  }
	case _MULTIANEWARRAY:	/* 197 */
	  {
	    idx = (uint16) ((getpc (0) << 8) | getpc (1));
	    low = (uint8) getpc (2);
	    CODE_DEBUG ("multinewarray %d %d\n", idx, low);
	    get_class_info (idx);
	    softcall_multianewarray (stack (low-1), low, stack (0), class_object ());
	    break;
	  }
	case _IFNULL:		/* 198 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc(1));
	    CODE_DEBUG ("ifnull %d\n", pc + idx);

	    cbranch_ref_const_eq (rstack (0), 0, reference_code_label (pc + idx));
	    POP (1);
	    break;
	  }
	case _IFNONNULL:	/* 199 */
	  {
	    idx = (int16) ((getpc (0) << 8) | getpc (1));
	    CODE_DEBUG ("ifnonnull\t%d\n", pc + idx);

	    cbranch_ref_const_ne (rstack (0), 0, reference_code_label (pc + idx));
	    POP (1);
	    break;
	  }
	case _GOTO_W:		/* 200 */
	  {
	    idx = (int32) ((getpc (0) << 24) | (getpc (1) << 16) |
			   (getpc (2) << 8) | getpc (3));
	    CODE_DEBUG ("goto_w %d\n", pc + idx);	    
	    branch_a (reference_code_label (pc + idx));
	    break;
	  }
	case _JSR_W:		/* 201 */
	  {
	    idx = (int32) ((getpc (0) << 24) | (getpc (1) << 16) |
			   (getpc (2) << 8) | getpc (3));
	    CODE_DEBUG ("jsr_w\t%d", pc + idx);
	    PUSH (1);
	    move_label_const (stack (0), reference_code_label (npc));
	    branch_a (reference_code_label (pc + idx));
	    break;
	  }
	case _BREAKPOINT:	/* 202 */
	  {
	    CODE_DEBUG ("breakpoint");
	    softcall_breakpoint ();
	  }
	default:
	  CODE_DEBUG ("Unknown bytecode %d [%s]", code[pc], insn_name[code[pc]]);
	  goto end;
	  break;
	}
    }

 end:
  return 0;
}

/* run_vm 󿡼 ⺻ exception  Ѵ.  */
int
setup_exception_handling (v, m, obj)
     struct VMexception *v;
     emethod_t *m;
     eobjectref_t *obj;
{
  erequest_t *er;
  
  if (!v || !m)
    return -1;

  er = FETCH_EREQUEST;
  if (!er)
    return -1;

  VMEXCEPTION_METHOD (v) = m;
  VMEXCEPTION_OBJECTREF (v) = obj;
  VMEXCEPTION_PC (v) = 0;
  VMEXCEPTION_PREV (v) = REQ_DATA_EXCEPTION (REQUEST_DATA (er));

  REQ_DATA_EXCEPTION (REQUEST_DATA (er)) = v;

  return 0;
}

int
cleanup_exception_handling (v)
     struct VMexception *v;
{
  erequest_t *er = FETCH_EREQUEST;

  if (!er)
    return -1;

  REQ_DATA_EXCEPTION (REQUEST_DATA (er)) = VMEXCEPTION_PREV (v);

  return 0;
}

static void only_debug_check_return_value (emethod_t *, evalue_t *);

static void
only_debug_check_return_value (m, ret)
     emethod_t *m ATTRIBUTE_UNUSED;
     evalue_t *ret ATTRIBUTE_UNUSED;
{

  /* nothing.  */
}

/* Method TM   ϱ ؼ, ʿ  ȯ ϰ,
    ̸  run_virtual_machine ()  ȣѴ.   Լ ȣ , 
     Argument  ޵ǰ ȴ.  

   ȣϰ ϴ method  AF_STATIC  ,

     ARG[0] = ȣϰ ϴ method  [0]
     ....

   ȣϰ ϴ method  AF_STATIC  ƴ ,

     ARG[0] =  environment   Object Ref
     ARG[1] = ȣϰ ϴ method  [0]

   , ̷ ARG  ȣϰ ϴ method  Local ν, ϵǰ Ǹ,
   ȣ ޴ method  ̷  ϰ ڵ带 ۼؾ Ѵ.  */
int
run_vm (m, ret, argv)
     emethod_t *m;
     evalue_t *argv, *ret;
{
  int i, argc = -1;
  evalue_t * volatile local;
  evalue_t * volatile sp;
  emethod_t *volatile const vm = m;
  uintp volatile npc = 0;
  enum env_access_flags af;
  struct VMexception ve;

  /* If this is native, then call the real function */
  if (METHOD_ACCESS_FLAGS (m) & AF_NATIVE)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		     "vm: call to native %s %s%s",
		     STRCONST_PTR (ENV_NAME (METHOD_ENV (m))),
		     STRCONST_PTR (METHOD_NAME (m)),
		     STRCONST_PTR (MTYPE_NAME (METHOD_TYPE (m))));

      if (METHOD_ACCESS_FLAGS (m) & AF_STATIC)
	{
	  if (call_method_a (m, m, NULL, ret, argv))
	    return -1;
	}
      else
	{
	  if (call_method_a (m, m, EVALUE_OBJECT (argv), ret, &argv[1]))
	    return -1;
	}

      env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		     "vm: return native %s %s%s, value = (type 0x%x) 0x%x",
		     STRCONST_PTR (ENV_NAME (METHOD_ENV (m))),
		     STRCONST_PTR (METHOD_NAME (m)),
		     STRCONST_PTR (MTYPE_NAME (METHOD_TYPE (m))),
		     EVALUE_TYPE (ret), EVALUE_INT (ret));

      return 0;
    }

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "vm: call %s %s%s",
		 STRCONST_PTR (ENV_NAME (METHOD_ENV (m))),
		 STRCONST_PTR (METHOD_NAME (m)),
		 STRCONST_PTR (MTYPE_NAME (METHOD_TYPE (m))));

  af = METHOD_ACCESS_FLAGS (m);

  /* Stack  local  ڸ  ** Ҵ״.  ̰ ѹ ,
      ̴. */
  local = (evalue_t *) alloca (sizeof (struct evalue) * (METHOD_LOCAL_SIZE (m) + METHOD_STACK_SIZE (m)));
  memset (local, '\0', (sizeof (struct evalue) * (METHOD_LOCAL_SIZE (m) + METHOD_STACK_SIZE (m))));

  setup_exception_handling (&ve, m, NULL);

  /*  exception table   , 츮 װ catch  غ ؾѴ.  */
  if (METHOD_EXCEPTION_TABLE (m))
    {
      if (setjmp (VMEXCEPTION_JMP_BUF (&ve)))
	{
	  erequest_t *er = FETCH_EREQUEST;

	  /* Exception  ߻Ͽ ,   ̴.  */
	  m = vm;
	  REQ_DATA_EXCEPTION (REQUEST_DATA (er)) = &ve;
	  npc = VMEXCEPTION_PC (&ve);
	  sp = &local[METHOD_LOCAL_SIZE (m)];
	  EVALUE_ADDRESS (sp) = (void *) REQ_DATA_EXCEPTION (REQUEST_DATA (er));
	  REQ_DATA_EXCEPTION (REQUEST_DATA (er)) = 0;
	  run_vm_core (m, ret, local, sp, npc, &ve);
	  goto end;
	}
    }

  argc = MTYPE_ARGC (METHOD_TYPE (m));
  if (argc < 0)
    return -1;

  /* AF_STATIC  , object ref   , AF_STATIC  ƴ ,
     ̰ ȣϴ caller  args  object ref  ڵ ־  ̴.  */
  argc += ((METHOD_ACCESS_FLAGS (m) & AF_STATIC) ? 0 : 1);

  /*  ARGS  ȣϰ ϴ method  Local  ٷ  ֵ, Local 
      Ѵ.  Argument  ޵  ׻ Local    ȴ.  */
  sp = local;
  for (i = 0; i < argc; i++) 
    *(sp++) = *(argv++);

  /* LOCAL  local variable  ϵ   ̸,  object  lock  ,
     local[0]  objectref   ̶   ̴.  */
  lock_synchronized (m, local);

  /*  -1  ̴  ó PUSH Ǿ  ġ Stack Slot  ù° ġ
     ϱ ̴.  Method   PUSH   , POP    ʴ´ٴ 
     ̴.  */
  sp = &local[METHOD_LOCAL_SIZE (m) - 1];
  run_vm_core (m, ret, local, sp, npc, &ve);

 end:
  unlock_synchronized (m, local);

  cleanup_exception_handling (&ve);

  only_debug_check_return_value (m, ret);

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "vm: return %s %s%s, value = (type 0x%x) 0x%x",
		 STRCONST_PTR (ENV_NAME (METHOD_ENV (m))),
		 STRCONST_PTR (METHOD_NAME (m)),
		 STRCONST_PTR (MTYPE_NAME (METHOD_TYPE (m))),
		 EVALUE_TYPE (ret), EVALUE_INT (ret));

  return 0;
}

static int
call_method_core (ci)
     struct emethod_callinfo *ci;
{
  emethod_t *m = (emethod_t *) CALLINFO_FUNCTION (ci);

  /*  method  native interface   , external.c  native () Լ
     ȣѴ.  ش object Ȥ library  load Ͽ, ش Լ ͸ Ѵ.  */ 
  if (METHOD_ACCESS_FLAGS (m) & AF_NATIVE)
    {
      /* TODO: system dependent call  ϴ  ߻ϴ exception  
	 ó ؾ Ѵ.  */
      
      if (!METHOD_TRANSLATED (m))
	{
	  void *func = call_native (m);
	  if (!func)
	    {
	      env_log_error (ENVLOG_MARK, ENVLOG_ERR, 
			     "Native ̺귯 ش Լ ã  ϴ.");

	      /* Native Լ ȣ  , ߸ return  Ǵ 
		   ġ̴.  */
	      memset (CALLINFO_RETURN (ci), '\0', sizeof (struct evalue));

	      return -1;
	    }
	  METHOD_NCODE_START (m) = func;
	  METHOD_ACCESS_FLAGS (m) |= AF_TRANSLATED;
	}

      CALLINFO_FUNCTION (ci) = METHOD_NCODE_START (m);

      if (METHOD_ACCESS_FLAGS (m) & AF_ENI)
	{
	  if (METHOD_ACCESS_FLAGS (m) & AF_STATIC)
	    {
	      CALLINFO_CALLSIZE_IDX (ci, 1) = PTR_TYPE_SIZE / SIZEOF_INT;
	      CALLINFO_ARGSIZE (ci) += CALLINFO_CALLSIZE_IDX (ci, 1);
	      CALLINFO_CALLTYPE_IDX (ci, 1) = 'L';
	      EVALUE_ADDRESS (&CALLINFO_ARGS_IDX (ci, 1)) = METHOD_ENV (m);
	    }
	  else
	    {
	      CALLINFO_NRARGS (ci) -= 1;
	      CALLINFO_ARGS (ci) += 1;
	      CALLINFO_CALLSIZE (ci) += 1;
	      CALLINFO_CALLTYPE (ci) += 1;
	    }
	  
	  /* Insert the ENI environment */
	  CALLINFO_CALLSIZE_IDX (ci, 0) = PTR_TYPE_SIZE / SIZEOF_INT;
	  CALLINFO_CALLTYPE_IDX (ci, 0) = 'L';
	  EVALUE_ADDRESS (&CALLINFO_ARGS_IDX (ci, 0)) = &eserver_ENINativeInterface; 
	  CALLINFO_ARGSIZE (ci) += CALLINFO_CALLSIZE_IDX (ci, 0);
	}
      else
	{
	  CALLINFO_NRARGS (ci) -= 2;
	  CALLINFO_ARGS (ci) += 2;
	  CALLINFO_CALLSIZE (ci) += 2;
	  CALLINFO_CALLTYPE (ci) += 2;
	}

      lock_synchronized (m, CALLINFO_ARGS (ci));

      /* Make the call - system dependent */
      SYSDEP_CALL_FUNCTION (ci);

      unlock_synchronized (m, CALLINFO_ARGS (ci));
    }
  else
    run_vm (m, CALLINFO_RETURN (ci), CALLINFO_ARGS (ci) + 2);

  return 0;
}

#define RESERVED_ARGS		2

/* Generic routine to call a native or method (array style).  

    method  static   ʾ , argv[0]  ݵ objectref Ȥ
   arrayref  Ѵ.
   
   TODO: struct call_method_info ü ȣ      Ȯϰ
   ؾ Ѵ.  */
int
call_method_a (m, func, obj, ret, argv)
     emethod_t *m;
     void *func, *obj;
     evalue_t *ret, *argv;
{
  int i = RESERVED_ARGS, j, s = 0;
  struct emethod_callinfo callinfo, *ci;
  evalue_t unused_ret = EVALUE_NULL;

  if (ret == NULL)
    ret = &unused_ret;

  ci = &callinfo;

  /* struct emethod_callinfo  Ѵ.  
     TODO: ⿡  ִ  4  2  Ȱ  Ȯ ǰ ʿϴ.  */
  CALLINFO_ARGS (ci) = (evalue_t *) alloca ((METHOD_ARGC (m) + RESERVED_ARGS + 2) * (sizeof (struct evalue) + 2));
  memset (CALLINFO_ARGS (ci), '\0', (METHOD_ARGC (m) + RESERVED_ARGS + 2) * (sizeof (struct evalue) + 2));
  CALLINFO_CALLSIZE (ci) = (char *) &CALLINFO_ARGS_IDX (ci, METHOD_ARGC (m) + RESERVED_ARGS + 2);
  CALLINFO_CALLTYPE (ci) = (char *) &CALLINFO_CALLSIZE_IDX (ci, METHOD_ARGC (m) + RESERVED_ARGS + 2);

  if (!(METHOD_ACCESS_FLAGS (m) & AF_STATIC))
    {
      CALLINFO_CALLSIZE_IDX (ci, i) = etype_slot[ETYPE_OBJECTREF];
      s += CALLINFO_CALLSIZE_IDX (ci, i);
      CALLINFO_CALLTYPE_IDX (ci, i) = 'L';
      EVALUE_OBJECT (&CALLINFO_ARGS_IDX (ci, i)) = obj;
      i++;
    }

  for (j = 0; j < METHOD_ARGC (m); i++, j++) 
    {
      CALLINFO_CALLTYPE_IDX (ci, i) = METHOD_ARGV_IDX_CHAR (m, j);
      switch (CALLINFO_CALLTYPE_IDX (ci, i)) 
	{
	case 'Z':
	  CALLINFO_CALLSIZE_IDX (ci, i) = etype_slot[ETYPE_BOOLEAN];
	  EVALUE_BOOLEAN (&CALLINFO_ARGS_IDX (ci, i)) = EVALUE_BOOLEAN (argv + j);
	  break;
	case 'S':
	  CALLINFO_CALLSIZE_IDX (ci, i) = etype_slot[ETYPE_SHORT];
	  EVALUE_SHORT (&CALLINFO_ARGS_IDX (ci, i)) = EVALUE_SHORT (argv + j);
	  break;
	case 'B':
	  CALLINFO_CALLSIZE_IDX (ci, i) = etype_slot[ETYPE_BYTE];
	  EVALUE_BYTE (&CALLINFO_ARGS_IDX (ci, i)) = EVALUE_BYTE (argv + j);
	  break;
	case 'C':
	  CALLINFO_CALLSIZE_IDX (ci, i) = etype_slot[ETYPE_CHAR];
	  EVALUE_CHAR (&CALLINFO_ARGS_IDX (ci, i)) = EVALUE_CHAR (argv + j);
	  break;
	case 'F':
	  CALLINFO_CALLSIZE_IDX (ci, i) = etype_slot[ETYPE_FLOAT];
	  EVALUE_FLOAT (&CALLINFO_ARGS_IDX (ci, i)) = EVALUE_FLOAT (argv + j);
	  break;
	case 'I':
	  CALLINFO_CALLSIZE_IDX (ci, i) = etype_slot[ETYPE_INT];
	  EVALUE_INT (&CALLINFO_ARGS_IDX (ci, i)) = EVALUE_INT (argv + j);
	  break;
	case 'D':
	case 'J':
	  CALLINFO_CALLSIZE_IDX (ci, i) = etype_slot[ETYPE_LONG];
	  CALLINFO_ARGS_IDX (ci, i) = argv[j];
	  break;
	case '[':
	  CALLINFO_CALLTYPE_IDX (ci, i) = 'L';
	  /* fall through */
	case 'L':
	  CALLINFO_CALLSIZE_IDX (ci, i) = etype_slot[ETYPE_OBJECTREF];
	  CALLINFO_ARGS_IDX (ci, i) = argv[j];
	  break;
	default:
	  ENV_ABORT();
	}
      s += CALLINFO_CALLSIZE_IDX (ci, i);
    }
  
  /* Return info */
  CALLINFO_RETTYPE (ci) = METHOD_RET_CHAR (m);
  switch (CALLINFO_RETTYPE (ci)) 
    {
    case '[':
      CALLINFO_RETTYPE (ci)  = 'L';
      /* fall through */
    case 'L':
      CALLINFO_RETSIZE (ci) = etype_slot[ETYPE_OBJECTREF];
      break;
    case 'V':
      CALLINFO_RETSIZE (ci) = etype_slot[ETYPE_VOID];
      break;
    case 'D':
    case 'J':
      CALLINFO_RETSIZE (ci) = etype_slot[ETYPE_LONG];
      break;
    default:
      CALLINFO_RETSIZE (ci) = etype_slot[ETYPE_INT];
      break;
    }
  
  /* Call info and arguments */
  CALLINFO_NRARGS (ci) = i;
  CALLINFO_ARGSIZE (ci) = s;
  CALLINFO_RETURN (ci) = ret;
  CALLINFO_FUNCTION (ci) = func;

  if (call_method_core (&callinfo))
    return -1;
  
  memset (CALLINFO_ARGS (ci), '\0', (METHOD_ARGC (m) + RESERVED_ARGS + 2) * (sizeof (struct evalue) + 2));
  memset (ci, 0, sizeof (struct emethod_callinfo));

  return 0;
}

/*
  @param meth the struct _methods of the method to be executed
  @param func the code that's to be executed
  @param obj  the object whose method is to be called (my be 0 iff method is static)
  @param ret  buffer for the return value of the method (may be 0 iff return type is void)
  @param args the arguments to be passed to the method
*/
int
call_method_v (m, func, obj, ret, argv)
     emethod_t *m;
     void *obj, *func;
     evalue_t *ret;
     va_list argv;
{
  int i = RESERVED_ARGS, j, s = 0;
  struct emethod_callinfo callinfo, *ci;
  evalue_t unused_ret = EVALUE_NULL;

  if (ret == NULL)
    ret = &unused_ret;

  ci = &callinfo;

  /* struct emethod_callinfo  Ѵ.  
     TODO: ⿡  ִ  4  2  Ȱ  Ȯ ǰ ʿϴ.  */
  CALLINFO_ARGS (ci) = (evalue_t *) alloca ((METHOD_ARGC (m) + RESERVED_ARGS + 2) * (sizeof (struct evalue) + 2));
  memset (CALLINFO_ARGS (ci), '\0', ((METHOD_ARGC (m) + RESERVED_ARGS + 2) * (sizeof (struct evalue) + 2)));
  CALLINFO_CALLSIZE (ci) = (char *) &CALLINFO_ARGS_IDX (ci, METHOD_ARGC (m) + RESERVED_ARGS + 2);
  CALLINFO_CALLTYPE (ci) = (char *) &CALLINFO_CALLSIZE_IDX (ci, METHOD_ARGC (m) + RESERVED_ARGS + 2);

  if (!(METHOD_ACCESS_FLAGS (m) & AF_STATIC))
    {
      CALLINFO_CALLSIZE_IDX (ci, i) = etype_slot[ETYPE_OBJECTREF];
      s += CALLINFO_CALLSIZE_IDX (ci, i);
      CALLINFO_CALLTYPE_IDX (ci, i) = 'L';
      EVALUE_OBJECT (&CALLINFO_ARGS_IDX (ci, i)) = obj;
      i++;
    }

  for (j = 0; j < METHOD_ARGC (m); i++, j++) 
    {
      CALLINFO_CALLTYPE_IDX (ci, i) = METHOD_ARGV_IDX_CHAR (m, j);
      switch (CALLINFO_CALLTYPE_IDX (ci, i)) 
	{
	case 'I':
	case 'Z':
	case 'S':
	case 'B':
	case 'C':
	  CALLINFO_CALLSIZE_IDX (ci, i) = etype_slot[ETYPE_INT];
	  EVALUE_INT (&CALLINFO_ARGS_IDX (ci, i)) = va_arg (argv, eint_t);
	  break;
	case 'F':
	  CALLINFO_CALLSIZE_IDX (ci, i) = etype_slot[ETYPE_FLOAT];
	  EVALUE_FLOAT (&CALLINFO_ARGS_IDX (ci, i)) = (edouble_t) va_arg (argv, edouble_t);
	  break;
	case 'D':
	  CALLINFO_CALLSIZE_IDX (ci, i) = etype_slot[ETYPE_DOUBLE];
	  EVALUE_DOUBLE (&CALLINFO_ARGS_IDX (ci, i)) = (edouble_t) va_arg (argv, edouble_t);
	  break;
	case 'J':
	  CALLINFO_CALLSIZE_IDX (ci, i) = etype_slot[ETYPE_LONG];
	  EVALUE_LONG (&CALLINFO_ARGS_IDX (ci, i)) = (elong_t) va_arg (argv, elong_t);
	  break;
	case '[':
	  CALLINFO_CALLTYPE_IDX (ci, i) = 'L';
	  /* fall through */
	case 'L':
	  CALLINFO_CALLSIZE_IDX (ci, i) = etype_slot[ETYPE_OBJECTREF];
	  EVALUE_ADDRESS (&CALLINFO_ARGS_IDX (ci, i)) = (eref_t) va_arg (argv, eref_t);
	  break;
	default:
	  ENV_ABORT();
	}
      s += CALLINFO_CALLSIZE_IDX (ci, i);
    }
  
  /* Return info */
  CALLINFO_RETTYPE (ci) = METHOD_RET_CHAR (m);
  switch (CALLINFO_RETTYPE (ci)) 
    {
    case '[':
      CALLINFO_RETTYPE (ci)  = 'L';
      /* fall through */
    case 'L':
      CALLINFO_RETSIZE (ci) = etype_slot[ETYPE_OBJECTREF];
      break;
    case 'V':
      CALLINFO_RETSIZE (ci) = etype_slot[ETYPE_VOID];
      break;
    case 'D':
    case 'J':
      CALLINFO_RETSIZE (ci) = etype_slot[ETYPE_LONG];
      break;
    default:
      CALLINFO_RETSIZE (ci) = etype_slot[ETYPE_INT];
      break;
    }
  
  /* Call info and arguments */
  CALLINFO_NRARGS (ci) = i;
  CALLINFO_ARGSIZE (ci) = s;
  CALLINFO_RETURN (ci) = ret;
  CALLINFO_FUNCTION (ci) = func;

  if (call_method_core (&callinfo))
    return -1;
  
  memset (CALLINFO_ARGS (ci), '\0', (METHOD_ARGC (m) + RESERVED_ARGS + 2) * (sizeof (struct evalue) + 2));
  memset (ci, 0, sizeof (struct emethod_callinfo));

  return 0;
}

eobjectref_t *
call_constructor_v (cname, e, type, args)
     const char *cname;
     env_t e;
     const char *type;
     va_list args;
{
  evalue_t ret = EVALUE_NULL;
  emethod_t *m;
  eobjectref_t *obj;

  if (!e)
    {
      /* Convert "." to "/" and lookup class */
      char *name = env_subst (cname, strlen (cname), "", '.', '/', "");
      e = lookup_env (name);
      env_free (name);
      if (!e)
	{
	  /* throwError (&info); */
	}
    }

  /* We cannot construct interfaces or abstract classes */
  if (ENV_IS_INTERFACE (e) || ENV_IS_ABSTRACT (e))
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "call_constructor_v: TODO");
      /* TODO: ϼؾ Ѵ.  */
    }

  m = find_method (e, new_strconst (e, "<init>", 6), new_strconst (e, type, strlen (type)));
  if (!m)
    {
      /* throw_exception (NoSuchMethodError ());  */
    }
  obj = new_objectref (e);

  if (call_method_v (m, m, obj, &ret, args))
    return NULL;

  return obj;
}

eobjectref_t *
call_constructor (const char *cname, env_t e, const char *type, ...)
{
  va_list args;
  eobjectref_t *obj;

  va_start (args, type);
  obj = call_constructor_v (cname, e, type, args);
  va_end (args);

  return obj;
}
