/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#ifndef ELANG_CONFIG_I386_LINUX_H
#define ELANG_CONFIG_I386_LINUX_H

#ifndef SYSDEP_CALL_FUNCTION
#define SYSDEP_CALL_FUNCTION(NODE)	i386_linux_call_function (NODE)
#endif

static inline void
i386_linux_call_function (struct emethod_callinfo *call)
{
  int nrargs = CALLINFO_NRARGS (call);
  
  /* Push all arguments into the stack, in last-to-first order.  This
     assumes that the second 32bit-word of longs and doubles is
     available as an additional int argument, as callMethod[AV]() do.  */
  while (nrargs > 0)
    asm volatile ("pushl %0" : : "m" (EVALUE_ADDRESS (&CALLINFO_ARGS_IDX (call, --nrargs))) : "sp");
  
  switch (CALLINFO_RETSIZE (call)) 
    {
    case 0:
      /* Must be void.  */
      ((void (*) (void)) (call->func)) ();
      break;
      
    case 1:
      if (CALLINFO_RETTYPE (call) == 'F')
	EVALUE_FLOAT (CALLINFO_RETURN (call)) = ((efloat_t (*)(void)) (call->func)) ();
      else /* Must be 32-bit or smaller int.  */
	EVALUE_INT (CALLINFO_RETURN (call)) = ((eint_t (*)(void)) (call->func)) ();
      break;
      
    default:
      /* It could've been `case 2;', but then we'd get an additional cmp
       * that we don't really need.  */
      if (CALLINFO_RETTYPE (call) == 'D')
	EVALUE_DOUBLE (CALLINFO_RETURN (call)) = ((edouble_t (*) (void)) (call->func)) ();
      else /* Must be elong_t.  */
	EVALUE_LONG (CALLINFO_RETURN (call)) = ((elong_t (*) (void)) (call->func)) ();
      break;
    }
  
  /* Adjust the stack pointer.  */
  asm volatile ("addl %0,%%esp" : :
		"r" (CALLINFO_ARGSIZE (call) * sizeof (eint_t)) : "cc", "sp");
}

#endif
