/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "http_protocol.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"
#include "util_script.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"
#include "config-info.h"

extern int econfparse (void);
static int skip_whitespace (char);
static int skip_line_comment (void);
static char get_effective_char (void);
static void handle_newline (void);
static int unescaped_terminator_p (const unsigned char *);
static void unterminated (int);
static void parse_string (struct econf_token *, char);
static etree get_identifier (unsigned char *, int);
static etree parse_identifier (void);
static etree lex_string (apr_pool_t *, const unsigned char *, unsigned int);
static int econf_hash_eq (const void *, const void *);
static void econf_hash_del (void *);

void econf_file_init (struct econf_file *, unsigned char *, size_t);
long econf_filesize (const char *);
econf_entry_t * econf_lookup (const char *);
int econf_entry_simplify (char *);
void econf_parse_number (struct econf_string *, char);
struct econf_token * econf_get_token (void);

/* empty_string  digit_vector  libetl  Ѵ.  */
extern const char empty_string[];
extern const char digit_vector[];

#define digit_string(d) (digit_vector + ((d) * 2))

/* econf  loading   ϴ  .  */
struct econf_file econf;
apr_pool_t *econf_mpool;
int econf_line;
vm_htab_t econf_htab;

/* ̰ LENGTH ̰ CONTENTS  ϴ ڿ constant  ҴϿ
   ȯմϴ.  LENGTH  -1 ̸ CONTENTS  null-terminated
   ڿ մϴ. ̴ strlen  Ͽ ˴ϴ. 
    ڿ constant   ҴǾٸ  纻   
   ȯ˴ϴ.  */

const char *
econf_alloc_string (p, contents, length)
     apr_pool_t *p;
     const char *contents;
     int length;
{
  if (length == -1)
    length = strlen (contents);

  if (length == 0)
    return empty_string;
  if (length == 1 && ISDIGIT (contents[0]))
    return digit_string (contents[0] - '0');

  return apr_pstrndup (p, contents, length);
}

#define is_vspace(x)    IS_VSPACE(x)
#define is_nvspace(x)   (IS_NVSPACE(x) && x != '\0')

/* whitespace  ǳʶٰ whitespace  ƴ  ڸ մϴ.
   Adjusts pfile->col_adjust to account for tabs.  ̰  token
    (ĭ)  ߸    Դϴ.  */
static int
skip_whitespace (c)
     char c;
{
  unsigned int warned = 0;

  do
    {
      /*  space  ׻ OK ̴.  */
      if (c == ' ')
        ;
      else if (c == '\t')
        ;
      /*  \f Ȥ \v \0  Ҵ.  */
      else if (c == '\0')
        {
          if (econf.cur - 1 == econf.base + econf.size)
            return 0;
          if (!warned)
            {
              warning ("null character(s) ignored");
              warned = 1;
            }
        }

      c = *econf.cur++;
    }
  /* 츮  non-vertical space  Ѵ.  ' ' \t \f \v \0.  */
  while (is_nvspace (c));

  econf.cur--;
  return 1;
}

static int
skip_line_comment ()
{
  char c;

  do
    {
      if (econf.cur == econf.base + econf.size)
        goto at_eof;

      c = *econf.cur++;
    }
  while (!is_vspace (c));

  /* Step back over the newline, except at EOF.  */
  while (*(econf.cur) == '\r' || *(econf.cur) == '\n')
    econf.cur++;

 at_eof:

  return 0;
}

static char
get_effective_char ()
{
  char next;

  econf.backup_to = econf.cur;
  next = *econf.cur++;

  return next;
}

/* Newline    ȣϸ, װ buffer->cur[-1]  Ѵٰ 
   Ѵ.  Newline () ٷ  ִ ڸ Ű buffer->cur 
     ȯǾ .  */
static void
handle_newline ()
{
  /* CR-LF  LF-CR  ٷ.  κ ٸ ͵ (, java)  CR-LF 
     ޾ ̸, ׷ ൿ ֵ ؾ ұ?  */
  if (econf.cur[-1] + econf.cur[0] == '\r' + '\n')
    econf.cur++;

  econf.line_base = econf.cur;
  econf_line++;
}

/*  ȣ preprocessing number  ȿ ˻Ѵ.  */
#define VALID_SIGN(c, prevc) \
  (((c) == '+' || (c) == '-') && \
   ((prevc) == 'e' || (prevc) == 'E' \
    || (((prevc) == 'p' || (prevc) == 'P'))))

#define is_numchar(x)   ISIDNUM(x)

void
econf_parse_number (struct econf_string *number, char c ATTRIBUTE_UNUSED)
{
  unsigned char *cur, *base;

  cur = econf.cur;
  while (is_numchar (*cur) || *cur == '.' || VALID_SIGN (*cur, *(cur-1)))
    cur++;

  base = econf.cur - 1;
  //fprintf (stderr, "<%.*s>", (int) (cur - base), base);
  number->text = base;
  number->len = cur - base;
  econf.cur = cur;
}

/* parse_string  ƾ.  */
static int
unescaped_terminator_p (dest)
     const unsigned char *dest;
{
  const unsigned char *start, *temp;

  start = econf.cur;

  /* An odd number of consecutive backslashes represents an escaped
     terminator.  */
  for (temp = dest; temp > start && temp[-1] == '\\'; temp--)
    ;

  return ((dest - temp) & 1) == 0;
}

/* Subroutine of parse_string.  Emits error for unterminated strings.  */
static void
unterminated (term)
     int term;
{
  error ("missing terminating %c character", term);
}

static void
parse_string (token, terminator)
     struct econf_token *token;
     char terminator;
{
  char c;
  unsigned char *cur, *dest, *limit;

  dest = econf.cur;
  cur = dest;
  limit = econf.base + econf.size;

  for (;;)
    {
      c = *econf.cur++;
      if (c == terminator)
        {
          if (unescaped_terminator_p (dest))
            break;
        }
      else if (is_vspace (c))
        {
          handle_newline ();
          c = '\n';
        }
      else if (c == '\0')
        {
          if (econf.cur - 1 == econf.base + econf.size)
            {
              unterminated (terminator);
              econf.cur--;
              break;
            }
        }
      *dest++ = c;
    }
  *dest = '\0';

  //fprintf (stderr, "parse_string: <%s>\n", cur);
  token->val.str.text = cur;
  token->val.str.len = dest - cur;

  econf.cur = dest + 1;
}

static etree
get_identifier (str, len)
     unsigned char *str;
     int len;
{
  etree s = etree_make_node (econf_mpool, ECONF_NAME_NODE);
  
  ETREE_IDENTIFIER_LENGTH (s) = len;
  ETREE_IDENTIFIER_POINTER (s) = econf_alloc_string (econf_mpool, str, len);

  return s;
}

/* Buffer κ IDENTIFIER  ؼѴ.  Identifier  ,  .  ϰ Ǵµ,
     ּҸ óϱ ؼ̴.  yylexname () Լ  ش  identifier
    address  ϰ ȴ.  */
static etree
parse_identifier ()
{
  unsigned char *cur, *base;
  etree result;

  cur = econf.cur;
  while (ISIDNUM (*cur) || *cur == '.')
    cur++;

  base = econf.cur - 1;
  result = get_identifier (base, cur - base);
  econf.cur = cur;

  //debug_etree (result);
  return result;
}

#define BACKUP() do {econf.cur = econf.backup_to;} while (0)

/* econf  ϳ token  о ش token ü ȯѴ.  */
struct econf_token *
econf_get_token (void)
{
  char c;
  struct econf_token *result;

  result = apr_pcalloc (econf_mpool, sizeof (struct econf_token));
  result->type = -1;

 fresh_line:

 skipped_white:
  c = *econf.cur++;

  switch (c)
    {
    case ' ': case '\t': case '\f': case '\v': case '\0':
      result->flags |= PREV_WHITE;
      if (skip_whitespace (c))
        goto skipped_white;

      /* EOF.  */
      result->type = ECONF_EOF;
      break;
    case '\r':
      c = get_effective_char ();
      if (c != '\n')
        {
          BACKUP ();
          goto fresh_line;
        }
    case '\n':
      c = get_effective_char ();
      BACKUP ();
      if (c == '\r' || c == '\n')
        goto fresh_line;
      handle_newline ();
      result->type = ECONF_PADDING;
      break;
    case '!':
      if (skip_line_comment ())
        warning ("multi-line comment");
      result->type = ECONF_PADDING;
      break;
    case '<':
      result->type = ECONF_ANGLE_START;
      break;
    case '/':
      result->type = ECONF_SLASH;
      break;
    case '>':
      result->type = ECONF_ANGLE_END;
      break;
    case ';':
      result->type = ECONF_STMT_SEPERATOR;
      break;
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      result->type = ECONF_NUMBER;
      econf_parse_number (&result->val.str, c);
      break;
    case '_':
    case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
    case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':
    case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
    case 's': case 't': case 'u': case 'v': case 'w': case 'x':
    case 'y': case 'z':
    case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
    case 'G': case 'H': case 'I': case 'J': case 'K': case 'L':
    case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
    case 'S': case 'T': case 'U': case 'V': case 'W': case 'X':
    case 'Y': case 'Z':
      result->type = ECONF_NAME;
      result->val.node = parse_identifier ();
      break;
    case '"':
      result->type = ECONF_STRING;
      parse_string (result, c);
      break;
    case '=':
      result->type = ECONF_EQ;
      break;
    default:
      result->type = ECONF_OTHER;
      result->val.c = c;
      break;
    }

  return result;
}

/* ECONF_STRING_CST node  ϰ ȯѴ.  */
static etree
lex_string (p, str, len)
     apr_pool_t *p;
     const unsigned char *str;
     unsigned int len;
{
  return etree_build_string (p, len, apr_pstrndup (p, str, len));
}

/* econf   .  */
extern etree lex_number (apr_pool_t *, const char *, unsigned int);

int
econf_lexer (value)
     etree *value;
{
  struct econf_token *tok;

 retry:
  do
    tok = econf_get_token ();
  while (tok->type == ECONF_PADDING);

  *value = ETREE_NULL_TREE;
  switch (tok->type)
    {
    case ECONF_STRING:
      *value = lex_string (econf_mpool, tok->val.str.text, tok->val.str.len);
      break;
    case ECONF_NUMBER:
      *value = lex_number (econf_mpool, (const char *)tok->val.str.text, tok->val.str.len);
      break;
    case ECONF_NAME:
      *value = tok->val.node;
      break;
    case ECONF_OTHER:
      if (ISGRAPH (tok->val.c))
        error ("stray '%c' in program", tok->val.c);
      else
        error ("stray '\\%o' '%d' in program", tok->val.c, tok->val.c);
      goto retry;
    default:
      break;
    }

  return tok->type;
}

/* Directory  buffer.  */
unsigned char econf_dir_buf[MAX_STRING_SIZE];
unsigned char *econf_dir;
int econf_dir_len;

/* Resolve "." and ".." elements of PATH by destructively modifying
   PATH and return non-zero if PATH has been modified, zero otherwise.

   The algorithm is in spirit similar to the one described in rfc1808,
   although implemented differently, in one pass.  To recap, path
   elements containing only "." are removed, and ".." is taken to mean
   "back up one element".  Single leading and trailing slashes are
   preserved.

   This function does not handle URL escapes explicitly.  If you're
   passing paths from URLs, make sure to unquote "%2e" and "%2E" to
   ".", so that this function can find the dots.  (Wget's URL parser
   calls reencode_escapes, which see.)

   For example, "a/b/c/./../d/.." will yield "a/b/".  More exhaustive
   test examples are provided below.  If you change anything in this
   function, run test_path_simplify to make sure you haven't broken a
   test case.  */

int
econf_entry_simplify (char *path)
{
  char *h, *t, *end;

  /* Preserve the leading '/'. */
  if (path[0] == '/')
    ++path;

  h = path;			/* hare */
  t = path;			/* tortoise */
  end = path + strlen (path);

  while (h < end)
    {
      /* Hare should be at the beginning of a path element. */

      if (h[0] == '.' && (h[1] == '/' || h[1] == '\0'))
	{
	  /* Ignore "./". */
	  h += 2;
	}
      else if (h[0] == '.' && h[1] == '.' && (h[2] == '/' || h[2] == '\0'))
	{
	  /* Handle "../" by retreating the tortoise by one path
	     element -- but not past beggining of PATH.  */
	  if (t > path)
	    {
	      /* Move backwards until T hits the beginning of the
		 previous path element or the beginning of path. */
	      for (--t; t > path && t[-1] != '/'; t--)
		;
	    }
	  h += 3;
	}
      else if (*h == '/')
	{
	  /* Ignore empty path elements.  Supporting them well is hard
	     (where do you save "http://x.com///y.html"?), and they
	     don't bring any practical gain.  Plus, they break our
	     filesystem-influenced assumptions: allowing them would
	     make "x/y//../z" simplify to "x/y/z", whereas most people
	     would expect "x/z".  */
	  ++h;
	}
      else
	{
	  /* A regular path element.  If H hasn't advanced past T,
	     simply skip to the next path element.  Otherwise, copy
	     the path element until the next slash.  */
	  if (t == h)
	    {
	      /* Skip the path element, including the slash.  */
	      while (h < end && *h != '/')
		t++, h++;
	      if (h < end)
		t++, h++;
	    }
	  else
	    {
	      /* Copy the path element, including the final slash.  */
	      while (h < end && *h != '/')
		*t++ = *h++;
	      if (h < end)
		*t++ = *h++;
	    }
	}
    }

  if (t != h)
    *t = '\0';

  return t != h;
}

void econf_add_dir (char *);

/* <DIR   , econf_dir_buf θ update Ѵ.  */
void
econf_dir_enter (str)
     etree str;
{
  snprintf (econf_dir, MAX_STRING_SIZE - econf_dir_len, "/%s", ETREE_STRING_POINTER (str));

  if (econf_entry_simplify (econf_dir_buf))
    econf_dir += strlen (econf_dir);
  else
    econf_dir += (ETREE_STRING_LENGTH (str) + 1);

  econf_dir_len = strlen (econf_dir_buf);
  econf_add_dir (econf_dir_buf);
}

/* </DIR   , econf_dir_buf θ update Ѵ.  */
void
econf_dir_exit (void)
{
  unsigned char *na = econf_dir;
  
  while (na > econf_dir_buf)
    {
      if (IS_DIR_SEPARATOR (na[0]))
        break;
      na--;
    }

  na[0] = '\0';
  econf_dir = na;
  econf_dir_len = strlen (econf_dir_buf);
}

/* Config  name  Ʈ  ߰Ѵ.    STRING Ȥ INTEGER  
    ϰ ִ.  */
void
econf_add_entry (name, value)
     etree name, value;
{
  econf_entry_t *entry, **slot;
  vm_hashval_t hash;
  char fullname[MAX_STRING_SIZE];

  if (!name || !value)
    return;

  snprintf (fullname, MAX_STRING_SIZE, "%s/%s", econf_dir_buf, ETREE_IDENTIFIER_POINTER (name));
  econf_entry_simplify (fullname);

  hash = vm_htab_hash_string (fullname);
  slot = (econf_entry_t **) vm_htab_find_slot_with_hash (econf_htab, fullname, hash, INSERT);
  entry = *slot;

  if (entry)
    return;

  entry = (econf_entry_t *) env_malloc (sizeof (struct econf_entry));
  ECONF_ENTRY_NAME (entry) = env_strdup (fullname);

  switch (ETREE_CODE (value))
    {
    case ECONF_STRING_CST:
      ECONF_ENTRY_TYPE (entry) = ECONF_TYPE_STRING;
      ECONF_ENTRY_STRING (entry) = env_strndup (ETREE_STRING_POINTER (value), ETREE_STRING_LENGTH (value));
      break;
    case ETREE_INTEGER_CST:
      ECONF_ENTRY_TYPE (entry) = ECONF_TYPE_INTEGER;
      ECONF_ENTRY_INTEGER (entry) = ((ETREE_INT_CST_HIGH (value) << 16) | ETREE_INT_CST_LOW (value));
      break;
    default:
      fprintf (stderr, "econf_add_entry: error.\n");
      break;
    }

  *slot = entry;
}

/* Config  name  Ʈ  ߰Ѵ.    STRING Ȥ INTEGER  
    ϰ ִ.  */
void
econf_add_switch (name, onoff)
     etree name;
     int onoff;
{
  econf_entry_t *entry, **slot;
  vm_hashval_t hash;
  char fullname[MAX_STRING_SIZE];

  if (!name)
    return;

  snprintf (fullname, MAX_STRING_SIZE, "%s/%s", econf_dir_buf, ETREE_IDENTIFIER_POINTER (name));
  econf_entry_simplify (fullname);

  hash = vm_htab_hash_string (fullname);
  slot = (econf_entry_t **) vm_htab_find_slot_with_hash (econf_htab, fullname, hash, INSERT);
  entry = *slot;

  if (entry)
    return;

  entry = (econf_entry_t *) env_malloc (sizeof (struct econf_entry));
  ECONF_ENTRY_NAME (entry) = env_strdup (fullname);

  ECONF_ENTRY_TYPE (entry) = ECONF_TYPE_SWITCH;
  ECONF_ENTRY_INTEGER (entry) = onoff;

  *slot = entry;
}

/* econf   entry Ȥ dir  ã´.  */
econf_entry_t *
econf_lookup (tp)
     const char *tp;
{
  econf_entry_t **slot;
  vm_hashval_t hash;

  hash = vm_htab_hash_string (tp);

  slot = (econf_entry_t **) vm_htab_find_slot_with_hash (econf_htab, tp, hash, NO_INSERT);

  if (slot)
    return *slot;

  return NULL;
}

/* ECONF Ʈ key  NAME   ȯѴ.  */
int
econf_get_integer (name)
     const char *name;
{
  econf_entry_t *e = econf_lookup (name);

  if (!e)
    return -1;

  if (ECONF_ENTRY_TYPE (e) != ECONF_TYPE_INTEGER)
    return -1;

  return ECONF_ENTRY_INTEGER (e);
}

/* ECONF Ʈ key  NAME  ڿ  ȯѴ.  */
char *
econf_get_string (name)
     const char *name;
{
  econf_entry_t *e = econf_lookup (name);

  if (!e)
    return NULL;

  if (ECONF_ENTRY_TYPE (e) != ECONF_TYPE_STRING)
    return NULL;

  return ECONF_ENTRY_STRING (e);
}

int
econf_get_switch (name)
     const char *name;
{
  econf_entry_t *e = econf_lookup (name);

  if (!e)
    return -1;

  if (ECONF_ENTRY_TYPE (e) != ECONF_TYPE_SWITCH)
    return -1;

  return ECONF_ENTRY_INTEGER (e);
}

/*  directory   Hash Table  ߰Ѵ.  */
void
econf_add_dir (name)
     char *name;
{
  econf_entry_t *entry, **slot;
  vm_hashval_t hash;

  if (!name)
    return;

  hash = vm_htab_hash_string (name);
  slot = (econf_entry_t **) vm_htab_find_slot_with_hash (econf_htab, name, hash, INSERT);
  entry = *slot;

  if (entry)
    return;

  entry = (econf_entry_t *) env_malloc (sizeof (struct econf_entry));
  ECONF_ENTRY_NAME (entry) = env_strdup (name);
  ECONF_ENTRY_TYPE (entry) = ECONF_TYPE_DIRECTORY;

  *slot = entry;
}

/*   ũ⸦ ȯѴ.  */
long
econf_filesize (const char *file)
{
  FILE *fp = fopen (file, "rb");
  long len = 0;

  if (fp != NULL)
    {
      if (fseek (fp, 0L, 2) != -1)
        len = ftell (fp);
      fclose (fp);
    }
  
  return len;
}

/* struct environment_file  ʱȭϱ  Լ̴.  */
void
econf_file_init (ef, buf, len)
     struct econf_file *ef;
     unsigned char* buf;
     size_t len;
{
  E_ASSERT (((buf == NULL) && (len == 0))
	      || ((buf != NULL) && (len != 0)));

  ef->base = ef->cur = buf;
  ef->size = len;
}

/* Hash Table  Ǵ ϼ ׽Ʈ Լν, ENTRY  
   Hash Table  ִ SLOT  ̰, X_ARG  vm_htab_find_slot_with_hash 
   ڷ  ̴.  */
static int
econf_hash_eq (entry, x_arg)
     const void *entry, *x_arg;
{
  econf_entry_t *sa;
  char *sb;
  
  sa = (econf_entry_t *) entry;
  sb = (char *) x_arg;

  if (sa && ECONF_ENTRY_NAME (sa) && sb 
      && !strcmp (ECONF_ENTRY_NAME (sa), sb))
    return 1;

  return 0;
}

/* Hash Table   Slot  ϵǾ ִ entry  delete   Ǵ Լ̴.
   vm_htab_create () Լ ȣ , Լ Ͱ  ϵȴ.  */
static void
econf_hash_del (entry)
     void *entry;
{
  econf_entry_t *e = (econf_entry_t *) entry;

  switch (ECONF_ENTRY_TYPE (e))
    {
    case ECONF_TYPE_STRING:
      env_free (ECONF_ENTRY_STRING (e));
      break;
    default:
      break;
    }

  env_free (ECONF_ENTRY_NAME (e));
  env_free (e);
}

/* eserver   eserver.conf  ؼؼ hash table  ä ִ´.  

   TODO:    ϱ ؼ ׿ ´ ̽ Ǿ  ̴.  */
int
option_init_econf (s, p)
     eserver_t *s;
     apr_pool_t *p;
{
  int i, len;
  char *buf, filepath[MAX_STRING_SIZE];
  FILE *fd;
  eoption_t *o = ESERVER_OPTION (s);

  econf_dir = econf_dir_buf;

  /* econf   ڵ  hash table  ϸ,  key  value 
     Ǿ ִ.  */
  econf_htab = OPTION_HASH_TABLE (o) =
    vm_htab_create (ECONF_HASH_INITIAL_SIZE, vm_htab_hash_string,
		    (vm_htab_eq) econf_hash_eq, 
		    (vm_htab_del) econf_hash_del);

  /* econf  ؼ    memory pool  Ҵϰ,  Լ  κп
     ش pool  Ѵ.  о econf   ؾ ϴ κ  ޸𸮸
     Ҵ ־ Ѵ.  */
  apr_pool_create (&econf_mpool, p);
  apr_pool_tag (econf_mpool, "econf mpool");

#define ESERVER_CONF_FILENAME		"eserver.conf"
  snprintf (filepath, MAX_STRING_SIZE, "%s/%s", APXS_CONFDIR, ESERVER_CONF_FILENAME);

  len = econf_filesize (filepath);
  if (!len)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, 
		     "%s   ʰų ũⰡ 0 Դϴ.", ESERVER_CONF_FILENAME);
      return 0;
    }

  buf = (char *) apr_pcalloc (econf_mpool, sizeof (char) * len + 1);
  fd  = fopen (filepath, "rb");
  if (!fd)
    return 0;

  i   = fread (buf, 1, len, fd);
  buf[i] = '\0';
  econf_file_init (&econf, buf, i);

  if (econfparse ())
    env_log_error (ENVLOG_MARK, ENVLOG_WARNING,  "warning: econfparse error.");
  
  apr_pool_destroy (econf_mpool);
  fclose (fd);
 
  return 0;
}
