/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#ifndef ESERVER_CONF_H
#define ESERVER_CONF_H

#define ECONF_HASH_INITIAL_SIZE		1000

enum econf_ttype
  {
    ECONF_PADDING,
    ECONF_NAME,
    ECONF_NUMBER,
    ECONF_STRING,
    ECONF_ANGLE_START,
    ECONF_ANGLE_END,
    ECONF_SLASH,
    ECONF_STMT_SEPERATOR,
    ECONF_EQ,
    ECONF_OTHER,
    ECONF_EOF,
    NUM_TOKEN_CODE
  };

/* Environment      Ǵ ü.  */

#define ECONF_FILE_BASE(NODE)		((NODE)->base)
#define ECONF_FILE_CUR(NODE)		((NODE)->cur)
#define ECONF_FILE_BACKUP_TO(NODE)	((NODE)->backup_to)
#define ECONF_FILE_LINE_BASE(NODE)	((NODE)->line_base)
#define ECONF_FILE_SIZE(NODE)		((NODE)->size)

struct econf_file
{
  unsigned char *base;
  unsigned char *cur;
  unsigned char *backup_to;
  unsigned char *line_base;
  size_t size;
};

/* NUMBER Ȥ STRING, CHAR, COMMENT token  (payload).  */
struct econf_string
{
  unsigned int len;
  const unsigned char *text;
};

/* econf_token ü  flag .  */
#define PREV_WHITE      (1 << 0) /*   token  whitespace .  */
#define DIGRAPH         (1 << 1) /*  digraph ̾ٸ.  */
#define STRINGIFY_ARG   (1 << 2) /*  ڿȭǾ ϴ macro argument .  */
#define PASTE_LEFT      (1 << 3) /*  ## operator  LHS ̶.  */
#define NAMED_OP        (1 << 4) /* C++ named operator .  */
#define NO_EXPAND       (1 << 5) /*  token  macro-expand  .  */
#define BOL             (1 << 6) /* () ۿ token.  */

struct econf_token
{
  enum econf_ttype type;
  unsigned char flags;          /* flag  -  . */

  union
  {
    char c;
    etree node;
    struct econf_string str;      /* ڿ Ȥ .  */
  } val;
};

enum econf_type
  {
    ECONF_TYPE_DIRECTORY,
    ECONF_TYPE_STRING,
    ECONF_TYPE_INTEGER,
    ECONF_TYPE_SWITCH
  };

#define ECONF_ENTRY_NAME(NODE)		((NODE)->name)
#define ECONF_ENTRY_TYPE(NODE)		((NODE)->type)
#define ECONF_ENTRY_STRING(NODE)	((NODE)->u.s)
#define ECONF_ENTRY_INTEGER(NODE)	((NODE)->u.i)

struct econf_entry
{
  char *name;

  enum econf_type type;

  union
  {
    char *s;
    int i;
  } u;
};

typedef struct econf_entry econf_entry_t;

extern void econf_dir_enter (etree);
extern void econf_dir_exit (void);
extern void econf_add_entry (etree, etree);
extern void econf_add_switch (etree, int);
extern int econf_get_integer (const char *);
extern char *econf_get_string (const char *);
extern int econf_get_switch (const char *);
extern const char *econf_alloc_string (apr_pool_t *, const char *, int);

extern int econf_lexer (etree *);
extern int option_init_econf (eserver_t *, apr_pool_t *);

#endif
