%{
/* Copyright 1999-2005 The Apache Software Foundation or its licensors, as
 * applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"          /* for get_remote_host */
#include "http_connection.h"
#include "http_protocol.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"
#include "util_script.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>            /* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"

/* Like YYERROR but do call yyerror.  */
#define YYERROR1 { yyerror ("syntax error"); YYERROR; }

/* Cause the "yydebug" variable to be defined.  */
#define YYDEBUG 1
%}

%union { enum etree_code code; etree ttype;
	 const char *filename; int lineno; }

%token IDENTIFIER
%token ON_T
%token OFF_T
%token DIR_T
%token CONSTANT
%token STRING

%type <ttype> IDENTIFIER CONSTANT STRING
%type <ttype> initval

%{
static int yylexname	  PARAMS ((void));
static void yyerror       PARAMS ((const char *));
static int  yylex         PARAMS ((void));
%}

%%
dir_expr:
	dir_start stmt_list dir_end
	;

stmt_list:
	stmt_list stmt
	| /* empty */
	;

stmt:
	IDENTIFIER '=' initval ';'
		{ econf_add_entry ($1, $3); }
	| IDENTIFIER '=' ON_T ';'
		{ econf_add_switch ($1, 0); }
	| IDENTIFIER '=' OFF_T ';'
		{ econf_add_switch ($1, 1); }
	| dir_expr
	;
	
dir_start:
	'<' DIR_T STRING '>'
		{ econf_dir_enter ($3); }
	;

dir_end:
	'<' '/' DIR_T '>'
		{ econf_dir_exit (); }
	;

initval:
	CONSTANT
	| STRING
	;
%%

static enum econf_ttype last_token;

/*  keyword   table.  */
struct resword
{
  const char *word;
  short yycode;
};

static const struct resword reswords[] =
{
  { "dir",	DIR_T},
  { "ON",	ON_T},
  { "OFF",	OFF_T},
};
#define N_reswords (sizeof reswords / sizeof (struct resword))

static int
yylexname (void)
{
  unsigned int i;

  for (i=0; i<N_reswords; i++)
    {
      if (!strncmp (reswords[i].word,
		    ETREE_IDENTIFIER_POINTER (yylval.ttype), 
		    ETREE_IDENTIFIER_LENGTH (yylval.ttype)))
	return reswords[i].yycode;
    }

  return IDENTIFIER;
}

static int
yylex ()
{
  last_token = econf_lexer (&yylval.ttype);
  
  switch (last_token)
    {
    case ECONF_ANGLE_START:
      return '<';
    case ECONF_SLASH:
      return '/';
    case ECONF_ANGLE_END:
      return '>';
    case ECONF_EQ:
      return '=';
    case ECONF_STMT_SEPERATOR:
      return ';';
    case ECONF_STRING:
      return STRING;
    case ECONF_NUMBER:
      return CONSTANT;
    case ECONF_NAME:
      return yylexname ();
    case ECONF_EOF:
      return 0;
    default:
      fprintf (stderr, " ó token = %d\n", last_token);
      break;
    }

  return 0;
}

static void
yyerror (msgid)
     const char *msgid;
{
  fprintf (stderr, "yyerror: error. msgid = %s", msgid);
}
