/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"

#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"

static void walk_string (gc_collector_t *, void *, void *, unsigned int);
static void walk_ref_array (gc_collector_t *, void *, void *, unsigned int);
static void walk_class (gc_collector_t *, void *, void *, unsigned int);
static void walk_object (gc_collector_t *, void *, void *, unsigned int);
static void destroy_string (gc_collector_t *, void *);
static void destroy_class (gc_collector_t *, void *);
static void destroy_lock (gc_collector_t *, void *);
static void finalize_object (gc_collector_t *, void *);

/* TODO:   Ű Ѵ.  */
static void
walk_string (gc, gc_info, base, size)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     void *gc_info ATTRIBUTE_UNUSED, *base ATTRIBUTE_UNUSED;
     unsigned int size ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "TODO: walk_string");
}

/* TODO:   Ű Ѵ.  */
static void
walk_ref_array (gc, gc_info, base, size)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     void *gc_info ATTRIBUTE_UNUSED, *base ATTRIBUTE_UNUSED;
     unsigned int size ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "TODO: walk_ref_array");
}

/* TODO:   Ű Ѵ.  */
static void
walk_class (gc, gc_info, base, size)
     gc_collector_t *gc;
     void *gc_info, *base;
     unsigned int size ATTRIBUTE_UNUSED;
{
  int i;
  env_t e = (env_t) base;

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "walk_class: `%s' state = %d", STRCONST_PTR (ENV_NAME (e)), ENV_STATE (e));

  /* !!! Ȳ!!!
     
     = ϰ óǰ ִ 
     
       (2005 08 20)
       alloc_env () Լ ҴǴ egc_malloc ó.   env_t  ٸ
       env_t   Ǵ κ ENV_SUPER ()  ENV_INTERFACE_IDX ()
        ϱ   κп  marking  ̷ ڵ带 Ͽ.
  */

  if (ENV_TYPE (e) == ENV_REO)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "TODO: walk_class reo");
      return;
    }
 
  if (ENV_SUPER (e))
    EGC_mark_object (gc, gc_info, ENV_SUPER (e));
  
  for (i=0; i<ENV_TOTAL_INTERFACE_NUM (e); i++)
    EGC_mark_object (gc, gc_info, ENV_INTERFACE_IDX (e, i));

  /* mark interfaces referenced by this class */
  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "TODO: walk_class ϼ  ȵ?");
}

/* TODO:   Ű Ѵ.  */
static void
walk_object (gc, gc_info, base, size)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     void *gc_info ATTRIBUTE_UNUSED, *base ATTRIBUTE_UNUSED;
     unsigned int size ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "TODO: walk_object");
}

/* TODO:   Ű Ѵ.  */
static void
destroy_string (gc, c)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     void *c ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "TODO: destroy_string");
}

/* TODO:   Ű Ѵ.  */
static void
destroy_class (gc, c)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     void *c;
{
  env_t e = (env_t) c;

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "destroy_class: `%s' state = %d", STRCONST_PTR (ENV_NAME (e)), ENV_STATE (e));

  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "TODO: destroy_class (0x%x)", c);
}

/* TODO:   Ű Ѵ.  */
static void
destroy_lock (gc, c)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     void *c ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "TODO: destroy_lock");
}

/* TODO:   Ű Ѵ.  */
static void
finalize_object (gc, ob)
     gc_collector_t *gc ATTRIBUTE_UNUSED;
     void *ob ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "TODO: finalize_object");
}

gc_collector_t *
initialize_gc (apr_pool_t *p)
{
  /* TODO:  boehm GC   ǵ Ǿ ִµ,  κп 
      ̷ Ѵ.  */
  gc_collector_t *gc = initialize_boehm_gc (p);
  struct gc_interface_ops *ops = EGC_COLLECTOR_OPERATION (gc);

  env_log_error (ENVLOG_MARK, ENVLOG_INFO, "GC ʱȭ ");
  
  ops->register_gc_type_by_index (gc, EGC_ALLOC_ELANG_STRING, walk_string, EGC_OBJECT_NORMAL, destroy_string, "env.core.string");
  ops->register_gc_type_by_index (gc, EGC_ALLOC_NOWALK, NULL, EGC_OBJECT_NORMAL, NULL, "other-nowalk");
  ops->register_gc_type_by_index (gc, EGC_ALLOC_NORMALOBJECT, walk_object, EGC_OBJECT_NORMAL, NULL, "obj-no-final");
  ops->register_gc_type_by_index (gc, EGC_ALLOC_PRIMARRAY, NULL, EGC_OBJECT_NORMAL, NULL, "prim-arrays");
  ops->register_gc_type_by_index (gc, EGC_ALLOC_REFARRAY, walk_ref_array, EGC_OBJECT_NORMAL, NULL, "ref-arrays");
  ops->register_gc_type_by_index (gc, EGC_ALLOC_CLASS, walk_class, EGC_OBJECT_NORMAL, destroy_class, "env.core.Class");
  ops->register_gc_type_by_index (gc, EGC_ALLOC_FINALIZEOBJECT, walk_object, finalize_object, NULL, "obj-final");
  ops->register_gc_type_by_index (gc, EGC_ALLOC_THREADCTX, NULL, EGC_OBJECT_NORMAL, NULL, "thread-ctxts");
  ops->register_gc_type_by_index (gc, EGC_ALLOC_INTERFACE, NULL, EGC_OBJECT_NORMAL, NULL, "interfaces");
  ops->register_gc_type_by_index (gc, EGC_ALLOC_INTERFACE_TABLE, NULL, EGC_OBJECT_NORMAL, NULL, "interface table");
  ops->register_gc_type_by_index (gc, EGC_ALLOC_BYTECODE, NULL, EGC_OBJECT_NORMAL, NULL, "java-bytecode");
  ops->register_gc_type_by_index (gc, EGC_ALLOC_LOCK, NULL, EGC_OBJECT_NORMAL, destroy_lock, "locks");
  
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_STATIC_THREADDATA, "thread-data");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_EXCEPTIONTABLE, "exc-table");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_STATICDATA, "static-data");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_CONSTANT, "constants");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_FIXED, "other-fixed");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_DISPATCHTABLE, "dtable");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_METHOD, "methods");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_FIELD, "fields");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_UTF8CONST, "utf8consts");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_REF, "gc-refs");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_CODEANALYSE, "code-analyse");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_CLASSPOOL, "class-pool");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_LINENRTABLE, "linenr-table");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_LOCALVARTABLE, "lvar-table");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_DECLAREDEXC, "declared-exc");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_CLASSMISC, "class-misc");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_VERIFIER, "verifier");
  ops->register_fixed_type_by_index (gc, EGC_ALLOC_NATIVELIB, "native-lib");

  env_log_error (ENVLOG_MARK, ENVLOG_INFO, "GC ʱȭ ");

  return gc;
}
