/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "http_protocol.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"
#include "util_script.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"
#include "env2etx.h"

static char * e2e_print_thisname (apr_pool_t *, env_t);
static char * e2e_print_supername (apr_pool_t *, env_t);
static char * e2e_print_envinfo (apr_pool_t *, env_t);
static char * e2e_print_interfaces (apr_pool_t *, env_t);
static char * e2e_print_fields (apr_pool_t *, env_t);
static char * e2e_print_methods (apr_pool_t *, env_t);

static char *
e2e_print_thisname (pool, e)
     apr_pool_t *pool;
     env_t e;
{
  return apr_pstrcat (pool, "\"", STRCONST_PTR (ENV_NAME (e)), "\" ", NULL);
}

static char *
e2e_print_supername (pool, e)
     apr_pool_t *pool;
     env_t e ATTRIBUTE_UNUSED;
{
  /* TODO:  super   ó ʱ  Ѵ.  Super   óǵ
     Ǿ  ̴.  */
  if (ENV_SUPER (e))
    return apr_pstrcat (pool, "\"", STRCONST_PTR (ENV_NAME (ENV_SUPER (e))), "\" ", NULL);
  else
    return apr_pstrcat (pool, "\"\" ", NULL);
}

static char *
e2e_print_envinfo (pool, e)
     apr_pool_t *pool;
     env_t e;
{
  const char *writer = "";
  const char *license = "";

  if (ENV_WRITER (e))
    writer = apr_pstrcat (pool, "(writer \"", ENV_WRITER (e), "\") ", NULL);

  if (ENV_LICENSE (e))
    license = apr_pstrcat (pool, "(license \"", ENV_LICENSE (e), "\") ", NULL);

  return apr_pstrcat (pool, "[", writer, license, "] ", NULL);
}

static char *
e2e_print_interfaces (pool, e)
     apr_pool_t *pool;
     env_t e ATTRIBUTE_UNUSED;
{
  return apr_pstrcat (pool, "[] ", NULL);
}

static char *
e2e_print_fields (pool, e)
     apr_pool_t *pool;
     env_t e;
{
  int i;
  const char *fb = "";
  efield_t *fld = ENV_FIELDS (e);

  for (i=0; i<ENV_FIELDS_COUNT (e); fld++, i++)
    {
      fb = apr_pstrcat (pool, fb,
			"(field ", 
			apr_itoa (pool, FIELD_ACCESS_FLAGS (fld)), " ",
			"\"", STRCONST_PTR (FIELD_NAME (fld)), "\" ",
			"\"", ETYPE_SIGNITURE (FIELD_TYPE (fld)), "\"", ") ", NULL);
    }

  return apr_pstrcat (pool, "[", fb, "] ", NULL);
}

static char *
e2e_print_methods (pool, e)
     apr_pool_t *pool;
     env_t e;
{
  int i;
  const char *mb = "";
  emethod_t *meth = ENV_METHODS (e);

  i = ENV_METHODS_COUNT (e);
  for (; --i >= 0; meth++)
    {
      mb = apr_pstrcat (pool, mb,
			"(method ", 
			apr_itoa (pool, METHOD_ACCESS_FLAGS (meth)), " ",
			"\"", STRCONST_PTR (METHOD_NAME (meth)), "\" ",
			"\"", STRCONST_PTR (MTYPE_NAME (METHOD_TYPE (meth))), "\" ", 
			apr_itoa (pool, METHOD_STACK_SIZE (meth)), " ",
			apr_itoa (pool, METHOD_LOCAL_SIZE (meth)), " ",
			apr_itoa (pool, METHOD_BCODE_LENGTH (meth)), " ",
			") ", NULL);
    }

  return apr_pstrcat (pool, "[", mb, "] ", NULL);
}

/* Environment  ETX  ȯ string buffer  ȯѴ.  */
const char *
env_to_etx (apr_pool_t *pool, env_t e)
{
  char keystr[ENVKEY_STRSIZE];

  if (!pool || !e)
    return "";

  GET_ENVKEY_STRING (ENV_KEY (e), keystr);

  return apr_pstrcat (pool, "(class ", 
		       "\"", keystr, "\" ",
		      apr_itoa (pool, ENV_ACCESS_FLAGS (e)), " ",
		      e2e_print_thisname (pool, e), 
		      e2e_print_supername (pool, e),
		      e2e_print_envinfo (pool, e),
		      e2e_print_interfaces (pool, e),
		      e2e_print_fields (pool, e),
		      e2e_print_methods (pool, e), ")", NULL);
}
