/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"

#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"

EREF_UNIQUE_IDX_TYPE make_randidx_toref (void);
eobjectref_t * lookup_objectref (EREF_UNIQUE_IDX_TYPE);
earrayref_t * lookup_arrayref (EREF_UNIQUE_IDX_TYPE);

/**
   ObjectRef  EOBJECTREF TREE  ã´.
 */
eobjectref_t *
lookup_objectref (uidx)
     EREF_UNIQUE_IDX_TYPE uidx;
{
  vm_splay_tree_node_t n;

  n = vm_splay_tree_lookup (VM_EOBJECTREF_TREE (ESERVER_VM (eserver)), (vm_splay_tree_key) uidx);
  if (n)
    return (eobjectref_t *) n->value;

  return NULL;
}

/**
   ArrayRef  EARRAYREF TREE  ã´.
 */
earrayref_t *
lookup_arrayref (uidx)
     EREF_UNIQUE_IDX_TYPE uidx;
{
  vm_splay_tree_node_t n;

  n = vm_splay_tree_lookup (VM_EARRAYREF_TREE (ESERVER_VM (eserver)), (vm_splay_tree_key) uidx);
  if (n)
    return (earrayref_t *) n->value;

  return NULL;
}

/**
   ObjectRef Ȥ ArrayRef   unique index  ϴ.
*/
EREF_UNIQUE_IDX_TYPE
make_randidx_toref (void)
{
  return rand ();
}

/**
   eclass   objectref  մϴ.

   TODO: ObjectRef  ,  ޸ Ҵ memory pool   ̷ ֱ 
   ε garbage collection  ߻ ʴ´.  ̸  ؾѴ.

     ϰ ִ mutex  쵵, memory pool  Ͽ , ̿
      ̷Ѵ.
*/
eobjectref_t *
new_objectref (e)
     env_t e;
{
  apr_status_t rv;
  eobjectref_t *obj;
  EREF_UNIQUE_IDX_TYPE uidx;

#if 0
  /* ߺǴ unique index  ϴ 캻 ,   Ѵ.  */
  while (1)
    {
      uidx = make_randidx_toref ();
      if (!lookup_objectref (uidx))
	break;
    }
#endif

  obj = (eobjectref_t *) apr_pcalloc (ENV_MPOOL (e), 
				      sizeof (struct eobjectref) + ENV_FIELDS_SIZE_IN_BYTE (e));
  if (!obj)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "new_objectref: out of memory.");
      return NULL;
    }

  /* Fill in object information */
  EOBJECTREF_MAGIC (obj) = 0xacacacac;
  EOBJECTREF_UIDX (obj) = uidx;
  EOBJECTREF_DTABLE (obj) = ENV_DISPATCH_TABLE (e);

  rv = apr_thread_mutex_create (&EOBJECTREF_MUTEX (obj), APR_THREAD_MUTEX_NESTED, ENV_MPOOL (e));
  if (rv != APR_SUCCESS)
    {
      env_log_error (ENVLOG_MARK, ENVLOG_ERR, "new_objectref: apr_thread_mutex_create error.");
      return NULL;
    }

#if 0
  vm_splay_tree_insert (VM_EOBJECTREF_TREE (ESERVER_VM (eserver)), (vm_splay_tree_key) uidx,
			(vm_splay_tree_value) obj);
#endif

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		"new_objectref: 0x%x (ũ %d) class %s", obj, ENV_FIELDS_SIZE_IN_BYTE (e),
		(e ? STRCONST_PTR (ENV_NAME (e)) : "<none>"));
  
  return obj;
}

/**
   Array Object  Ͽ ȯѴ. arrayref  Ѵ.
   Allocate a new array, of whatever types.  
*/
earrayref_t *
new_arrayref (type, count)
     etype_t *type;
     size_t count;
{
  EREF_UNIQUE_IDX_TYPE uidx;
  earrayref_t *array = NULL;
  size_t total_count;

#if 0  
  /* ߺǴ unique index  ϴ 캻 ,   Ѵ.  */
  while (1)
    {
      uidx = make_randidx_toref ();
      if (!lookup_arrayref (uidx))
	break;
    }
#endif

  total_count = (ETYPE_SIZE (type) * count) + EARRAYREF_DATA_OFFSET;

  if (total_count > count)
    {
      array = env_calloc (1, total_count);
      EARRAYREF_MAGIC (array) = 0xbdbdbdbd;
      EARRAYREF_TYPE (array) = type;
      EARRAYREF_SIZE (array) = count;
      EARRAYREF_UIDX (array) = uidx;

#if 0
      vm_splay_tree_insert (VM_EARRAYREF_TREE (ESERVER_VM (eserver)), (vm_splay_tree_key) uidx,
			    (vm_splay_tree_value) array);
#endif
    }

  env_log_error (ENVLOG_MARK, ENVLOG_INFO,
		 "new_arrayref: 0x%x, type \"%s\", count %d, total byte %d", 
		 array, etype_name[ETYPE_KIND (type)], count, total_count);
  
  return array;
}

/**
   new_anewarray

   TODO:  ϼ ʾҴ!!!!  ڵ Ȯ 𸣰ڱ.  
*/
void *
new_anewarray (e, size)
     env_t e ATTRIBUTE_UNUSED;
     int size ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "new_anewarray: TODO");

  return NULL;
}

