/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#ifndef ESERVER_H
#define ESERVER_H

#define E_MAGIC_TYPE		"application/x-httpd-env"

/* Environment methods */
enum 
  {
    /* ----[ Instruction Part ]----  */
    E_M_GETFIELD = 0,
    E_M_PUTFIELD,
    E_M_INVOKEVIRTUAL,
    E_M_INVOKESPECIAL,
    E_M_INVOKESTATIC,
    E_M_INVOKEINTERFACE,
    /* ----[ Operation Part ]---- */
    /* uri   environment file  path  Ȯϰ ش environment 
       LOAD Ѵ.  */
    E_M_LOAD,
    /* uri   environment  Ȯ , load  Ǿ  , load
        , ش environment  ETX  ȯѴ.  */
    E_M_INFO,
    E_M_LAST
  };

/* Լ  Ȥ .  */
#define SUCCESS		0
#define FAILURE		-1	/* this MUST stay a negative number, or it may affect functions! */

/* Environment ɼ.  */
#define ENV_ON		1
#define ENV_OFF		0

#ifndef PATH_MAX
#define PATH_MAX 1024
#endif

#ifndef MAXPATHLEN
# ifdef PATH_MAX
#  define MAXPATHLEN PATH_MAX
# else
#  define MAXPATHLEN 256    /* Should be safe for any weird systems that do not define it */
# endif
#endif

#define MAX_STRING_SIZE		(1024 * 8)

#ifdef ENV_WIN32
#include "readdir.h"
#include <sys/utime.h>
/* mode_t isn't defined on Windows */
typedef unsigned short mode_t;

#define DEFAULT_SLASH '\\'
#define DEFAULT_DIR_SEPARATOR	';'
#define IS_SLASH(c)	((c) == '/' || (c) == '\\')
#define IS_SLASH_P(c)	(*(c) == '/' || \
        (*(c) == '\\' && !IsDBCSLeadByte(*(c-1))))

/* COPY_WHEN_ABSOLUTE also takes path as argument because netware needs it
 * to account for volume name that is unique to NetWare absolute paths
 */
#define COPY_WHEN_ABSOLUTE(path) 2
#define IS_UNC_PATH(path, len) \
	(len >= 2 && IS_SLASH(path[0]) && IS_SLASH(path[1]))
#define IS_ABSOLUTE_PATH(path, len) \
	(len >= 2 && ((isalpha(path[0]) && path[1] == ':') || IS_UNC_PATH(path, len)))

#elif defined(NETWARE)
#ifdef HAVE_DIRENT_H
#include <dirent.h>
#endif

#define DEFAULT_SLASH '/'
#define DEFAULT_DIR_SEPARATOR	';'
#define IS_SLASH(c)	((c) == '/' || (c) == '\\')
#define COPY_WHEN_ABSOLUTE(path) \
    (strchr(path, ':') - path + 1)  /* Take the volume name which ends with a colon */
#define IS_ABSOLUTE_PATH(path, len) \
    (strchr(path, ':') != NULL) /* Colon indicates volume name */

#else
#ifdef HAVE_DIRENT_H
#include <dirent.h>
#endif

#define DEFAULT_SLASH '/'

#ifdef __riscos__
#define DEFAULT_DIR_SEPARATOR  ';'
#else
#define DEFAULT_DIR_SEPARATOR  ':'
#endif

#define IS_SLASH(c)	((c) == '/')
#define IS_SLASH_P(c)	(*(c) == '/')

#endif

/*
  Environment Option.
  Environment Module   
*/

#define OPTION_HASH_TABLE(NODE)			((NODE)->eoption_htab)
#define OPTION_HOSTNAME(NODE)			((NODE)->hostname)
#define OPTION_PORT(NODE)			((NODE)->port)
#define OPTION_ERE_ROOT(NODE)			((NODE)->ere_root)
#define OPTION_ERROR_LOG(NODE)			((NODE)->error_log)
#define OPTION_LOG_NAME(NODE)			((NODE)->logname)
#define OPTION_LOG_LEVEL(NODE)			((NODE)->loglevel)
#define OPTION_UPLOAD_TMP_DIR(NODE)		((NODE)->upload_tmp_dir)
#define OPTION_UPLOAD_MAX_FILESIZE(NODE)	((NODE)->upload_max_filesize)
#define OPTION_FILE_UPLOADS(NODE)		((NODE)->file_uploads)
#define OPTION_LIBPATH(NODE)			((NODE)->libpath)

struct eoption
{
  /* econf  Ͽ ɼ  丮   Ǵ hash table.  */
  vm_htab_t eoption_htab;

  const char *hostname;
  apr_port_t port;

  /* Environment Runtime File  ִ root .  */
  const char *ere_root;

  /* Error Log  description.  */
  apr_file_t *error_log;
  /* Error Log  ̸.  */
  const char *logname;
  /* Error Log  .  */
  int loglevel;

  /* Upload   ӽ  .  */
  char *upload_tmp_dir;
  /* Upload   ִ ũ.  */
  long upload_max_filesize;

  char *libpath;

  /* POST  ؼ  ε尡 ϵ Ϸ  1 ̾ Ѵ.  */
  unsigned file_uploads : 1;	/* True or False.  */
  unsigned tmp_unused : 31;
};

typedef struct eoption eoption_t;

/* Environment Virtual Machine.  */
#define ESERVER_MEMORY_POOL(NODE)	((NODE)->ep)
#define ESERVER_METHOD(NODE)		((NODE)->methods)
#define ESERVER_METHOD_ARRAY(NODE, IDX)	((NODE)->methods[(IDX)])
#define ESERVER_VM(NODE)		((NODE)->vm)
#define ESERVER_OPTION(NODE)		((NODE)->options)

struct eserver_info
{
  /* Eserver  õ ޸ pool.  Main memory pool  ӵȴ.  ޸ pool
       ڼ  Ѵٸ Ʒ . 

     Eserver α׷  Apache  Ǿ Ǳ , ޸  
     ġ κ-ڽ  ޸ pool  Ѵ.  ̰ ü 帧 ´
     ̶ ߴ.

     Eserver α׷ Apache   κν load  Ǹ, Apache main 
     memory pool   ϳ pool  Ѵ.  װ Ʒ Ǿ ִ EP
     ̴.  ̰  õ   Ÿ  ִ.  ش 
     Ʒ .

       1. Eserver ü.
       2. VM ü

      request   Ǵ ü request_rec *r  default pool  ϵ
     Ѵ.  */
  apr_pool_t *ep;

  /* E-Lang  ϴ method .  mod_e.c  e_init_method () Լ ʱȭ
     ̷.  */
  int methods[E_M_LAST];

  /* Eserver   κ Virtual Machine Part. 
     VM ü.  Thread-safe ϰ Ǿ Ѵ.  */
  vm_t *vm;

  /* Eserver Option    ü .  */
  eoption_t *options;
};

typedef struct eserver_info eserver_t;

extern eserver_t *eserver;

extern void e_handle_aborted_connection (void);

#endif
