/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#ifndef ENV_EURI_H
#define ENV_EURI_H

/*************************************************************************
 * Environment URI.
 *************************************************************************/

/* e://localhost/lang/env/String.e("n7J2znmJv5pCocpie/1eATzhXBY=")->method (I)V */
#define APR_URI_ENV_DEFAULT_PORT	8889

typedef struct euri euri_t;

#define EURI_SCHEME(NODE)		((NODE)->scheme)
#define EURI_HOSTINFO(NODE)		((NODE)->hostinfo)
#define EURI_USER(NODE)			((NODE)->user)
#define EURI_PASSWORD(NODE)		((NODE)->password)
#define EURI_HOSTNAME(NODE)		((NODE)->hostname)
#define EURI_PORT_STRING(NODE)		((NODE)->port_str)
#define EURI_PATH(NODE)			((NODE)->path)
#define EURI_KEY(NODE)			((NODE)->key)
#define EURI_QUERY(NODE)		((NODE)->query)
#define EURI_FRAGMENT(NODE)		((NODE)->fragment)
#define EURI_HOSTENT(NODE)		((NODE)->hostent)
#define EURI_PORT(NODE)			((NODE)->port)
#define EURI_IS_INITIALIZED(NODE)	((NODE)->is_initialized)
#define EURI_DNS_LOOKED_UP(NODE)	((NODE)->dns_looked_up)
#define EURI_DNS_RESOLVED(NODE)		((NODE)->dns_resolved)

/* A structure to encompass all of the fields in a uri.  */
struct euri
{
  /** scheme ("http"/"ftp"/...) */
  char *scheme;
  /** combined [user[:password]\@]host[:port] */
  char *hostinfo;
  /** user name, as in http://user:passwd\@host:port/ */
  char *user;
  /** password, as in http://user:passwd\@host:port/ */
  char *password;
  /** hostname from URI (or from Host: header) */
  char *hostname;
  /** port string (integer representation is in "port") */
  char *port_str;
  /** the request path (or "/" if only scheme://host was given) */
  char *path;
  envkey_t *key;
  /** Everything after a '?' in the path, if present */
  char *query;
  /** Trailing "#fragment" string, if present */
  char *fragment;
  
  /** structure returned from gethostbyname() */
  struct hostent *hostent;
  
  /** The port number, numeric, valid only if port_str != NULL */
  unsigned short port;
  
  /** has the structure been initialized */
  unsigned is_initialized:1;
  
  /** has the DNS been looked up yet */
  unsigned dns_looked_up:1;
  /** has the dns been resolved yet */
  unsigned dns_resolved:1;
};

extern int euri_parse (apr_pool_t *, const char *, euri_t *);

#endif

