/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#ifndef ZEND_LLIST_H
#define ZEND_LLIST_H

#include <stdlib.h>

typedef struct _zend_llist_element 
{
  struct _zend_llist_element *next;
  struct _zend_llist_element *prev;
  char data[1]; /* Needs to always be last in the struct */
} zend_llist_element;

typedef void (*llist_dtor_func_t) (void *);
typedef int (*llist_compare_func_t) (const zend_llist_element **, const zend_llist_element **);
typedef void (*llist_apply_with_args_func_t) (void *data, int num_args, va_list args);
typedef void (*llist_apply_with_arg_func_t) (void *data, void *arg);
typedef void (*llist_apply_func_t) (void *);

typedef struct _zend_llist 
{
  zend_llist_element *head;
  zend_llist_element *tail;
  size_t size;
  size_t count;
  llist_dtor_func_t dtor;
  zend_llist_element *traverse_ptr;
} zend_llist;

typedef zend_llist_element* zend_llist_position;

extern void zend_llist_init (zend_llist *l, size_t size, llist_dtor_func_t dtor);
extern void zend_llist_add_element (zend_llist *l, void *element);
extern void zend_llist_prepend_element (zend_llist *l, void *element);
extern void zend_llist_del_element (zend_llist *l, void *element, int (*compare) (void *element1, void *element2));
extern void zend_llist_destroy (zend_llist *l);
extern void zend_llist_clean (zend_llist *l);
extern void *zend_llist_remove_tail (zend_llist *l);
extern void zend_llist_copy (zend_llist *dst, zend_llist *src);
extern void zend_llist_apply (zend_llist *l, llist_apply_func_t func);
extern void zend_llist_apply_with_del (zend_llist *l, int (*func)(void *data));
extern void zend_llist_apply_with_argument (zend_llist *l, llist_apply_with_arg_func_t func, void *arg);
extern void zend_llist_apply_with_arguments (zend_llist *l, llist_apply_with_args_func_t func, int num_args, ...);
extern int zend_llist_count (zend_llist *l);
extern void zend_llist_sort (zend_llist *l, llist_compare_func_t comp_func);

/* traversal */
extern void *zend_llist_get_first_ex (zend_llist *l, zend_llist_position *pos);
extern void *zend_llist_get_last_ex (zend_llist *l, zend_llist_position *pos);
extern void *zend_llist_get_next_ex (zend_llist *l, zend_llist_position *pos);
extern void *zend_llist_get_prev_ex (zend_llist *l, zend_llist_position *pos);

#define zend_llist_get_first(l) zend_llist_get_first_ex(l, NULL)
#define zend_llist_get_last(l) zend_llist_get_last_ex(l, NULL)
#define zend_llist_get_next(l) zend_llist_get_next_ex(l, NULL)
#define zend_llist_get_prev(l) zend_llist_get_prev_ex(l, NULL)

#endif
