/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "http_protocol.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"
#include "util_script.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero() */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor() */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"
#include "util_time.h"

typedef struct {
  const char *t_name;
  int t_val;
} TRANS;

static const TRANS priorities[] = {
    {"emerg",   ENVLOG_EMERG},
    {"alert",   ENVLOG_ALERT},
    {"crit",    ENVLOG_CRIT},
    {"error",   ENVLOG_ERR},
    {"warn",    ENVLOG_WARNING},
    {"notice",  ENVLOG_NOTICE},
    {"info",    ENVLOG_INFO},
    {"debug",   ENVLOG_DEBUG},
    {NULL,      -1},
};

void 
env_log_error_core (const char *file, int line, int level,
		    const char *fmt, va_list args)
{
  apr_size_t len;
  apr_file_t *logf = NULL;
  char errstr[MAX_STRING_LEN];
  eoption_t *o = ESERVER_OPTION (eserver);
  int level_and_mask = level & APLOG_LEVELMASK;

  logf = OPTION_ERROR_LOG (o);

  errstr[0] = '[';
  ap_recent_ctime (errstr + 1, apr_time_now ());
  errstr[1 + APR_CTIME_LEN - 1] = ']';
  errstr[1 + APR_CTIME_LEN    ] = ' ';
  len = 1 + APR_CTIME_LEN + 1;

  len += apr_snprintf (errstr + len, MAX_STRING_LEN - len,
		       "[%s] ", priorities[level_and_mask].t_name);

  len += apr_snprintf (errstr + len, MAX_STRING_LEN - len, "<%s:%d> %d -> ", file, line, level);
  len += apr_vsnprintf (errstr + len, MAX_STRING_LEN - len, fmt, args);
  len += apr_snprintf (errstr + len, MAX_STRING_LEN - len, "\n");

  apr_file_puts (errstr, logf);
  apr_file_flush (logf);
}

/* LOG Ͽ  ޼ Ѵ.  */
void
env_log_error (const char *file, int line, int level, const char *fmt, ...)
{
  va_list args;

  va_start (args, fmt);
  env_log_error_core (file, line, level, fmt, args);
  va_end (args);
}
