/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "http_protocol.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"
#include "util_script.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"

/**
    ü  Ǿ ִ Eserver      ִ.   ʱȭ
   ̷  type  ϰ    ʴ´.
*/
eserver_t *eserver;

apr_pool_t * get_eserver_pool (void);

static void option_init_common (eserver_t *, server_rec *, apr_pool_t *);
static apr_status_t  e_server_cleanup (void *);
static void pre_option_init (eserver_t *, server_rec *, apr_pool_t *, eoption_t *);
static void post_option_init (eserver_t *, server_rec *, apr_pool_t *, eoption_t *);
static int e_response_process (erequest_t *);

apr_pool_t *
get_eserver_pool (void)
{
  return ESERVER_MEMORY_POOL (eserver);
}

/**
   ER  ׿ ִ eresponse_info_t ü  û ڿ Ѹ.  ȯϴ
    HTTP Status Code Number ̴.
*/
static int
e_response_process (er)
     erequest_t *er;
{
  eresponse_info_t *r = RESPONSE_INFO (er);

  /*  Ѹ.  */

  return RES_STATUS (r);
}

/**
   Request    , ȣȴ.  
*/
int
erequest_shutdown (er, r)
     erequest_t *er;
     request_rec *r ATTRIBUTE_UNUSED;
{
  /* Request     eresponse_info_t ü ׿ ִ 
     Ѹ.  */
  int rescode = e_response_process (er);
  erequest_info_t *i = REQUEST_INFO (er);

  /* ==   ==  */
  if (RCONTENT_ARGS (i))
    vm_htab_delete (RCONTENT_ARGS (i));
  if (RCONTENT_RFC1867_UPLOADED_FILES (i))
    vm_htab_delete (RCONTENT_RFC1867_UPLOADED_FILES (i));

  /*  thread  Ҵ  ȸ resource  Ѵ.  */
  tsrm_destory ();

  return rescode;
}

/**
   Connection  ߰  , óؾ ϴ ϵ Ѵ.
*/
void
e_handle_aborted_connection (void)
{
  return;
}

/**
   eserver.conf  б  ⺻   ִ ɼ  մϴ.
*/
static void
pre_option_init (s, server, p, o)
     eserver_t *s ATTRIBUTE_UNUSED;
     server_rec *server;
     apr_pool_t *p;
     eoption_t *o;
{
  core_server_config *sconf = ap_get_module_config (server->module_config, &core_module);
  
  /* ⺻ ERE .  */
  OPTION_ERE_ROOT (o) = (char *) apr_pstrdup (p, sconf->ap_document_root);

  /* Hostname  Port  */
  OPTION_HOSTNAME (o) = apr_pstrdup (p, server->server_hostname);
  OPTION_PORT (o) = server->port;

  OPTION_UPLOAD_TMP_DIR (o) = (char *) "/tmp";
  /* Upload ִ ũ.  ⺻ 8 ް Ʈ.  */
  OPTION_UPLOAD_MAX_FILESIZE (o) = 8 * 1024 * 1024;
  OPTION_FILE_UPLOADS (o) = ENV_ON;

  OPTION_LOG_NAME (o) = ap_server_root_relative (p, (char *) "logs/elang_log");
  OPTION_LOG_LEVEL (o) = DEFAULT_ENVLOGLEVEL;

  /* TREE   Ǵ node  .  Ʒ option_init_econf
      etree node  ϱ    Ϸؾ Ѵ.  */
  build_common_etree_nodes (p);
}

/**
    Լ ȣ  eserver.conf  ؼ   Դϴ.  
   conf Ͽ   о  Ž Ѵ.
*/
static void
post_option_init (s, server, p, o)
     eserver_t *s ATTRIBUTE_UNUSED;
     server_rec *server;
     apr_pool_t *p;
     eoption_t *o;
{
  int i;
  char *str;
  core_server_config *sconf = ap_get_module_config (server->module_config, &core_module);

  if ((str = econf_get_string ("/eserver/hostname")))
    OPTION_HOSTNAME (o) = str;

  if ((i = econf_get_integer ("/eserver/port")) != -1)
    OPTION_PORT (o) = i;

  /* ⺻ ERE .  */
  if ((str = econf_get_string ("/eRE/rootpath")))
    OPTION_ERE_ROOT (o) = apr_pstrcat (p, sconf->ap_document_root, str, NULL);

  if ((str = econf_get_string ("/eserver/upload_tmpdir")))
    OPTION_UPLOAD_TMP_DIR (o) = str;

  /* Upload ִ ũ.  ⺻ 8 ް Ʈ.  */
  if ((i = econf_get_integer ("/eserver/upload_max_filesize")) != -1)
    OPTION_UPLOAD_MAX_FILESIZE (o) = i;
  if ((i = econf_get_switch ("/eserver/upload_enable")))
    OPTION_FILE_UPLOADS (o) = i;

  if ((str = econf_get_string ("/eserver/error_log_name")))
    OPTION_LOG_NAME (o) = ap_server_root_relative (p, str);
  if ((i = econf_get_integer ("/eserver/error_log_level")) != -1)
    OPTION_LOG_LEVEL (o) = i;

  if (apr_file_open (&OPTION_ERROR_LOG (o), OPTION_LOG_NAME (o),
		     APR_APPEND|APR_WRITE|APR_CREATE, APR_OS_DEFAULT,
		     p) != APR_SUCCESS)
    return;

  if ((str = econf_get_string ("/eserver/libpath")))
    OPTION_LIBPATH (o) = ap_server_root_relative (p, str);
}

/**
   Environment  ϴ option  մϴ.
*/
static void
option_init_common (s, server, p)
     eserver_t *s;
     server_rec *server;
     apr_pool_t *p;
{
  eoption_t *o;

  ESERVER_OPTION (s) = o = (eoption_t *) apr_pcalloc (p, sizeof (struct eoption));

  /* econf  б ⺻ eoption  ʱȭմϴ.  */
  pre_option_init (s, server, p, o);

  /* econf  ؼϿ ⺻ option  Ѵ.  */
  option_init_econf (s, p);

  /* option_init_econf ()    ݿѴ.  */
  post_option_init (s, server, p, o);
}

static apr_status_t 
e_server_cleanup (data)
     void *data;
{
  eserver_t *s;
  eoption_t *o;

  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "e_server_cleanup: ");
  
  s = (eserver_t *) data;
  o = ESERVER_OPTION (s);
  /*  eserver_t  , Memory Pool   ޸𸮰 Ǳ ,
      ޸𸮸  ʿ , ο Ǵ Non-Memory-Pool
       , ⼭   ʿ䰡  ̴.    α 
     fclose    Ͽ Ѵ.  */

  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "e_server_cleanup: ");

  apr_file_close (OPTION_ERROR_LOG (o));

  return APR_SUCCESS;
}

/**
    Լ Environment     ü ʱȭѴ.
*/
int
e_init_main (p, plog, ptemp, s)
     apr_pool_t *p, *plog ATTRIBUTE_UNUSED, *ptemp ATTRIBUTE_UNUSED;
     server_rec *s;
{
  eserver = (eserver_t *) apr_pcalloc (p, sizeof (struct eserver_info));

  /*   eserver   cleanup  Ѵ.  */
  apr_pool_cleanup_register (p, (void *) eserver, e_server_cleanup, apr_pool_cleanup_null);

  apr_pool_create (&ESERVER_MEMORY_POOL (eserver), p);
  apr_pool_tag (ESERVER_MEMORY_POOL (eserver), "E-Server");

  /* Part 1 - Option.  */
  option_init_common (eserver, s, ESERVER_MEMORY_POOL (eserver));

  /* Part 2 - EVM.
     VM  ʱȭ ̷  κ óѴ.  */
  vm_init_common (eserver, ESERVER_MEMORY_POOL (eserver));

  return OK;
}
