/* mod_e.c - apache 2.x  module 
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "http_protocol.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"
#include "util_script.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"

static int e_pre_config (apr_pool_t *, apr_pool_t *, apr_pool_t *);
static void e_init_method (apr_pool_t *);
static apr_status_t e_server_shutdown (void *);
static int e_post_config (apr_pool_t *, apr_pool_t *, apr_pool_t *, server_rec *);
static int erequest_begin (erequest_t *, request_rec *, eserver_context_t *);
static erequest_t * init_erequest (request_rec *);
static void e_register_hooks (apr_pool_t *);

static int 
e_pre_config (p, plog, ptemp)
     apr_pool_t *p ATTRIBUTE_UNUSED, *plog ATTRIBUTE_UNUSED, *ptemp ATTRIBUTE_UNUSED;
{
#ifndef ENV_THREADSAFE
  int threaded_mpm;
  
  ap_mpm_query (AP_MPMQ_IS_THREADED, &threaded_mpm);
  if (threaded_mpm) 
    {
      ap_log_error (APLOG_MARK, APLOG_CRIT, 0, 0, 
		    "Apache is running a threaded MPM, but your E Module is not compiled to be threadsafe.");
      return DONE;
    }
#endif

  /* VM  ʿ pre config  Ѵ.  */
  vm_pre_config ();

  return OK;
}

/**
    Apache HTTP method  Ѵ.  Ʒ  method  e  ⿡
   ó ̴.   Լ ȣǱ  ݵ e_init_main () Լ ȣǾ 
    eserver  ʱȭ Ǿ Ѵ.
*/
static void
e_init_method (p)
     apr_pool_t *p;
{
  /* Register Environment methods - Instruction PART.  */
  ESERVER_METHOD_ARRAY (eserver, E_M_GETFIELD) = ap_method_register (p, "EGETFIELD");
  ESERVER_METHOD_ARRAY (eserver, E_M_PUTFIELD) = ap_method_register (p, "EPUTFIELD");
  ESERVER_METHOD_ARRAY (eserver, E_M_INVOKEVIRTUAL) = ap_method_register (p, "EINVOKEVIRTUAL");
  ESERVER_METHOD_ARRAY (eserver, E_M_INVOKESPECIAL) = ap_method_register (p, "EINVOKESPECIAL");
  ESERVER_METHOD_ARRAY (eserver, E_M_INVOKESTATIC) = ap_method_register (p, "EINVOKESTATIC");
  ESERVER_METHOD_ARRAY (eserver, E_M_INVOKEINTERFACE) = ap_method_register (p, "EINVOKEINTERFACE");

  /* Register Environment methods - Operation PART.  */
  ESERVER_METHOD_ARRAY (eserver, E_M_LOAD) = ap_method_register (p, "ELOAD");
  ESERVER_METHOD_ARRAY (eserver, E_M_INFO) = ap_method_register (p, "EINFO");
}

static void
finish_eserver_struct (void)
{
  vm_t *vm = ESERVER_VM (eserver);
  eoption_t *o = ESERVER_OPTION (eserver);

  /* VM_T ü */
  vm_splay_tree_delete (VM_E_TREE (vm));
  vm_splay_tree_delete (VM_EOBJECTREF_TREE (vm));
  vm_splay_tree_delete (VM_EARRAYREF_TREE (vm));
  vm_htab_delete (VM_ETYPES_HASH (vm));
  apr_thread_mutex_destroy (VM_GLOBAL_MUTEX (vm));

  /* EOPTION_T ü */
  vm_htab_delete (OPTION_HASH_TABLE (o));

}

extern apr_thread_t *finalmon, *gcmon;
static void
finish_final_and_gc_thread (void)
{
  apr_status_t rv, trv;

  rv = apr_thread_join (&trv, finalmon);
  rv = apr_thread_join (&trv, gcmon);  
}

/**
     ۵ǰ ִ Ȯ  , eserver  shutdown  , ȣ
   Լν eserver ü Ҵϰ ִ  ޸𸮸 Ѵ.
   
   TODO:  ۵  ִ Ȯؾ Ѵ.
 */
static apr_status_t
e_server_shutdown (tmp)
     void *tmp ATTRIBUTE_UNUSED;
{
  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "e_server_shutdown: ");

  finish_tsrm ();
  finish_final_and_gc_thread ();
  finish_eserver_struct ();

  env_log_error (ENVLOG_MARK, ENVLOG_ERR, "e_server_shutdown: ");

  return APR_SUCCESS;
}

static int
e_post_config (p, plog, ptemp, s)
     apr_pool_t *p, *plog, *ptemp;
     server_rec *s;
{
  void *data = NULL;
  const char *userdata_key = "apache2hook_post_config";

  /* Apache will load, unload and then reload a DSO module. This
   * prevents us from starting PHP until the second load. */
  apr_pool_userdata_get (&data, userdata_key, s->process->pool);
  if (!data) 
    {
      /* We must use set() here and *not* setn(), otherwise the
	 static string pointed to by userdata_key will be mapped
	 to a different location when the DSO is reloaded and the
	 pointers won't match, causing get() to return NULL when
	 we expected it to return non-NULL. */
      apr_pool_userdata_set ((const void *) 1, userdata_key, apr_pool_cleanup_null, s->process->pool);

      return OK;
    }

  /* Thread-Safe Resource Manager ʱȭ.  */
  init_tsrm (TSRM_DEBUG_NOLEVEL, NULL);

  /* Environment Module ʱȭ.  */
  e_init_main (p, plog, ptemp, s);

  /*  method .   Լ ݵ e_init_main () Լ ȣ ڿ ȣǾ Ѵ.  */
  e_init_method (p);

  apr_pool_cleanup_register (p, NULL, e_server_shutdown, apr_pool_cleanup_null);
  ap_add_version_component (p, "elang/0.1");

  env_log_error (ENVLOG_MARK, ENVLOG_INFO, " û ");

  return OK;
}

static int 
erequest_begin (er, r, ctx)
     erequest_t *er;
     request_rec *r;
     eserver_context_t *ctx ATTRIBUTE_UNUSED;
{
  char *p;
  erequest_info_t *i = REQUEST_INFO (er);

  RCONTENT_METHOD (i) = r->method;
  RCONTENT_METHOD_NUMBER (i) = r->method_number;
  RCONTENT_TYPE (i) = apr_table_get (r->headers_in, "Content-Type");
  p = (char *) apr_table_get (r->headers_in, "Content-Length");
  RCONTENT_LENGTH (i) = (p ? atoi (p) : 0);

  RCONTENT_QUERY_STRING (i) = apr_pstrdup (r->pool, r->args);
  
  apr_table_unset (r->headers_out, "Content-Length");
  apr_table_unset (r->headers_out, "Last-Modified");
  apr_table_unset (r->headers_out, "Expires");
  apr_table_unset (r->headers_out, "ETag");

  return e_request_startup (er);
}

/**
   Request   , Memory Pool Ž ȣǴ callback Լ.  ̰  
   eserver  Server Context κ NULL  .  ޸𸮴 Memory Pool  ҴǾ
   ֱ    ʿ .
*/
static apr_status_t 
e_server_context_cleanup (void *data)
{
  void **d = data;

  *d = NULL;

  return APR_SUCCESS;
}

/**
   EREQUEST  õ   ش.  
*/
static erequest_t *
init_erequest (r)
     request_rec *r;
{
  erequest_t *er;

  er = (erequest_t *) tsrm_malloc (r->pool, sizeof (struct eserver_request));

  /* erequest_t  ʱȭѴ.  */
  REQUEST_POOL (er) = r->pool;
  REQUEST_SERVER_CONTEXT (er) = NULL;

  /* erequest_info_t  ʱȭѴ.  */
  REQUEST_INFO (er) = (erequest_info_t *) apr_pcalloc (REQUEST_POOL (er), sizeof (struct request_info));
  RCONTENT_READ_POST_BYTES (REQUEST_INFO (er)) = 0;
  RCONTENT_RFC1867_UPLOADED_FILES (REQUEST_INFO (er)) = NULL;

  RCONTENT_CALL_ARGC (REQUEST_INFO (er)) = -1;
  RCONTENT_CALL_ARGV (REQUEST_INFO (er)) = NULL;
  RCONTENT_INTERPRETREF_LIST (REQUEST_INFO (er)) = NULL;

  REQUEST_DATA (er) = (erequest_data_t *) apr_pcalloc (REQUEST_POOL (er), sizeof (struct request_data));

  RESPONSE_INFO (er) = (eresponse_info_t *) apr_pcalloc (REQUEST_POOL (er), sizeof (struct response_info));
  RES_STATUS (RESPONSE_INFO (er)) = OK;

  return er;
}

/** 
   E-Lang  ȣ  , óϰ  handler.   ƾ ̴.  
*/
static int 
e_handler (request_rec *r)
{
  int rescode = OK;
  apr_status_t rv;
  apr_bucket *bucket;
  apr_bucket_brigade *brigade;
  request_rec *parent_req = NULL;
  eserver_context_t *ctx;
  erequest_t *er;

  if (strcmp (r->handler, E_MAGIC_TYPE))
    return DECLINED;

  if (r->finfo.filetype == 0)
    return HTTP_NOT_FOUND;

  if (r->finfo.filetype == APR_DIR)
    return HTTP_FORBIDDEN;

  /* Setup the CGI variables if this is the main request */
  if (r->main == NULL
      /* .. or if the sub-request envinronment differs from the main-request. */ 
      || r->subprocess_env != r->main->subprocess_env) 
    {
      /* setup standard CGI variables */
      ap_add_common_vars (r);
      ap_add_cgi_vars (r);
    }

  er = init_erequest (r);

  ctx = REQUEST_SERVER_CONTEXT (er);
  if (!ctx)
    {
      ctx = REQUEST_SERVER_CONTEXT (er) = apr_pcalloc (r->pool, sizeof (struct eserver_context));
      /*  request   , REQUEST_SERVER_CONTEXT (er)
	  ڵ  ǵ cleanup  Ѵ.  */
      apr_pool_cleanup_register (r->pool, 
				(void *) &REQUEST_SERVER_CONTEXT (er),
				e_server_context_cleanup, apr_pool_cleanup_null);
      CONTEXT_REQUEST (ctx) = r;
      CONTEXT_BRIGADE (ctx) = brigade =
	apr_brigade_create (r->pool, r->connection->bucket_alloc);

      if (erequest_begin (er, r, ctx) != SUCCESS)
	{
	  /* TODO: н   ۼؾ Ѵ.  */
	}
    }
  else
    {
      parent_req = CONTEXT_REQUEST (ctx);
      CONTEXT_REQUEST (ctx) = r;
      brigade = CONTEXT_BRIGADE (ctx);
    }

  /*  GET Ȥ POST data   request info ü  ߴ.
       loader   óѴ.  */
  rv = e_execute_loader (er);

  /* .  */
  if (!parent_req) 
    {
      /* TODO: Request  Ѵ.  */
      rescode = erequest_shutdown (er, r);
      CONTEXT_REQUEST_PROCESSED (ctx) = 1;
      bucket = apr_bucket_eos_create (r->connection->bucket_alloc);
      APR_BRIGADE_INSERT_TAIL (brigade, bucket);
      
      rv = ap_pass_brigade (r->output_filters, brigade);
      if (rv != APR_SUCCESS || r->connection->aborted) 
	e_handle_aborted_connection ();
      apr_brigade_cleanup (brigade);
    }
  else 
    CONTEXT_REQUEST (ctx) = parent_req;

  return rescode;
}

static void 
e_register_hooks (p)
     apr_pool_t *p ATTRIBUTE_UNUSED;
{
  /* Config  ̷   ʿ䰡 ִ   츦 Ͽ,
     pre_config hook  Ѵ.  
     TODO:  ʱȭ 켱  ̰,   ؾ   
     ϴ  ñϴ.  */
  ap_hook_pre_config (e_pre_config, NULL, NULL, APR_HOOK_MIDDLE);

  /* Config  ̷ Ŀ  ʿ䰡 ִ ۿ  hook  Ѵ.  */
  ap_hook_post_config (e_post_config, NULL, NULL, APR_HOOK_MIDDLE);

  ap_hook_handler (e_handler, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA e_module = 
  {
    STANDARD20_MODULE_STUFF,
    NULL,			/* create per-directory config structure */
    NULL,			/* merge per-directory config structures */
    NULL,			/* create per-server config structure */
    NULL,			/* merge per-server config structures */
    NULL,			/* command apr_table_t */
    e_register_hooks		/* register hooks */
  };
