/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "apr.h"
#include "apr_portable.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE

#include "ap_config.h"
#include "httpd.h"
#include "mpm_default.h"
#include "http_main.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"		/* for get_remote_host */
#include "http_connection.h"
#include "http_protocol.h"
#include "scoreboard.h"
#include "ap_mpm.h"
#include "unixd.h"
#include "mpm_common.h"
#include "ap_listen.h"
#include "ap_mmn.h"
#include "apr_poll.h"
#include "util_script.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero() */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor() */
#endif

#include <signal.h>
#include <sys/times.h>

#include "vm.h"

#include <limits.h>

#define QSORT_STACK_SIZE	(sizeof(size_t) * CHAR_BIT)

static void 
_zend_qsort_swap (void *a, void *b, size_t siz)
{
  register char  *tmp_a_char;
  register char  *tmp_b_char;
  register int   *tmp_a_int;
  register int   *tmp_b_int;
  register size_t i;
  int             t_i;
  char            t_c;
  
  tmp_a_int = (int *) a;
  tmp_b_int = (int *) b;
  
  for (i = sizeof(int); i <= siz; i += sizeof(int)) 
    {
      t_i = *tmp_a_int;
      *tmp_a_int++ = *tmp_b_int;
      *tmp_b_int++ = t_i;
    }
  
  tmp_a_char = (char *) tmp_a_int;
  tmp_b_char = (char *) tmp_b_int;
  
  for (i = i - sizeof(int) + 1; i <= siz; ++i) 
    {
      t_c = *tmp_a_char;
      *tmp_a_char++ = *tmp_b_char;
      *tmp_b_char++ = t_c;
    }
}

void 
zend_qsort (void *base, size_t nmemb, size_t siz, compare_func_t compare)
{
  void           *begin_stack[QSORT_STACK_SIZE];
  void           *end_stack[QSORT_STACK_SIZE];
  register char  *begin;
  register char  *end;
  register char  *seg1;
  register char  *seg2;
  register char  *seg2p;
  register int    loop;
  uint            offset;
  
  begin_stack[0] = (char *) base;
  end_stack[0]   = (char *) base + ((nmemb - 1) * siz);
  
  for (loop = 0; loop >= 0; --loop) 
    {
      begin = begin_stack[loop];
      end   = end_stack[loop];
      
      while (begin < end) 
	{
	  offset = (end - begin) >> 1;
	  _zend_qsort_swap(begin, begin + (offset - (offset % siz)), siz);
	  
	  seg1 = begin + siz;
	  seg2 = end;
	  
	  while (1) 
	    {
	      for (; seg1 < seg2 && compare(begin, seg1) > 0;
		   seg1 += siz);
	      
	      for (; seg2 >= seg1 && compare(seg2, begin) > 0;
		   seg2 -= siz);
	      
	      if (seg1 >= seg2)
		break;
	      
	      _zend_qsort_swap(seg1, seg2, siz);
	      
	      seg1 += siz;
	      seg2 -= siz;
	    }
	  
	  _zend_qsort_swap(begin, seg2, siz);
	  
	  seg2p = seg2;
	  
	  if ((seg2p - begin) <= (end - seg2p)) 
	    {
	      if ((seg2p + siz) < end) 
		{
		  begin_stack[loop] = seg2p + siz;
		  end_stack[loop++] = end;
		}
	      end = seg2p - siz;
	    }
	  else
	    {
	      if ((seg2p - siz) > begin) 
		{
		  begin_stack[loop] = begin;
		  end_stack[loop++] = seg2p - siz;
		}
	      begin = seg2p + siz;
	    }
	}
    }
}
