/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#ifndef E_REQUEST_H
#define E_REQUEST_H

typedef struct request_info erequest_info_t;
typedef struct request_data erequest_data_t;
typedef struct response_info eresponse_info_t;
typedef struct eserver_request erequest_t;
typedef struct env_post_entry env_post_entry_t;
typedef struct rfc1867_uploaded_files_constant rfc1867_uploaded_files_t;

/* Brigade κ ѹ  Block ũ.  */
#define REQUEST_POST_BLOCK_SIZE		(1024 * 4)

/*
  Environment  ϴ Content-Type .
*/
#define DEFAULT_POST_CONTENT_TYPE	"application/x-www-form-urlencoded"
/* RFC1867  */
#define MULTIPART_CONTENT_TYPE		"multipart/form-data"

/* ü ؼ ʿ ArrayRef   ڷ .  */
#define INTERPRETREF_ARG_INDEX(NODE)	((NODE)->argidx)
#define INTERPRETREF_INDEX(NODE)	((NODE)->idx)
#define INTERPRETREF_VALUE(NODE)	((NODE)->value)
#define INTERPRETREF_NEXT(NODE)		((NODE)->next)

struct interpretref
{
  /* Argument  index ȣ.  */
  short argidx;
  /* ArrayRef  index.  */
  int idx;

  evalue_t *value;

  struct interpretref *next;
};

#define RCONTENT_METHOD(NODE)			((NODE)->method)
#define RCONTENT_METHOD_NUMBER(NODE)		((NODE)->method_number)
#define RCONTENT_TYPE(NODE)			((NODE)->content_type)
#define RCONTENT_LENGTH(NODE)			((NODE)->content_length)
#define RCONTENT_QUERY_STRING(NODE)		((NODE)->query_string)
#define RCONTENT_POST_ENTRY(NODE)		((NODE)->post_entry)
#define RCONTENT_POST_DATA(NODE)		((NODE)->post_data)
#define RCONTENT_POST_DATA_ARRAY(NODE, IDX)	((NODE)->post_data[(IDX)])
#define RCONTENT_POST_DATA_LENGTH(NODE)		((NODE)->post_data_length)
#define RCONTENT_READ_POST_BYTES(NODE)		((NODE)->read_post_bytes)
#define RCONTENT_ARGS_HASH_INITIAL_SIZE		1000
#define RCONTENT_ARGS(NODE)			((NODE)->args)
#define RFC1867_UPLOADED_FILES_HASH_INITIAL_SIZE	1000
#define RCONTENT_RFC1867_UPLOADED_FILES(NODE)	((NODE)->rfc1867_uploaded_files)
/* Request  target  ׻ environment file  ̴.  ̰ ش environment
     pointer   ִ.  , 

   GETFIELD /lang/env/String.e

    /lang/env/String.e   environment structure ̴.  */
#define RCONTENT_ENV(NODE)			((NODE)->e)
/*  request  call instruction  operation  , ؼ  
   VM  ȣϰ Ǵµ,     ȿ óϱ ؼ ARGC  ARGV  ⿡
    ´.  ̰ Query String  ó  ǰ ȴ.  */
#define RCONTENT_CALL_ARGC(NODE)		((NODE)->call_argc)
#define RCONTENT_CALL_ARGV(NODE)		((NODE)->call_argv)
#define RCONTENT_CALL_ARGV_IDX(NODE, IDX)	((NODE)->call_argv[(IDX)])
#define RCONTENT_BAILOUT_SET(NODE)		((NODE)->bailout_set)
#define RCONTENT_BAILOUT(NODE)			((NODE)->bailout)
#define RCONTENT_INTERPRETREF_LIST(NODE)	((NODE)->ia)

struct request_info
{
  /* Request Method  Ÿ.  */
  const char *method;
  int method_number;
  const char *content_type;
  int content_length;

  char *query_string;

  env_post_entry_t *post_entry;
  /*  û ޵ POST buffer.  */
  char *post_data;
  uint post_data_length;
  int read_post_bytes;

  vm_htab_t args;
  /* Upload   /tmporary 丮 ϰ Ǵµ, ش  θ
      ִ hash table ̴.  */
  vm_htab_t rfc1867_uploaded_files;

  /*  request  Ǿ ִ environment structure .  */
  env_t e;

  /* Request  call instruction  , RCONTENT_ARGS ()  ׿  
     ⿡  ǰ ȴ.  */
  int call_argc;
  evalue_t **call_argv;

  /*  ڰ [T<number>    type  ڷ  , ̰ ؼؼ
     ߿  ArrayRef  Ͽ Ѵ.  Ʒ  linked list ̴.  */
  struct interpretref *ia;

  jmp_buf bailout;

  unsigned bailout_set : 1;
  unsigned unused : 31;
};

#define REQ_DATA_EXCEPTION(NODE)		((NODE)->exception)

/* eserver  ʿ per-request  ͸  ü.  */
struct request_data
{
  /* ENI  ü ; Ѵ.  */
  //eni_t eni;

  /* Exception ó  ʿ ͵ Ʒ ; Ѵ.

     TODO:   ġ struct VMexception *  ϴ   
      ִ.     óϰ ϴ exception  per-request 
     ַ ϱ ̴.   exception  ߻ , request  óϴٰ
      thread  ߻ϴ exception    Ȯϰ ǵ 
     Ͽ.   eserver  ,  ü thread    
     ۼϿ.   δ apache server ü multi-thread ȯ̱ 
     Ư thread ȯ   ̴.

     ߿    ü ҽ apache    𸥴.  ׷ 
     ⸦...
  */
  struct VMexception *exception;
};

#define RES_STATUS(NODE)		((NODE)->status)

/* Request   û   ִ ü̴.  */
struct response_info
{
  /* HTTP Status Code Number */
  int status;

  /*  Ǵ  ʾ, Reponse Body    ǵ ̴.  */
};

/* struct eserver_request (erequest_t) ü multi-thread ȯ̵ one-process
   ȯ̵     Ǵ üν   request  о װ
   ؼ    ̴.     ޸𸮴 request_rec *r 
   default pool  Ͽ ҴǸ, pool  cleanup    cleanup Ǿ Ѵ.
   ֳϸ   ȸ ̱ ̴.  Multi-thread ȯ濡   
   thread key  ̿Ͽ  thread  request    ֵ Ǿ Ѵ.  */

#define REQUEST_POOL(NODE)		((NODE)->rp)
#define REQUEST_SERVER_CONTEXT(NODE)	((NODE)->server_context)
#define REQUEST_INFO(NODE)		((NODE)->info)
#define REQUEST_DATA(NODE)		((NODE)->data)
#define RESPONSE_INFO(NODE)		((NODE)->resinfo)

/* Apache κ ޵Ǵ request_rec ü   ü.   ü
    ߿ üν ϳ û     ,  û 
   response ü   ִ.   response ü  Ͽ HTTP status
   ȣ ϰ Ǵ ̴.  */		
struct eserver_request
{
  /*  RP  ܼ request_rec *r  default pool  ̴.  ,
     request ü ҴǴ ޸𸮴  default pool   Ǹ,
     request   ,  ڵ ȴ.  */
  apr_pool_t *rp;

  eserver_context_t *server_context;

  /*  ޵ request    ִ.  */
  erequest_info_t *info;

  /* eserver  ʿ per-request  ͸  ü.  */
  erequest_data_t *data;

  /*  request   ó      ִ.  */
  eresponse_info_t *resinfo;
};

/*
  Post  óϱ  ü.  
*/

#define POST_ENTRY_CONTENT_TYPE(NODE)		((NODE)->content_type)
#define POST_ENTRY_CONTENT_TYPE_LENGTH(NODE)	((NODE)->content_type_len)
#define POST_ENTRY_READER(NODE)			((NODE)->post_reader)
#define POST_ENTRY_HANDLER(NODE)		((NODE)->post_handler)

/* Content-Type   POST data   ó ׸  ü.  */
struct env_post_entry 
{
  const char *content_type;
  int  content_type_len;
  void (*post_reader) (erequest_t *);
  void (*post_handler) (erequest_t *, const char *, void *);
};

#define RFC1867_HASH(NODE)		((NODE)->hash)
#define RFC1867_BUF(NODE)		((NODE)->buf)
#define RFC1867_LENGTH(NODE)		((NODE)->len)

/* RFC1867  ǵǾ ִ multipart/form-data  óϱ  ü Ϻ.
   ε  ̸ (⼭  ̸ temporary  ̸)    
   ִ.

   (Ʈ)  eserver  ޵Ǵ   ̸     ʰ ִ.  */
   
struct rfc1867_uploaded_files_constant
{
  int hash;
  char *buf;
  int len;
};

/* evalue  Լ.  */
extern evalue_t *new_evalue (apr_pool_t *, enum evalue_type, ...);

/* request  Լ.  */
extern int e_request_startup (erequest_t *);
extern int erequest_shutdown (erequest_t *, request_rec *);
extern eargs_t * lookup_erequest_args (erequest_t *, const char *);

#endif
