/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#ifndef VM_COMMON_H
#define VM_COMMON_H

/* Native Interface   ִ ̺귯  ̸.  */
#define NATIVELIBRARY		"libere"

#define	MAXLIBS		16
#define	MAXLIBPATH	1024

#ifndef LIBRARYPREFIX
#define	LIBRARYPREFIX	"lib"
#endif

/* libltdl  , ڵ suffix  ̱  Ű  ʾƵ ȴ.  */
#ifndef LIBRARYSUFFIX
#define LIBRARYSUFFIX	""
#endif

#define LT_NON_POSIX_NAMESPACE
#include "ltdl.h"

#define LIBRARYHANDLE	lt_dlhandle

static inline void
nativelib_init (void)
{
  lt_dlinit ();
}

static inline LIBRARYHANDLE
nativelib_load (const char *libname)
{
  return lt_dlopenext (libname);
}

static inline void
nativelib_unload (LIBRARYHANDLE handle)
{
  lt_dlclose (handle);
}

static inline void *
nativelib_getsymbol (LIBRARYHANDLE handle,
		     const char *symname)
{
  return lt_dlsym (handle, symname);
}

static inline const char *
nativelib_geterror (void)
{
  return lt_dlerror ();
}

/* Load  Native Interface ü.  */
struct _lib_handler
{
  LIBRARYHANDLE desc;
  char *name;

  struct _lib_handler *next;
};

/* TODO: Ʒ ϴ envkey   Ȯ ʽϴ.  ̰ ϰ ذ
    ã ֽϴ.  */
#define ENV_STRING_TYPE			"Le://localhost/env/core/string('n7J2znmJv5pCocpie/1eATzhXBY=');"
#define ENV_OBJECT_TYPE			"Le://localhost/env/core/Object('n7J2znmJv5pCocpie/1eATzhXBY=');"

/* 64 should put us on the safe side */
#define	MAXMARGS		64

#define CALLINFO_FUNCTION(NODE)			((NODE)->func)
#define CALLINFO_ARGS(NODE)			((NODE)->args)
#define CALLINFO_ARGS_IDX(NODE, IDX)		((NODE)->args[(IDX)])
#define CALLINFO_RETURN(NODE)			((NODE)->ret)
#define CALLINFO_NRARGS(NODE)			((NODE)->nrargs)
#define CALLINFO_ARGSIZE(NODE)			((NODE)->argsize)
#define CALLINFO_RETSIZE(NODE)			((NODE)->retsize)
#define CALLINFO_RETTYPE(NODE)			((NODE)->rettype)
#define CALLINFO_CALLSIZE(NODE)			((NODE)->callsize)
#define CALLINFO_CALLSIZE_IDX(NODE, IDX)	((NODE)->callsize[(IDX)])
#define CALLINFO_CALLTYPE(NODE)			((NODE)->calltype)
#define CALLINFO_CALLTYPE_IDX(NODE, IDX)	((NODE)->calltype[(IDX)])

/* VM  Ư method  ȣϱ ؼ ϴ  ü.  */
struct emethod_callinfo 
{
  void *func;
  evalue_t *args;
  evalue_t *ret;
  int nrargs;
  int argsize;
  char retsize;
  char rettype;
  char *callsize;
  char *calltype;
};

/* ó exception  鿡 ؼ ϰ Ѵ.  */
enum exception_kind
  {
    THROW_AbstractMethodError,
    THROW_ArithmeticException,
    THROW_ArrayIndexOutOfBoundsException,
    THROW_ArrayStoreException,
    THROW_ClassCastException,
    THROW_IllegalAccessError,
    THROW_IllegalMonitorStateException,
    THROW_IllegalThreadStateException,
    THROW_IncompatibleClassChangeError,
    THROW_InstantiationException,
    THROW_LinkageError,
    THROW_NegativeArraySizeException,
    THROW_NoClassDefFoundError,
    THROW_NoSuchFieldError,
    THROW_NoSuchMethodError,
    THROW_NullPointerException,
    THROW_OutOfMemoryError,
    THROW_StackOverflowError,
    THROW_ThreadDeath,
    THROW_UnsatisfiedLinkError,
    THROW_VirtualMachineError
  };

#define VM_CURRENT_UNIQUE_ID(NODE)	((NODE)->current_unique_id)
#define VM_GLOBAL_MEMORY_POOL(NODE)	((NODE)->gmp)
#define VM_E_TREE(NODE)			((NODE)->etree)
#define VM_EOBJECTREF_TREE(NODE)	((NODE)->eobjectref_tree)
#define VM_EARRAYREF_TREE(NODE)		((NODE)->earrayref_tree)
#define VM_GLOBAL_MUTEX(NODE)		((NODE)->mutex)
#define ETYPES_HASH_INITIAL_SIZE	1000
#define VM_ETYPES_HASH(NODE)		((NODE)->etypes)
#define VM_GC(NODE)			((NODE)->collector)
#define VM_NATIVE_LIBS(NODE)		((NODE)->libs)

/* Environment Virtual Machine  ü   ü.  */
struct vm
{
  /*  environment  ڽ unique id   Ǵµ,   ؼ environment 
     ο unique id    id  .  */
  int current_unique_id;

  /* VM   ü memory pool.  */
  apr_pool_t *gmp;

  /* VM  load Ǵ eclass   splay tree.  */
  vm_splay_tree_t etree;

  /* Objectref  Arrayref   splay tree.  */
  vm_splay_tree_t eobjectref_tree;
  vm_splay_tree_t earrayref_tree;

  /* VM  õ operation  locking  ʿ κп  mutex.  */
  apr_thread_mutex_t *mutex;

  /* Environment  ϴ TYPE 鿡    ִ Hash table .  */
  vm_htab_t etypes;

  /* VM garbage collector   interface pointer.  */
  gc_collector_t *collector;

  struct _lib_handler *libs;
};

typedef struct vm vm_t;

#include "eserver.h"

/* VM ʱȭ Լ.  */
AP_DECLARE (void) vm_pre_config (void);
AP_DECLARE (void) vm_init_common (eserver_t *, apr_pool_t *);
AP_DECLARE (apr_status_t) vm_mpm_query (int, int *);

/* String Hash  ƾ Լ .  */
AP_DECLARE (int) strconst_hash_eq (const void *, const void *);
AP_DECLARE (void) strconst_hash_del (void *);
AP_DECLARE (int) strconstcmp (strconst_t *, strconst_t *);
AP_DECLARE (strconst_t *) new_strconst (env_t, const char *, int);

/* Environment .  */
AP_DECLARE (env_t) get_envstruct (envkey_t *);
AP_DECLARE (env_t) lookup_envstruct (envkey_t *);
AP_DECLARE (env_t) get_envstruct_with_idx (int, env_t);
AP_DECLARE (emethod_t *) find_method (env_t, strconst_t *, strconst_t *);
AP_DECLARE (efield_t *) lookup_field (env_t, strconst_t *, int);

/* Environment Type .  */
AP_DECLARE (etype_t *) get_etype_with_kind (enum evalue_type);
AP_DECLARE (etype_t *) get_etype (char *);
AP_DECLARE (etype_t *) lookup_etype (char *);
AP_DECLARE (int) get_field (int, env_t, int, struct efield_info *);
AP_DECLARE (int) get_method_type_environment (int, env_t, int, int, struct emethod_info *);

/* Environment Struct   ϵ lock Ѵ.  */
AP_DECLARE (void) lock_envstruct (env_t);
/* Environment Struct   Ѵ.  */
AP_DECLARE (void) unlock_envstruct (env_t);
extern env_t lookup_env (const char *);

extern void lock_eobjectref (eobjectref_t *);
extern void unlock_eobjectref (eobjectref_t *);
extern void lock_synchronized (emethod_t *, evalue_t *);
extern void unlock_synchronized (emethod_t *, evalue_t *);
extern void *lookup_interfacemethod (eobjectref_t *, env_t, int);

extern void soft_athrow (eobjectref_t *);
extern void * soft_checkcast (env_t, eobjectref_t *);
extern int soft_instanceof (env_t, eobjectref_t *);
extern void * soft_multianewarray (env_t, int, evalue_t *);
extern void throw_error (enum exception_kind);
extern void throw_exception (eobjectref_t *);
extern void env_abort (void);
extern void *call_native (emethod_t *);

#endif
