/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#ifndef APACHE_VM_SCOREBOARD_H
#define APACHE_VM_SCOREBOARD_H

#ifdef __cplusplus
extern "C" {
#endif

#ifdef HAVE_SYS_TIMES_H
#include <sys/time.h>
#include <sys/times.h>
#elif defined(TPF)
#include <time.h>
#endif

#include "ap_config.h"
#include "apr_hooks.h"
#include "apr_thread_proc.h"
#include "apr_portable.h"
#include "apr_shm.h"

/* Scoreboard file, if there is one */
#ifndef DEFAULT_VM_SCOREBOARD
#define DEFAULT_VM_SCOREBOARD "logs/apache_vm_runtime_status"
#endif

typedef struct vm_sb_handle_t vm_sb_handle_t;

AP_DECLARE(int) vm_exists_scoreboard_image(void);
AP_DECLARE(void) vm_increment_counts(vm_sb_handle_t *sbh, request_rec *r);

int vm_create_scoreboard(apr_pool_t *p, ap_scoreboard_e t);
apr_status_t vm_reopen_scoreboard(apr_pool_t *p, apr_shm_t **shm, int detached);
void vm_init_scoreboard(void *shared_score);
AP_DECLARE(int) vm_calc_scoreboard_size(void);
apr_status_t vm_cleanup_scoreboard(void *d);

AP_DECLARE(void) vm_create_sb_handle(vm_sb_handle_t **new_sbh, apr_pool_t *p,
                                     int child_num, int thread_num);
    
AP_DECLARE(int) find_child_by_pid(apr_proc_t *pid);
AP_DECLARE(int) vm_update_child_status(vm_sb_handle_t *sbh, int status, request_rec *r);
AP_DECLARE(int) vm_update_child_status_from_indexes(int child_num, int thread_num,
                                                    int status, request_rec *r);
void vm_time_process_request(vm_sb_handle_t *sbh, int status);

AP_DECLARE(worker_score *) vm_get_scoreboard_worker(int x, int y);
AP_DECLARE(process_score *) vm_get_scoreboard_process(int x);
AP_DECLARE(global_score *) vm_get_scoreboard_global(void);

AP_DECLARE_DATA extern scoreboard *vm_scoreboard_image;
AP_DECLARE_DATA extern const char *vm_scoreboard_fname;
AP_DECLARE_DATA extern int vm_extended_status;

AP_DECLARE_DATA extern ap_generation_t volatile vm_my_generation;

/* Hooks */
/**
  * Hook for post scoreboard creation, pre mpm.
  * @param p       Apache pool to allocate from.
  * @param sb_type 
  * @ingroup hooks
  * @return OK or DECLINE on success; anything else is a error
  */  
AP_DECLARE_HOOK(int, pre_vm_mpm, (apr_pool_t *p, ap_scoreboard_e sb_type))

#ifdef __cplusplus
}
#endif

#endif	/* !APACHE_VM_SCOREBOARD_H */
