/* Ȯ  hash table  datatype.
   Copyright (C) 1999, 2000 Free Software Foundation, Inc.
   Contributed by Vladimir Makarov (vmakarov@cygnus.com).

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/* This package implements basic hash table functionality.  It is possible
   to search for an entry, create an entry and destroy an entry.

   Elements in the table are generic pointers.

   The size of the table is not fixed; if the occupancy of the table
   grows too high the hash table will be expanded.

   The abstract data implementation is based on generalized Algorithm D
   from Knuth's book "The art of computer programming".  Hash table is
   expanded by creation of new hash table and transferring elements from
   the old table to the new table.  */

#ifndef __HASHTAB_H__
#define __HASHTAB_H__

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include <ansidecl.h>

/* The type for a hash code.  */
typedef unsigned int hashval_t;

/* Callback Լ  type .  */

/* Table entry  hash  մϴ.  */
typedef hashval_t (*htab_hash) PARAMS ((const void *));

/*  entry  table entry  Ѵ.  table  ̹ 
   ִ entry  ׻  ;ϸ, ׷ ι° element  ٸ
   type   ִ.  ( ̷  htab_find  htab_find_slot
      ;  hash value  ޾Ƶ   ִ 
    ݵ ؾѴ.).  */
typedef int (*htab_eq) PARAMS ((const void *, const void *));

/* live element  hash table  ŵ  ȣǴ cleanup 
   Լ.  */
typedef void (*htab_del) PARAMS ((void *));
  
/*  live element  htab_traverse   ȣǴ Լ.  ù°
   ڴ element  slot (   htab_clear_slot  ޵
    ) ̰, ι° ڴ htab_traverse   ٷ 
   ̴.  Ⱦ⸦   1,   0  ȯѴ.  */
typedef int (*htab_trav) PARAMS ((void **, void *));

/* Hash table  Ʒ type ̴.   type  ü () 
   hash table   ʿ  ʴ´.  hash table  
     Ʒ ޵Ǿ ִ Լ鸸 ؼ Ǿ 
   Ѵ.  */

struct htab
{
  /* Hash Լ .  */
  htab_hash hash_f;

  /*  Լ .  */
  htab_eq eq_f;

  /* Cleanup Լ .  */
  htab_del del_f;

  /* Table  ڽ.  */
  PTR *entries;

  /* Hash table   (entry ) ũ */
  size_t size;

  /*  element  Ͽ  element   */
  size_t n_elements;

  /* Table    element   */
  size_t n_deleted;

  /*  Ҵ debugging  ˴ϴ.  ̰  hash table 
      `htab_find_slot'  ȣ  Ƚ̴.  */
  unsigned int searches;

  /*  Ҵ debugging  ˴ϴ.  ̰  hash table
       ð  collision  ̴.  */
  unsigned int collisions;

  /* 츮 ޸ Ҵ ϴ Լ ȣ⿡  NULL  ȯϴ 
     ȴٸ 0  ƴ  .  */
  int return_allocation_failure;
};

typedef struct htab *htab_t;

/* An enum saying whether we insert into the hash table or not.  */
enum insert_option {NO_INSERT, INSERT};

/* The prototypes of the package functions. */

extern htab_t	htab_create	PARAMS ((size_t, htab_hash,
					 htab_eq, htab_del));

/* This function is like htab_create, but may return NULL if memory
   allocation fails, and also signals that htab_find_slot_with_hash and
   htab_find_slot are allowed to return NULL when inserting.  */
extern htab_t	htab_try_create	PARAMS ((size_t, htab_hash,
					 htab_eq, htab_del));
extern void	htab_delete	PARAMS ((htab_t));
extern void	htab_empty	PARAMS ((htab_t));

extern PTR	htab_find	PARAMS ((htab_t, const void *));
extern PTR     *htab_find_slot	PARAMS ((htab_t, const void *,
					 enum insert_option));
extern PTR	htab_find_with_hash	  PARAMS ((htab_t, const void *,
						   hashval_t));
extern PTR     *htab_find_slot_with_hash  PARAMS ((htab_t, const void *,
						   hashval_t,
						   enum insert_option));
extern void	htab_clear_slot	PARAMS ((htab_t, void **));
extern void	htab_remove_elt	PARAMS ((htab_t, void *));

extern void	htab_traverse	PARAMS ((htab_t, htab_trav, void *));

extern size_t	htab_size	PARAMS ((htab_t));
extern size_t	htab_elements	PARAMS ((htab_t));
extern double	htab_collisions	PARAMS ((htab_t));

/* A hash function for pointers.  */
extern htab_hash htab_hash_pointer;

/* An equality function for pointers.  */
extern htab_eq htab_eq_pointer;

/* A hash function for null-terminated strings.  */
extern hashval_t htab_hash_string PARAMS ((const PTR));

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __HASHTAB_H */
