%{
#include "config.h"
#include "i386.h"
#include "hwint.h"
#include <stdio.h>
#include <ansidecl.h>
#include <stdlib.h>
#include <string.h>
#include "libiberty.h"
#include "safe-ctype.h"
#include "system.h"
#include "hashtab.h"
#include "obstack.h"
#include "hashtable.h"
#include "machmode.h"
#include "tree.h"
#include "stringpool.h"
#include "esm.h"
#include "errors.h"
#include "fold-const.h"

/* Like YYERROR but do call yyerror.  */
#define YYERROR1 { yyerror ("syntax error"); YYERROR; }

/* Cause the "yydebug" variable to be defined.  */
#define YYDEBUG 1

%}

%start program

%union { enum tree_code code; tree ttype;
         enum esm_access acc;
         const char *filename; int lineno; }

%token IDENTIFIER
%token CONSTANT
%token STRING
%token INSTRUCTION
%token FROM
%token IS
%token TO
%token USING
%token DSET
%token DEFAULT
%token DEND
%token DLINE
%token DCATCH
%token DMETHOD
%token DTHROWS
%token DLIMIT
%token DSUPER
%token DLICENSE
%token DWRITER
%token DACCESS
%token DMATCH
%token DPACKAGE
%token ENVIRONMENT_PATH
%token ENVIRONMENT_PREFIX
%token TABLESWITCH
%token LOOKUPSWITCH
%token METHOD
%token COLON
%token DENVIRONMENT
%token DFIELD
%token DSOURCE
%token DVAR
%token DINTERFACE
%token ACCESS_PUBLIC
%token ACCESS_PRIVATE
%token ACCESS_PROTECTED
%token ACCESS_STATIC
%token ACCESS_FINAL
%token ACCESS_SYNCHRONIZED
%token ACCESS_VOLATILE
%token ACCESS_TRANSIENT
%token ACCESS_NATIVE
%token ACCESS_INTERFACE
%token ACCESS_ABSTRACT
%token DIMPLEMENTS
%token POINTSAT
%token SHA1_STRING
%token SHA1DATA
%token IPADDRESS
%token ADDRESS
%token METHODS_TYPE
%token INSTRUCTION
%token FIELD_TYPE
%token MINUS
%token SEPERATOR

%type <acc> access access_list access_item access_items
%type <ttype> source_spec writer_spec program_info license_spec
%type <ttype> identifier IDENTIFIER STRING ADDRESS ENVIRONMENT_PATH ENVIRONMENT_PREFIX
%type <ttype> address env_spec environment_name super_spec access_exprs
%type <ttype> access_expr program_header method_or_field_name
%type <ttype> implements_list implements_spec fields field_list field_spec
%type <ttype> FIELD_TYPE acchost optional_default item CONSTANT
%type <ttype> defmethod METHODS_TYPE simple_instruction INSTRUCTION instruction
%type <ttype> complex_instruction var_expr limit_expr line_expr throws_expr
%type <ttype> catch_expr set_expr methods method_list statements statement
%type <ttype> lookup_default lookup_entry lookup_list method_spec error
%type <ttype> table_list table_entry table_default table_args
%type <ttype> lookup table lookup_args label directive stmnt program
%type <ttype> env implements matchs match_list match_spec
%type <ttype> package_spec field_prototype method_prototype method_or_field_path
%type <ttype> host_address environment_path environment_segment sha1_string
%type <code> MINUS

%{
enum esm_access tmpacc;
int insn_offset;
int num_switch_entry;
struct labels *insn_labels;
static void yyerror	  PARAMS ((const char *));
static int  yylex	  PARAMS ((void));
%}

%%
program:
        	{ $<ttype>$ = NULL_TREE; }
	program_info matchs program_header env 
		{ $$ = build_program ($2, $3, $4, $5);
		  if (finish_program ($$))
		    error ("finish_program: error.\n"); }
	;

env:
	implements fields methods
		{ $$ = build_body ($1, $2, $3); 
		  TREE_BODY_FIELDS_COUNT ($$) = count_fields ($2);
		  TREE_BODY_METHODS_COUNT ($$) = count_methods ($3); }
	;

program_header:
	package_spec env_spec super_spec
		{ $$ = build_header ($1, $2, $3); }
	;

/* .packet env.lang    óѴ.  */
package_spec:
	DPACKAGE STRING SEPERATOR
		{ process_package_directive ($2);
		  $$ = $2; }
	| /* empty */
		{ $$ = NULL_TREE; }
	;

program_info:
	writer_spec license_spec source_spec
		{ $$ = build_info ($1, $2, $3); 
		  constant_process (CONSTANT_Unknown, $$); }
	;

writer_spec:
	DWRITER STRING SEPERATOR
		{ $$ = $2; }
	;

license_spec:
	DLICENSE STRING SEPERATOR
		{ $$ = $2; }
	;
source_spec:
	DSOURCE STRING SEPERATOR
		{ $$ = $2; 
		  cp_add (CONSTANT_Utf8, strdup ("SourceFile")); }
	| /* empty */
		{ $$ = NULL_TREE; }
	;

identifier:
	IDENTIFIER
		{ $$ = build_tree_list (NULL_TREE, $1); }
	;

acchost:
	'{' access_exprs '}'
		{ $$ = $2; }
	| /* empty */
		{ $$ = NULL_TREE; }
	;

env_spec:
	  DENVIRONMENT access environment_name acchost SEPERATOR
	  	{ $$ = build_env ($2, $3, $4);
		  /*  environment    current_envname  Ѵ.  ̰
		     ٸ method   environment   , Ǳ ؼ̴.  */
		  current_envname = TREE_VALUE ($3);
		  TREE_ENVIRONMENT_ACCESS_HOST_COUNT ($$) = count_access_hosts ($4);
		  constant_process (CONSTANT_Unknown, $$); }
	| DINTERFACE access environment_name acchost SEPERATOR
		{ $$ = build_interface ($2, $3, $4); }
	;

/* Environment Name  .  */
environment_name:
	  identifier
	/* Environment URL  Ѵ.  */
	| address
	;

access:
	access_list
	;

access_list:
		{ $$ = 0; tmpacc = 0; }	
	access_items
		{ $$ = tmpacc; }
	| /* empty */
		{ $$ = 0; }
	;

access_items:
	access_items access_item 
	| access_item
	;

access_item:
	ACCESS_PUBLIC
		{ tmpacc |= ACC_PUBLIC; }
	| ACCESS_PRIVATE
		{ tmpacc |= ACC_PRIVATE; }
	| ACCESS_PROTECTED
		{ tmpacc |= ACC_PROTECTED; }
	| ACCESS_STATIC
		{ tmpacc |= ACC_STATIC; }
	| ACCESS_FINAL
		{ tmpacc |= ACC_FINAL; }
	| ACCESS_SYNCHRONIZED
		{ tmpacc |= ACC_SYNCHRONIZED; }
	| ACCESS_VOLATILE
		{ tmpacc |= ACC_VOLATILE; }
	| ACCESS_TRANSIENT
		{ tmpacc |= ACC_TRANSIENT; }
	| ACCESS_NATIVE
		{ tmpacc |= ACC_NATIVE; }
	| ACCESS_INTERFACE
		{ tmpacc |= ACC_INTERFACE; }
	| ACCESS_ABSTRACT
		{ tmpacc |= ACC_ABSTRACT; }
	;

address:
	ENVIRONMENT_PREFIX host_address '/' environment_path
		{ $$ = simplify_environment_name (build_address ($2, $4)); }
	;

host_address:
	ADDRESS
		{ $$ = build_ipport ($1, NULL_TREE); }
	| ADDRESS ':' CONSTANT
		{ $$ = build_ipport ($1, $3); }
	;
	
environment_path:
	environment_segment
	| environment_segment '/' environment_path
		{ $$ = chainon ($1, build_tree_list (NULL_TREE, $3)); }
	;

environment_segment:
	identifier
		{ $$ = build_segment ($1, NULL_TREE); }
	| identifier '(' sha1_string ')'
		{ $$ = build_segment ($1, $3); }
	;

sha1_string:
	STRING
	;

super_spec:
	DSUPER address SEPERATOR
		{ $$ = $2; }
	| /* empty */
		{ $$ = NULL_TREE; }
	;

implements:
	implements_list 
	| /* empty */
		{ $$ = NULL_TREE; }
	;

implements_list:
	implements_list implements_spec 
		{ $$ = build_tree_list ($1, $2); }
	| implements_spec
	;

implements_spec:
	DIMPLEMENTS environment_name SEPERATOR
		{ $$ = $2; }
	;

fields:
	field_list
		{ $$ = $1; }
	| /* empty */
		{ $$ = NULL_TREE; }
	;

field_list:
	field_list field_spec 
		{ $$ = chainon ($1, build_tree_list (NULL_TREE, $2)); }
	| field_spec
	;

field_spec:
	DFIELD access identifier FIELD_TYPE acchost optional_default SEPERATOR
		{ $$ = build_field ($2, $3, $4, $5, $6); 
		  TREE_FIELD_ACCESS_HOST_COUNT ($$) = count_access_hosts ($5);
		  constant_process (CONSTANT_Unknown, $$); }
	;

matchs:
	match_list
		{ $$ = $1; }
	| /* empty */
		{ $$ = NULL_TREE; }
	;

match_list:
	match_list match_spec 
		{ $$ = chainon ($1, build_tree_list (NULL_TREE, $2)); }
	| match_spec
	;

match_spec:
	DMATCH environment_name identifier SEPERATOR
		{ $$ = build_match ($2, $3);
		  process_match ($$); }
	;


optional_default:
	'=' item
		{ $$ = $2; }
	| /* empty */
		{ $$ = NULL_TREE; }
	;

item:
	CONSTANT
	| STRING
	;

methods:
	method_list
	| /* empty */
		{ $$ = NULL_TREE; }
	;

method_list:
	method_list method_spec 
		{ $$ = chainon ($1, build_tree_list (NULL_TREE, $2)); }
	| method_spec
	;

method_spec:
	defmethod statements endmethod
		{ $$ = build_method ($1, $2);
		  TREE_METHOD_ACCESS_HOST_COUNT ($$) = count_access_hosts (TREE_METHOD_ACCESS_HOST ($1));
		  TREE_METHOD_INSN_LENGTH ($$) = count_insn_length ($2);
		  TREE_METHOD_MAX_STACK ($$) = count_max_stack ($2, 0);
		  TREE_METHOD_MAX_LOCALS ($$) = count_max_locals ($1, $2);
		  TREE_METHOD_ATTRIBUTE_COUNT ($$) = count_attribute ($2);
		  TREE_METHOD_LABEL_LIST ($$) = insn_labels;
		  constant_process (CONSTANT_Unknown, $$);
		  }
	| defmethod endmethod
		{ $$ = build_method ($1, NULL_TREE);
		  TREE_METHOD_ACCESS_HOST_COUNT ($$) = count_access_hosts (TREE_METHOD_ACCESS_HOST ($1));
		  TREE_METHOD_MAX_LOCALS ($$) = count_max_locals ($1, NULL_TREE);
		  constant_process (CONSTANT_Unknown, $$); }
	;

defmethod:
	DMETHOD access method_or_field_name METHODS_TYPE acchost SEPERATOR
		{ $$ = build_method_header ($2, $3, $4, $5); 
		  /* init Ȥ clinit, main  access flag   üũѴ.  */
		  check_reserved_method ($$);
		  insn_offset = 0;
		  insn_labels = NULL; }
	;

method_or_field_name:
	identifier
	| '<' identifier '>'
		{ $$ = process_method_name ($2); }
	;
	
endmethod:
	DEND METHOD SEPERATOR
	;

statements:
	statements statement
		{ $$ = chainon ($1, build_tree_list (NULL_TREE, $2)); }
	| statement
	;

statement:
	stmnt SEPERATOR
		{ $$ = $1; }
	;

stmnt:
	instruction
		{ process_insn ($1); 
		  num_switch_entry = 0; }
	| directive
	| error
	| label
	;

label:
	identifier COLON
		{ $$ = build_label ($1, insn_offset); 
		  add_label (&insn_labels, $1, insn_offset); }
	;

directive:
	DVAR var_expr
		{ $$ = $2; }
	| DLIMIT limit_expr
		{ $$ = $2; }
	| DLINE line_expr
		{ $$ = $2; }
	| DTHROWS throws_expr
		{ $$ = $2; }
	| DCATCH catch_expr
		{ $$ = $2; }
	| DSET set_expr
		{ $$ = $2; }
	;

access_exprs:
	access_exprs access_expr
		{ $$ = build_tree_list ($1, $2); }
	| access_expr
	;

access_expr:
	STRING address SEPERATOR
		{ $$ = build_access_expr ($1, $2); }
	| SEPERATOR STRING address SEPERATOR
		{ $$ = build_access_expr ($2, $3); }
	;

var_expr:
	CONSTANT IS identifier identifier FROM identifier TO identifier
		{ $$ = build_var ($1, $3, $4, $6, $8); }
	| CONSTANT IS identifier identifier
		{ $$ = build_var ($1, $3, $4, NULL_TREE, NULL_TREE); }
	;

limit_expr:
	identifier CONSTANT
		{ $$ = build_limit ($1, $2); }
	;

line_expr:
	CONSTANT
		{ $$ = build_line ($1); }
	;

throws_expr:
	environment_name
		{ $$ = build_throws ($1); }
	;

catch_expr:
	environment_name FROM identifier TO identifier USING identifier
		{ $$ = build_catch ($1, $3, $5, $7); }
	;

set_expr:
	identifier CONSTANT
		{ $$ = build_set ($1, $2); }
	| identifier identifier
		{ $$ = build_set ($1, $2); }
	| identifier STRING
		{ $$ = build_set ($1, $2); }
	;

instruction:
	simple_instruction
	| complex_instruction

method_or_field_path:
	environment_name POINTSAT method_or_field_name
		{ $$ = build_url ($1, $3); }
	;

method_prototype:
	method_or_field_path METHODS_TYPE
		{ $$ = build_methodref ($1, $2); }
	;

field_prototype:
	method_or_field_path FIELD_TYPE
		{ $$ = build_fieldref ($1, $2); }
	;

simple_instruction:
	/* aload_0
	   iconst_0
	     ɾ óϱ ؼ.  */
	INSTRUCTION
		{ $$ = build_insn ($1, NULL_TREE, NULL_TREE); 
		  TREE_INSN_CODE_OFFSET ($$) = insn_offset;
		  insn_offset += TREE_INSN_LENGTH (TREE_INSN_CODE ($$)); }
	| INSTRUCTION FIELD_TYPE
		{ $$ = build_insn ($1, $2, NULL_TREE);
		  TREE_INSN_CODE_OFFSET ($$) = insn_offset;
		  insn_offset += TREE_INSN_LENGTH (TREE_INSN_CODE ($$)); }
	| INSTRUCTION address 
		{ $$ = build_insn ($1, $2, NULL_TREE);
		  TREE_INSN_CODE_OFFSET ($$) = insn_offset;
		  insn_offset += TREE_INSN_LENGTH (TREE_INSN_CODE ($$)); }
	/* putfield        String->value   I
	     ɾ óϱ ؼ.  */
	| INSTRUCTION field_prototype
		{ $$ = build_insn ($1, $2, NULL_TREE);
		  TREE_INSN_CODE_OFFSET ($$) = insn_offset;
		  insn_offset += TREE_INSN_LENGTH (TREE_INSN_CODE ($$)); }
	/* invokevirtual   String-><init>  ()V
	     ɾ óϱ ؼ.  */
	| INSTRUCTION method_prototype
		{ $$ = build_insn ($1, $2, NULL_TREE);
		  TREE_INSN_CODE_OFFSET ($$) = insn_offset;
		  insn_offset += TREE_INSN_LENGTH (TREE_INSN_CODE ($$)); }
	| INSTRUCTION CONSTANT
		{ $$ = build_insn ($1, $2, NULL_TREE);
		  TREE_INSN_CODE_OFFSET ($$) = insn_offset;
		  insn_offset += TREE_INSN_LENGTH (TREE_INSN_CODE ($$)); }
	| INSTRUCTION CONSTANT CONSTANT
		{ $$ = build_insn ($1, $2, $3);
		  TREE_INSN_CODE_OFFSET ($$) = insn_offset;
		  insn_offset += TREE_INSN_LENGTH (TREE_INSN_CODE ($$)); }
	| INSTRUCTION CONSTANT MINUS CONSTANT
		{ $$ = build_insn ($1, $2, fold (build_minus ($4)));
		  TREE_INSN_CODE_OFFSET ($$) = insn_offset;
		  insn_offset += TREE_INSN_LENGTH (TREE_INSN_CODE ($$)); }
	| INSTRUCTION MINUS CONSTANT
		{ $$ = build_insn ($1, fold (build_minus ($3)), NULL_TREE);
		  TREE_INSN_CODE_OFFSET ($$) = insn_offset;
		  insn_offset += TREE_INSN_LENGTH (TREE_INSN_CODE ($$)); }
	/* newarray        "char"
	     instruction  óϱ ؼ.  */
	| INSTRUCTION STRING
		{ $$ = build_insn ($1, $2, NULL_TREE);
		  TREE_INSN_CODE_OFFSET ($$) = insn_offset;
		  insn_offset += TREE_INSN_LENGTH (TREE_INSN_CODE ($$)); }
	/* INSTRUCTION  identifier   , 󺧷 ؾ  ̴.  
	   ifeq		simple_label */
	| INSTRUCTION identifier
		{ $$ = build_insn ($1, $2, NULL_TREE);
		  TREE_INSN_CODE_OFFSET ($$) = insn_offset;
		  insn_offset += TREE_INSN_LENGTH (TREE_INSN_CODE ($$)); }
	/* iload	%1 */
	| INSTRUCTION '%' CONSTANT
		{ $$ = build_insn ($1, build_local ($3), NULL_TREE);
		TREE_INSN_CODE_OFFSET ($$) = insn_offset;
		insn_offset += TREE_INSN_LENGTH (TREE_INSN_CODE ($$)); }
	/* iinc		%1	1 */
	| INSTRUCTION '%' CONSTANT CONSTANT
		{ $$ = build_insn ($1, build_local ($3), $4);
		TREE_INSN_CODE_OFFSET ($$) = insn_offset;
		insn_offset += TREE_INSN_LENGTH (TREE_INSN_CODE ($$)); }
	/* iinc		%1	-1 */
	| INSTRUCTION '%' CONSTANT MINUS CONSTANT
		{ $$ = build_insn ($1, build_local ($3), fold (build_minus ($5)));
		TREE_INSN_CODE_OFFSET ($$) = insn_offset;
		insn_offset += TREE_INSN_LENGTH (TREE_INSN_CODE ($$)); }
	;

complex_instruction:
	LOOKUPSWITCH lookup
		{ $$ = $2; }
	| TABLESWITCH table
		{ $$ = $2; }
	;

lookup:
	lookup_args lookup_list lookup_default
		{ $$ = build_switch_lookup ($1, $2, $3); 
		  TREE_SWITCH_LOOKUP_CODE_OFFSET ($$) = insn_offset;
		  TREE_SWITCH_LOOKUP_CASE_NUMBER ($$) = num_switch_entry;
		  insn_offset += calculate_switch ($$, num_switch_entry); }
	;

lookup_args:
	SEPERATOR
		{ $$ = NULL_TREE; }
	;

lookup_list:
	lookup_list lookup_entry
		{ $$ = chainon ($1, build_tree_list (NULL_TREE, $2)); 
		  num_switch_entry++; }
	| lookup_entry
		{ num_switch_entry++; }

lookup_entry:
	CONSTANT COLON identifier SEPERATOR
		{ $$ = build_case ($1, $3); }
	;

lookup_default:
	DEFAULT COLON identifier
		{ $$ = build_case (NULL_TREE, $3); }
	;

table:
	table_args table_list table_default
		{ $$ = build_switch_table ($1, $2, $3); 
		  TREE_SWITCH_TABLE_CODE_OFFSET ($$) = insn_offset;
		  insn_offset += calculate_switch ($$, num_switch_entry); }
	;

table_args:
	CONSTANT SEPERATOR
		{ $$ = build_switch_args ($1, NULL_TREE); }
	| CONSTANT CONSTANT SEPERATOR
		{ $$ = build_switch_args ($1, $2); }
	;

table_list:
	table_list table_entry 
		{ $$ = chainon ($1, build_tree_list (NULL_TREE, $2)); 
		  num_switch_entry++; }
	| table_entry
		{ num_switch_entry++; }
	;

table_entry:
	identifier SEPERATOR
		{ $$ = build_case (NULL_TREE, $1); }
	;

table_default:
	DEFAULT COLON identifier
		{ $$ = build_case (NULL_TREE, $3); }
	;
%%

static enum esm_ttype last_token;

static int
yylexname ()
{
  if (TREE_IS_RESERVED_WORD (yylval.ttype))
    {
      if (!strcmp (IDENTIFIER_POINTER (yylval.ttype), "lookupswitch"))
	return LOOKUPSWITCH;
      if (!strcmp (IDENTIFIER_POINTER (yylval.ttype), "tableswitch"))
	return TABLESWITCH;

      return TREE_RESERVED_WORD_CODE (yylval.ttype);
    }

  if (TREE_IS_RESERVED_INSN (yylval.ttype))
    return INSTRUCTION;

  /*  identifier  "method"  , Ư ó!  */
  if (!strcmp (IDENTIFIER_POINTER (yylval.ttype), "method"))
    return METHOD;

  /*  ڿ ADDRESS  IDENTIFIER  Ѵ.  ϱ 
     localhost  Ȥ identifier  . ڰ   , ADDRESS 
     Ѵ.  */
  if (strstr (IDENTIFIER_POINTER (yylval.ttype), ".") 
      || !strncmp (IDENTIFIER_POINTER (yylval.ttype), "localhost", 9))
    return ADDRESS;

  return IDENTIFIER;
}

static int
yylex()
{
  last_token = esm_lexer (&yylval.ttype);

  switch (last_token)
    {
    case ESM_STARTP:
      return '{';
    case ESM_ENDP:
      return '}';
    case ESM_ANGLE_START:
      return '<';
    case ESM_ANGLE_END:
      return '>';
    case ESM_SEPERATOR:
      return SEPERATOR;
    case ESM_MINUS:
      return MINUS;
    case ESM_COLON:
      return COLON;
    case ESM_STRING:
      return STRING;
    case ESM_FIELD_TYPE:
      return FIELD_TYPE;
    case ESM_NUMBER:
      return CONSTANT;
    case ESM_DEREF:
      return POINTSAT;
    case ESM_METHOD_TYPE:
      return METHODS_TYPE;
    case ESM_ADDRESS:
      return ADDRESS;
    case ESM_NAME:
      return yylexname ();
    case ESM_EOF:
      return 0;
    case ESM_PREFIX:
      return ENVIRONMENT_PREFIX;
    case ESM_PATH_SEPERATOR:
      return '/';
    case ESM_PAR_START:
      return '(';
    case ESM_PAR_END:
      return ')';
    case ESM_EQ:
      return '=';
    case ESM_LOCALS:
      return '%';
    default:
      fprintf (stderr, " ó! token = %d\n", last_token);
      abort ();
      break;
    }

  /* NOT REACHED.  */
  return 0;
}

#define NAME(type) esm_type2name (type)

static void
yyerror (msgid)
     const char *msgid;
{
  const char *string = msgid;

  if (last_token == ESM_EOF)
    error ("%s at end of input", string);
  else if (last_token == ESM_CHAR)
    {
      unsigned int val = TREE_INT_CST_LOW (yylval.ttype);
      const char *const ell = (last_token == ESM_CHAR) ? "" : "L";
      if (val <= UCHAR_MAX && ISGRAPH (val))
        error ("%s before %s'%c'", string, ell, val);
      else
	error ("%s before %s'\\x%x'", string, ell, val);
    }

  else if (last_token == ESM_STRING)
    error ("%s before string constant", string);
  else if (last_token == ESM_NUMBER)
    error ("%s before numeric constant", string);
  else if (last_token == ESM_NAME)
    error ("%s before \"%s\"", string, IDENTIFIER_POINTER (yylval.ttype));
  else
    error ("%s before '%s' token", string, NAME(last_token));
}
