#include "eas.h"
#include "machmode.h"
#include "tree.h"
#include "stringpool.h"
#include "esm.h"
#include "system.h"
#include "sha1.h"

/*  doubleword  doubleword  negate Ѵ.
   װ sign     operation  overflow ̸, 0  ƴ
    ȯѴ.
    argument  L1  H1   `HOST_WIDE_INT'  ־.
    *LV  *HV   `HOST_WIDE_INT'  ȴ.  */

static int
neg_double (l1, h1, lv, hv)
     unsigned HOST_WIDE_INT l1;
     HOST_WIDE_INT h1;
     unsigned HOST_WIDE_INT *lv;
     HOST_WIDE_INT *hv;
{
  if (l1 == 0)
    {
      *lv = 0;
      *hv = - h1;
      return (*hv & h1) < 0;
    }
  else
    {
      *lv = -l1;
      *hv = ~h1;
      return 0;
    }
}

/* Type   ʴ constant   bit  0 Ȥ 1  ν 
   constant T  type   T  ȿϵ .

    signed overflow  ߻ϸ 1  ׷  0  ȯѴ.  
   OVERFLOW  0  ƴϸ, signed overfow  T  ϴ  ̹ ߻
   ̴. ׷ װ ܼ ø.

   ( ) CHECK_FLOAT_VALUE  ȣν type  real constant T 
    ȿϵ .  */

int
force_fit_type (t, overflow)
     tree t;
     int overflow;
{
  unsigned HOST_WIDE_INT low;
  HOST_WIDE_INT high;
  unsigned int prec;

  if (TREE_CODE (t) == REAL_CST)
    {
      return overflow;
    }
  else if (TREE_CODE (t) != INTEGER_CST)
    return overflow;

  low = TREE_INT_CST_LOW (t);
  high = TREE_INT_CST_HIGH (t);

  if (TREE_CODE (t) == INTEGER_CST)
    prec = HOST_BITS_PER_WIDE_INT;
  else
    prec = 32;

  /*  type  precision   bit   Ѵ.  */

  if (prec == 2 * HOST_BITS_PER_WIDE_INT)
    ;
  else if (prec > HOST_BITS_PER_WIDE_INT)
    TREE_INT_CST_HIGH (t)
      &= ~((HOST_WIDE_INT) (-1) << (prec - HOST_BITS_PER_WIDE_INT));
  else
    {
      TREE_INT_CST_HIGH (t) = 0;
      if (prec < HOST_BITS_PER_WIDE_INT)
	TREE_INT_CST_LOW (t) &= ~((unsigned HOST_WIDE_INT) (-1) << prec);
    }

  /*   sign bit  Ǿ ִٸ, sign  ȮѴ.  */
  if (prec != 2 * HOST_BITS_PER_WIDE_INT
      && (prec > HOST_BITS_PER_WIDE_INT
	  ? 0 != (TREE_INT_CST_HIGH (t)
		  & ((HOST_WIDE_INT) 1
		     << (prec - HOST_BITS_PER_WIDE_INT - 1)))
	  : 0 != (TREE_INT_CST_LOW (t)
		  & ((unsigned HOST_WIDE_INT) 1 << (prec - 1)))))
    {
      /* Value is negative:
	 set to 1 all the bits that are outside this type's precision.  */
      if (prec > HOST_BITS_PER_WIDE_INT)
	TREE_INT_CST_HIGH (t)
	  |= ((HOST_WIDE_INT) (-1) << (prec - HOST_BITS_PER_WIDE_INT));
      else
	{
	  TREE_INT_CST_HIGH (t) = -1;
	  if (prec < HOST_BITS_PER_WIDE_INT)
	    TREE_INT_CST_LOW (t) |= ((unsigned HOST_WIDE_INT) (-1) << prec);
	}
    }

  /*  signed overflow  ߻Ͽٸ 0  ƴ  ȯѴ.  */
  return
    ((overflow | (low ^ TREE_INT_CST_LOW (t)) | (high ^ TREE_INT_CST_HIGH (t)))
     != 0);
}

tree
fold (expr)
     tree expr;
{
  tree t = expr;
  tree arg0 = NULL_TREE, arg1 = NULL_TREE;
  enum tree_code code = TREE_CODE (t);
  int kind = TREE_CODE_CLASS (code);
  /* WINS   operand   switch   
     0  ƴ   ̴.  */
  int wins = 1;

  if (IS_EXPR_CODE_CLASS (kind) || kind == 'r')
    {
      int len = first_rtl_op (code);
      int i;
      for (i = 0; i < len; i++)
	{
	  tree op = TREE_OPERAND (t, i);
	  tree subop;

	  if (op == 0)
	    continue;		/* Valid for CALL_EXPR, at least.  */

	  subop = op;

	  if (TREE_CODE (subop) != INTEGER_CST
	      && TREE_CODE (subop) != REAL_CST)
	    /* Note that TREE_CONSTANT isn't enough:
	       static var addresses are constant but we can't
	       do arithmetic on them.  */
	    wins = 0;

	  if (i == 0)
	    arg0 = op;
	  else if (i == 1)
	    arg1 = op;
	}
    }

  switch (code)
    {
    case INTEGER_CST:
      return t;
    case NEGATE_EXPR:
      if (wins)
	{
	  if (TREE_CODE (arg0) == INTEGER_CST)
	    {
	      unsigned HOST_WIDE_INT low;
	      HOST_WIDE_INT high;
	      int overflow = neg_double (TREE_INT_CST_LOW (arg0),
					 TREE_INT_CST_HIGH (arg0),
					 &low, &high);
	      t = build_int_2 (low, high);
	      TREE_OVERFLOW (t)	= (TREE_OVERFLOW (arg0) | force_fit_type (t, overflow));
	      TREE_CONSTANT_OVERFLOW (t) = TREE_OVERFLOW (t) | TREE_CONSTANT_OVERFLOW (arg0);
	    }
	}
      else if (TREE_CODE (arg0) == NEGATE_EXPR)
	return TREE_OPERAND (arg0, 0);

      return t;
    default:
      error ("fold: óؾ ϴ tree node.\n");
      debug_tree (t);
      return t;
    }
}
