#ifndef ESERVER_HASHTABLE_H
#define ESERVER_HASHTABLE_H

typedef struct ht_identifier ht_identifier;
struct ht_identifier
{
  unsigned int len;
  const unsigned char *str;
};

#define HT_LEN(NODE) ((NODE)->len)
#define HT_STR(NODE) ((NODE)->str)

typedef struct ht hash_table;
typedef struct ht_identifier *hashnode;

enum ht_lookup_option {HT_NO_INSERT = 0, HT_ALLOC, HT_ALLOCED};

/* Cpplib  Front end   ĺ hash table. */
struct ht
{
  /* ĺڵ ⼭ Ҵ˴ϴ.  */
  struct obstack stack;

  hashnode *entries;
  /* Call back. */
  hashnode (*alloc_node) PARAMS ((hash_table *));

  unsigned int nslots;		/* Entry 迭  slot .  */
  unsigned int nelements;	/* (ִ|Ȱϴ) element  .  */

  /* ̺  ġ.  */
  unsigned int searches;
  unsigned int collisions;
};

extern void gcc_obstack_init PARAMS ((struct obstack *));

/* Initialise the hashtable with 2 ^ order entries.  */
extern hash_table *ht_create PARAMS ((unsigned int order));

/* Frees all memory associated with a hash table.  */
extern void ht_destroy PARAMS ((hash_table *));

extern hashnode ht_lookup PARAMS ((hash_table *, const unsigned char *,
				   unsigned int, enum ht_lookup_option));

/* For all nodes in TABLE, make a callback.  The callback takes
   TABLE->PFILE, the node, and a PTR, and the callback sequence stops
   if the callback returns zero.  */
typedef int (*ht_cb) PARAMS ((hashnode, const void *));
extern void ht_forall PARAMS ((hash_table *, ht_cb, const void *));

/* Dump allocation statistics to stderr.  */
extern void ht_dump_statistics PARAMS ((hash_table *));

/* Approximate positive square root of a host double.  This is for
   statistical reports, not code generation.  */
extern double approx_sqrt PARAMS ((double));

#endif
