#include "eas.h"
#include "machmode.h"
#include "tree.h"
#include "stringpool.h"
#include "esm.h"
#include "esm.tab.h"
#include "system.h"
#include "sha1.h"
#include "real.h"

extern struct labels *insn_labels;

int yyparse ();
extern int yydebug;
int opt_debug_all;

/* Constant Pool */
struct constant_pool *root_cp = NULL;
/* Constant Pool's unique number  */
int cp_count = 1;

/* Name for output file of assembly code, specified with -o.  */

char *env_file_name;
static int version_flag;
static FILE *env_out_file;
static char *filename;
int file_offset;
int line;

static void process_options PARAMS ((void));

/* Dump output file   ⺻ ̸  ̸.  */

const char *dump_base_name;

/* Environment      Ǵ ü.  */
struct environment_file 
{
  unsigned char *base;
  unsigned char *cur;
  unsigned char *backup_to;
  unsigned char *line_base;
  size_t size;
};

/*  keyword   table.  */
struct resword
{
  const char *word;
  unsigned int code;
  unsigned int disable;
};

static const struct resword reswords[] =
{
  { ".catch", DCATCH, 0 },
  { ".environment", DENVIRONMENT, 0 },
  { ".super", DSUPER, 0 },
  { ".end", DEND, 0 },
  { ".field", DFIELD, 0 },
  { ".implements", DIMPLEMENTS, 0 },
  { ".interface", DINTERFACE, 0 },
  { ".limit", DLIMIT, 0 },
  { ".line", DLINE, 0 },
  { ".method", DMETHOD, 0 },
  { ".match", DMATCH, 0 },
  { ".set", DSET, 0 },
  { ".source", DSOURCE, 0 },
  { ".throws", DTHROWS, 0 },
  { ".var", DVAR, 0 },
  { ".license", DLICENSE, 0 },
  { ".writer", DWRITER, 0 },
  { ".package", DPACKAGE, 0 },
  { "from", FROM, 0 },
  { "method", METHOD, 0 },
  { "to", TO, 0 },
  { "is", IS, 0 },
  { "using", USING, 0 },
  { "tableswitch", TABLESWITCH, 0 },
  { "lookupswitch", LOOKUPSWITCH, 0 },
  { "default", DEFAULT, 0 },
  { "public", ACCESS_PUBLIC, 0 },
  { "private", ACCESS_PRIVATE, 0 },
  { "protected", ACCESS_PROTECTED, 0 },
  { "static", ACCESS_STATIC, 0 },
  { "final", ACCESS_FINAL, 0 },
  { "synchronized", ACCESS_SYNCHRONIZED, 0 },
  { "volatile", ACCESS_VOLATILE, 0 },
  { "transient", ACCESS_TRANSIENT, 0 },
  { "native", ACCESS_NATIVE, 0 },
  { "interface", ACCESS_INTERFACE, 0 },
  { "abstract", ACCESS_ABSTRACT, 0 },
};

#define N_reswords (sizeof reswords / sizeof (struct resword))

/*  keyword   table.  */
struct resword_insn
{
  const char *word;
  enum insn_code bytecode;
  int oplen;
};

static const struct resword_insn resword_insns[] =
{
#define DEF_CODE(ENUM, NAME, NUMBER, FORMAT, CLASS, OPLEN)   { NAME, ENUM, OPLEN },
#include "code.def"
#undef DEF_CODE
};

#define N_resword_insns (sizeof resword_insns / sizeof (struct resword_insn))

static const char *token_spellings[] =
{
#define DEF_TOKEN(ENUM, NAME)   NAME ,
#include "token.def"
#undef DEF_TOKEN
};

/* machine mode  Ǿ  machine mode  ̸
   մϴ.  ̸ "mode" ڸ  ʽϴ.  */

#define DEF_MACHMODE(SYM, NAME, CLASS, BITSIZE, SIZE, UNIT, WIDER, INNER)  NAME,

const char * const mode_name[NUM_MACHINE_MODES] = {
#include "machmode.def"
};

#undef DEF_MACHMODE

/* machine mode  Ǿ  GET_MODE_CLASS   class
   mode  մϴ.  */

#define DEF_MACHMODE(SYM, NAME, CLASS, BITSIZE, SIZE, UNIT, WIDER, INNER)  CLASS,

const enum mode_class mode_class[NUM_MACHINE_MODES] = {
#include "machmode.def"
};

#undef DEF_MACHMODE

/* machine mode  Ǿ  mode  ̸ Ʈ ũ
   մϴ. GET_MODE_BITSIZE  ̰ մϴ.  */

#define DEF_MACHMODE(SYM, NAME, CLASS, BITSIZE, SIZE, UNIT, WIDER, INNER)  BITSIZE,

const unsigned short mode_bitsize[NUM_MACHINE_MODES] = {
#include "machmode.def"
};

#undef DEF_MACHMODE

/* machine mode  Ǿ  mode  ̸ Ʈ ũ
   մϴ. GET_MODE_SIZE  ̰ մϴ.  */

#define DEF_MACHMODE(SYM, NAME, CLASS, BITSIZE, SIZE, UNIT, WIDER, INNER)  SIZE,

const unsigned char mode_size[NUM_MACHINE_MODES] = {
#include "machmode.def"
};

#undef DEF_MACHMODE

/* machine mode  Ǿ  mode  subunit  ̸
   մϴ. GET_MODE_UNIT_SIZE  ̰ մϴ.  */

#define DEF_MACHMODE(SYM, NAME, CLASS, BITSIZE, SIZE, UNIT, WIDER, INNER)  UNIT,

const unsigned char mode_unit_size[NUM_MACHINE_MODES] = {
#include "machmode.def"		/* machine mode   ȭǾ  */
};

#undef DEF_MACHMODE

/* Machine mode  Ǿ      mode 
   ݴϴ.  (QI -> HI -> SI -> DI, Ÿ )  Widening multiply
   instruction  ̰ մϴ.  */

#define DEF_MACHMODE(SYM, NAME, CLASS, BITSIZE, SIZE, UNIT, WIDER, INNER)  \
  (unsigned char) WIDER,

const unsigned char mode_wider_mode[NUM_MACHINE_MODES] = {
#include "machmode.def"		/* machine mode   ȭǾ  */
};

#undef DEF_MACHMODE

#define DEF_MACHMODE(SYM, NAME, CLASS, BITSIZE, SIZE, UNIT, WIDER, INNER)  \
  ((BITSIZE) >= HOST_BITS_PER_WIDE_INT) ? ~(unsigned HOST_WIDE_INT) 0 : ((unsigned HOST_WIDE_INT) 1 << (BITSIZE)) - 1,

/* Machine mode  ĵǾ, mode  significant bit  mask  
   Ѵ.

   Indexed by machine mode, gives mask of significant bits in mode.  */

const unsigned HOST_WIDE_INT mode_mask_array[NUM_MACHINE_MODES] = {
#include "machmode.def"
};

#undef DEF_MACHMODE

#define DEF_MACHMODE(SYM, NAME, CLASS, BITSIZE, SIZE, UNIT, WIDER, INNER) INNER,

/* Machine mode  Ǿ  Vector type  inner element
    mode  մϴ.  */

const enum machine_mode inner_mode_array[NUM_MACHINE_MODES] = {
#include "machmode.def"
};

/* Mode class  зǾ   class     mode 
   ˷ݴϴ.  Q mode  ׻ ʺ 1 (Ҽ (complex)  ؼ
   2) ̱  ٸ mode   ﷡ ׷  ϴ.

   츮 MODE_INT  BImode ſ QImode  ϴ  ˰
   ʽÿ.  ̷ ϴ  middle end  츮  
   ʴ ü ׿  ͵鿡 bitfields   õ
    ֱ Դϴ.   target md file  BImode widget 
   ؾ մϴ.  */

const enum machine_mode class_narrowest_mode[(int) MAX_MODE_CLASS] = {
    /* MODE_RANDOM */		VOIDmode,
    /* MODE_INT */		QImode,
    /* MODE_FLOAT */		QFmode,
    /* MODE_PARTIAL_INT */	PQImode,
    /* MODE_CC */		CCmode,
    /* MODE_COMPLEX_INT */	CQImode,
    /* MODE_COMPLEX_FLOAT */	QCmode,
    /* MODE_VECTOR_INT */	V2QImode,
    /* MODE_VECTOR_FLOAT */	V2SFmode
};

struct matchs
{
  tree match;

  struct matchs *next;
};

struct environment_file ef_info;

struct matchs *match_root;

/* .package "eng.lang"  eng.lang  eng/lang/   θ 
   .  */
char *package_path;

/*  óϰ ִ environment  ̸  ִ.  */
tree current_envname;

/* SWITCH ɾ ̸ Ѵ.  */
int
calculate_switch (t, num_entry)
     tree t;
     int num_entry;
{
  int oplen = 1;	/* oplen  1  ϴ  opcode   ̴.  */
  int offset = TREE_SWITCH_LOOKUP_CODE_OFFSET (t) + oplen;

  if (num_entry < 0)
    {
      error ("߸ entry ũ.\n");
      return -1;
    }

  switch (TREE_CODE (t))
    {
    case SWITCH_LOOKUP:
      {
	int i, padnum = (offset % 4) ? 4 - (offset % 4) : 0;

	oplen += padnum;
	oplen += 4;	/* Default Value */
	oplen += 4;	/* Entry Size */	    
	for (i=0; i<num_entry; i++)
	  {
	    oplen += 4;	/* Key Value */
	    oplen += 4;	/* Jump Address */
	  }

	TREE_SWITCH_LOOKUP_CODE_LENGTH (t) = oplen;
#if 0
	error ("calculate_switch: ϼؾ. Number Entry = %d, Offset = %d, Pad Num = %d Oplen = %d\n", 
	       num_entry, TREE_SWITCH_LOOKUP_CODE_OFFSET (t), padnum, oplen);
#endif
	break;
      }
    default:
      error ("calculate_switch: ϼؾ. Number Entry = %d, Offset = %d\n", 
	     num_entry, TREE_SWITCH_LOOKUP_CODE_OFFSET (t));
      debug_tree (t);
      break;
    }

  return oplen;
}

const char *
esm_type2name (type)
     enum esm_ttype type;
{
  return (const char *) token_spellings[type];
}

int
find_label (list, name)
     struct labels *list;
     tree name;
{
  struct labels *tmp = list;

  if (!list)
    return -1;

  if (TREE_CODE (name) != IDENTIFIER_NODE)
    {
      error ("find_label:  NAME  IDENTIFIER_NODE  ƴմϴ.\n");
      return -1;
    }

  while (tmp)
    {
      if (!strcmp (IDENTIFIER_POINTER (tmp->name), IDENTIFIER_POINTER (name)))
	return tmp->code_offset;
      tmp = tmp->next;
    }

  return -1;
}

int
add_label (list, name, code_offset)
     struct labels **list;
     tree name;
     int code_offset;
{
  int p;
  struct labels *n;
  
  p = find_label (*list, TREE_VALUE (name));
  if (p != -1)
    {
      error ("տ ̹ label  ǵǾ .\n");
      return -1;
    }
  
  n = (struct labels *) malloc (sizeof (struct labels));
  n->name = TREE_VALUE (name);
  n->code_offset = code_offset;
  n->next = *list;
  *list = n;

  return 0;
}

/*  reserved method  , Access Flag  ˻Ѵ.  */
void
check_reserved_method (t)
     tree t;
{
  const char *name = IDENTIFIER_POINTER (TREE_VALUE (TREE_METHOD_NAME (t)));

  if (!strcmp (name, "main") 
      && (!((TREE_METHOD_ACCESS (t) & ACC_PUBLIC) 
	   && (TREE_METHOD_ACCESS (t) & ACC_STATIC))
	  || (TREE_METHOD_ACCESS (t) & (~(ACC_PUBLIC | ACC_STATIC)))))
    warning ("Main Method  Access Flag  public static  ƴմϴ.\n");

  if (!strcmp (name, "<init>") 
      && (!(TREE_METHOD_ACCESS (t) & ACC_PUBLIC)
	  || (TREE_METHOD_ACCESS (t) & (~ACC_PUBLIC))))
    warning ("<init> Method  Access Flag  public  ƴմϴ.\n");

  if (!strcmp (name, "<clinit>") 
      && (!(TREE_METHOD_ACCESS (t) & ACC_STATIC)
	  || (TREE_METHOD_ACCESS (t) & (~ACC_STATIC))))
    warning ("<clinit> Method  Access Flag  public  ƴմϴ.\n");
}

#define MAX_COLUMN		36

/*
 - split - divide a string into fields, like awk split()
 = int split(char *string, char *fields[], int nfields, char *sep);
 */
int                             /* number of fields, including overflow */
split(string, fields, nfields, sep)
     char *string;
     char *fields[];                 /* list is not NULL-terminated */
     int nfields;                    /* number of entries available in fields[] */
     char *sep;                      /* "" white, "c" single char, "ab" [ab]+ */
{
  register char *p = string;
  register char c;                        /* latest character */
  register char sepc = sep[0];
  register char sepc2;
  register int fn;
  register char **fp = fields;
  register char *sepp;
  register int trimtrail;

  /* white space */
  if (sepc == '\0')
    {
      while ((c = *p++) == ' ' || c == '\t')
        continue;
      p--;
      trimtrail = 1;
      sep = " \t";    /* note, code below knows this is 2 long */
      sepc = ' ';
    }
  else
    trimtrail = 0;
  sepc2 = sep[1];         /* now we can safely pick this up */

  /* catch empties */
  if (*p == '\0')
    return(0);

  /* single separator */
  if (sepc2 == '\0')
    {
      fn = nfields;
      for (;;)
        {
          *fp++ = p;
          fn--;
          if (fn == 0)
            break;
          while ((c = *p++) != sepc)
            if (c == '\0')
              return(nfields - fn);
          *(p-1) = '\0';
        }
      /* we have overflowed the fields vector -- just count them */
      fn = nfields;
      for (;;)
        {
          while ((c = *p++) != sepc)
            if (c == '\0')
              return(fn);
          fn++;
        }
      /* not reached */
    }

  /* two separators */
  if (sep[2] == '\0')
    {
      fn = nfields;
      for (;;)
        {
          *fp++ = p;
          fn--;
          while ((c = *p++) != sepc && c != sepc2)
            if (c == '\0')
              {
                if (trimtrail && **(fp-1) == '\0')
                  fn++;
                return(nfields - fn);
              }
          if (fn == 0)
            break;
          *(p-1) = '\0';
          while ((c = *p++) == sepc || c == sepc2)
            continue;
          p--;
        }
      /* we have overflowed the fields vector -- just count them */
      fn = nfields;
      while (c != '\0')
        {
          while ((c = *p++) == sepc || c == sepc2)
            continue;
          p--;
          fn++;
          while ((c = *p++) != '\0' && c != sepc && c != sepc2)
            continue;
        }
      /* might have to trim trailing white space */
      if (trimtrail)
        {
          p--;
          while ((c = *--p) == sepc || c == sepc2)
            continue;
          p++;
          if (*p != '\0')
            {
              if (fn == nfields+1)
                *p = '\0';
              fn--;
            }
        }
      return(fn);
    }

  /* n separators */
  fn = 0;
  for (;;)
    {
      if (fn < nfields)
        *fp++ = p;
      fn++;
      for (;;)
        {
          c = *p++;
          if (c == '\0')
            return(fn);
          sepp = sep;
          while ((sepc = *sepp++) != '\0' && sepc != c)
            continue;
          if (sepc != '\0')       /* it was a separator */
            break;
        }
      if (fn < nfields)
        *(p-1) = '\0';
      for (;;) {
        c = *p++;
        sepp = sep;
        while ((sepc = *sepp++) != '\0' && sepc != c)
          continue;
        if (sepc == '\0')       /* it wasn't a separator */
          break;
      }
      p--;
    }

  /* not reached */
}

struct matchs *
find_match (t)
     tree t;
{
  struct matchs *a = match_root;

  while (a)
    {
      if (t == a->match)
	break;
      a = a->next;
    }

  return a;
}

struct matchs *
find_match_by_name (name)
     char *name;
{
  struct matchs *a = match_root;

  while (a)
    {
      if (!strcmp (name, IDENTIFIER_POINTER (TREE_SEGMENT_NAME (TREE_EADDRESS_PATHANDSHA1 (TREE_MATCH_STRING (a->match))))))
	break;
      a = a->next;
    }

  return a;
}

struct matchs *
find_match_by_name_with_length (name, len)
     char *name;
     int len;
{
  struct matchs *a = match_root;
  char tmpbuf[1024];

  //fprintf (stderr, "find_match_by_name_with_length: %.*s\n", len, name);
  strncpy (tmpbuf, name, len);
  tmpbuf[len] = '\0';

  while (a)
    {
      if (!strcmp (tmpbuf, IDENTIFIER_POINTER (TREE_SEGMENT_NAME (TREE_EADDRESS_PATHANDSHA1 (TREE_MATCH_STRING (a->match))))))
	break;
      a = a->next;
    }

  return a;
}

int
insert_match (t)
     tree t;
{
  struct matchs *a;

  a = (struct matchs *) malloc (sizeof (struct matchs));
  if (!a)
    return -1;

  a->match = t;
  a->next = match_root;
  match_root = a;

  return 0;
}

void
print_cp (void)
{
  struct constant_pool *tm = root_cp;

  fprintf (stderr, "Print Constant Pool\n");
  while (tm)
    {
      switch (CONSTANT_TYPE (tm))
	{
	case CONSTANT_Environment:
	  fprintf (stderr, "CONSTANT INDEX = 0x%x\tCONSTANT_Environment\t", CONSTANT_INDEX (tm));
	  fprintf (stderr, "ARG0 = 0x%x, ARG1 = 0x%x\n", CONSTANT_VARG0 (tm), CONSTANT_VARG1 (tm));
	  break;
	case CONSTANT_Fieldref:
	  fprintf (stderr, "CONSTANT INDEX = 0x%x\tCONSTANT_Fieldref\t", CONSTANT_INDEX (tm));
	  fprintf (stderr, "ARG0 = 0x%x, ARG1 = 0x%x\n", CONSTANT_VARG0 (tm), CONSTANT_VARG1 (tm));
	  break;
	case CONSTANT_Methodref:
	  fprintf (stderr, "CONSTANT INDEX = 0x%x\tCONSTANT_Methodref\t", CONSTANT_INDEX (tm));
	  fprintf (stderr, "ARG0 = 0x%x, ARG1 = 0x%x\n", CONSTANT_VARG0 (tm), CONSTANT_VARG1 (tm));
	  break;
	case CONSTANT_InterfaceMethodref:
	  error ("print_cp: ó ٶ\n");
	  break;
	case CONSTANT_String:
	  fprintf (stderr, "CONSTANT INDEX = 0x%x\tCONSTANT_String\t\t", CONSTANT_INDEX (tm));
	  fprintf (stderr, "ARG0 = 0x%x\n", CONSTANT_VARG0 (tm));
	  break;
	case CONSTANT_Integer:
	  fprintf (stderr, "CONSTANT INDEX = 0x%x\tCONSTANT_Integer\t", CONSTANT_INDEX (tm));
	  fprintf (stderr, "VALUE = %d\n", CONSTANT_INTEGER_VALUE_ARG0 (tm));
	  break;
	case CONSTANT_Float:
	  fprintf (stderr, "CONSTANT INDEX = 0x%x\tCONSTANT_Float\t\t", CONSTANT_INDEX (tm));
	  fprintf (stderr, "VALUE = 0x%lx\n", CONSTANT_LONG_VALUE_ARG0 (tm));
	  break;
	case CONSTANT_Long:
	  fprintf (stderr, "CONSTANT INDEX = 0x%x\tCONSTANT_Long\t\t", CONSTANT_INDEX (tm));
	  fprintf (stderr, "VALUE = ");
	  fprintf (stderr, HOST_WIDE_INT_PRINT_DOUBLE_HEX,
		   CONSTANT_INTEGER_VALUE_ARG0 (tm), CONSTANT_INTEGER_VALUE_ARG1 (tm));
	  fprintf (stderr, "\n");
	  break;
	case CONSTANT_Double:
	  fprintf (stderr, "CONSTANT INDEX = 0x%x\tCONSTANT_Double\t\t", CONSTANT_INDEX (tm));
	  fprintf (stderr, "VALUE = 0x%lx%lx\n", CONSTANT_LONG_VALUE_ARG0 (tm), CONSTANT_LONG_VALUE_ARG1 (tm));
	  break;
	case CONSTANT_Sha1AndName:
	  fprintf (stderr, "CONSTANT INDEX = 0x%x\tCONSTANT_Sha1AndName\t", CONSTANT_INDEX (tm));
	  fprintf (stderr, "ARG0 = 0x%x, ARG1 = 0x%x\n", CONSTANT_VARG0 (tm), CONSTANT_VARG1 (tm));
	  break;	  
	case CONSTANT_NameAndType:
	  fprintf (stderr, "CONSTANT INDEX = 0x%x\tCONSTANT_NameAndType\t", CONSTANT_INDEX (tm));
	  fprintf (stderr, "ARG0 = 0x%x, ARG1 = 0x%x\n", CONSTANT_VARG0 (tm), CONSTANT_VARG1 (tm));
	  break;
	case CONSTANT_Utf8:
	  fprintf (stderr, "CONSTANT INDEX = 0x%x\tCONSTANT_Utf8\t\t", CONSTANT_INDEX (tm));
	  fprintf (stderr, "STRING = %s [%d]\n", CONSTANT_VSTRING (tm), strlen (CONSTANT_VSTRING (tm)));
	  break;
	case CONSTANT_Sha1:
	  {
	    int i;
	    fprintf (stderr, "CONSTANT INDEX = 0x%x\tCONSTANT_Sha1\t\t", CONSTANT_INDEX (tm));
	    fprintf (stderr, "SHA1 = ");
	    if (CONSTANT_VSTRING (tm))
	      {
		for (i=0; i<SHA_DIGESTSIZE; i++)
		  fprintf (stderr, "%x ", CONSTANT_VSTRING (tm)[i] & 0xff);
	      }
	    else
	      fprintf (stderr, "NULL");
	    fprintf (stderr, "\n");
	  }
	  break;
	case CONSTANT_Sha1This:
	  fprintf (stderr, "CONSTANT INDEX = 0x%x\tCONSTANT_Sha1This\n", CONSTANT_INDEX (tm));
	  break;	  
	case CONSTANT_Accessref:
	  fprintf (stderr, "CONSTANT INDEX = 0x%x\tCONSTANT_Accessref\t", CONSTANT_INDEX (tm));
	  fprintf (stderr, "ARG0 = 0x%x, ARG1 = 0x%x\n", CONSTANT_VARG0 (tm), CONSTANT_VARG1 (tm));
	  break;
	case CONSTANT_Host:
	  fprintf (stderr, "CONSTANT INDEX = 0x%x\tCONSTANT_Host\t\t", CONSTANT_INDEX (tm));
	  fprintf (stderr, "ARG0 = 0x%x, ARG1 = 0x%x\n", CONSTANT_VARG0 (tm), CONSTANT_VARG1 (tm));
	  break;
	default:
	  error ("print_cp: ó ٶ %d\n", CONSTANT_TYPE (tm));
	  break;
	}
      tm = CONSTANT_NEXT (tm);
    }
}

/* T1  T2  ִ SHA1  Ѵ.    0 , ٸ -1  ȯѴ.  */
int
strcmp_sha1 (t1, t2)
     char *t1, *t2;
{
  int i, done = 0;
 
  if (!t1 || !t2)
    return -1;

  for (i=0; i<SHA_DIGESTSIZE; i++)
    {
      if (t1[i] != t2[i])
	done = 1;
    }

  return done;
}

/* Constant Pool  ϳ constant  Ѵ.  CONSTANT_Utf8, CONSTANT_Sha1  , 
   ش ڿ ۿ  ͸  ִµ,   ڿ ۸  ҴϿ 
   ϰ ִ´.  */
int
cp_add (enum constant type, ...)
{
  va_list ap;
  HOST_WIDE_INT arg0 = -1, arg1 = -1;
  long larg0 = -1, larg1 = -1;
  char *str = NULL;
  struct constant_pool *root, *new;
 
  /* ã  */
  va_start(ap, type);
  switch (type)
    {
    case CONSTANT_Environment:
      {
	arg0 = va_arg (ap, HOST_WIDE_INT);
	arg1 = va_arg (ap, HOST_WIDE_INT);
	root = root_cp;
	while (root)
	  {
	    if (CONSTANT_TYPE (root) == type 
		&& CONSTANT_VARG0 (root) == arg0 
		&& CONSTANT_VARG1 (root) == arg1)
	      return CONSTANT_INDEX (root);
	    root = CONSTANT_NEXT (root);
	  }
	break;
      }
    case CONSTANT_InterfaceMethodref:
      error ("cp_add: ó ٶ\n");
      break;
    case CONSTANT_String:
      {
	arg0 = va_arg (ap, HOST_WIDE_INT);
	root = root_cp;
	while (root)
	  {
	    if (CONSTANT_TYPE (root) == type && CONSTANT_VARG0 (root) == arg0)
	      return CONSTANT_INDEX (root);
	    root = CONSTANT_NEXT (root);
	  }
	break;
      }
    case CONSTANT_Integer:
      {
	arg0 = va_arg (ap, HOST_WIDE_INT);
	root = root_cp;
	while (root)
	  {
	    if (CONSTANT_TYPE (root) == type && CONSTANT_INTEGER_VALUE_ARG0 (root) == arg0)
	      return CONSTANT_INDEX (root);
	    root = CONSTANT_NEXT (root);
	  }
	break;
      }
    case CONSTANT_Float:
      {
	larg0 = va_arg (ap, long);
	root = root_cp;
	while (root)
	  {
	    if (CONSTANT_TYPE (root) == type 
		&& CONSTANT_LONG_VALUE_ARG0 (root) == larg0)
	      return CONSTANT_INDEX (root);
	    root = CONSTANT_NEXT (root);
	  }
	break;
      }
    case CONSTANT_Long:
      {
	arg0 = va_arg (ap, HOST_WIDE_INT);
	arg1 = va_arg (ap, HOST_WIDE_INT);
	root = root_cp;
	while (root)
	  {
	    if (CONSTANT_TYPE (root) == type 
		&& CONSTANT_INTEGER_VALUE_ARG0 (root) == arg0 
		&& CONSTANT_INTEGER_VALUE_ARG1 (root) == arg1)
	      return CONSTANT_INDEX (root);
	    root = CONSTANT_NEXT (root);
	  }
	break;
      }
    case CONSTANT_Double:
      {
	larg0 = va_arg (ap, long);
	larg1 = va_arg (ap, long);
	root = root_cp;
	while (root)
	  {
	    if (CONSTANT_TYPE (root) == type 
		&& CONSTANT_LONG_VALUE_ARG0 (root) == larg0 
		&& CONSTANT_LONG_VALUE_ARG1 (root) == larg1)
	      return CONSTANT_INDEX (root);
	    root = CONSTANT_NEXT (root);
	  }
	break;
      }
    case CONSTANT_Utf8:
      {
	str = va_arg (ap, char *);
	if (!str)
	  {
	    error ("cp_add: CONSTANT_Utf8   string  NULL ̴.");
	    return -1;
	  }
	root = root_cp;
	while (root)
	  {
	    if (CONSTANT_TYPE (root) == type && !strcmp (CONSTANT_VSTRING (root), str))
	      return CONSTANT_INDEX (root);
	    root = CONSTANT_NEXT (root);
	  }
	break;
      }
    case CONSTANT_Sha1:
      {
	str = va_arg (ap, char *);
	root = root_cp;
	while (root && str)
	  {
	    if (CONSTANT_TYPE (root) == type && !strcmp_sha1 (CONSTANT_VSTRING (root), str))
	      return CONSTANT_INDEX (root);
	    root = CONSTANT_NEXT (root);
	  }
	break;
      }
    case CONSTANT_Sha1This:
      {
	root = root_cp;
	while (root)
	  {
	    if (CONSTANT_TYPE (root) == type)
	      return CONSTANT_INDEX (root);
	    root = CONSTANT_NEXT (root);
	  }
	break;
      }      
    case CONSTANT_Sha1AndName:
    case CONSTANT_Fieldref:
    case CONSTANT_Methodref:
    case CONSTANT_NameAndType:
    case CONSTANT_Accessref:
    case CONSTANT_Host:
      {
	arg0 = va_arg (ap, HOST_WIDE_INT);
	arg1 = va_arg (ap, HOST_WIDE_INT);
	root = root_cp;
	while (root)
	  {
	    if (CONSTANT_TYPE (root) == type 
		&& CONSTANT_VARG0 (root) == arg0 && CONSTANT_VARG1 (root) == arg1)
	      return CONSTANT_INDEX (root);
	    root = CONSTANT_NEXT (root);
	  }
	break;
      }
    default:
      error ("cp_add: ó ٶ\n");
      break;
    }
  va_end (ap);

  /*   */
  new = (struct constant_pool *) malloc (sizeof (struct constant_pool));
  memset (new, '\0', sizeof (struct constant_pool));
  CONSTANT_INDEX (new) = cp_count++;
  CONSTANT_TYPE (new) = type;

  switch (CONSTANT_TYPE (new))
    {
    case CONSTANT_Environment:
      CONSTANT_VARG0 (new) = arg0;
      CONSTANT_VARG1 (new) = arg1;
      break;
    case CONSTANT_InterfaceMethodref:
      error ("cp_add: ó ٶ\n");
      break;
    case CONSTANT_String:
      CONSTANT_VARG0 (new) = arg0;
      break;
    case CONSTANT_Integer:
      CONSTANT_INTEGER_VALUE_ARG0 (new) = arg0;
      break;
    case CONSTANT_Float:
      CONSTANT_LONG_VALUE_ARG0 (new) = larg0;
      break;
    case CONSTANT_Long:
      CONSTANT_INTEGER_VALUE_ARG0 (new) = arg0;
      CONSTANT_INTEGER_VALUE_ARG1 (new) = arg1;
      break;
    case CONSTANT_Double:
      CONSTANT_LONG_VALUE_ARG0 (new) = larg0;
      CONSTANT_LONG_VALUE_ARG1 (new) = larg1;
      break;
    case CONSTANT_Utf8:
      {
	int len = strlen (str);
	
	CONSTANT_VSTRING (new) = (char *) malloc (len + 1);
	memcpy (CONSTANT_VSTRING (new), str, len);
	CONSTANT_VSTRING (new)[len] = '\0';
	break;
      }
    case CONSTANT_Sha1:
      {
	CONSTANT_VSTRING (new) = (char *) malloc (SHA_DIGESTSIZE);
	memcpy (CONSTANT_VSTRING (new), str, SHA_DIGESTSIZE);
	break;
      }
    case CONSTANT_Sha1AndName:
    case CONSTANT_Fieldref:
    case CONSTANT_NameAndType:
    case CONSTANT_Methodref:
    case CONSTANT_Accessref:
    case CONSTANT_Host:
      CONSTANT_VARG0 (new) = arg0;
      CONSTANT_VARG1 (new) = arg1;
      break;
    case CONSTANT_Sha1This:
      break;
    default:
      error ("cp_add: ó ٶ\n");
      break;
    }
  
  CONSTANT_NEXT (new) = root_cp;
  root_cp = new;

  return CONSTANT_INDEX (new);
}

int
make_CONSTANT_Environment (t)
     tree t;
{
  if (TREE_CODE (t) != ADDR)
    {
      error ("make_CONSTANT_Environment: ߸ node.\n");
      return -1;
    }

  return cp_add (CONSTANT_Environment,
		 cp_add (CONSTANT_Host,
			 cp_add (CONSTANT_Utf8, IDENTIFIER_POINTER (TREE_EADDRESS_HOSTNAME_IP (t))),
			 cp_add (CONSTANT_Integer, DEFAULT_ESERVER_PORT)),
		 cp_add (CONSTANT_Sha1AndName, 
			 cp_add (CONSTANT_Sha1, TREE_EADDRESS_PATHANDSHA1_SHA1_BINARY (t)), 
			 cp_add (CONSTANT_Utf8, IDENTIFIER_POINTER (TREE_EADDRESS_PATHANDSHA1_ABSOLUTE_NAME (t)))));
}

int
calculate_access_perm (t)
     tree t;
{
  int r = 0, w = 0, x = 0;
  const char *p;

  if (TREE_CODE (t) != STRING_CST)
    {
      fprintf (stderr, "calculate_access_perm: ߸ node.\n");
      return -1;
    }

  p = TREE_STRING_POINTER (t);

  if (p[0] == 'r')
    r = 4;
  if (p[1] == 'w')
    w = 2;
  if (p[2] == 'x')
    x = 1;

  return r + w + x;
}

int
do_access_hosts (t)
     tree t;
{
  if (!t || (t && TREE_CODE (t) != ACCESS_EXPR))
    {
      error ("do_access_hosts: ߸ node.\n");
      return -1;
    }

  return cp_add (CONSTANT_Accessref,
		 cp_add (CONSTANT_Integer, calculate_access_perm (TREE_ACCESS_EXPR_FLAG (t))),
		 make_CONSTANT_Environment (TREE_ACCESS_EXPR_ADDRESS (t)));
}

/* Constant Pool  environment name  ,  Ѵ.
   츮 CONSTANT_Environment   , ؾ ϴµ, CONSTANT_Environment
    CONSTANT_Utf8  CONSTANT_Utf8  Ǿ ִ.  */
int
do_this_environment (str)
     char *str;
{
  int i, exist_host = 0;

  /* STR  host ̸ ִ ȮѴ.  */
  for (i=0; i<strlen (str); i++)
    {
      if (str[i] == ':')
	{
	  exist_host = i;
	  break;
	}
    }

  if (exist_host)
    {
      /* HOST ̸   */
      error ("do_this_environment: Host ̸  츦, óؾ Ѵ.\n");
    }
  else
    {
      char envpath[1024];
      
      if (package_path)
	{
	  strcpy (envpath, package_path);
	  strcat (envpath, str);
	}
      else
	strcpy (envpath, str);

      return cp_add (CONSTANT_Environment, 
		     cp_add (CONSTANT_Host, 
			     cp_add (CONSTANT_Utf8, strdup ("localhost")),
			     cp_add (CONSTANT_Integer, DEFAULT_ESERVER_PORT)),
		     cp_add (CONSTANT_Sha1AndName, 
			     cp_add (CONSTANT_Sha1This), cp_add (CONSTANT_Utf8, envpath)));
    }

  return -1;
}

enum attribute_type
  {
    CODE_attribute_type = 0x1
  };

int
count_attribute2 (t)
     tree t;
{
  int i = 0;

  if (!t)
    return 0;

  switch (TREE_CODE (t))
    {
    case TREE_LIST:
      i |= count_attribute2 (TREE_PURPOSE (t));
      i |= count_attribute2 (TREE_VALUE (t));
      i |= count_attribute2 (TREE_CHAIN (t));
      break;
    case IDENTIFIER_NODE:
      if (TREE_IS_RESERVED_INSN (t))
	i |= CODE_attribute_type;
      else
	error ("count_attribute2: INSN  ߸ .\n");
      break;
    case LIMIT:
      i |= count_attribute2 (TREE_CHAIN (t));
      break;
    case INSN:
      if (TREE_IS_RESERVED_INSN (TREE_INSN_CODE (t)))
	i |= CODE_attribute_type;
      else
	error ("count_attribute2: INSN  ߸ .\n");
      i |= count_attribute2 (TREE_CHAIN (t));
      break;
    case LABEL:
      i |= count_attribute2 (TREE_CHAIN (t));
      break;
    default:
      error ("count_attribute2: ߸ NODE.\n");
      debug_tree (t);
      break;
    }

  return i;
}

int
count_attribute (t)
     tree t;
{
  int count = 0, i, attr = count_attribute2 (t);

  for (i = 0; i < sizeof (int) * 8; i++)
    {
      if (attr & 0x1)
	count++;
      attr >>= 1;
    }

  //fprintf (stderr, "attribute count = %d\n", count);

  return count;
}

/*  Լ , ɾ ̸ ȯѴ.  */
int
count_insn_length (t)
     tree t;
{
  int i = 0;

  if (!t)
    return 0;

  switch (TREE_CODE (t))
    {
    case TREE_LIST:
      i += count_insn_length (TREE_PURPOSE (t));
      i += count_insn_length (TREE_VALUE (t));
      i += count_insn_length (TREE_CHAIN (t));
      break;
    case LIMIT:
      i += count_insn_length (TREE_CHAIN (t));
      break;
    case INSN:
      if (TREE_IS_RESERVED_INSN (TREE_INSN_CODE (t)))
	{
	  switch (TREE_BYTE_CODE (TREE_INSN_CODE (t)))
	    {
	    case _TABLESWITCH:
	    case _LOOKUPSWITCH:
	      error ("count_insn_length:    ġ.\n");
	      break;
	    default:
	      i += TREE_INSN_LENGTH (TREE_INSN_CODE (t));
	      break;
	    }
	}
      i += count_insn_length (TREE_CHAIN (t));
      break;
    case SWITCH_TABLE:
      error ("count_insn_length: óؾ ϴ SWITCH_TABLE.\n");
      debug_tree (t);
      break;
    case SWITCH_LOOKUP:
      i += TREE_SWITCH_LOOKUP_CODE_LENGTH (t);
      break;
    case LABEL:
      break;
    default:
      error ("count_insn_length: ߸ NODE.\n");
      debug_tree (t);
      break;
    }

  return i;
}

/*
 * Calculate size (in words) of a signature item.
 */
int
sizeof_type_char (ch, want_wide_refs)
     char ch;
     int want_wide_refs;
{
  switch (ch)
    {
    case 'V':
      return 0;
    case 'I':
    case 'Z':
    case 'S':
    case 'B':
    case 'C':
    case 'F':
      return 1;
      break;
    case 'D':
    case 'J':
      return 2;
      break;
    case '[':
    case 'L':
      return want_wide_refs ? sizeof (void*) / sizeof(int) : 1;
    }

  return -1;
}

/*
 * Calculate size (in words) of a signature item and move *strp so
 * that it points to the next element of the signature
 */
int
sizeof_type_item (const char** strp, int want_wide_refs)
{
  int count;
  const char* str;
  
  for (str = *strp; ; str++) 
    {
      count = sizeof_type_char (*str, want_wide_refs);
      if (count == -1) 
	{
	  switch (*str) 
	    {
	    case '(':
	      continue;
	    case 0:
	    case ')':
	      break;
	    default:
	      count = -1;	/* avoid compiler warning */
	      abort ();
	    }
	} 
      else 
	{
	  while (*str == '[')
	    ++str;
	  if (*str == 'L') 
	    {
	      while (*str != ';') 
		str++;
	    }
	}
      
      *strp = str + 1;
      return (count);
    }
}

/*
 * Calculate size of data item based on signature and move *strp so
 * that it points to the next element of the signature.
 */
int
sizeof_type (const char** strp, int want_wide_refs)
{
  int count;
  int c;
  
  count = 0;
  for (;;) 
    {
      c = sizeof_type_item (strp, want_wide_refs);
      if (c == -1)
	return (count);

      count += c;
    }

  return -1;
}

/* Determine the number of arguments and return values from the
   method signature.  */
void
count_ins_and_outs (const char* str, short* ins, short* outs, char* outtype)
{
  *ins = sizeof_type (&str, 0);
  *outtype = str[0];
  *outs = sizeof_type (&str, 0);
}

/*
 * Count the number of arguments in a signature.
 */
int
count_args_in_type (type)
     const char *type;
{
  int nargs = 0;
  
  while (sizeof_type_item (&type, 0) != -1) 
    ++nargs;

  return (nargs);
}

#define MAX_LOCAL_VARIABLES		256

void 
count_max_locals_part2 (t, local)
     tree t;
     int *local;
{
  if (!t)
    return;

  switch (TREE_CODE (t))
    {
    case TREE_LIST:
      count_max_locals_part2 (TREE_PURPOSE (t), local);
      count_max_locals_part2 (TREE_VALUE (t), local);
      count_max_locals_part2 (TREE_CHAIN (t), local);
      break;
    case LIMIT:
      count_max_locals_part2 (TREE_CHAIN (t), local);
      break;
    case INSN:
      if (TREE_IS_RESERVED_INSN (TREE_INSN_CODE (t)))
	{
	  switch (TREE_BYTE_CODE (TREE_INSN_CODE (t)))
	    {
	    case _ALOAD_0:
	    case _ILOAD_0:
	    case _LLOAD_0:	/* opnum = 30 */
	    case _FLOAD_0:	/* opnum = 34 */
	    case _DLOAD_0:	/* opnum = 38 */
	    case _ISTORE_0:
	    case _LSTORE_0:	/* opnum = 63 */
	    case _FSTORE_0:	/* opnum = 67 */
	    case _DSTORE_0:	/* opnum = 71 */
	      local[0] = 1;
	      break;
	    case _ALOAD_1:
	    case _ILOAD_1:
	    case _LLOAD_1:	/* opnum = 31 */
	    case _FLOAD_1:	/* opnum = 35 */
	    case _DLOAD_1:	/* opnum = 39 */
	    case _ISTORE_1:
	    case _LSTORE_1:	/* opnum = 64 */
	    case _FSTORE_1:	/* opnum = 68 */
	    case _DSTORE_1:	/* opnum = 72 */
	      local[1] = 1;
	      break;
	    case _ALOAD_2:
	    case _ILOAD_2:
	    case _LLOAD_2:	/* opnum = 32 */
	    case _FLOAD_2:	/* opnum = 36 */
	    case _DLOAD_2:	/* opnum = 40 */
	    case _ISTORE_2:
	    case _LSTORE_2:	/* opnum = 65 */
	    case _FSTORE_2:	/* opnum = 69 */
	    case _DSTORE_2:	/* opnum = 73 */
	      local[2] = 1;
	      break;
	    case _ALOAD_3:
	    case _ILOAD_3:
	    case _LLOAD_3:	/* opnum = 33 */
	    case _FLOAD_3:	/* opnum = 37 */
	    case _DLOAD_3:	/* opnum = 41 */
	    case _ISTORE_3:
	    case _LSTORE_3:	/* opnum = 66 */
	    case _FSTORE_3:	/* opnum = 70 */
	    case _DSTORE_3:	/* opnum = 74 */
	      local[3] = 1;
	      break;
	    case _ILOAD:	/* opnum = 21 */
	    case _LLOAD:	/* opnum = 22 */
	    case _FLOAD:	/* opnum = 23 */
	    case _DLOAD:	/* opnum = 24 */
	    case _ALOAD:	/* opnum = 25 */
	    case _ISTORE:	/* opnum = 54 */
	    case _LSTORE:	/* opnum = 55 */
	    case _FSTORE:	/* opnum = 56 */
	    case _DSTORE:	/* opnum = 57 */
	    case _ASTORE:	/* opnum = 58 */
	    case _IINC:		/* opnum = 132 */
	    case _RET:		/* opnum = 169 */
	      if (TREE_CODE (TREE_INSN_OP1 (t)) != LOCALS)
		{
		  error ("op1  LOCALS node  ƴմϴ.\n");
		  debug_tree (TREE_INSN_OP1 (t));
		}
	      if (TREE_CODE (TREE_LOCAL_CONSTANT (TREE_INSN_OP1 (t))) != INTEGER_CST)
		error ("op1  constant  INTEGER_CST  ƴմϴ.\n");
	      if (TREE_INT_CST_LOW (TREE_LOCAL_CONSTANT (TREE_INSN_OP1 (t))) < 0
		  || TREE_INT_CST_LOW (TREE_LOCAL_CONSTANT (TREE_INSN_OP1 (t))) > 255)
		error ("op1  constant  256  Ůϴ.\n");

	      local[TREE_INT_CST_LOW (TREE_LOCAL_CONSTANT (TREE_INSN_OP1 (t)))] = 1;
	      break;
	    case _NOP:
	    case _ACONST_NULL:	/* opnum = 0x1 */
	    case _ICONST_M1:
	    case _ICONST_0:
	    case _ICONST_1:
	    case _ICONST_2:
	    case _ICONST_3:
	    case _ICONST_4:
	    case _ICONST_5:
	    case _LCONST_0:	/* opnum = 9 */
	    case _LCONST_1:	/* opnum = 10 */
	    case _FCONST_0:	/* opnum = 11 */
	    case _FCONST_1:	/* opnum = 12 */
	    case _FCONST_2:	/* opnum = 13 */
	    case _DCONST_0:	/* opnum = 14 */
	    case _DCONST_1:	/* opnum = 15 */
	    case _BIPUSH:	/* opnum = 16 */
	    case _SIPUSH:	/* opnum = 17 */
	    case _LDC:		/* opnum = 18 */
	    case _LDC_W:	/* opnum = 19 */
	    case _LDC2_W:	/* opnum = 20 */
	    case _IALOAD:	/* opnum = 46 */
	    case _LALOAD:	/* opnum = 47 */
	    case _FALOAD:	/* opnum = 48 */
	    case _DALOAD:	/* opnum = 49 */
	    case _AALOAD:	/* opnum = 50 */
	    case _BALOAD:	/* opnum = 51 */
	    case _CALOAD:	/* opnum = 52 */
	    case _SALOAD:	/* opnum = 53 */
	    case _IASTORE:	/* opnum = 79 */
	    case _LASTORE:	/* opnum = 80 */
	    case _FASTORE:	/* opnum = 81 */
	    case _DASTORE:	/* opnum = 82 */
	    case _AASTORE:	/* opnum = 83 */
	    case _BASTORE:	/* opnum = 84 */
	    case _CASTORE:	/* opnum = 85 */
	    case _SASTORE:	/* opnum = 86 */
	    case _POP:		/* opnum = 87 */
	    case _POP2:		/* opnum = 88 */
	    case _DUP:		/* opnum = 89 */
	    case _DUP_X1:	/* opnum = 90 */
	    case _DUP_X2:	/* opnum = 91 */
	    case _DUP2:		/* opnum = 92 */
	    case _DUP2_X1:	/* opnum = 93 */
	    case _DUP2_X2:	/* opnum = 94 */
	    case _SWAP:		/* opnum = 95 */
	    case _IADD:		/* opnum = 96 */
	    case _LADD:		/* opnum = 97 */
	    case _FADD:		/* opnum = 98 */
	    case _DADD:		/* opnum = 99 */
	    case _ISUB:		/* opnum = 100 */
	    case _LSUB:		/* opnum = 101 */
	    case _FSUB:		/* opnum = 102 */
	    case _DSUB:		/* opnum = 103 */
	    case _IMUL:		/* opnum = 104 */
	    case _LMUL:		/* opnum = 105 */
	    case _FMUL:		/* opnum = 106 */
	    case _DMUL:		/* opnum = 107 */
	    case _IDIV:		/* opnum = 108 */
	    case _LDIV:		/* opnum = 109 */
	    case _FDIV:		/* opnum = 110 */
	    case _DDIV:		/* opnum = 111 */
	    case _IREM:		/* opnum = 112 */
	    case _LREM:		/* opnum = 113 */
	    case _FREM:		/* opnum = 114 */
	    case _DREM:		/* opnum = 115 */
	    case _INEG:		/* opnum = 116 */
	    case _LNEG:		/* opnum = 117 */
	    case _FNEG:		/* opnum = 118 */
	    case _DNEG:		/* opnum = 119 */
	    case _ISHL:		/* opnum = 120 */
	    case _LSHL:		/* opnum = 121 */
	    case _ISHR:		/* opnum = 122 */
	    case _LSHR:		/* opnum = 123 */	      
	    case _IUSHR:	/* opnum = 124 */
	    case _LUSHR:	/* opnum = 125 */
	    case _IAND:		/* opnum = 126 */
	    case _LAND:		/* opnum = 127 */
	    case _IOR:		/* opnum = 128 */
	    case _LOR:		/* opnum = 129 */
	    case _IXOR:		/* opnum = 130 */
	    case _LXOR:		/* opnum = 131 */
	    case _I2L:		/* opnum = 133 */
	    case _I2F:		/* opnum = 134 */
	    case _I2D:		/* opnum = 135 */
	    case _L2I:		/* opnum = 136 */
	    case _L2F:		/* opnum = 137 */
	    case _L2D:		/* opnum = 138 */
	    case _F2I:		/* opnum = 139 */
	    case _F2L:		/* opnum = 140 */
	    case _F2D:		/* opnum = 141 */
	    case _D2I:		/* opnum = 142 */
	    case _D2L:		/* opnum = 143 */
	    case _D2F:		/* opnum = 144 */
	    case _I2B:		/* opnum = 145 */
	    case _I2C:		/* opnum = 146 */
	    case _I2S:		/* opnum = 147 */
	    case _LCMP:		/* opnum = 148 */
	    case _FCMPL:	/* opnum = 149 */
	    case _FCMPG:	/* opnum = 150 */
	    case _DCMPL:	/* opnum = 151 */
	    case _DCMPG:	/* opnum = 152 */
	    case _IFEQ:		/* opnum = 153 */
	    case _IFNE:		/* opnum = 154 */
	    case _IFLT:		/* opnum = 155 */
	    case _IFGE:		/* opnum = 156 */
	    case _IFGT:		/* opnum = 157 */
	    case _IFLE:		/* opnum = 158 */
	    case _IF_ICMPEQ:	/* opnum = 159 */
	    case _IF_ICMPNE:	/* opnum = 160 */
	    case _IF_ICMPLT:	/* opnum = 161 */
	    case _IF_ICMPGE:	/* opnum = 162 */
	    case _IF_ICMPGT:	/* opnum = 163 */
	    case _IF_ICMPLE:	/* opnum = 164 */
	    case _IF_ACMPEQ:	/* opnum = 165 */
	    case _IF_ACMPNE:	/* opnum = 166 */
	    case _GOTO:		/* opnum = 167 */
	    case _JSR:		/* opnum = 168 */
	    case _JSR_W:	/* opnum = 201 */
	    case _NEWARRAY:	/* opnum = 188 */
	    case _IRETURN:	/* opnum = 172 */
	    case _LRETURN:	/* opnum = 173 */
	    case _FRETURN:	/* opnum = 174 */
	    case _DRETURN:	/* opnum = 175 */
	    case _ARETURN:	/* opnum = 176 */
	    case _RETURN:	/* opnum = 177 */
	    case _GETSTATIC:	/* opnum = 178 */
	    case _PUTSTATIC:	/* opnum = 179 */
	    case _GETFIELD:	/* opnum = 180 */
	    case _PUTFIELD:	/* opnum = 181 */
	    case _INVOKEVIRTUAL:	/* opnum = 182 */
	    case _INVOKESPECIAL:	/* opnum = 183 */
	    case _INVOKESTATIC:		/* opnum = 184 */
	      break;
	      /* TODO: DUP ø  computational type  ؾ ϴ ͵ ؾ Ѵ.  */
	    default:
	      warning ("count_max_locals_part2: code [%s] update ʿ.\n", IDENTIFIER_POINTER (t));
	      break;
	    }
	}
      else
	error ("count_max_locals_part2: INSN  ߸ .\n");

      count_max_locals_part2 (TREE_CHAIN (t), local);
      break;
    case SWITCH_LOOKUP:
    case SWITCH_TABLE:
    case LABEL:
      break;
    default:
      error ("count_max_locals_part2: ߸ NODE.\n");
      debug_tree (t);
      break;
    }
}

int
count_max_locals (def, stmt)
     tree def, stmt;
{
  int count=0, tmp, i, max = 0;
  int *p;
  int local[MAX_LOCAL_VARIABLES];

  p = local;
  memset (local, '\0', sizeof (int) * MAX_LOCAL_VARIABLES);

  if (!(TREE_METHOD_ACCESS (def) & ACC_STATIC))
    local[count++] = 1;

  tmp = count_args_in_type (IDENTIFIER_POINTER (TREE_METHOD_TYPE (def)));
  for (i=0; i<tmp; i++)
    local[count++] = 1;

  count_max_locals_part2 (stmt, p);

  for (i=0; i<MAX_LOCAL_VARIABLES; i++)
    {
      if (local[i] == 1)
	max = i + 1;
    }

  return max;
}

/*      +--------------+-----------------------+----------+
	| Actual Type  | Computational Type    | Category |
	+--------------+-----------------------+----------+
	 boolean 	int			category 1
	 byte		int			category 1
	 char		int			category 1
	 short		int			category 1
	 int		int			category 1
	 float		float			category 1
	 reference 	reference		category 1
	 returnAddress 	returnAddress		category 1
	 long		long			category 2
	 double		double			category 2
*/
enum vstack_value_type
  {
    STACK_UNKNOWN	= 0x0,
    STACK_BOOLEAN	= 0x1,
    STACK_BYTE		= 0x2,
    STACK_CHAR		= 0x4,
    STACK_SHORT		= 0x8,
    STACK_INTEGER	= 0x10,
    STACK_FLOAT		= 0x20,
    STACK_LONG		= 0x40,
    STACK_DOUBLE	= 0x80,
    STACK_ARRAYREF	= 0x100,
    STACK_OBJECTREF	= 0x200,
    STACK_RETURNADDRESS = 0x400
  };

#define STACK_ALL	(STACK_BOOLEAN | STACK_BYTE | STACK_CHAR | STACK_SHORT | \
                         STACK_INTEGER | STACK_FLOAT | STACK_LONG | STACK_DOUBLE | \
                         STACK_ARRAYREF | STACK_OBJECTREF)

#define STACK_COMPUTATIONAL_CATEGORY1	\
                        (STACK_BOOLEAN | STACK_BYTE | STACK_CHAR | STACK_SHORT | \
                         STACK_INTEGER | STACK_FLOAT | STACK_ARRAYREF | STACK_OBJECTREF)

#define STACK_COMPUTATIONAL_CATEGORY2	(STACK_LONG | STACK_DOUBLE)

#define VSTACK_TYPE(NODE)		((NODE)->type)
#define VSTACK_NEXT(NODE)		((NODE)->next)

struct vstack
{
  int index;
  enum vstack_value_type type;

  struct vstack *next;
};

int vstack_index;
struct vstack *vstack_top;

void
push_vstack (type)
     enum vstack_value_type type;
{
  struct vstack *tmp;
  
  tmp = (struct vstack *) malloc (sizeof (struct vstack));
  tmp->index = vstack_index++;
  tmp->type = type;
  tmp->next = vstack_top;

  vstack_top = tmp;
}

void
push_vstack_field_type_char (ch)
     char ch;
{
  switch (ch)
    {
    case 'I':
      push_vstack (STACK_INTEGER);
      break;
    case 'Z':
      push_vstack (STACK_BOOLEAN);
      break;
    case 'S':
      push_vstack (STACK_SHORT);
      break;
    case 'B':
      push_vstack (STACK_BYTE);
      break;
    case 'C':
      push_vstack (STACK_CHAR);
      break;
    case 'F':
      push_vstack (STACK_FLOAT);
      break;
    case 'D':
      push_vstack (STACK_DOUBLE);
      break;
    case 'J':
      push_vstack (STACK_LONG);
      break;
    case '[':
      push_vstack (STACK_ARRAYREF);
      break;
    case 'L':
      push_vstack (STACK_OBJECTREF);
      break;
    default:
      error ("push_vstack_field_type_char: ߸ field type.\n");
      break;
    }
}

void
pop_vstack (type)
     enum vstack_value_type type;	/*   */
{
  //struct vstack **tmp = &vstack_top;
  
  if (vstack_top)
    {
      vstack_top = vstack_top->next;
      vstack_index--;
    }
}

/* Constant Pool  index ȣ IDX   ã ȯѴ.  */
struct constant_pool *
get_cp_with_index (root, idx)
     struct constant_pool *root;
     int idx;
{
  struct constant_pool *tm = root;

  while (tm)
    {
      if (CONSTANT_INDEX (tm) == idx)
	return tm;

      tm = CONSTANT_NEXT (tm);
    }

  return NULL;
}

/* T  statements   TREE node ̸, ̸ ̿Ͽ  method    ִ
   ִ stack  Ѵ.  */
int 
count_max_stack (t, max)
     tree t;
     int max;
{
  int i = max;

  if (!t)
    return 0;

  switch (TREE_CODE (t))
    {
    case TREE_LIST:
      i += count_max_stack (TREE_PURPOSE (t), max);
      i += count_max_stack (TREE_VALUE (t), max);
      i += count_max_stack (TREE_CHAIN (t), max);
      break;
    case LIMIT:
      i += count_max_stack (TREE_CHAIN (t), max);
      break;
    case INSN:
      if (TREE_IS_RESERVED_INSN (TREE_INSN_CODE (t)))
	{
	  switch (TREE_BYTE_CODE (TREE_INSN_CODE (t)))
	    {
	    case _INVOKEVIRTUAL:	/* opnum = 182 */
	    case _INVOKESPECIAL:	/* opnum = 183 */
	      i -= count_args_in_type (IDENTIFIER_POINTER (TREE_METHODREF_TYPE (TREE_INSN_OP1 (t))));
	      /* ߰ objectref  pop ȴ.  */
	      pop_vstack (STACK_OBJECTREF);
	      i += -1;
	      break;
	    case _INVOKESTATIC:		/* opnum = 184 */
	      i -= count_args_in_type (IDENTIFIER_POINTER (TREE_METHODREF_TYPE (TREE_INSN_OP1 (t))));
	      break;
	      /* TODO: DUP ø  computational type  ؾ ϴ ͵ ؾ Ѵ.  */
	    case _ICONST_0:
	    case _ICONST_1:
	    case _ICONST_2:
	    case _ICONST_3:
	    case _ICONST_4:
	    case _ICONST_5:
	    case _ILOAD_0:
	    case _ILOAD_1:
	    case _ILOAD_2:
	    case _ILOAD_3:
	    case _ILOAD:
	      push_vstack (STACK_INTEGER);
	      i += 1;
	      break;
	    case _ACONST_NULL:	/* opnum = 0x1 */
	    case _ALOAD_0:
	    case _ALOAD_1:
	    case _ALOAD_2:
	    case _ALOAD_3:
	    case _ALOAD:
	      push_vstack (STACK_OBJECTREF);
	      i += 1;
	      break;
	    case _LCONST_0:	/* opnum = 9 */
	    case _LCONST_1:	/* opnum = 10 */
	    case _LLOAD_0:	/* opnum = 30 */
	    case _LLOAD_1:	/* opnum = 31 */
	    case _LLOAD_2:	/* opnum = 32 */
	    case _LLOAD_3:	/* opnum = 33 */
	    case _LLOAD:
	      push_vstack (STACK_LONG);
	      i += 2;
	      break;
	    case _FCONST_0:	/* opnum = 11 */
	    case _FCONST_1:	/* opnum = 12 */
	    case _FCONST_2:	/* opnum = 13 */
	    case _FLOAD:	/* opnum = 23 */
	    case _FLOAD_0:	/* opnum = 34 */
	    case _FLOAD_1:	/* opnum = 35 */
	    case _FLOAD_2:	/* opnum = 36 */
	    case _FLOAD_3:	/* opnum = 37 */
	      push_vstack (STACK_FLOAT);
	      i += 1;
	      break;
	    case _DCONST_0:	/* opnum = 14 */
	    case _DCONST_1:	/* opnum = 15 */
	    case _DLOAD:	/* opnum = 24 */
	    case _DLOAD_0:	/* opnum = 38 */
	    case _DLOAD_1:	/* opnum = 39 */
	    case _DLOAD_2:	/* opnum = 40 */
	    case _DLOAD_3:	/* opnum = 41 */
	      push_vstack (STACK_DOUBLE);
	      i += 2;
	      break;
	    case _BIPUSH:	/* opnum = 16 */
	      push_vstack (STACK_BYTE);
	      i += 1;
	      break;
	    case _SIPUSH:	/* opnum = 17 */
	      push_vstack (STACK_SHORT);
	      i += 1;
	      break;
	    case _LDC:		/* opnum = 18 */
	    case _LDC_W:	/* opnum = 19 */
	    case _LDC2_W:	/* opnum = 19 */
	      {
		struct constant_pool *tm = get_cp_with_index (root_cp, TREE_CONSTANT_POOL_NUMBER (t));

		switch (CONSTANT_TYPE (tm))
		  {
		  case CONSTANT_String:
		    push_vstack (STACK_OBJECTREF);
		    i += 1;
		    break;
		  case CONSTANT_Integer:
		    push_vstack (STACK_INTEGER);
		    i += 1;
		    break;
		  case CONSTANT_Float:
		    push_vstack (STACK_FLOAT);
		    i += 1;
		    break;
		  case CONSTANT_Long:
		    if (TREE_BYTE_CODE (TREE_INSN_CODE (t)) != _LDC2_W)
		      error ("count_max_stack: ߸ LDC2_W ɾ.\n");
		    push_vstack (STACK_LONG);
		    i += 2;
		  case CONSTANT_Double:
		    if (TREE_BYTE_CODE (TREE_INSN_CODE (t)) != _LDC2_W)
		      error ("count_max_stack: ߸ LDC2_W ɾ.\n");
		    push_vstack (STACK_DOUBLE);
		    i += 2;
		    break;
		  default:
		    warning ("count_max_stack: óؾ LDC   . type = %d\n", CONSTANT_TYPE (tm));
		  }
		break;
	      }
	    case _IALOAD:	/* opnum = 46 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_ARRAYREF);
	      push_vstack (STACK_INTEGER);
	      i += -1;
	      break;
	    case _LALOAD:	/* opnum = 47 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_ARRAYREF);
	      push_vstack (STACK_LONG);
	      break;
	    case _FALOAD:	/* opnum = 48 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_ARRAYREF);
	      push_vstack (STACK_FLOAT);
	      i += -1;
	      break;
	    case _DALOAD:	/* opnum = 49 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_ARRAYREF);
	      push_vstack (STACK_DOUBLE);
	      break;
	    case _AALOAD:	/* opnum = 50 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_ARRAYREF);
	      push_vstack (STACK_OBJECTREF|STACK_ARRAYREF);
	      i += -1;
	      break;
	    case _BALOAD:	/* opnum = 51 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_ARRAYREF);
	      push_vstack (STACK_BYTE|STACK_BOOLEAN);
	      i += -1;
	      break;
	    case _CALOAD:	/* opnum = 52 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_ARRAYREF);
	      push_vstack (STACK_CHAR);
	      i += -1;
	      break;
	    case _SALOAD:	/* opnum = 53 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_ARRAYREF);
	      push_vstack (STACK_SHORT);
	      i += -1;
	      break;
	    case _ISTORE:	/* opnum = 54 */
	    case _ISTORE_0:
	    case _ISTORE_1:
	    case _ISTORE_2:
	    case _ISTORE_3:
	      pop_vstack (STACK_INTEGER);
	      i += -1;
	      break;
	    case _LSTORE:	/* opnum = 55 */
	    case _LSTORE_0:	/* opnum = 63 */
	    case _LSTORE_1:	/* opnum = 64 */
	    case _LSTORE_2:	/* opnum = 65 */
	    case _LSTORE_3:	/* opnum = 66 */
	      pop_vstack (STACK_LONG);
	      i += -2;
	      break;
	    case _FSTORE:	/* opnum = 56 */
	    case _FSTORE_0:	/* opnum = 67 */
	    case _FSTORE_1:	/* opnum = 68 */
	    case _FSTORE_2:	/* opnum = 69 */
	    case _FSTORE_3:	/* opnum = 70 */
	      pop_vstack (STACK_FLOAT);
	      i += -1;
	      break;
	    case _DSTORE:	/* opnum = 57 */
	    case _DSTORE_0:	/* opnum = 71 */
	    case _DSTORE_1:	/* opnum = 72 */
	    case _DSTORE_2:	/* opnum = 73 */
	    case _DSTORE_3:	/* opnum = 74 */	      
	      pop_vstack (STACK_DOUBLE);
	      i += -2;
	      break;
	    case _ASTORE:	/* opnum = 58 */
	    case _ASTORE_0:	/* opnum = 75 */
	    case _ASTORE_1:	/* opnum = 76 */
	    case _ASTORE_2:	/* opnum = 77 */
	    case _ASTORE_3:	/* opnum = 78 */
	      pop_vstack (STACK_OBJECTREF);
	      i += -1;
	      break;
	    case _IASTORE:	/* opnum = 79 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_ARRAYREF);
	      pop_vstack (STACK_INTEGER);
	      i += -3;
	      break;
	    case _LASTORE:	/* opnum = 80 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_ARRAYREF);
	      pop_vstack (STACK_LONG);
	      i += -4;
	      break;
	    case _FASTORE:	/* opnum = 81 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_ARRAYREF);
	      pop_vstack (STACK_FLOAT);
	      i += -3;
	      break;
	    case _DASTORE:	/* opnum = 82 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_ARRAYREF);
	      pop_vstack (STACK_DOUBLE);
	      i += -4;
	      break;
	    case _AASTORE:	/* opnum = 83 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_ARRAYREF);
	      pop_vstack (STACK_OBJECTREF|STACK_ARRAYREF);
	      i += -3;
	      break;
	    case _BASTORE:	/* opnum = 84 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_ARRAYREF);
	      pop_vstack (STACK_BYTE|STACK_BOOLEAN);
	      i += -3;
	      break;
	    case _CASTORE:	/* opnum = 85 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_ARRAYREF);
	      pop_vstack (STACK_CHAR);
	      i += -3;
	      break;
	    case _SASTORE:	/* opnum = 86 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_ARRAYREF);
	      pop_vstack (STACK_SHORT);
	      i += -3;
	      break;
	    case _POP:		/* opnum = 87 */
	      pop_vstack (STACK_ALL);
	      i += -1;
	      break;
	    case _POP2:		/* opnum = 88 */
	      if (!vstack_top)
		error ("vstack  NULL Դϴ.");
	      if (VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY2)
		{
		  pop_vstack (STACK_COMPUTATIONAL_CATEGORY2);
		  i += -1;
		}
	      else
		{
		  if (!VSTACK_NEXT (vstack_top))
		    error ("vstack  next  NULL Դϴ.");
		  if (!(VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY1)
		      || !(VSTACK_TYPE (VSTACK_NEXT (vstack_top)) & STACK_COMPUTATIONAL_CATEGORY1))
		    error ("pop2   operand stack  computational type  category1   ʽϴ.");
		  pop_vstack (STACK_COMPUTATIONAL_CATEGORY1);
		  pop_vstack (STACK_COMPUTATIONAL_CATEGORY1);
		  i += -2;
		}
	      break;
	    case _DUP:		/* opnum = 89 */
	      if (VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY2)
		error ("dup ɾ  stack  COMPUTATIONAL CATEGORY1  ƴմϴ.\n");
	      push_vstack (VSTACK_TYPE (vstack_top));
	      i += 1;
	      break;
	    case _DUP_X1:	/* opnum = 90 */
	      {
		enum vstack_value_type t0, t1;

		if (!vstack_top)
		  error ("vstack  NULL Դϴ.");
		if (!VSTACK_NEXT (vstack_top))
		  error ("vstack  next  NULL Դϴ.");
		if (!(VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY1)
		    || !(VSTACK_TYPE (VSTACK_NEXT (vstack_top)) & STACK_COMPUTATIONAL_CATEGORY1))
		  error ("pop2   operand stack  computational type  category1   ʽϴ.");
		
		t0 = VSTACK_TYPE (vstack_top);
		t1 = VSTACK_TYPE (VSTACK_NEXT (vstack_top));
		VSTACK_TYPE (vstack_top) = t1;
		VSTACK_TYPE (VSTACK_NEXT (vstack_top))= t0;

		push_vstack (t0);
		i += 1;
		break;
	      }
	    case _DUP_X2:	/* opnum = 91 */
	      if (!vstack_top)
		error ("vstack  NULL Դϴ.");
	      if (!VSTACK_NEXT (vstack_top))
		error ("vstack  next  NULL Դϴ.");
	      
	      if ((VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY1)
		  && (VSTACK_TYPE (VSTACK_NEXT (vstack_top)) & STACK_COMPUTATIONAL_CATEGORY2))
		{
		  /* ..., value2, value1 => ..., value1, value2, value1 */
		  enum vstack_value_type t0, t1;

		  t0 = VSTACK_TYPE (vstack_top);
		  t1 = VSTACK_TYPE (VSTACK_NEXT (vstack_top));
		  pop_vstack (t0);
		  pop_vstack (t1);
		  push_vstack (t0);
		  push_vstack (t1);
		  push_vstack (t0);
		  i += 1;
		}
	      else if ((VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY1)
		       && (VSTACK_TYPE (VSTACK_NEXT (vstack_top)) & STACK_COMPUTATIONAL_CATEGORY1)
		       && (VSTACK_TYPE (VSTACK_NEXT (VSTACK_NEXT (vstack_top))) & STACK_COMPUTATIONAL_CATEGORY2))
		{
		  /* ..., value3, value2, value1 => ..., value1, value3, value2, value1 */
		  enum vstack_value_type t0, t1, t2;

		  t0 = VSTACK_TYPE (vstack_top);
		  t1 = VSTACK_TYPE (VSTACK_NEXT (vstack_top));
		  t2 = VSTACK_TYPE (VSTACK_NEXT (VSTACK_NEXT (vstack_top)));

		  pop_vstack (t0);
		  pop_vstack (t1);
		  pop_vstack (t2);
		  push_vstack (t0);
		  push_vstack (t2);
		  push_vstack (t1);
		  push_vstack (t0);
		  i += 1;
		}
	      else
		error ("DUP_X2  ߸ Դϴ.\n");
	      break;
	    case _DUP2:		/* opnum = 92 */
	      if (VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY2)
		{
		  /* ..., value => ..., value, value */
		  push_vstack (VSTACK_TYPE (vstack_top));
		  i += 1;
		}
	      else if ((VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY1)
		       && (VSTACK_TYPE (VSTACK_NEXT (vstack_top)) & STACK_COMPUTATIONAL_CATEGORY1))
		{
		  push_vstack (VSTACK_TYPE (VSTACK_NEXT (vstack_top)));
		  push_vstack (VSTACK_TYPE (vstack_top));
		  i += 2;
		}
	      else
		error ("DUP2  ߸ Դϴ.\n");
	      break;
	    case _DUP2_X1:	/* opnum = 93 */
	      if ((VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY2)
		  && (VSTACK_TYPE (VSTACK_NEXT (vstack_top)) & STACK_COMPUTATIONAL_CATEGORY1))
		{
		  enum vstack_value_type t0, t1;

		  t0 = VSTACK_TYPE (vstack_top);
		  t1 = VSTACK_TYPE (VSTACK_NEXT (vstack_top));
		  pop_vstack (t0);
		  pop_vstack (t1);
		  push_vstack (t0);
		  push_vstack (t1);
		  push_vstack (t0);
		  i += 1;
		}
	      else if ((VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY1)
		       && (VSTACK_TYPE (VSTACK_NEXT (vstack_top)) & STACK_COMPUTATIONAL_CATEGORY1)
		       && (VSTACK_TYPE (VSTACK_NEXT (VSTACK_NEXT (vstack_top))) & STACK_COMPUTATIONAL_CATEGORY1))
		{
		  /* ..., value3, value2, value1 => ..., value1, value3, value2, value1 */
		  enum vstack_value_type t0, t1, t2;

		  t0 = VSTACK_TYPE (vstack_top);
		  t1 = VSTACK_TYPE (VSTACK_NEXT (vstack_top));
		  t2 = VSTACK_TYPE (VSTACK_NEXT (VSTACK_NEXT (vstack_top)));

		  pop_vstack (t0);
		  pop_vstack (t1);
		  pop_vstack (t2);
		  push_vstack (t1);
		  push_vstack (t0);
		  push_vstack (t2);
		  push_vstack (t1);
		  push_vstack (t0);
		  i += 2;
		}
	      else
		error ("DUP2_X1  ߸ Դϴ.\n");
	      break;
	    case _DUP2_X2:	/* opnum = 94 */
	      if ((VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY2)
		  && (VSTACK_TYPE (VSTACK_NEXT (vstack_top)) & STACK_COMPUTATIONAL_CATEGORY2))
		{
		  /* form 4 */
		  enum vstack_value_type t0, t1;

		  t0 = VSTACK_TYPE (vstack_top);
		  t1 = VSTACK_TYPE (VSTACK_NEXT (vstack_top));
		  pop_vstack (t0);
		  pop_vstack (t1);
		  push_vstack (t0);
		  push_vstack (t1);
		  push_vstack (t0);
		  i += 1;
		}
	      else if ((VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY1)
		       && (VSTACK_TYPE (VSTACK_NEXT (vstack_top)) & STACK_COMPUTATIONAL_CATEGORY1)
		       && (VSTACK_TYPE (VSTACK_NEXT (VSTACK_NEXT (vstack_top))) & STACK_COMPUTATIONAL_CATEGORY2))
		{
		  /* form 3 */
		  enum vstack_value_type t0, t1, t2;

		  t0 = VSTACK_TYPE (vstack_top);
		  t1 = VSTACK_TYPE (VSTACK_NEXT (vstack_top));
		  t2 = VSTACK_TYPE (VSTACK_NEXT (VSTACK_NEXT (vstack_top)));

		  pop_vstack (t0);
		  pop_vstack (t1);
		  pop_vstack (t2);
		  push_vstack (t1);
		  push_vstack (t0);
		  push_vstack (t2);
		  push_vstack (t1);
		  push_vstack (t0);
		  i += 2;
		}
	      else if ((VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY2)
		       && (VSTACK_TYPE (VSTACK_NEXT (vstack_top)) & STACK_COMPUTATIONAL_CATEGORY1)
		       && (VSTACK_TYPE (VSTACK_NEXT (VSTACK_NEXT (vstack_top))) & STACK_COMPUTATIONAL_CATEGORY1))
		{
		  /* form 2 */
		  enum vstack_value_type t0, t1, t2;

		  t0 = VSTACK_TYPE (vstack_top);
		  t1 = VSTACK_TYPE (VSTACK_NEXT (vstack_top));
		  t2 = VSTACK_TYPE (VSTACK_NEXT (VSTACK_NEXT (vstack_top)));

		  pop_vstack (t0);
		  pop_vstack (t1);
		  pop_vstack (t2);
		  push_vstack (t0);
		  push_vstack (t2);
		  push_vstack (t1);
		  push_vstack (t0);
		  i += 1;
		}
	      else if ((VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY1)
		       && (VSTACK_TYPE (VSTACK_NEXT (vstack_top)) & STACK_COMPUTATIONAL_CATEGORY1)
		       && (VSTACK_TYPE (VSTACK_NEXT (VSTACK_NEXT (vstack_top))) & STACK_COMPUTATIONAL_CATEGORY1)
		       && (VSTACK_TYPE (VSTACK_NEXT (VSTACK_NEXT (VSTACK_NEXT (vstack_top)))) & STACK_COMPUTATIONAL_CATEGORY1))
		{
		  /* form 1 */
		  enum vstack_value_type t0, t1, t2, t3;

		  t0 = VSTACK_TYPE (vstack_top);
		  t1 = VSTACK_TYPE (VSTACK_NEXT (vstack_top));
		  t2 = VSTACK_TYPE (VSTACK_NEXT (VSTACK_NEXT (vstack_top)));
		  t3 = VSTACK_TYPE (VSTACK_NEXT (VSTACK_NEXT (VSTACK_NEXT (vstack_top))));

		  pop_vstack (t0);
		  pop_vstack (t1);
		  pop_vstack (t2);
		  pop_vstack (t3);
		  push_vstack (t1);
		  push_vstack (t0);
		  push_vstack (t3);
		  push_vstack (t2);
		  push_vstack (t1);
		  push_vstack (t0);
		  i += 2;
		}
	      else
		error ("DUP2_X2  ߸ Դϴ.\n");
	      break;
	    case _SWAP:		/* opnum = 95 */
	      {
		enum vstack_value_type t0, t1;

		if ((VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY2)
		    || (VSTACK_TYPE (VSTACK_NEXT (vstack_top)) & STACK_COMPUTATIONAL_CATEGORY2))
		  error ("SWAP  ߸ . computational category  Ѵ type 1  Ѵ.\n");

		t0 = VSTACK_TYPE (vstack_top);
		t1 = VSTACK_TYPE (VSTACK_NEXT (vstack_top));

		pop_vstack (t0);
		pop_vstack (t1);
		push_vstack (t0);
		push_vstack (t1);
		break;
	      }
	    case _IADD:		/* opnum = 96 */
	    case _ISUB:		/* opnum = 100 */
	    case _IMUL:		/* opnum = 104 */
	    case _IDIV:		/* opnum = 108 */
	    case _IREM:		/* opnum = 112 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_INTEGER);
	      push_vstack (STACK_INTEGER);
	      i += -1;
	      break;
	    case _LADD:		/* opnum = 97 */
	    case _LSUB:		/* opnum = 101 */
	    case _LMUL:		/* opnum = 105 */
	    case _LDIV:		/* opnum = 109 */
	    case _LREM:		/* opnum = 113 */
	      pop_vstack (STACK_LONG);
	      pop_vstack (STACK_LONG);
	      push_vstack (STACK_LONG);
	      i += -2;
	      break;
	    case _FADD:		/* opnum = 98 */
	    case _FSUB:		/* opnum = 102 */
	    case _FMUL:		/* opnum = 106 */
	    case _FDIV:		/* opnum = 110 */
	    case _FREM:		/* opnum = 114 */
	      pop_vstack (STACK_FLOAT);
	      pop_vstack (STACK_FLOAT);
	      push_vstack (STACK_FLOAT);
	      i += -1;
	      break;
	    case _DADD:		/* opnum = 99 */
	    case _DSUB:		/* opnum = 103 */
	    case _DMUL:		/* opnum = 107 */
	    case _DDIV:		/* opnum = 111 */
	    case _DREM:		/* opnum = 115 */
	      pop_vstack (STACK_DOUBLE);
	      pop_vstack (STACK_DOUBLE);
	      push_vstack (STACK_DOUBLE);
	      i += -2;
	      break;
	    case _INEG:		/* opnum = 116 */
	    case _LNEG:		/* opnum = 117 */
	    case _FNEG:		/* opnum = 118 */
	    case _DNEG:		/* opnum = 119 */
	      /* .NEG  , û Ǵ  .  */
	      break;
	    case _ISHL:		/* opnum = 120 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_INTEGER);
	      push_vstack (STACK_INTEGER);
	      i += -1;
	      break;
	    case _LSHL:		/* opnum = 121 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_LONG);
	      push_vstack (STACK_LONG);
	      i += -1;
	      break;
	    case _ISHR:		/* opnum = 122 */
	    case _IUSHR:	/* opnum = 124 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_INTEGER);
	      push_vstack (STACK_INTEGER);
	      i += -1;
	      break;
	    case _LSHR:		/* opnum = 123 */	  
	    case _LUSHR:	/* opnum = 125 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_LONG);
	      push_vstack (STACK_LONG);
	      i += -1;
	      break;
	    case _IAND:		/* opnum = 126 */
	    case _IOR:		/* opnum = 128 */
	    case _IXOR:		/* opnum = 130 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_INTEGER);
	      push_vstack (STACK_INTEGER);
	      i += -1;
	      break;
	    case _LAND:		/* opnum = 127 */
	    case _LOR:		/* opnum = 129 */
	    case _LXOR:		/* opnum = 131 */
	      pop_vstack (STACK_LONG);
	      pop_vstack (STACK_LONG);
	      push_vstack (STACK_LONG);
	      i += -2;
	      break;
	    case _I2L:		/* opnum = 133 */
	      pop_vstack (STACK_INTEGER);
	      push_vstack (STACK_LONG);
	      i += 1;
	      break;
	    case _I2F:		/* opnum = 134 */
	      pop_vstack (STACK_INTEGER);
	      push_vstack (STACK_FLOAT);
	      break;
	    case _I2D:		/* opnum = 135 */
	      pop_vstack (STACK_INTEGER);
	      push_vstack (STACK_DOUBLE);
	      i += 1;
	      break;
	    case _L2I:		/* opnum = 136 */
	      pop_vstack (STACK_LONG);
	      push_vstack (STACK_INTEGER);
	      i += -1;
	      break;
	    case _L2F:		/* opnum = 137 */
	      pop_vstack (STACK_LONG);
	      push_vstack (STACK_FLOAT);
	      i += -1;
	      break;
	    case _L2D:		/* opnum = 138 */
	      pop_vstack (STACK_LONG);
	      push_vstack (STACK_DOUBLE);
	      break;
	    case _F2I:		/* opnum = 139 */
	      pop_vstack (STACK_FLOAT);
	      push_vstack (STACK_INTEGER);
	      break;
	    case _F2L:		/* opnum = 140 */
	      pop_vstack (STACK_FLOAT);
	      push_vstack (STACK_LONG);
	      i += 1;
	      break;
	    case _F2D:		/* opnum = 141 */
	      pop_vstack (STACK_FLOAT);
	      push_vstack (STACK_DOUBLE);
	      i += 1;
	      break;
	    case _D2I:		/* opnum = 142 */
	      pop_vstack (STACK_DOUBLE);
	      push_vstack (STACK_INTEGER);
	      i += -1;
	      break;
	    case _D2L:		/* opnum = 143 */
	      pop_vstack (STACK_DOUBLE);
	      push_vstack (STACK_LONG);
	      break;
	    case _D2F:		/* opnum = 144 */
	      pop_vstack (STACK_DOUBLE);
	      push_vstack (STACK_FLOAT);
	      i += -1;
	      break;
	    case _I2B:		/* opnum = 145 */
	      pop_vstack (STACK_INTEGER);
	      push_vstack (STACK_BYTE);
	      break;
	    case _I2C:		/* opnum = 146 */
	      pop_vstack (STACK_INTEGER);
	      push_vstack (STACK_CHAR);
	      break;
	    case _I2S:		/* opnum = 147 */
	      pop_vstack (STACK_INTEGER);
	      push_vstack (STACK_SHORT);
	      break;
	    case _LCMP:		/* opnum = 148 */
	      pop_vstack (STACK_LONG);
	      pop_vstack (STACK_LONG);
	      push_vstack (STACK_INTEGER);
	      i += -3;
	      break;
	    case _FCMPL:	/* opnum = 149 */
	    case _FCMPG:	/* opnum = 150 */
	      pop_vstack (STACK_FLOAT);
	      pop_vstack (STACK_FLOAT);
	      push_vstack (STACK_INTEGER);
	      i += -1;
	      break;
	    case _DCMPL:	/* opnum = 151 */
	    case _DCMPG:	/* opnum = 152 */
	      pop_vstack (STACK_DOUBLE);
	      pop_vstack (STACK_DOUBLE);
	      push_vstack (STACK_INTEGER);
	      i += -3;
	      break;
	    case _IFEQ:		/* opnum = 153 */
	    case _IFNE:		/* opnum = 154 */
	    case _IFLT:		/* opnum = 155 */
	    case _IFGE:		/* opnum = 156 */
	    case _IFGT:		/* opnum = 157 */
	    case _IFLE:		/* opnum = 158 */
	      pop_vstack (STACK_INTEGER);
	      i += -1;
	      break;
	    case _IF_ICMPEQ:	/* opnum = 159 */
	    case _IF_ICMPNE:	/* opnum = 160 */
	    case _IF_ICMPLT:	/* opnum = 161 */
	    case _IF_ICMPGE:	/* opnum = 162 */
	    case _IF_ICMPGT:	/* opnum = 163 */
	    case _IF_ICMPLE:	/* opnum = 164 */
	      pop_vstack (STACK_INTEGER);
	      pop_vstack (STACK_INTEGER);
	      i += -2;
	      break;
	    case _IF_ACMPEQ:	/* opnum = 165 */
	    case _IF_ACMPNE:	/* opnum = 166 */
	      pop_vstack (STACK_ARRAYREF|STACK_OBJECTREF);
	      pop_vstack (STACK_ARRAYREF|STACK_OBJECTREF);
	      i += -2;
	      break;
	    case _JSR:		/* opnum = 168 */
	    case _JSR_W:	/* opnum = 201 */
	      push_vstack (STACK_RETURNADDRESS);
	      i += 1;
	      break;
	    case _IRETURN:	/* opnum = 172 */
	      pop_vstack (STACK_INTEGER);
	      i += -1;
	      break;
	    case _LRETURN:	/* opnum = 173 */
	      pop_vstack (STACK_LONG);
	      i += -2;
	      break;
	    case _FRETURN:	/* opnum = 174 */
	      pop_vstack (STACK_FLOAT);
	      i += -1;
	      break;
	    case _DRETURN:	/* opnum = 175 */
	      pop_vstack (STACK_DOUBLE);
	      i += -2;
	      break;
	    case _ARETURN:	/* opnum = 176 */
	      pop_vstack (STACK_OBJECTREF|STACK_ARRAYREF);
	      i += -1;
	      break;
	    case _GETSTATIC:	/* opnum = 178 */
	      {
		const char *fp = IDENTIFIER_POINTER (TREE_FIELDREF_TYPE (TREE_INSN_OP1 (t)));
		int s;

		push_vstack (STACK_ALL);
		s = sizeof_type_char (fp[0], 0);
		if (s == 1)
		  i += 1;
		else if (s == 2)
		  i += 2;
		else
		  {
		    warning ("count_max_stack: ߸ TYPE SIZE.\n");
		    debug_tree (t);
		  }
		break;
	      }
	    case _PUTSTATIC:	/* opnum = 179 */
	      if ((VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY2))
		{
		  pop_vstack (STACK_COMPUTATIONAL_CATEGORY2);
		  i += -2;
		}
	      else
		{	
		  pop_vstack (STACK_COMPUTATIONAL_CATEGORY1);
		  i += -1;
		}
	      break;
	    case _GETFIELD:
	      {
		const char *fp = IDENTIFIER_POINTER (TREE_FIELDREF_TYPE (TREE_INSN_OP1 (t)));
		int s;

		pop_vstack (STACK_OBJECTREF);
		push_vstack_field_type_char (fp[0]);
		s = sizeof_type_char (fp[0], 0);
		if (s == 2)
		  i += 1;
		break;
	      }
	    case _PUTFIELD:
	      if ((VSTACK_TYPE (vstack_top) & STACK_COMPUTATIONAL_CATEGORY2))
		{
		  pop_vstack (STACK_COMPUTATIONAL_CATEGORY2);
		  pop_vstack (STACK_OBJECTREF);
		  i += -3;
		}
	      else
		{	
		  pop_vstack (STACK_COMPUTATIONAL_CATEGORY1);
		  pop_vstack (STACK_OBJECTREF);
		  i += -2;
		}
	      break;
	    case _IINC:		/* opnum = 132 */
	    case _GOTO:		/* opnum = 167 */
	    case _RET:		/* opnum = 169 */
	    case _RETURN:	/* opnum = 177 */
	    case _NOP:		/* opnum = 0 */
	    case _NEWARRAY:
	      break;
	    default:
	      warning ("count_max_stack: code [%s] update ʿ.\n", IDENTIFIER_POINTER (TREE_INSN_CODE (t)));
	      debug_tree (t);
	      break;
	    }
	}
      i += count_max_stack (TREE_CHAIN (t), max);
      break;
    case SWITCH_LOOKUP:
    case SWITCH_TABLE:
      i += -1;
      break;
    case LABEL:
      break;
    default:
      error ("count_max_stack: ߸ NODE.\n");
      debug_tree (t);
      break;
    }

  return (i > max) ? i : max;
}

int
count_access_hosts (t)
     tree t;
{
  int i = 0;

  if (!t)
    return 0;

  switch (TREE_CODE (t))
    {
    case TREE_LIST:
      i += count_access_hosts (TREE_PURPOSE (t));
      i += count_access_hosts (TREE_VALUE (t));
      i += count_access_hosts (TREE_CHAIN (t));
      break;
    case ACCESS_EXPR:
      return 1;
    default:
      error ("count_access_hosts: ߸ NODE.\n");
      debug_tree (t);
      break;
    }

  return i;
}

int
count_fields (t)
     tree t;
{
  int i = 0;

  if (!t)
    return 0;

  switch (TREE_CODE (t))
    {
    case TREE_LIST:
      i += count_fields (TREE_PURPOSE (t));
      i += count_fields (TREE_VALUE (t));
      i += count_fields (TREE_CHAIN (t));
      break;
    case FIELD:
      i += 1;
      i += count_fields (TREE_CHAIN (t));
      break;
    default:
      error ("count_fields: ߸ NODE.\n");
      break;
    }

  return i;
}

int
count_methods (t)
     tree t;
{
  int i = 0;

  if (!t)
    return 0;

  switch (TREE_CODE (t))
    {
    case TREE_LIST:
      i += count_methods (TREE_PURPOSE (t));
      i += count_methods (TREE_VALUE (t));
      i += count_methods (TREE_CHAIN (t));
      break;
    case METHODS:
      i += 1;
      i += count_methods (TREE_CHAIN (t));
      break;
    default:
      error ("count_methods: ߸ NODE.\n");
      break;
    }

  return i;
}

int
constant_process (type, t)
     enum constant type;
     tree t;
{
  if (!t)
    return -1;

  TREE_CONSTANT_POOL_TYPE (t) = type;

  switch (TREE_CODE (t))
    {
    case ENVIRONMENT:
      constant_process (CONSTANT_Environment, TREE_ENVIRONMENT_NAME (t));
      constant_process (CONSTANT_Accessref, TREE_ENVIRONMENT_ACCESS_HOST (t));
      break;
    case TREE_LIST:
      constant_process (type, TREE_PURPOSE (t));
      constant_process (type, TREE_VALUE (t));
      break;
    case FIELD:
      constant_process (CONSTANT_Accessref, TREE_FIELD_ACCESS_HOST (t));
      constant_process (CONSTANT_Utf8, TREE_FIELD_NAME (t));
      constant_process (CONSTANT_Utf8, TREE_FIELD_TYPE (t));
      break;
    case METHODS:
      constant_process (CONSTANT_Accessref, TREE_METHOD_ACCESS_HOST (t));
      constant_process (CONSTANT_Utf8, TREE_METHOD_NAME (t));
      constant_process (CONSTANT_Utf8, TREE_METHOD_TYPE (t));
      break;
    case INFO:
      if (TREE_INFO_FILENAME (t))
	TREE_CONSTANT_POOL_NUMBER (TREE_INFO_FILENAME (t))
	  = cp_add (CONSTANT_Utf8, TREE_STRING_POINTER (TREE_INFO_FILENAME (t)));
      break;
    case STRING_CST:
      if (type == CONSTANT_Utf8)
	TREE_CONSTANT_POOL_NUMBER (t) = cp_add (CONSTANT_Utf8, TREE_STRING_POINTER (t));
      else
	warning ("constant_process: ó ʿ.");      
      break;
    case ACCESS_EXPR:
      TREE_CONSTANT_POOL_NUMBER (t) = do_access_hosts (t);      
      break;
    case IDENTIFIER_NODE:
      if (type == CONSTANT_Utf8)
	TREE_CONSTANT_POOL_NUMBER (t) = cp_add (CONSTANT_Utf8, IDENTIFIER_POINTER (t));
      else if (type == CONSTANT_Environment)
	TREE_CONSTANT_POOL_NUMBER (t) = do_this_environment (IDENTIFIER_POINTER (t));
      else
	warning ("constant_process: ó ʿ.");
      break;
    default:
      error ("constant_process: ó    = %s\n", tree_code_name[(TREE_CODE (t))]);
      break;
    }

  return -1;
}

/* struct environment_file  ʱȭϱ  Լ̴.  */
void 
environment_file_init (ef, buf, len)
     struct environment_file *ef;
     unsigned char* buf;
     size_t len;
{
  assert (((buf == NULL) && (len == 0))
	  || ((buf != NULL) && (len != 0)));
  
  ef->base = ef->cur = buf;
  ef->size = len;
}

long 
fileSize (const char *file)
{
  FILE *fp = fopen(file, "rb");
  long len = 0;
  
  if (fp != NULL)
    {
      if (fseek (fp, 0L, 2) != -1)
	len = ftell (fp);
      fclose (fp);
    }
  
  return len;
}

void
init_reserved_words (void)
{
  int i;
  tree id;

  for (i=0; i<N_reswords; i++)
    {
      id = get_identifier (reswords[i].word);
      TREE_RESERVED_WORD_CODE (id) = reswords[i].code;
      TREE_IS_RESERVED_WORD (id) = 1;
    }

  for (i=0; i<N_resword_insns; i++)
    {
      id = get_identifier (resword_insns[i].word);
      TREE_BYTE_CODE (id) = resword_insns[i].bytecode;
      TREE_INSN_LENGTH (id) = resword_insns[i].oplen;
      TREE_IS_RESERVED_INSN (id) = 1;
    }
}

void
init_esm (void)
{
  init_stringpool ();

  init_reserved_words ();

  build_common_tree_nodes ();
}

/* NUMBER Ȥ STRING, CHAR, COMMENT token  (payload).  */
struct esm_string
{
  unsigned int len;
  const unsigned char *text;
};

/* esm_token ü  flag .  */
#define PREV_WHITE	(1 << 0) /*   token  whitespace .  */
#define DIGRAPH		(1 << 1) /*  digraph ̾ٸ.  */
#define STRINGIFY_ARG	(1 << 2) /*  ڿȭǾ ϴ macro argument .  */
#define PASTE_LEFT	(1 << 3) /*  ## operator  LHS ̶.  */
#define NAMED_OP	(1 << 4) /* C++ named operator .  */
#define NO_EXPAND	(1 << 5) /*  token  macro-expand  .  */
#define BOL		(1 << 6) /* () ۿ token.  */

struct esm_token
{
  enum esm_ttype type;
  unsigned char flags;		/* flag  -  . */

  union
  {
    char c;
    tree node;
    struct esm_string str;	/* ڿ Ȥ .  */
  } val;
};

#define is_vspace(x)	IS_VSPACE(x)
#define is_nvspace(x)	(IS_NVSPACE(x) && x != '\0')

/* whitespace  ǳʶٰ whitespace  ƴ  ڸ մϴ.
   Adjusts pfile->col_adjust to account for tabs.  ̰  token
    (ĭ)  ߸    Դϴ.  */
static int
skip_whitespace (c)
     char c;
{
  unsigned int warned = 0;

  do
    {
      /*  space  ׻ OK ̴.  */
      if (c == ' ')
	;
      else if (c == '\t')
	;
      /*  \f Ȥ \v \0  Ҵ.  */
      else if (c == '\0')
	{
	  if (ef_info.cur - 1 == ef_info.base + ef_info.size)
	    return 0;
	  if (!warned)
	    {
	      warning ("null character(s) ignored");
	      warned = 1;
	    }
	}

      c = *ef_info.cur++;
    }
  /* 츮  non-vertical space  Ѵ.  ' ' \t \f \v \0.  */
  while (is_nvspace (c));

  ef_info.cur--;
  return 1;
}

static char
get_effective_char ()
{
  char next;

  ef_info.backup_to = ef_info.cur;
  next = *ef_info.cur++;

  return next;
}

/* Newline    ȣϸ, װ buffer->cur[-1]  Ѵٰ 
   Ѵ.  Newline () ٷ  ִ ڸ Ű buffer->cur 
     ȯǾ .  */
static void
handle_newline ()
{
  /* CR-LF  LF-CR  ٷ.  κ ٸ ͵ (, java)  CR-LF 
     ޾ ̸, ׷ ൿ ֵ ؾ ұ?  */
  if (ef_info.cur[-1] + ef_info.cur[0] == '\r' + '\n')
    ef_info.cur++;

  ef_info.line_base = ef_info.cur;
  line++;
}

#if 0
static int
skip_block_comment ()
{
  char c = EOF, prevc = EOF;

  while (ef_info.cur != ef_info.base + ef_info.size)
    {
      prevc = c, c = *ef_info.cur++;

      /* People like decorating comments with '*', so check for '/'
	 instead for efficiency.  */
      if (c == '/')
	{
	  if (prevc == '*')
	    break;
	}
      else if (is_vspace (c))
	handle_newline ();
      else if (c == '\t')
	;
    }

  return c != '/' || prevc != '*';
}
#endif

static int
skip_line_comment ()
{
  char c;

  do
    {
      if (ef_info.cur == ef_info.base + ef_info.size)
	goto at_eof;

      c = *ef_info.cur++;
    }
  while (!is_vspace (c));

  /* Step back over the newline, except at EOF.  */
  while (*(ef_info.cur) == '\r' || *(ef_info.cur) == '\n')
    ef_info.cur++;

 at_eof:

  return 0;
}

/* Buffer κ IDENTIFIER  ؼѴ.  Identifier  ,  .  ϰ Ǵµ,
     ּҸ óϱ ؼ̴.  yylexname () Լ  ش  identifier
    address  ϰ ȴ.  */
static tree
parse_identifier ()
{
  unsigned char *cur, *base;
  tree result;

  cur = ef_info.cur;
  while (ISIDNUM (*cur) || *cur == '.')
    cur++;

  base = ef_info.cur - 1;
  result = get_identifier_with_length (base, cur - base);
  ef_info.cur = cur;
  
  //debug_tree (result);
  return result;
}

static tree
parse_address ()
{
  unsigned char *cur, *base;
  tree result;

  cur = ef_info.cur;
  while (ISIDNUM (*cur) || *cur == '.' || *cur == ':' || *cur == '+' || *cur == '/'
	 || *cur == '=' || *cur == '<' || *cur == '>')
    cur++;

  base = ef_info.cur - 1;
  //fprintf (stderr, "<%.*s>", (int) (cur - base), base);
  result = get_identifier_with_length (base, cur - base);
  ef_info.cur = cur;
  
  return result;
}

static tree
parse_type ()
{
  unsigned char *cur, *base;
  tree result;

  cur = ef_info.cur;
  while (ISIDNUM (*cur) || *cur == '.' || *cur == ':' || *cur == '+' || *cur == '/'
	 || *cur == '=' || *cur == '[' || *cur == ')' || *cur == ';')
    cur++;

  base = ef_info.cur - 1;
  //fprintf (stderr, "<%.*s>", (int) (cur - base), base);
  result = get_identifier_with_length (base, cur - base);
  ef_info.cur = cur;
  
  return result;
}

tree
print_match_url (t)
     tree t;
{
  int len = 0;
  char tmpbuf[1024];
  tree host, path;

  host = TREE_EADDRESS_HOSTNAME (TREE_MATCH_STRING (t));
  path = TREE_EADDRESS_PATHANDSHA1 (TREE_MATCH_STRING (t));

  len += sprintf (tmpbuf, "e://%s/", IDENTIFIER_POINTER (TREE_ADDRESS_IP (host)));
  len += sprintf (tmpbuf+len, "%s(\"%s\")", 
		  IDENTIFIER_POINTER (TREE_SEGMENT_ABSOLUTE_NAME (path)),
		  IDENTIFIER_POINTER (TREE_SEGMENT_SHA1 (path)));

#if 0
  fprintf (stderr, "print_match_url: %s", tmpbuf);
#endif

  return get_identifier (tmpbuf);
}

static tree
parse_arraytype (char c)
{
  unsigned char *cur, *base;
  tree result;

  cur = ef_info.cur;
  switch (c)
    {
    case '[':
    case 'L':
      {
	while (ISIDNUM (*cur) || *cur == '.' || *cur == ':' || *cur == '+' || *cur == '/'
	       || *cur == '=' || *cur == '[' || *cur == ')' || *cur == ';')
	  cur++;
	break;
      }
    default:
      //cur++;
      break;
    }

  base = ef_info.cur - 1;
  //fprintf (stderr, "<%.*s>", (int) (cur - base), base);
  {
    if (c == 'L')
      {
	if (!strncmp (base + 1, ENVIRON_ADDRESS_REF_PREFIX, ENVIRON_ADDRESS_REF_PREFIX_LEN))
	  {
	    /* e://    */
	    result = get_identifier_with_length (base, cur - base);
	  }
	else
	  {
	    int kk, is_location = 0;
	    for (kk=0; kk<cur - base; kk++)
	      {
		if (base[kk] == ';')
		  {
		    is_location = 1;
		    break;
		  }
	      }

	    /*  ';'      L.... type  ƴϱ , IDENTIFIER  ٽ
	       ؼؾ Ѵ.  */
	    if (is_location)
	      {
		/*   -2   ;  ϱ ؼ ̴.  */
		struct matchs *t = find_match_by_name_with_length (base + 1, cur - base - 2);
		char tmpbuf[1024];
		int a = 0;
		if (!t)
		  {
		    error (".match  ǵ .\n");
		    return NULL;
		  }
		a = sprintf (tmpbuf, "L%s;", IDENTIFIER_POINTER (print_match_url (t->match)));
		result = get_identifier_with_length (tmpbuf, a);
	      }
	    else
	      return parse_identifier ();
	  }
      }
    else
      result = get_identifier_with_length (base, cur - base);
  }

  ef_info.cur = cur;
  
  return result;
}

/*  ȣ preprocessing number  ȿ ˻Ѵ.  */
#define VALID_SIGN(c, prevc) \
  (((c) == '+' || (c) == '-') && \
   ((prevc) == 'e' || (prevc) == 'E' \
    || (((prevc) == 'p' || (prevc) == 'P'))))

#define is_numchar(x)	ISIDNUM(x)

void
parse_number (struct esm_string *number, char c)
{
  unsigned char *cur, *base;
  
  cur = ef_info.cur;
  while (is_numchar (*cur) || *cur == '.' || VALID_SIGN (*cur, *(cur-1)))
    cur++;

  base = ef_info.cur - 1;
  //fprintf (stderr, "<%.*s>", (int) (cur - base), base);
  number->text = base;
  number->len = cur - base;
  ef_info.cur = cur;
}

/* parse_string  ƾ.  */
static int
unescaped_terminator_p (dest)
     const unsigned char *dest;
{
  const unsigned char *start, *temp;

  start = ef_info.cur;

  /* An odd number of consecutive backslashes represents an escaped
     terminator.  */
  for (temp = dest; temp > start && temp[-1] == '\\'; temp--)
    ;

  return ((dest - temp) & 1) == 0;
}

/* Subroutine of parse_string.  Emits error for unterminated strings.  */
static void
unterminated (term)
     int term;
{
  error ("missing terminating %c character", term);
}

static void
parse_string (token, terminator)
     struct esm_token *token;
     char terminator;
{
  char c;
  unsigned char *cur, *dest, *limit;

  dest = ef_info.cur;
  cur = dest;
  limit = ef_info.base + ef_info.size;

  for (;;)
    {
      c = *ef_info.cur++;
      if (c == terminator)
	{
	  if (unescaped_terminator_p (dest))
	    break;
	}
      else if (is_vspace (c))
	{	  	      
	  handle_newline ();
	  c = '\n';
	}
      else if (c == '\0')
	{
	  if (ef_info.cur - 1 == ef_info.base + ef_info.size)
	    {
	      unterminated (terminator);
	      ef_info.cur--;
	      break;
	    }
	}
      *dest++ = c;
    }
  *dest = '\0';
  
  //fprintf (stderr, "parse_string: <%s>\n", cur);
  token->val.str.text = cur;
  token->val.str.len = dest - cur;

  ef_info.cur = dest + 1;
}

#define BACKUP() do {ef_info.cur = ef_info.backup_to;} while (0)

struct esm_token *
esm_get_token (void)
{
  char c;
  struct esm_token *result;

  result = malloc (sizeof (struct esm_token));
  result->type = -1;

 fresh_line:

 skipped_white:
  c = *ef_info.cur++;

  switch (c)
    {
    case ' ': case '\t': case '\f': case '\v': case '\0':
      result->flags |= PREV_WHITE;
      if (skip_whitespace (c))
	goto skipped_white;

      /* EOF.  */
      result->type = ESM_EOF;
      break;
    case '\r':
      c = get_effective_char ();
      if (c != '\n')
	{
	  BACKUP ();
	  goto fresh_line;
	}
    case '\n':
      c = get_effective_char ();
      BACKUP ();
      if (c == '\r' || c == '\n')
	goto fresh_line;	  
      handle_newline ();
      result->type = ESM_SEPERATOR;
      break;
    case '!':
      if (skip_line_comment ())
	warning ("multi-line comment");
      result->type = ESM_PADDING;
      break;
    case '/':
      result->type = ESM_PATH_SEPERATOR;
      break;
    case '%':
      result->type = ESM_LOCALS;
      break;
    case '(':
      c = get_effective_char ();
      BACKUP ();
      if (c == '"')
	result->type = ESM_PAR_START;
      else
	{
	  result->type = ESM_METHOD_TYPE;
	  result->val.node = parse_type ();
	}
      break;
    case ')':
      result->type = ESM_PAR_END;
      break;
    case ':':
      c = get_effective_char ();
      if (ISIDNUM (c))
	{
	  result->type = ESM_ADDRESS;
	  result->val.node = parse_address ();
	}
      else
	{
	  BACKUP ();
	  result->type = ESM_COLON;
	}
      break;
    case '-':
      c = get_effective_char ();
      
      if (c == '>')
	result->type = ESM_DEREF;
      else
	{
	  BACKUP ();
	  result->type = ESM_MINUS;
	}
      break;
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      result->type = ESM_NUMBER;
      parse_number (&result->val.str, c);
      break;
    case '.':
      /* . ڷ  ,  ̴.   , yylexname ()
          ó ȴ.  */
      result->type = ESM_NAME;
      result->val.node = parse_identifier ();
      break;
    case '_':
    case 'a': case 'b': case 'c': case 'd':           case 'f':
    case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':
    case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
    case 's': case 't': case 'u': case 'v': case 'w': case 'x':
    case 'y': case 'z':
    case 'A': case 'E': case 'G': case 'H': case 'K':
    case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
    case 'T': case 'U': case 'W': case 'X': case 'Y':
    start_ident:
      result->type = ESM_NAME;
      result->val.node = parse_identifier ();
      break;
    case 'e':
      c = get_effective_char ();
      if (c == ':')
	{
	  c = get_effective_char ();
	  if (c == '/')
	    {
	      c = get_effective_char ();
	      if (c == '/')
		{
		  result->type = ESM_PREFIX;
		  break;
		}
	      else
		{
		  BACKUP ();
		}
	    }
	  else
	    {
	      BACKUP ();
	    }
	}
      else
	{
	  BACKUP ();
	  goto start_ident;
	}
      break;
    case 'B': case 'C': case 'D': case 'F': case 'I': case 'S': 
    case 'V': case 'Z': case 'J':
      c = get_effective_char ();
      BACKUP ();
      if (ISIDNUM (c))
	goto start_ident;
      /* Ʒ !!  */
    case 'L': 
    case '[':
      {
	result->type = ESM_FIELD_TYPE;
	result->val.node = parse_arraytype (c);
	if (c == 'L')
	  {
	    if (!strstr (IDENTIFIER_POINTER (result->val.node), ";"))
	      result->type = ESM_NAME;
	  }
	break;
      }
    case '\'':
    case '"':
      result->type = c == '"' ? ESM_STRING: ESM_CHAR;
      parse_string (result, c);
      break;
    case '{':
      result->type = ESM_STARTP;
      break;
    case '}':
      result->type = ESM_ENDP;
      break;
    case '<':
      result->type = ESM_ANGLE_START;
      break;
    case '>':
      result->type = ESM_ANGLE_END;
      break;
    case '=':
      result->type = ESM_EQ;
      break;
    default:
      fprintf (stderr, "OTHER = %c\n", c);
      result->type = ESM_OTHER;
      result->val.c = c;
    }

  return result;
}

char *
esm_strndup (const char *s, size_t length)
{
  char *p;

  p = (char *) malloc (length+1);
  if (!p)
    return (char *) NULL;

  memcpy (p, s, length);
  p[length] = 0;

  return p;
}

static tree
lex_string (str, len)
     const unsigned char *str;
     unsigned int len;
{
  return build_string (len, esm_strndup (str, len));
}

struct pf_args
{
  /* Input */
  const char *str;
  int fflag;
  int lflag;
  int base;
  /* Output */
  int conversion_errno;
  REAL_VALUE_TYPE value;
  tree type;
};

static int float_handler_set;
int float_handled;
jmp_buf float_handler;

/* Signal    ´.  */

static void
float_signal (signo)
     /* If this is missing, some compilers complain.  */
     int signo ATTRIBUTE_UNUSED;
{
  if (float_handled == 0)
    error ("internal error: %s", strsignal (signo));
  float_handled = 0;

  /* On System-V derived systems, we must reinstall the signal handler.
     This is harmless on BSD-derived systems.  */
  signal (SIGFPE, float_signal);
  longjmp (float_handler, 1);
}

/* floating arithmetic error  ߻  longjmp    Ѵ.
    HANDLER  0  ,  ̻  ٷ  ǹѴ.  */

static void
set_float_handler (handler)
     jmp_buf handler;
{
  float_handled = (handler != 0);
  if (handler)
    memcpy (float_handler, handler, sizeof (float_handler));

  if (float_handled && ! float_handler_set)
    {
      signal (SIGFPE, float_signal);
      float_handler_set = 1;
    }
}

/* ̰ arithmetic exception  ų  ִ code   wrapper 
   function ̴.  ش code  function pointer FN  ޵Ǹ,
   ϳ argument DATA  ǳٴ.  DATA   function FN  
   real input  output  ϰ ִ.   Լ  longjmp  ȣǾ
   (, exception  ߻Ǹ) 0  (н) ȯѴ.  ׷  
   () 1  ȯѴ.  */

int
do_float_handler (fn, data)
     void (*fn) PARAMS ((PTR));
     PTR data;
{
  jmp_buf buf;

  if (setjmp (buf))
    {
      /* 츮 Լ fn ()  exception  ߻  longjmp ()  
	  Ѵ.  */
      set_float_handler (NULL);
      return 0;
    }

  set_float_handler (buf);
  (*fn) (data);
  set_float_handler (NULL);
  return 1;
}

static void
parse_float (data)
  PTR data;
{
  struct pf_args * args = (struct pf_args *) data;
  const char *typename;

  args->conversion_errno = 0;
  args->type = double_type_node;
  typename = "double";

  /* The second argument, machine_mode, of REAL_VALUE_ATOF
     tells the desired precision of the binary result
     of decimal-to-binary conversion.  */

  if (args->fflag)
    {
      if (args->lflag)
	error ("both 'f' and 'l' suffixes on floating constant");

      args->type = float_type_node;
      typename = "float";
    }
  else if (args->lflag)
    {
      args->type = long_double_type_node;
      typename = "long double";
    }
  errno = 0;
  if (args->base == 16)
    args->value = REAL_VALUE_HTOF (args->str, TYPE_MODE (args->type));
  else
    args->value = REAL_VALUE_ATOF (args->str, TYPE_MODE (args->type));

  args->conversion_errno = errno;
  /* A diagnostic is required here by some ISO C testsuites.
     This is not pedwarn, because some people don't want
     an error for this.  */
  if (REAL_VALUE_ISINF (args->value))
    warning ("floating point number exceeds range of '%s'", typename);
}
 
/* Node   type  մϴ.  */
static tree
lex_number (str, len)
     const char *str;
     unsigned int len;
{
  int count = 0;
  int base = 10;
  tree type, value;
  int c;
  int largest_digit = 0;
  int numdigits = 0;
  int overflow = 0;
  const char *p;
  enum anon1 { NOT_FLOAT = 0, AFTER_POINT, AFTER_EXPON } floatflag = NOT_FLOAT;

  /* 츮   part   HOST_BITS_PER_CHAR bit 鸸 Ѵ.
     Part 迭 ä Ʒ code  host int  host char  ּ ι̻
     ṵ̃, HOST_BITS_PER_WIDE_INT  HOST_BITS_PER_CHAR  ¦ ̶
     Ѵ.   HOST_WIDE_INT  츮   ִ  ū int literal
     ̴.  Ʒ overflow  ϱ ؼ, Part  (TOTAL_PARTS) 
       HOST_WIDE_INT  bit   ִµ ʿ Ȯ part
      number ߸ Ѵ.  */
#define TOTAL_PARTS ((HOST_BITS_PER_WIDE_INT / HOST_BITS_PER_CHAR) * 2)
  unsigned int parts[TOTAL_PARTS];

  /*     ȭѴ.  */
  if (len == 1)
    {
      if (*str == '0')
	return build_int_2 (*str - '0', 0);
      else if (*str == '1')
	return build_int_2 (*str - '0', 0);
      else
	return build_int_2 (*str - '0', 0);
    }

  for (count = 0; count < TOTAL_PARTS; count++)
    parts[count] = 0;

  /*   len  >1  ˷.  */
  p = str;

  if (len > 2 && str[0] == '0' && (str[1] == 'x' || str[1] == 'X'))
    {
      base = 16;
      p = str + 2;
    }
  /* ISDIGIT check  ν 0   suffix   ȥϴ   ̴.  */
  else if (str[0] == '0' && ISDIGIT (str[1]))
    {
      base = 8;
      p = str + 1;
    }

  do
    {
      c = *p++;

      if (c == '.')
	{
	  if (floatflag == AFTER_POINT)
	    error ("too many decimal points in floating constant");
	  else if (floatflag == AFTER_EXPON)
	    error ("decimal point in exponent - impossible!");
	  else
	    floatflag = AFTER_POINT;

	  if (base == 8)
	    base = 10;
	}
      else if (c == '_')
	/* ߿ Ȯ ɼ , number   _  Ѵ.
	   Cosmetic grouping  ϴµ,  , 0x8000_0000 == 0x80000000
	   ̴.  ټ б   ִ.  Ada  ̰ ϳ?  */
	error ("underscore in number");
      else
	{
	  int n;
	  /* Decial point  ƴϴ.  Digit (Ƹ hex digit)  Ѵ.  */

	  if (ISDIGIT (c)
	      || (base == 16 && ISXDIGIT (c)))
	    {
	      n = hex_value (c);
	    }
	  else if (base <= 10 && (c == 'e' || c == 'E'))
	    {
	      base = 10;
	      floatflag = AFTER_EXPON;
	      break;
	    }
	  else if (base == 16 && (c == 'p' || c == 'P'))
	    {
	      floatflag = AFTER_EXPON;
	      break;   /* Exponent   */
	    }
	  else
	    {
	      p--;
	      break;  /* start of suffix */
	    }

	  if (n >= largest_digit)
	    largest_digit = n;
	  numdigits++;

	  for (count = 0; count < TOTAL_PARTS; count++)
	    {
	      parts[count] *= base;
	      if (count)
		{
		  parts[count]
		    += (parts[count-1] >> HOST_BITS_PER_CHAR);
		  parts[count-1]
		    &= (1 << HOST_BITS_PER_CHAR) - 1;
		}
	      else
		parts[0] += n;
	    }

	  /*  -- part  overflow  ߻  (host char
	        ִ ͺ ũٸ) ü number  overflow Ǿ.
	     ̰ ϰ -- part  truncate Ѵ.  */
	  if (parts[TOTAL_PARTS - 1] >> HOST_BITS_PER_CHAR)
	    {
	      overflow = 1;
	      parts[TOTAL_PARTS - 1] &= (1 << HOST_BITS_PER_CHAR) - 1;
	    }
	}
    }
  while (p < str + len);

  /* ̰ `int i = 0x;'   Է   ߻Ѵ.  */
  if (numdigits == 0)
    error ("numeric constant with no digits");

  if (largest_digit >= base)
    error ("numeric constant contains digits beyond the radix");

  if (floatflag != NOT_FLOAT)
    {
      int imag, fflag, lflag, conversion_errno;
      char *copy;
      REAL_VALUE_TYPE real;
      struct pf_args args;

      /* Read explicit exponent if any, and put it in tokenbuf.  */
      if ((base == 10 && ((c == 'e') || (c == 'E')))
	  || (base == 16 && (c == 'p' || c == 'P')))
	{
	  if (p < str + len)
	    c = *p++;
	  if (p < str + len && (c == '+' || c == '-'))
	    c = *p++;
	  /* Exponent is decimal, even if string is a hex float.  */
	  if (! ISDIGIT (c))
	    error ("floating constant exponent has no digits");
	  while (p < str + len && ISDIGIT (c))
	    c = *p++;
	  if (! ISDIGIT (c))
	    p--;
	}

      /*  float constant  Ѵ; 츮 parse_float  ǳ
	 ڿ  suffix  ԵǱ  ʴ´.  */
      copy = (char *) malloc (p - str + 1);
      memcpy (copy, str, p - str);
      copy[p - str] = '\0';

      /*  suffix  ؼѴ.  */
      fflag = lflag = imag = 0;
      while (p < str + len)
	switch (*p++)
	  {
	  case 'f': case 'F':
	    if (fflag)
	      error ("more than one 'f' suffix on floating constant");

	    fflag = 1;
	    break;

	  case 'l': case 'L':
	    if (lflag)
	      error ("more than one 'l' suffix on floating constant");

	    lflag = 1;
	    break;

	  case 'i': case 'I':
	  case 'j': case 'J':
	    if (imag)
	      error ("more than one 'i' or 'j' suffix on floating constant");

	    imag = 1;
	    break;

	  default:
	    error ("invalid suffix on floating constant");
	  }

      /* parse_float()   Է° Ѵ. */
      args.str = copy;
      args.fflag = fflag;
      args.lflag = lflag;
      args.base = base;

      /* ڿ double  ϴµ, overflow  ˻Ѵ.  */
      if (do_float_handler (parse_float, (PTR) &args))
	{
	  /* parse_float()  output  ޴´.  */
	  real = args.value;
	}
      else
	/* We got an exception from parse_float() */
	error ("floating constant out of range");

      conversion_errno = args.conversion_errno;
      type = args.type;
      
      /*  type   ̿Ͽ node  Ѵ.  */
      if (imag)
	{
	  error (" Ҽ ó Ѵ. TODO: ó ƾ ߰ ʿ.");
	  value = NULL;
	}
      else
	value = build_real (type, real);
    }
  else
    {
      tree trad_type, ansi_type;
      int high, low;
      int spec_unsigned = 0;
      int spec_long = 0;
      int spec_long_long = 0;
      int spec_imag = 0;
      int suffix_lu = 0;
      int warn = 0, i;

      trad_type = ansi_type = type = NULL_TREE;
      while (p < str + len)
	{
	  c = *p++;
	  switch (c)
	    {
	    case 'u': case 'U':
	      if (spec_unsigned)
		error ("two 'u' suffixes on integer constant");

	      spec_unsigned = 1;
	      if (spec_long)
		suffix_lu = 1;
	      break;

	    case 'l': case 'L':
	      if (spec_long)
		{
		  if (spec_long_long)
		    error ("three 'l' suffixes on integer constant");
		  else if (suffix_lu)
		    error ("'lul' is not a valid integer suffix");
		  else if (c != spec_long)
		    error ("'Ll' and 'lL' are not valid integer suffixes");
		  spec_long_long = 1;
		}
	      spec_long = c;
	      break;

	    case 'i': case 'I': case 'j': case 'J':
	      if (spec_imag)
		error ("more than one 'i' or 'j' suffix on integer constant");
	      spec_imag = 1;
	      break;

	    default:
	      error ("invalid suffix on integer constant");
	    }
	}

      /*  literal  overflow Ǿٸ,  ׿  Ѵ.  */
      if (overflow)
	{
	  warn = 1;
	  warning ("integer constant is too large for this configuration of the compiler - truncated to %d bits", 
		   HOST_BITS_PER_WIDE_INT * 2);
	}

      /* ̰ 츮  ׻ ̶ Ƿ ȭȴ.  */
      
      high = low = 0;

      for (i = 0; i < HOST_BITS_PER_WIDE_INT / HOST_BITS_PER_CHAR; i++)
	{
	  high |= ((int) parts[i + (HOST_BITS_PER_WIDE_INT
					      / HOST_BITS_PER_CHAR)]
		   << (i * HOST_BITS_PER_CHAR));
	  low |= (int) parts[i] << (i * HOST_BITS_PER_CHAR);
	}

      value = build_int_2 (low, high);
      TREE_TYPE (value) = long_long_unsigned_type_node;

      /* Calculate the ISO type.  */
      if (! spec_long && ! spec_unsigned
	  && int_fits_type_p (value, integer_type_node))
	ansi_type = integer_type_node;
      else if (! spec_long && (base != 10 || spec_unsigned)
	       && int_fits_type_p (value, unsigned_type_node))
	ansi_type = unsigned_type_node;
      else if (! spec_unsigned && !spec_long_long
	       && int_fits_type_p (value, long_integer_type_node))
	ansi_type = long_integer_type_node;
      else if (! spec_long_long
	       && int_fits_type_p (value, long_unsigned_type_node))
	ansi_type = long_unsigned_type_node;
      else if (! spec_unsigned
	       && int_fits_type_p (value, long_long_integer_type_node))
	ansi_type = long_long_integer_type_node;
      else if (int_fits_type_p (value, long_long_unsigned_type_node))
	ansi_type = long_long_unsigned_type_node;

      type = ansi_type;

      if (base == 10 && ! spec_unsigned && TREE_UNSIGNED (type))
	warning ("decimal constant is so large that it is unsigned");

      if (spec_imag)
	{
	  if (TYPE_PRECISION (type)
	      <= TYPE_PRECISION (integer_type_node))
	    error (" Ҽ ó Ѵ. TODO: ó ƾ ߰ ʿ.");
	  else
	    error ("complex integer constant is too wide for 'complex int'");
	}
      else
	TREE_TYPE (value) = type;
    }

  if (p < str + len)
    error ("missing white space after number '%.*s'", (int) (p - str), str);

  return value;
}

int
esm_lexer (tree *value)
{
  struct esm_token *tok;

 retry:
  do
    tok = esm_get_token ();
  while (tok->type == ESM_PADDING);

  *value = NULL_TREE;
  switch (tok->type)
    {
    case ESM_STRING:
      *value = lex_string (tok->val.str.text, tok->val.str.len);
      break;
    case ESM_NUMBER:
      *value = lex_number ((const char *)tok->val.str.text, tok->val.str.len);
      break;
    case ESM_CHAR:
    case ESM_FIELD_TYPE:
    case ESM_METHOD_TYPE:
    case ESM_ADDRESS:
    case ESM_NAME:
      *value = tok->val.node;
      break;
    case ESM_OTHER:
      if (ISGRAPH (tok->val.c))
	error ("stray '%c' in program", tok->val.c);
      else
	error ("stray '\\%o' '%d' in program", tok->val.c, tok->val.c);
      goto retry;
    default:
      break;
    }

  return tok->type;
}

/*  I  FP  1 Ʈ .  */
void
write_i1byte (FILE *fp, int i)
{
  char tmp[1];

  tmp[0] = (char) i & 0xff;
  fwrite (tmp, sizeof (char), 1, fp);
  file_offset += 1;
}

/*  I  FP  2 Ʈ .  */
void
write_i2byte (FILE *fp, int i)
{
  char tmp[2];

  tmp[0] = (char) ((i >> 8) & 0xff);
  tmp[1] = (char) (i & 0xff);

  fwrite (tmp, sizeof (char), 2, fp);
  file_offset += 2;
}

/*  I  FP  4 Ʈ .  */
void
write_i4byte (FILE *fp, int i)
{
  char tmp[4];

  tmp[0] = (char) ((i >> 24) & 0xff);
  tmp[1] = (char) ((i >> 16) & 0xff);
  tmp[2] = (char) ((i >> 8) & 0xff);
  tmp[3] = (char) ((i) & 0xff);

  fwrite (tmp, sizeof (char), 4, fp);
  file_offset += 4;
}

/* float  F  FP  4 Ʈ .  */
void
write_f4byte (FILE *fp, long f)
{
  fwrite (&f, sizeof (long), 1, fp);
  file_offset += 4;
}

/* double  D  FP  8 Ʈ .  */
void
write_d8byte (FILE *fp, long d1, long d2)
{
  fwrite (&d1, sizeof (long), 1, fp);
  fwrite (&d2, sizeof (long), 1, fp);
  file_offset += 8;
}

/*
 * Skip a certain amount of data in a file, either by seeking if
 * possible, or if that fails then by reading and discarding.
 */
static void 
write_skip (unsigned long dist, FILE *fp) 
{
  char buffer[256];		       /* should fit on most stacks :-) */
  
  file_offset += dist;
  /*
   * Got to be careful with fseek: at least one fseek I've tried
   * doesn't approve of SEEK_CUR. So I'll use SEEK_SET and
   * ftell... horrible but apparently necessary.
   */
  if (fseek (fp, dist+ftell(fp), SEEK_SET)) 
    {
      while (dist > 0) 
	{
	  unsigned long len = (dist < sizeof(buffer) ?
			       dist : sizeof(buffer));
	  if (fread (buffer, 1, len, fp) < len) 
	    {
	      perror("fread");
	      exit(1);
	    }
	  dist -= len;
	}
    }
}

void
write_info (t)
     tree t;
{
  //debug_tree (t);
  //fprintf (stderr, "Writer Info: [%s] [%d]\n", TREE_STRING_POINTER (TREE_INFO_WRITER (t)), TREE_STRING_LENGTH (TREE_INFO_WRITER (t)));

  write_i1byte (env_out_file, TREE_STRING_LENGTH (TREE_INFO_WRITER (t)));
  file_offset += fprintf (env_out_file, "%.*s", TREE_STRING_LENGTH (TREE_INFO_WRITER (t)), TREE_STRING_POINTER (TREE_INFO_WRITER (t)));
  write_i1byte (env_out_file, TREE_STRING_LENGTH (TREE_INFO_LICENSE (t)));
  file_offset += fprintf (env_out_file, "%.*s", TREE_STRING_LENGTH (TREE_INFO_LICENSE (t)), TREE_STRING_POINTER (TREE_INFO_LICENSE (t)));
}

void
write_constant_pool (fp)
     FILE *fp;
{
  /* Write Constant Pool Number */
  write_i2byte (fp, cp_count);
  
  /* Write each constant pool */
  {
    struct constant_pool *tm = root_cp;
    while (tm)
      {
	write_i1byte (fp, CONSTANT_TYPE (tm));
	switch (CONSTANT_TYPE (tm))
	  {
	  case CONSTANT_String:
	    write_i2byte (fp, CONSTANT_VARG0 (tm));
	    break;
	  case CONSTANT_Fieldref:
	  case CONSTANT_Accessref:
	  case CONSTANT_NameAndType:
	  case CONSTANT_Sha1AndName:
	  case CONSTANT_Methodref:
	  case CONSTANT_Environment:
	  case CONSTANT_Host:
	    write_i2byte (fp, CONSTANT_VARG0 (tm));
	    write_i2byte (fp, CONSTANT_VARG1 (tm));
	    break;
	  case CONSTANT_InterfaceMethodref:
	    error ("write_constant_pool: ó ٶ\n");
	    break;
	    error ("write_constant_pool: ó ٶ\n");
	    break;
	  case CONSTANT_Integer:
	    write_i4byte (fp, CONSTANT_INTEGER_VALUE_ARG0 (tm));
	    break;
	  case CONSTANT_Float:
	    write_f4byte (fp, CONSTANT_LONG_VALUE_ARG0 (tm));
	    break;
	  case CONSTANT_Long:
	    write_d8byte (fp, CONSTANT_INTEGER_VALUE_ARG0 (tm), CONSTANT_INTEGER_VALUE_ARG1 (tm));
	    break;
	  case CONSTANT_Double:
	    write_d8byte (fp, CONSTANT_LONG_VALUE_ARG0 (tm), CONSTANT_LONG_VALUE_ARG1 (tm));
	    break;
	  case CONSTANT_Utf8:
	    {
	      int j;
	      write_i2byte (fp, strlen (CONSTANT_VSTRING (tm)));
	      
	      for (j=0; j<strlen (CONSTANT_VSTRING (tm)); j++)
		write_i1byte (fp, (int) (CONSTANT_VSTRING (tm)[j]));
	      
	    }
	    break;
	  case CONSTANT_Sha1:
	    {
	      int j;
	      if (CONSTANT_VSTRING (tm))
		{
		  for (j=0; j<SHA_DIGESTSIZE; j++)
		    write_i1byte (fp, (int) (CONSTANT_VSTRING (tm)[j]));
		}
	      else
		{
		  for (j=0; j<SHA_DIGESTSIZE; j++)
		    write_i1byte (fp, (int) 0);	  
		}
	      
	    }
	    break;
	  case CONSTANT_Sha1This:
	    break;
	  default:
	    error ("write_constant_pool: ó ٶ\n");
	    break;
	  }
	tm = CONSTANT_NEXT (tm);
      }
  }
}

void
write_access_host (fp, t)
     FILE *fp;
     tree t;
{
  if (!t)
    return;

  switch (TREE_CODE (t))
    {
    case TREE_LIST:
      write_access_host (fp, TREE_PURPOSE (t));
      write_access_host (fp, TREE_VALUE (t));
      write_access_host (fp, TREE_CHAIN (t));
      break;
    case ACCESS_EXPR:
      if (TREE_CONSTANT_POOL_TYPE (t) == CONSTANT_Accessref)
	write_i2byte (fp, TREE_CONSTANT_POOL_NUMBER (t));
      else
	error ("write_access_host: Access Host  ߸ type [%d].\n", TREE_CONSTANT_POOL_TYPE (t));
      break;
    default:
      error ("write_access_host: ߸ NODE.\n");
      break;
    }
}

void
write_environment_name (fp, t)
     FILE *fp;
     tree t;
{
  if (!t)
    return;

  switch (TREE_CODE (t))
    {
    case TREE_LIST:
      write_environment_name (fp, TREE_PURPOSE (t));
      write_environment_name (fp, TREE_VALUE (t));
      write_environment_name (fp, TREE_CHAIN (t));
      break;
    case IDENTIFIER_NODE:
      if (TREE_CONSTANT_POOL_TYPE (t) == CONSTANT_Environment)
	write_i2byte (fp, TREE_CONSTANT_POOL_NUMBER (t));
      else
	error ("write_environment_name: Environment  ߸ type.\n");
      break;
    default:
      error ("write_environment_name: ߸ NODE.\n");
      break;
    }
}

void
write_access_rights (t, fp)
     tree t;
     FILE *fp;
{
  tree env = TREE_HEADER_ENVIRONMENT (t);
  tree super = TREE_HEADER_SUPER (t);

  write_i2byte (fp, TREE_ENVIRONMENT_ACCESS (env));
  //write_i2byte (fp, TREE_ENVIRONMENT_ACCESS_HOST_COUNT (env));
  //write_access_host (fp, TREE_ENVIRONMENT_ACCESS_HOST (env));
  write_environment_name (fp, TREE_ENVIRONMENT_NAME (env));
  if (super)
    write_i2byte (fp, make_CONSTANT_Environment (super));
  else
    write_i2byte (fp, 0);
}

void
write_interfaces (fp)
     FILE *fp;
{
  write_i2byte (fp, 0);
}

void
write_utf8 (fp, t)
     FILE *fp;
     tree t;
{
  if (!t)
    return;

  switch (TREE_CODE (t))
    {
    case TREE_LIST:
      write_utf8 (fp, TREE_PURPOSE (t));
      write_utf8 (fp, TREE_VALUE (t));
      write_utf8 (fp, TREE_CHAIN (t));
      break;
    case IDENTIFIER_NODE:
      if (TREE_CONSTANT_POOL_TYPE (t) == CONSTANT_Utf8)
	write_i2byte (fp, TREE_CONSTANT_POOL_NUMBER (t));
      else
	error ("write_utf8: utf8  ߸ type.\n");
      break;
    default:
      error ("write_utf8: ߸ NODE.\n");
      break;
    }
}

void
write_attributes (fp, t)
     FILE *fp;
     tree t;
{
  if (!t)
    {
      warning ("write_attributes: TREE node  NULL ̴.");
      return;
    }
  
  switch (TREE_CODE (t))
    {
    case FIELD:
      /*  FIELD   VALUE  Ǿ  , ̰  Ѵ.  */
      if (TREE_FIELD_VALUE (t))
	{
	  /* TODO: VALUE   ,  ڵ 󿡼 ConstantValue  óǵ
	     Ǿ , ٸ 찡 ߻ , ̸  ó  ֵ Ǿ
	     Ѵ.  */
	  write_i2byte (fp, 1);		/* Write Attribute Count */
	  /* Write Attibute Name  */
	  write_i2byte (fp, cp_add (CONSTANT_Utf8, strdup ("ConstantValue")));
	  /*  ConstantValue  Ű ؼ  2 Ʈ
	     ʿϱ , 2  .  */
	  write_i2byte (fp, 2);		/* Write Byte Count */
	  write_i2byte (fp, TREE_CONSTANT_POOL_NUMBER (TREE_FIELD_VALUE (t)));
	}
      else
	write_i2byte (fp, 0);
      break;
    default:
      warning ("write_attributes: ߸ node.");
      break;
    }  
}

void
write_each_fields (fp, t)
     FILE *fp;
     tree t;
{
  if (!t)
    return;

  switch (TREE_CODE (t))
    {
    case TREE_LIST:
      write_each_fields (fp, TREE_PURPOSE (t));
      write_each_fields (fp, TREE_VALUE (t));
      write_each_fields (fp, TREE_CHAIN (t));
      break;
    case FIELD:
      write_i2byte (fp, TREE_FIELD_ACCESS (t));
      //write_i2byte (fp, TREE_FIELD_ACCESS_HOST_COUNT (t));
      //write_access_host (fp, TREE_FIELD_ACCESS_HOST (t));
      write_utf8 (fp, TREE_FIELD_NAME (t));
      write_utf8 (fp, TREE_FIELD_TYPE (t));
      write_attributes (fp, t);
      write_each_fields (fp, TREE_CHAIN (t));
      break;
    default:
      error ("write_each_fields: ߸ NODE.\n");
      break;
    }
}

void
write_fields (t, fp)
     tree t;
     FILE *fp;
{
  tree f = TREE_BODY_FIELDS (t);

  write_i2byte (fp, TREE_BODY_FIELDS_COUNT (t));
  write_each_fields (fp, f);
}

void
write_switch_cases (fp, mth, swt, t)
     FILE *fp;
     tree mth, swt, t;
{
  if (!t)
    return;

  switch (TREE_CODE (t))
    {
    case TREE_LIST:
      write_switch_cases (fp, mth, swt, TREE_PURPOSE (t));
      write_switch_cases (fp, mth, swt, TREE_VALUE (t));
      write_switch_cases (fp, mth, swt, TREE_CHAIN (t));
      break;
    case CASE:
      {
	int tmp;
	struct labels *label = TREE_METHOD_LABEL_LIST (mth);

	tree val = TREE_CASE_VALUE (t);
	write_i4byte (fp, TREE_INT_CST_LOW (val));
	
	tmp = find_label (label, TREE_VALUE (TREE_CASE_NAME (t)));
	write_i4byte (fp, tmp - TREE_SWITCH_LOOKUP_CODE_OFFSET (swt));

	write_switch_cases (fp, mth, swt, TREE_CHAIN (t));
	break;
      }
    default:
      error ("write_switch_cases: ߸ node.\n");
      break;
    }
}

void
write_insns (fp, mth, t)
     FILE *fp;
     tree mth, t;
{
  tree op1, op2;

  if (!t)
    return;

  switch (TREE_CODE (t))
    {
    case TREE_LIST:
      write_insns (fp, mth, TREE_PURPOSE (t));
      write_insns (fp, mth, TREE_VALUE (t));
      write_insns (fp, mth, TREE_CHAIN (t));
      break;
    case IDENTIFIER_NODE:
      break;
    case LIMIT:
      write_insns (fp, mth, TREE_CHAIN (t));
      break;
    case INSN:
      {
	write_i1byte (fp, TREE_BYTE_CODE (TREE_INSN_CODE (t)));
	
	op1 = TREE_INSN_OP1 (t);
	op2 = TREE_INSN_OP2 (t);

	switch (TREE_BYTE_CODE (TREE_INSN_CODE (t)))
	  {
	  case _NOP:		/* opnum = 0 */
	  case _ACONST_NULL:	/* opnum = 1 */
	  case _ICONST_M1:	/* opnum = 2 */
	  case _ICONST_0:	/* opnum = 3 */
	  case _ICONST_1:	/* opnum = 4 */
	  case _ICONST_2:	/* opnum = 5 */
	  case _ICONST_3:	/* opnum = 6 */
	  case _ICONST_4:	/* opnum = 7 */
	  case _ICONST_5:	/* opnum = 8 */
	  case _LCONST_0:	/* opnum = 9 */
	  case _LCONST_1:	/* opnum = 10 */
	  case _FCONST_0:	/* opnum = 11 */
	  case _FCONST_1:	/* opnum = 12 */
	  case _FCONST_2:	/* opnum = 13 */
	  case _DCONST_0:	/* opnum = 14 */
	  case _DCONST_1:	/* opnum = 15 */
	  case _ILOAD_0:
	  case _ILOAD_1:
	  case _ILOAD_2:
	  case _ILOAD_3:
	  case _LLOAD_0:
	  case _LLOAD_1:
	  case _LLOAD_2:
	  case _LLOAD_3:
	  case _FLOAD_0:	/* opnum = 34 */
	  case _FLOAD_1:	/* opnum = 35 */
	  case _FLOAD_2:	/* opnum = 36 */
	  case _FLOAD_3:	/* opnum = 37 */
	  case _DLOAD_0:	/* opnum = 38 */
	  case _DLOAD_1:	/* opnum = 39 */
	  case _DLOAD_2:	/* opnum = 40 */
	  case _DLOAD_3:	/* opnum = 41 */
	  case _ALOAD_0:
	  case _ALOAD_1:
	  case _ALOAD_2:
	  case _ALOAD_3:
	  case _IALOAD:		/* opnum = 46 */
	  case _LALOAD:		/* opnum = 47 */
	  case _FALOAD:		/* opnum = 48 */
	  case _DALOAD:		/* opnum = 49 */
	  case _AALOAD:		/* opnum = 50 */
	  case _BALOAD:		/* opnum = 51 */
	  case _CALOAD:		/* opnum = 52 */
	  case _SALOAD:		/* opnum = 53 */
	  case _ISTORE_0:
	  case _ISTORE_1:
	  case _ISTORE_2:
	  case _ISTORE_3:
	  case _LSTORE_0:	/* opnum = 63 */
	  case _LSTORE_1:	/* opnum = 64 */
	  case _LSTORE_2:	/* opnum = 65 */
	  case _LSTORE_3:	/* opnum = 66 */
	  case _FSTORE_0:	/* opnum = 67 */
	  case _FSTORE_1:	/* opnum = 68 */
	  case _FSTORE_2:	/* opnum = 69 */
	  case _FSTORE_3:	/* opnum = 70 */
	  case _DSTORE_0:	/* opnum = 71 */
	  case _DSTORE_1:	/* opnum = 72 */
	  case _DSTORE_2:	/* opnum = 73 */
	  case _DSTORE_3:	/* opnum = 74 */
	  case _ASTORE_0:	/* opnum = 75 */
	  case _ASTORE_1:	/* opnum = 76 */
	  case _ASTORE_2:	/* opnum = 77 */
	  case _ASTORE_3:	/* opnum = 78 */
	  case _IASTORE:	/* opnum = 79 */
	  case _LASTORE:	/* opnum = 80 */
	  case _FASTORE:	/* opnum = 81 */
	  case _DASTORE:	/* opnum = 82 */
	  case _AASTORE:	/* opnum = 83 */
	  case _BASTORE:	/* opnum = 84 */
	  case _CASTORE:	/* opnum = 85 */
	  case _SASTORE:	/* opnum = 86 */
	  case _POP:		/* opnum = 87 */
	  case _POP2:		/* opnum = 88 */
	  case _DUP:		/* opnum = 89 */
	  case _DUP_X1:		/* opnum = 90 */
	  case _DUP_X2:		/* opnum = 91 */
	  case _DUP2:		/* opnum = 92 */
	  case _DUP2_X1:	/* opnum = 93 */
	  case _DUP2_X2:	/* opnum = 94 */
	  case _SWAP:		/* opnum = 95 */
	  case _IADD:		/* opnum = 96 */
	  case _LADD:		/* opnum = 97 */
	  case _FADD:		/* opnum = 98 */
	  case _DADD:		/* opnum = 99 */
	  case _ISUB:		/* opnum = 100 */
	  case _LSUB:		/* opnum = 101 */
	  case _FSUB:		/* opnum = 102 */
	  case _DSUB:		/* opnum = 103 */
	  case _IMUL:		/* opnum = 104 */
	  case _LMUL:		/* opnum = 105 */
	  case _FMUL:		/* opnum = 106 */
	  case _DMUL:		/* opnum = 107 */
	  case _IDIV:		/* opnum = 108 */
	  case _LDIV:		/* opnum = 109 */
	  case _FDIV:		/* opnum = 110 */
	  case _DDIV:		/* opnum = 111 */
	  case _IREM:		/* opnum = 112 */
	  case _LREM:		/* opnum = 113 */
	  case _FREM:		/* opnum = 114 */
	  case _DREM:		/* opnum = 115 */
	  case _INEG:		/* opnum = 116 */
	  case _LNEG:		/* opnum = 117 */
	  case _FNEG:		/* opnum = 118 */
	  case _DNEG:		/* opnum = 119 */
	  case _ISHL:		/* opnum = 120 */
	  case _LSHL:		/* opnum = 121 */
	  case _ISHR:		/* opnum = 122 */
	  case _LSHR:		/* opnum = 123 */
	  case _IUSHR:		/* opnum = 124 */
	  case _LUSHR:		/* opnum = 125 */
	  case _IAND:		/* opnum = 126 */
	  case _LAND:		/* opnum = 127 */
	  case _IOR:		/* opnum = 128 */
	  case _LOR:		/* opnum = 129 */
	  case _IXOR:		/* opnum = 130 */
	  case _LXOR:		/* opnum = 131 */
	  case _I2L:		/* opnum = 133 */
	  case _I2F:		/* opnum = 134 */
	  case _I2D:		/* opnum = 135 */
	  case _L2I:		/* opnum = 136 */
	  case _L2F:		/* opnum = 137 */
	  case _L2D:		/* opnum = 138 */
	  case _F2I:		/* opnum = 139 */
	  case _F2L:		/* opnum = 140 */
	  case _F2D:		/* opnum = 141 */
	  case _D2I:		/* opnum = 142 */
	  case _D2L:		/* opnum = 143 */
	  case _D2F:		/* opnum = 144 */
	  case _I2B:		/* opnum = 145 */
	  case _I2C:		/* opnum = 146 */
	  case _I2S:		/* opnum = 147 */
	  case _LCMP:		/* opnum = 148 */
	  case _FCMPL:		/* opnum = 149 */
	  case _FCMPG:		/* opnum = 150 */
	  case _DCMPL:		/* opnum = 151 */
	  case _DCMPG:		/* opnum = 152 */
	  case _IRETURN:	/* opnum = 172 */
	  case _LRETURN:	/* opnum = 173 */
	  case _FRETURN:	/* opnum = 174 */
	  case _DRETURN:	/* opnum = 175 */
	  case _ARETURN:	/* opnum = 176 */
	  case _RETURN:		/* opnum = 177 */
	    break;
	  case _BIPUSH:		/* opnum = 16 */
	    write_i1byte (fp, TREE_INT_CST_LOW (op1));
	    break;
	  case _SIPUSH:		/* opnum = 17 */
	    write_i2byte (fp, TREE_INT_CST_LOW (op1));
	    break;
	  case _LDC:		/* opnum = 18 */
	    if (TREE_CONSTANT_POOL_NUMBER (TREE_INSN_CODE (t)) < 256)
	      write_i1byte (fp, TREE_CONSTANT_POOL_NUMBER (t));
	    else
	      error ("write_insns: LDC_W óٶ.\n");
	    break;
	  case _LDC_W:		/* opnum = 19 */
	  case _LDC2_W:		/* opnum = 20 */
	    write_i2byte (fp, TREE_CONSTANT_POOL_NUMBER (t));
	    break;
	  case _ILOAD:		/* opnum = 21 */
	  case _LLOAD:		/* opnum = 22 */
	  case _FLOAD:		/* opnum = 23 */
	  case _DLOAD:		/* opnum = 24 */
	  case _ALOAD:		/* opnum = 25 */
	  case _ISTORE:		/* opnum = 54 */
	  case _LSTORE:		/* opnum = 55 */
	  case _FSTORE:		/* opnum = 56 */
	  case _DSTORE:		/* opnum = 57 */
	  case _ASTORE:		/* opnum = 58 */
	  case _RET:		/* opnum = 169 */
	    write_i1byte (fp, TREE_INT_CST_LOW (TREE_LOCAL_CONSTANT (op1)));
	    break;
	  case _NEWARRAY:
	    write_i1byte (fp, TREE_CONSTANT_POOL_NUMBER (t));
	    break;
	  case _GETSTATIC:	/* opnum = 178 */
	  case _PUTSTATIC:	/* opnum = 179 */
	  case _GETFIELD:	/* opnum = 180 */
	  case _PUTFIELD:	/* opnum = 181 */
	  case _INVOKEVIRTUAL:	/* opnum = 182 */
	  case _INVOKESPECIAL:	/* opnum = 183 */
	  case _INVOKESTATIC:	/* opnum = 184 */
	    write_i2byte (fp, TREE_CONSTANT_POOL_NUMBER (t));
	    break;
	  case _IINC:		/* opnum = 132 */
	    write_i1byte (fp, TREE_INT_CST_LOW (TREE_LOCAL_CONSTANT (op1)));
	    write_i1byte (fp, TREE_INT_CST_LOW (op2));
	    break;
	  case _IFEQ:		/* opnum = 153 */
	  case _IFNE:		/* opnum = 154 */
	  case _IFLT:		/* opnum = 155 */
	  case _IFGE:		/* opnum = 156 */
	  case _IFGT:		/* opnum = 157 */
	  case _IFLE:		/* opnum = 158 */
	  case _IF_ICMPEQ:	/* opnum = 159 */
	  case _IF_ICMPNE:	/* opnum = 160 */
	  case _IF_ICMPLT:	/* opnum = 161 */
	  case _IF_ICMPGE:	/* opnum = 162 */
	  case _IF_ICMPGT:	/* opnum = 163 */
	  case _IF_ICMPLE:	/* opnum = 164 */
	  case _IF_ACMPEQ:	/* opnum = 165 */
	  case _IF_ACMPNE:	/* opnum = 166 */
	  case _GOTO:		/* opnum = 167 */
	  case _JSR:		/* opnum = 168 */
	  case _JSR_W:		/* opnum = 201 */
	    {
	      int diff, offset;

	      offset = find_label (TREE_METHOD_LABEL_LIST (mth), TREE_VALUE (op1));

	      if (offset == -1)
		error ("write_insns: ش   ʽϴ.\n");

	      diff = offset - TREE_INSN_CODE_OFFSET (t);

	      if (TREE_BYTE_CODE (TREE_INSN_CODE (t)) == _JSR_W)
		write_i4byte (fp, diff);
	      else
		write_i2byte (fp, diff);
	      break;
	    }
	    break;
	  default:
	    error ("write_insns: óؾ ϴ insn [%s]\n", resword_insns[TREE_BYTE_CODE (TREE_INSN_CODE (t))].word);
	    debug_tree (t);
	    break;
	  }
	write_insns (fp, mth, TREE_CHAIN (t));
	break;
      }
    case SWITCH_LOOKUP:
      {
	int offset = TREE_SWITCH_LOOKUP_CODE_OFFSET (t) + 1;
	int i, tmp, padnum = (offset % 4) ? 4 - (offset % 4) : 0;
	struct labels *label = TREE_METHOD_LABEL_LIST (mth);

	/* Write opcode.  */
	write_i1byte (fp, (int) _LOOKUPSWITCH);

	/* Write Null pad bytes.  */
	for (i=0; i<padnum; i++)
	  write_i1byte (fp, 0);
	
	/* Write Default Address.  */
	tmp = find_label (label, TREE_VALUE (TREE_CASE_NAME (TREE_SWITCH_LOOKUP_DEFAULT (t))));
	if (tmp == -1)
	  error ("write_insns: find_label () Լ ȯ -1 ̴.\n");
	write_i4byte (fp, tmp - TREE_SWITCH_LOOKUP_CODE_OFFSET (t));

	/* Write Case Number.  */
	write_i4byte (fp, TREE_SWITCH_LOOKUP_CASE_NUMBER (t));

	/* Write Key and Address.  */
	write_switch_cases (fp, mth, t, TREE_SWITCH_LOOKUP_LIST (t));
	break;
      }
    case LABEL:
      break;
    default:
      error ("write_insns: ߸ NODE.\n");
      debug_tree (t);
      break;
    }
}

void
write_each_methods (fp, t)
     FILE *fp;
     tree t;
{
  if (!t)
    return;

  switch (TREE_CODE (t))
    {
    case TREE_LIST:
      write_each_methods (fp, TREE_PURPOSE (t));
      write_each_methods (fp, TREE_VALUE (t));
      write_each_methods (fp, TREE_CHAIN (t));
      break;
    case METHODS:
      write_i2byte (fp, TREE_METHOD_ACCESS (t));
      //write_i2byte (fp, TREE_METHOD_ACCESS_HOST_COUNT (t));
      //write_access_host (fp, TREE_METHOD_ACCESS_HOST (t));
      write_utf8 (fp, TREE_METHOD_NAME (t));
      write_utf8 (fp, TREE_METHOD_TYPE (t));
      write_i2byte (fp, TREE_METHOD_ATTRIBUTE_COUNT (t));
      if (TREE_METHOD_INSN_LENGTH (t))
	{
	  /* Attribute Name */
	  write_i2byte (fp, cp_add (CONSTANT_Utf8, strdup ("Code")));
	  /* Attribute Length :    */
	  write_i4byte (fp, 2+2+4+TREE_METHOD_INSN_LENGTH (t)+2+2);
	  write_i2byte (fp, TREE_METHOD_MAX_STACK (t));
	  write_i2byte (fp, TREE_METHOD_MAX_LOCALS (t));
	  write_i4byte (fp, TREE_METHOD_INSN_LENGTH (t));
	  write_insns (fp, t, TREE_METHOD_STATEMENTS (t));
	  /* Handler Count */
	  write_i2byte (fp, 0);
	  /* Attribute Count */
	  write_i2byte (fp, 0);
	}
      write_each_methods (fp, TREE_CHAIN (t));
      break;
    default:
      break;
    }
}

void
write_methods (t, fp)
     tree t;
     FILE *fp;
{
  tree m = TREE_BODY_METHODS (t);

  write_i2byte (fp, TREE_BODY_METHODS_COUNT (t));
  write_each_methods (fp, m);  
}

void
write_extra_attribute (fp, t)
     FILE *fp;
     tree t;
{
  if (TREE_INFO_FILENAME (t))
    {
      write_i2byte (fp, 1);
      write_i2byte (fp, cp_add (CONSTANT_Utf8, strdup ("SourceFile")));
      write_i4byte (fp, 0);
      write_i2byte (fp, TREE_CONSTANT_POOL_NUMBER (TREE_INFO_FILENAME (t)));
    }
  else
    {
      if (1)
        {
          /* start - ӽ  ڵԴϴ.  */
          write_i2byte (fp, 1);
          write_i2byte (fp, cp_add (CONSTANT_Utf8, strdup ("gnu.gcj.gcj-compiled")));
          write_i4byte (fp, 0);
        }
      else
        {
          /* end - ӽ  ڵԴϴ.  */
          write_i2byte (fp, 0);
        }
    }
}

void
carve_unique_id (void)
{
  int len, i, size;
  char *buf, tmp[256];
  char sha1[SHA_DIGESTSIZE];
  FILE *fp;
  
  len = fileSize (env_file_name);
  fp = fopen (env_file_name, "r+");

  memset (sha1, '\0', SHA_DIGESTSIZE);

  buf = (char *) malloc (sizeof (char) * len);
  i   = fread (buf, 1, len, fp);
  if (len != i)
    error ("carve_unique_id: error.\n");

  ap_sha1_base64 (buf, len, tmp);
#if DEBUG
  fprintf (stderr, "carve_unique_id: i = %d len = %d, SHA1BASE64 ڿ : %s %d\n", i, len, tmp, buf[len - 2]);
#endif
  size = ap_base64decode_binary (sha1, tmp);

  (void) fseek (fp, 8L, SEEK_SET);

  for (i = 0; i < size; i++)
    write_i1byte (fp, sha1[i]);

  //free (buf);

  fclose (fp);
}

void
before_write_environment (void)
{
  struct constant_pool *cp, *next=NULL, *prev=NULL;
  
  for (cp = root_cp; cp; cp = next)
    {
      next = CONSTANT_NEXT (cp);
      CONSTANT_NEXT (cp) = prev;
      prev = cp;
    }

  root_cp = prev;
}

enum newarray_enum
  {
    T_BOOLEAN 	= 4,
    T_CHAR 	= 5,
    T_FLOAT 	= 6,
    T_DOUBLE 	= 7,
    T_BYTE 	= 8,
    T_SHORT 	= 9,
    T_INT 	= 10,
    T_LONG 	= 11
  };

/* Environment path   ȯѴ.  package_path  Ǿ  , ƴ 
    Ͽ ڿ ȯǾ Ѵ.  */
char *
make_envpath (t)
     tree t;
{
  char tmp[1024];

  if (TREE_CODE (t) != IDENTIFIER_NODE)
    {
      warning ("make_envpath: TREE node  IDENTIFIER_NODE  ƴմϴ.\n");
      return NULL;
    }

  if (package_path && !strcmp (IDENTIFIER_POINTER (t), IDENTIFIER_POINTER (current_envname)))
    {
      strcpy (tmp, package_path);
      strcat (tmp, IDENTIFIER_POINTER (t));
    }
  else
    strcpy (tmp, IDENTIFIER_POINTER (t));

  return strdup (tmp);
}

struct newarray
{
  char *name;
  enum newarray_enum type;
};

struct newarray newarray_strings[] = 
  {
    { "boolean",	T_BOOLEAN },
    { "char",		T_CHAR },
    { "float",		T_FLOAT },
    { "double",		T_DOUBLE },
    { "byte",		T_BYTE },
    { "short",		T_SHORT },
    { "int",		T_INT },
    { "long",		T_LONG },
    { NULL, },
  };

/*  INSN  ؼǾ , ȣǴ  INSN  ó־ ϴ ͵
   óѴ.   óϴ operation  Ʒ  ͵ ִ.  

     1. Constant Pool  ϼѴ.  */
void
process_insn (t)
     tree t;
{
  tree code, op1, op2;

  if (TREE_CODE (t) != INSN)
    return;

  code = TREE_INSN_CODE (t);
  op1 = TREE_INSN_OP1 (t);
  op2 = TREE_INSN_OP2 (t);

  switch (TREE_BYTE_CODE (code))
    {
    case _NOP:
    case _ACONST_NULL:
    case _ICONST_M1:	/* opnum = 2 */
    case _ICONST_0:
    case _ICONST_1:
    case _ICONST_2:
    case _ICONST_3:
    case _ICONST_4:
    case _ICONST_5:
    case _LCONST_0:	/* opnum = 9 */
    case _LCONST_1:	/* opnum = 10 */
    case _FCONST_0:	/* opnum = 11 */
    case _FCONST_1:	/* opnum = 12 */
    case _FCONST_2:	/* opnum = 13 */
    case _DCONST_0:	/* opnum = 14 */
    case _DCONST_1:	/* opnum = 15 */
    case _BIPUSH:	/* opnum = 16 */
    case _SIPUSH:	/* opnum = 17 */
    case _ILOAD:	/* opnum = 21 */
    case _LLOAD:	/* opnum = 22 */
    case _FLOAD:	/* opnum = 23 */
    case _DLOAD:	/* opnum = 24 */
    case _ALOAD:	/* opnum = 25 */
    case _ILOAD_0:	/* opnum = 26 */
    case _ILOAD_1:	/* opnum = 27 */
    case _ILOAD_2:	/* opnum = 28 */
    case _ILOAD_3:	/* opnum = 29 */
    case _LLOAD_0:	/* opnum = 30 */
    case _LLOAD_1:	/* opnum = 31 */
    case _LLOAD_2:	/* opnum = 32 */
    case _LLOAD_3:	/* opnum = 33 */
    case _FLOAD_0:	/* opnum = 34 */
    case _FLOAD_1:	/* opnum = 35 */
    case _FLOAD_2:	/* opnum = 36 */
    case _FLOAD_3:	/* opnum = 37 */
    case _DLOAD_0:	/* opnum = 38 */
    case _DLOAD_1:	/* opnum = 39 */
    case _DLOAD_2:	/* opnum = 40 */
    case _DLOAD_3:	/* opnum = 41 */
    case _ALOAD_0:	/* opnum = 42 */
    case _ALOAD_1:	/* opnum = 43 */
    case _ALOAD_2:	/* opnum = 44 */
    case _ALOAD_3:	/* opnum = 45 */
    case _IALOAD:	/* opnum = 46 */
    case _LALOAD:	/* opnum = 47 */
    case _FALOAD:	/* opnum = 48 */
    case _DALOAD:	/* opnum = 49 */
    case _AALOAD:	/* opnum = 50 */
    case _BALOAD:	/* opnum = 51 */
    case _CALOAD:	/* opnum = 52 */
    case _SALOAD:	/* opnum = 53 */
    case _ISTORE:	/* opnum = 54 */
    case _LSTORE:	/* opnum = 55 */
    case _FSTORE:	/* opnum = 56 */
    case _DSTORE:	/* opnum = 57 */
    case _ASTORE:	/* opnum = 58 */
    case _ISTORE_0:	/* opnum = 59 */
    case _ISTORE_1:
    case _ISTORE_2:
    case _ISTORE_3:
    case _LSTORE_0:	/* opnum = 63 */
    case _LSTORE_1:	/* opnum = 64 */
    case _LSTORE_2:	/* opnum = 65 */
    case _LSTORE_3:	/* opnum = 66 */
    case _FSTORE_0:	/* opnum = 67 */
    case _FSTORE_1:	/* opnum = 68 */
    case _FSTORE_2:	/* opnum = 69 */
    case _FSTORE_3:	/* opnum = 70 */
    case _DSTORE_0:	/* opnum = 71 */
    case _DSTORE_1:	/* opnum = 72 */
    case _DSTORE_2:	/* opnum = 73 */
    case _DSTORE_3:	/* opnum = 74 */
    case _ASTORE_0:	/* opnum = 75 */
    case _ASTORE_1:	/* opnum = 76 */
    case _ASTORE_2:	/* opnum = 77 */
    case _ASTORE_3:	/* opnum = 78 */
    case _IASTORE:	/* opnum = 79 */
    case _LASTORE:	/* opnum = 80 */
    case _FASTORE:	/* opnum = 81 */
    case _DASTORE:	/* opnum = 82 */
    case _AASTORE:	/* opnum = 83 */
    case _BASTORE:	/* opnum = 84 */
    case _CASTORE:	/* opnum = 85 */
    case _SASTORE:	/* opnum = 86 */
    case _POP:		/* opnum = 87 */
    case _POP2:		/* opnum = 88 */
    case _DUP:		/* opnum = 89 */
    case _DUP_X1:	/* opnum = 90 */
    case _DUP_X2:	/* opnum = 91 */
    case _DUP2:		/* opnum = 92 */
    case _DUP2_X1:	/* opnum = 93 */
    case _DUP2_X2:	/* opnum = 94 */
    case _SWAP:		/* opnum = 95 */
    case _IADD:		/* opnum = 96 */
    case _LADD:		/* opnum = 97 */
    case _FADD:		/* opnum = 98 */
    case _DADD:		/* opnum = 99 */
    case _ISUB:		/* opnum = 100 */
    case _LSUB:		/* opnum = 101 */
    case _FSUB:		/* opnum = 102 */
    case _DSUB:		/* opnum = 103 */
    case _IMUL:		/* opnum = 104 */
    case _LMUL:		/* opnum = 105 */
    case _FMUL:		/* opnum = 106 */
    case _DMUL:		/* opnum = 107 */
    case _IDIV:		/* opnum = 108 */
    case _LDIV:		/* opnum = 109 */
    case _FDIV:		/* opnum = 110 */
    case _DDIV:		/* opnum = 111 */
    case _IREM:		/* opnum = 112 */
    case _LREM:		/* opnum = 113 */
    case _FREM:		/* opnum = 114 */
    case _DREM:		/* opnum = 115 */
    case _INEG:		/* opnum = 116 */
    case _LNEG:		/* opnum = 117 */
    case _FNEG:		/* opnum = 118 */
    case _DNEG:		/* opnum = 119 */
    case _ISHL:		/* opnum = 120 */
    case _LSHL:		/* opnum = 121 */	      
    case _ISHR:		/* opnum = 122 */
    case _LSHR:		/* opnum = 123 */	      
    case _IUSHR:	/* opnum = 124 */
    case _LUSHR:	/* opnum = 125 */	      
    case _IAND:		/* opnum = 126 */
    case _LAND:		/* opnum = 127 */
    case _IOR:		/* opnum = 128 */
    case _LOR:		/* opnum = 129 */
    case _IXOR:		/* opnum = 130 */
    case _LXOR:		/* opnum = 131 */
    case _IINC:		/* opnum = 132 */
    case _I2L:		/* opnum = 133 */
    case _I2F:		/* opnum = 134 */
    case _I2D:		/* opnum = 135 */
    case _L2I:		/* opnum = 136 */
    case _L2F:		/* opnum = 137 */
    case _L2D:		/* opnum = 138 */
    case _F2I:		/* opnum = 139 */
    case _F2L:		/* opnum = 140 */
    case _F2D:		/* opnum = 141 */
    case _D2I:		/* opnum = 142 */
    case _D2L:		/* opnum = 143 */
    case _D2F:		/* opnum = 144 */
    case _I2B:		/* opnum = 145 */
    case _I2C:		/* opnum = 146 */
    case _I2S:		/* opnum = 147 */
    case _LCMP:		/* opnum = 148 */
    case _FCMPL:	/* opnum = 149 */
    case _FCMPG:	/* opnum = 150 */
    case _DCMPL:	/* opnum = 151 */
    case _DCMPG:	/* opnum = 152 */
    case _IFEQ:		/* opnum = 153 */
    case _IFNE:		/* opnum = 154 */
    case _IFLT:		/* opnum = 155 */
    case _IFGE:		/* opnum = 156 */
    case _IFGT:		/* opnum = 157 */
    case _IFLE:		/* opnum = 158 */
    case _IF_ICMPEQ:	/* opnum = 159 */
    case _IF_ICMPNE:	/* opnum = 160 */
    case _IF_ICMPLT:	/* opnum = 161 */
    case _IF_ICMPGE:	/* opnum = 162 */
    case _IF_ICMPGT:	/* opnum = 163 */
    case _IF_ICMPLE:	/* opnum = 164 */
    case _IF_ACMPEQ:	/* opnum = 165 */
    case _IF_ACMPNE:	/* opnum = 166 */
    case _GOTO:		/* opnum = 167 */
    case _JSR:		/* opnum = 168 */
    case _RET:		/* opnum = 169 */
    case _JSR_W:	/* opnum = 201 */
    case _IRETURN:	/* opnum = 172 */
    case _LRETURN:	/* opnum = 173 */
    case _FRETURN:	/* opnum = 174 */
    case _DRETURN:	/* opnum = 175 */
    case _ARETURN:	/* opnum = 176 */
    case _RETURN:	/* opnum = 177 */
      break;
    case _LDC:		/* opnum = 18 */
    case _LDC_W:	/* opnum = 19 */
    case _LDC2_W:	/* opnum = 20 */
      if (TREE_BYTE_CODE (code) == _LDC2_W 
	  && (TYPE_MODE (TREE_TYPE (op1)) == SImode || TYPE_MODE (TREE_TYPE (op1)) == SFmode))
	{
	  error ("process_insn: LDC2_W   ߸ type Դϴ.\n");
	  debug_tree (op1);
	}

      switch (TREE_CODE (op1))
	{
	case STRING_CST:
	  TREE_CONSTANT_POOL_NUMBER (t) = 
	    cp_add (CONSTANT_String, 
		    cp_add (CONSTANT_Utf8, TREE_STRING_POINTER (op1)));
	  break;
	case INTEGER_CST:
	  if (!TREE_TYPE (op1))
	    {
	      error ("process_insn: op1  type  NULL Դϴ.\n");
	      debug_tree (op1);
	      break;
	    }

	  switch (TYPE_MODE (TREE_TYPE (op1)))
	    {
	    case SImode:	/* 4 bytes integer */
	      TREE_CONSTANT_POOL_NUMBER (t) = cp_add (CONSTANT_Integer, TREE_INT_CST_LOW (op1));
	      break;
	    case DImode:	/* 8 bytes integer */
	      TREE_CONSTANT_POOL_NUMBER (t) = cp_add (CONSTANT_Long, TREE_INT_CST_HIGH (op1), TREE_INT_CST_LOW (op1));
	      break;
	    default:
	      warning ("process_insn:      type Դϴ.  mode = %s\n",
		       mode_name[TYPE_MODE (TREE_TYPE (op1))]);
	      break;
	    }
	  break;
	case REAL_CST:
	  {
	    long l[2];

	    if (!TREE_TYPE (op1))
	      {
		error ("process_insn: op1  type  NULL Դϴ.\n");
		debug_tree (op1);
		break;
	      }

	    switch (TYPE_MODE (TREE_TYPE (op1)))
	      {
	      case SFmode:	/* 4 bytes */
		REAL_VALUE_TO_TARGET_SINGLE (TREE_REAL_CST (op1), l[0]);
		TREE_CONSTANT_POOL_NUMBER (t) = cp_add (CONSTANT_Float, l[0]);
		break;
	      case DFmode:	/* 8 bytes */
		REAL_VALUE_TO_TARGET_DOUBLE (TREE_REAL_CST (op1), l);
		TREE_CONSTANT_POOL_NUMBER (t) = cp_add (CONSTANT_Double, l[0], l[1]);
		break;
	      default:
		warning ("process_insn:    Ǽ  type Դϴ.  mode = %s\n",
			 mode_name[TYPE_MODE (TREE_TYPE (op1))]);
		break;
	      }
	    break;
	  }
	default:
	  error ("process_insn: óؾ ϴ LDC ڵ \n");
	  debug_tree (t);
	}
      break;
    case _NEWARRAY:
      {
	int i;
	int got_type = 0;
	for (i=0; newarray_strings[i].name != NULL; i++)
	  {
	    if (!strncmp (newarray_strings[i].name, TREE_STRING_POINTER (op1), TREE_STRING_LENGTH (op1)))
	      {
		/* NEWARRAY  , TREE_CONSTANT_POOL_NUMBER (code)  ش type ȣ .  */
		TREE_CONSTANT_POOL_NUMBER (t) = newarray_strings[i].type;
		got_type = 1;
		break;
	      }
	  }
	if (got_type == 0)
	  error ("process_insn:  type  ã  .\n");
	break;
      }
    case _GETSTATIC:		/* opnum = 178 */
    case _PUTSTATIC:		/* opnum = 179 */
    case _GETFIELD:		/* opnum = 180 */
    case _PUTFIELD:		/* opnum = 181 */
      {
	tree url, type, ename, mfname;

	url = TREE_FIELDREF_URL (op1);
	type = TREE_FIELDREF_TYPE (op1);

	ename = TREE_ENVIRONMENT_URL_NAME (url);
	mfname = TREE_ENVIRONMENT_URL_FIELDORMETHOD_NAME (url);

	switch (TREE_CODE (ename))
	  {
	  case ADDR:
	    {
	      int eidx;
	      char *envpath = make_envpath (TREE_SEGMENT_ABSOLUTE_NAME (TREE_EADDRESS_PATHANDSHA1 (ename)));

	      if (TREE_SEGMENT_SHA1 (TREE_EADDRESS_PATHANDSHA1 (ename)))
		{
		  eidx = 
		    cp_add (CONSTANT_Environment,
			    cp_add (CONSTANT_Host,
				    cp_add (CONSTANT_Utf8, IDENTIFIER_POINTER (TREE_ADDRESS_IP (TREE_EADDRESS_HOSTNAME (ename)))),
				    cp_add (CONSTANT_Integer, DEFAULT_ESERVER_PORT)),
			    cp_add (CONSTANT_Sha1AndName, 
				    cp_add (CONSTANT_Sha1, TREE_SEGMENT_SHA1_BINARY (TREE_EADDRESS_PATHANDSHA1 (ename))),
				    cp_add (CONSTANT_Utf8, envpath)));
		}
	      else
		{
		  eidx = 
		    cp_add (CONSTANT_Environment,
			    cp_add (CONSTANT_Host,
				    cp_add (CONSTANT_Utf8, IDENTIFIER_POINTER (TREE_ADDRESS_IP (TREE_EADDRESS_HOSTNAME (ename)))),
				    cp_add (CONSTANT_Integer, DEFAULT_ESERVER_PORT)),
			    cp_add (CONSTANT_Sha1AndName, 
				    cp_add (CONSTANT_Sha1This),
				    cp_add (CONSTANT_Utf8, envpath)));
		}
	      
	      /* CONSTANT_NameAndType ϼ  ƾ.  */
	      TREE_CONSTANT_POOL_NUMBER (t) = 
		cp_add (CONSTANT_Fieldref, 
			eidx,
			cp_add (CONSTANT_NameAndType, 
				cp_add (CONSTANT_Utf8, IDENTIFIER_POINTER (TREE_VALUE (mfname))),
				cp_add (CONSTANT_Utf8, IDENTIFIER_POINTER (type))));
	      break;
	    }
	  case TREE_LIST:
	    {
	      /* CONSTANT_NameAndType ϼ  ƾ.  */
	      TREE_CONSTANT_POOL_NUMBER (t) = 
		cp_add (CONSTANT_Fieldref, 
			cp_add (CONSTANT_Environment,
				cp_add (CONSTANT_Host,
					cp_add (CONSTANT_Utf8, "localhost"),
					cp_add (CONSTANT_Integer, DEFAULT_ESERVER_PORT)),
				cp_add (CONSTANT_Sha1AndName, 
					cp_add (CONSTANT_Sha1This),
					cp_add (CONSTANT_Utf8, make_envpath (TREE_VALUE (ename))))),
			cp_add (CONSTANT_NameAndType, 
				cp_add (CONSTANT_Utf8, IDENTIFIER_POINTER (TREE_VALUE (mfname))),
				cp_add (CONSTANT_Utf8, IDENTIFIER_POINTER (type))));
	      break;
	    }
	  default:
	    break;
	  }
	break;
      }
    case _INVOKEVIRTUAL:	/* opnum = 182 */
    case _INVOKESPECIAL:	/* opnum = 183 */
    case _INVOKESTATIC:		/* opnum = 184 */
      {
	tree url, type, ename, mfname;

	url = TREE_METHODREF_URL (op1);
	type = TREE_METHODREF_TYPE (op1);

	ename = TREE_ENVIRONMENT_URL_NAME (url);
	mfname = TREE_ENVIRONMENT_URL_FIELDORMETHOD_NAME (url);

	switch (TREE_CODE (ename))
	  {
	  case ADDR:
	    {
	      int eidx;
	      char *envpath = make_envpath (TREE_SEGMENT_ABSOLUTE_NAME (TREE_EADDRESS_PATHANDSHA1 (ename)));

	      if (TREE_SEGMENT_SHA1 (TREE_EADDRESS_PATHANDSHA1 (ename)))
		{
		  eidx = 
		    cp_add (CONSTANT_Environment,
			    cp_add (CONSTANT_Host,
				    cp_add (CONSTANT_Utf8, IDENTIFIER_POINTER (TREE_ADDRESS_IP (TREE_EADDRESS_HOSTNAME (ename)))),
				    cp_add (CONSTANT_Integer, DEFAULT_ESERVER_PORT)),
			    cp_add (CONSTANT_Sha1AndName, 
				    cp_add (CONSTANT_Sha1, TREE_SEGMENT_SHA1_BINARY (TREE_EADDRESS_PATHANDSHA1 (ename))),
				    cp_add (CONSTANT_Utf8, envpath)));
		}
	      else
		{
		  eidx = 
		    cp_add (CONSTANT_Environment,
			    cp_add (CONSTANT_Host,
				    cp_add (CONSTANT_Utf8, IDENTIFIER_POINTER (TREE_ADDRESS_IP (TREE_EADDRESS_HOSTNAME (ename)))),
				    cp_add (CONSTANT_Integer, DEFAULT_ESERVER_PORT)),
			    cp_add (CONSTANT_Sha1AndName, 
				    cp_add (CONSTANT_Sha1This),
				    cp_add (CONSTANT_Utf8, envpath)));
		}
	      
	      /* CONSTANT_NameAndType ϼ  ƾ.  */
	      TREE_CONSTANT_POOL_NUMBER (t) = 
		cp_add (CONSTANT_Methodref, 
			eidx,
			cp_add (CONSTANT_NameAndType, 
				cp_add (CONSTANT_Utf8, IDENTIFIER_POINTER (TREE_VALUE (mfname))),
				cp_add (CONSTANT_Utf8, IDENTIFIER_POINTER (type))));
	      break;
	    }
	  case TREE_LIST:
	    /* CONSTANT_NameAndType ϼ  ƾ.  */
	    TREE_CONSTANT_POOL_NUMBER (t) = 
	      cp_add (CONSTANT_Methodref, 
		      cp_add (CONSTANT_Environment,
			      cp_add (CONSTANT_Host,
				      cp_add (CONSTANT_Utf8, "localhost"),
				      cp_add (CONSTANT_Integer, DEFAULT_ESERVER_PORT)),
			      cp_add (CONSTANT_Sha1AndName, 
				      cp_add (CONSTANT_Sha1This),
				      cp_add (CONSTANT_Utf8, make_envpath (TREE_VALUE (ename))))),
		      cp_add (CONSTANT_NameAndType, 
			      cp_add (CONSTANT_Utf8, IDENTIFIER_POINTER (TREE_VALUE (mfname))),
			      cp_add (CONSTANT_Utf8, IDENTIFIER_POINTER (type))));
	    break;
	  default:
	    break;
	  }
	break;
      }
    default:
      error ("process_insn: óؾ ϴ ڵ  [%s]\n", IDENTIFIER_POINTER (code));
      debug_tree (t);
    }
}

/* T  , MATCH tree node ̴.  */
int
process_match (t)
     tree t;
{
  struct matchs *a;

  a = find_match (t);
  if (!a)
    insert_match (t);

  return 0;
}

#define ENVIRONMENT_MAGIC		0xcafebeaf
#define ENVIRONMENT_MAJOR		1
#define ENVIRONMENT_MINOR		0

void
psuper (tree t)
{
  tree super = TREE_HEADER_SUPER (t);

  if (super)
    make_CONSTANT_Environment (super);
}

int
finish_program (t)
     tree t;
{
  cp_add (CONSTANT_Utf8, strdup ("gnu.gcj.gcj-compiled"));
  cp_add (CONSTANT_Utf8, strdup ("Code"));
  psuper (TREE_PROGRAM_HEADER (t));

  before_write_environment ();

  write_i4byte (env_out_file, ENVIRONMENT_MAGIC);
  write_i2byte (env_out_file, ENVIRONMENT_MINOR);
  write_i2byte (env_out_file, ENVIRONMENT_MAJOR);

  /*  ϰ ϴ  SHA1    ȮѴ.  */
  write_skip (20, env_out_file);

  write_info (TREE_PROGRAM_INFO (t));
  write_constant_pool (env_out_file);
  write_access_rights (TREE_PROGRAM_HEADER (t), env_out_file);

  write_interfaces (env_out_file);

  write_fields (TREE_PROGRAM_BODY (t), env_out_file);
  write_methods (TREE_PROGRAM_BODY (t), env_out_file);
  write_extra_attribute (env_out_file, TREE_PROGRAM_INFO (t));

  fclose (env_out_file);

  carve_unique_id ();

  if (opt_debug_all)
    {
      debug_tree (t);
      print_cp ();
    }

  return 0;
}


typedef struct
{
  const char *const string;
  int *const variable;
  const int on_value;
  const char *const description;
}
lang_independent_options;

/* Table of language-specific options.  */

static const struct lang_opt
{
  const char *const option;
  const char *const description;
}
documented_lang_options[] =
{
  { "-Wall", 
    "Enable most warning messages" },
};

/* toplev_main  argument  纻 */
int save_argc;
char **save_argv;

/*  0  ƴ  츮 option  ؼ  Ĩϴ.  */
static int exit_after_options = 0;

/* Print various extra warnings.  -W.  */

int extra_warnings = 0;

/* Name with which this program was invoked.  */

static const char *programname;

static void
display_help ()
{
  printf ("Usage: %s [options] file...\n", programname);
  fputs ("Options:\n", stdout);

  printf ("  -d[letters]              Enable dumps from specific passes of the compiler\n");
  fputs  ("  --help                   Display this information\n", stdout);

  printf ("\
\nOptions starting with -g, -f, -m, -O, -W, or --param are automatically\n\
 passed on to the various sub-processes invoked by %s.  In order to pass\n\
 other options on to these processes the -W<letter> options must be used.\n\
", programname);

  /* The rest of the options are displayed by invocations of the various
     sub-processes.  */

}

/* dump file  ǥմϴ.  */

struct dump_file_info
{
  /* The unique extension to apply, e.g. ".jump".  */
  const char *const extension;

  /* The -d<c> character that enables this dump file.  */
  char const debug_switch;

  /* True if there is a corresponding graph dump file.  */
  char const graph_dump_p;

  /* True if the user selected this dump.  */
  char enabled;

  /* True if the files have been initialized (ie truncated).  */
  char initialized;
};

enum dump_file_index
{
  DFI_test,
  DFI_test1,
  DFI_MAX
};

static struct dump_file_info dump_file[DFI_MAX] =
{
  { "test",	'r', 0, 0, 0 },
  { "test1",	'e', 1, 0, 0 },
};

/* Parse a -d... command line switch.  */

static void
decode_d_option (arg)
     const char *arg;
{
  int i, c, matched;

  while (*arg)
    switch (c = *arg++)
      {
      case 'a':
	opt_debug_all = 1;
	break;
      case 'y':
	yydebug = 1;
	break;
      default:
	matched = 0;
	for (i = 0; i < (int) DFI_MAX; ++i)
	  if (c == dump_file[i].debug_switch)
	    {
	      dump_file[i].enabled = 1;
	      matched = 1;
	    }

	if (! matched)
	  warning ("unrecognized essembly debugging option: %c", c);
	break;
      }
}

/* Print version information to FILE.
   Each line begins with INDENT (for the case where FILE is the
   assembler output file).  */

static void
print_version (file, indent)
     FILE *file;
     const char *indent;
{
#ifndef VERSION
#define VERSION "[?]"
#endif
  fprintf (file, "esm version %s\n", VERSION);
}

/* language-independent option ν argv  ù° ڸ صմϴ.
   ص ڿ  ǵ ݴϴ. */

static unsigned int
decode_option (argc, argv)
     int argc;
     char **argv;
{
  char *arg = argv[0];

  if (arg[0] != '-' || arg[1] == 0)
    {
      if (arg[0] == '+')
	return 0;

      filename = arg;

      return 1;
    }

  arg++;

  if (!strcmp (arg, "-help"))
    {
      display_help ();
      exit_after_options = 1;
    }

  if (!strcmp (arg, "-version"))
    {
      print_version (stderr, "");
      exit_after_options = 1;
    }

  if (*arg == 'Y')
    arg++;

  switch (*arg)
    {
    default:
      return 0;
    case 'd':
      decode_d_option (arg + 1);
      break;
    case 'o':
      if (arg[1] == 0)
	{
	  if (argc == 1)
	    return 0;

	  env_file_name = argv[1];
	  return 2;
	}
      return 0;
    case 'v':
      if (!strcmp (arg, "version"))
	version_flag = 1;
      else
	return 0;
      break;
    }

  return 1;
}

/* command line option  ؼϰ default falg  մϴ.
    Լ language-independent option-independent ʱȭĿ
   ȣ˴ϴ. ׸ ּ option processing  մϴ.
   Outputting diagnostics  ʱȭ Ǿ GC  identifier
   hashtable  ʱȭ  ʾҽϴ. */
static void
parse_options_and_default_flags (argc, argv)
     int argc;
     char **argv;
{
  int i;

  /* md  comment ν arg  ϱ  츦 
      Ӵϴ. */
  save_argc = argc;
  save_argv = argv;

  /* Ϲ command line ġ ص մϴ. */
  for (i = 1; i < argc;)
    {
      int processed;

      processed = decode_option (argc - i, argv + i);

      if (processed)
	i += processed;
      else
	{
	  const char *option = NULL;
	  const char *lang = NULL;
	  unsigned int j;

	  /* Command line switch   󿡼  
             ɼ , ٸ  Ǵ  
             ֽϴ.  Ϸ   ȣȯ ϱ
              (   ޼  ) ̷ ɼ
             ⼭ ˻մϴ.    ´  ã 
             extra_warnings  Ǿ ִ ο   ޼
             ϸ ׷   츮   ɼ
             մϴ.  */
	  for (j = 0; j < ARRAY_SIZE (documented_lang_options); j++)
	    {
	      option = documented_lang_options[j].option;

	      if (option == NULL)
		lang = documented_lang_options[j].description;
	      else if (! strncmp (argv[i], option, strlen (option)))
		break;
	    }

	  if (j != ARRAY_SIZE (documented_lang_options))
	    {
	      if (extra_warnings)
		{
		  warning ("ignoring command line option '%s'", argv[i]);
		  if (lang)
		    warning
		      ("(it is valid for %s but not the selected language)",
		       lang);
		}
	    }
	  else if (argv[i][0] == '-' && argv[i][1] == 'g')
	    warning ("`%s': unknown or unsupported -g option", &argv[i][2]);
	  else
	    error ("unrecognized option `%s'", argv[i]);

	  i++;
	}
    }

  if (!filename)
    {
      printf ("%s: no input files.\n", programname);
      exit_after_options = 1;
    }
}

/* Strip off a legitimate source ending from the input string NAME of
   length LEN.  Rather than having to know the names used by all of
   our front ends, we strip off an ending of a period followed by
   up to five characters.  (Java uses ".class".)  */

void
strip_off_ending (name, len)
     char *name;
     int len;
{
  int i;
  for (i = 2;  i < 6 && len > i;  i++)
    {
      if (name[len - i] == '.')
	{
	  name[len - i] = '\0';
	  break;
	}
    }
}

/*  code output  ϴ. 
   Do this even if -fsyntax-only is
   on, because then the driver will have provided the name of a
   temporary file or bit bucket for us.  NAME is the file specified on
   the command line, possibly NULL.  */
static void
init_env_output (name)
     const char *name;
{
  if (dump_base_name == 0)
    dump_base_name = name ? name : "esdump";
  
  if (env_file_name == 0)
    {
      int len = strlen (dump_base_name);
      char *dumpname = (char *) xmalloc (len + 6);
      memcpy (dumpname, dump_base_name, len + 1);
      strip_off_ending (dumpname, len);
      strcat (dumpname, ".e");
      env_file_name = dumpname;
    }
  if (!strcmp (env_file_name, "-"))
    env_out_file = stdout;
  else
    env_out_file = fopen (env_file_name, "w");
  if (env_out_file == 0)
    error ("can't open %s for writing", env_file_name);
}

/* ؼ option  óմϴ. */
static void
process_options ()
{
  /*  츮  (verbose mode) ,  
     ϴ  ɼ flag  մϴ.  */
  if (version_flag)
    print_version (stderr, "");
}


int
toplev_main (argc, argv)
     int argc;
     char *argv[];
{
  int a, i, len;
  char *buf;
  FILE *fd;
  const char *p;

  //GC_find_leak = 1;
  progname = argv[0];

  p = argv[0] + strlen (argv[0]);
  while (p != argv[0] && !IS_DIR_SEPARATOR (p[-1]))
    --p;
  programname = p;

  parse_options_and_default_flags (argc, argv);

  /*  command line switch ó .  */
  process_options ();

  if (!exit_after_options)
    {
      init_env_output (filename);

      init_esm ();
      
      len = fileSize (filename);
      buf = (char *) malloc (sizeof (char) * len + 1);
      fd  = fopen (filename, "rb");
      i   = fread (buf, 1, len, fd);
      buf[i] = '\0';
      environment_file_init (&ef_info, buf, i);

      if ((a = yyparse ()))
	warning ("yyparse error.\n");

      free (buf);
    }
  
  CHECK_LEAKS ();

  return 0;
}

int
main (argc, argv)
     int argc;
     char *argv[];
{
  toplev_main (argc, argv);

  return 0;
}
