#include "eas.h"
#include "machmode.h"
#include "real.h"
#include "tree.h"
#include "stringpool.h"

/* ̹  node  hash table  Ѵ.
   ׷ node  ݺ ʴ´; brief cross-reference  ȴ.  */

#define HASH_SIZE		37

struct bucket
{
  tree node;
  struct bucket *next;
};

static struct bucket **table;

/*  · node  ϴµ, code  address, name  ش.  */

void
print_node_brief (file, prefix, node, indent)
     FILE *file;
     const char *prefix;
     tree node;
     int indent;
{
  char class;

  if (node == 0)
    return;

  class = TREE_CODE_CLASS (TREE_CODE (node));

  /* Always print the slot this node is in, and its code, address and
     name if any.  */
  if (indent > 0)
    fprintf (file, " ");
  fprintf (file, "%s <%s ", prefix, tree_code_name[(int) TREE_CODE (node)]);
  fprintf (file, "%p", (char *) node);

  if (TREE_CODE (node) == IDENTIFIER_NODE)
    fprintf (file, " %s", IDENTIFIER_POINTER (node));

  fprintf (file, ">");
}

void
indent_to (file, column)
     FILE *file;
     int column;
{
  int i;

  /* ̰  ̱⿡, 䱸 column ŭ 鿩 Ѵ.  */
  if (column > 0)
    fprintf (file, "\n");
  for (i = 0; i < column; i++)
    fprintf (file, " ");
}

/* Print the node NODE in full on file FILE, preceded by PREFIX,
   starting in column INDENT.  */

void
print_node (file, prefix, node, indent)
     FILE *file;
     const char *prefix;
     tree node;
     int indent;
{
  int i, hash, len;
  char class;
  struct bucket *b;
  enum machine_mode mode;

  if (node == 0)
    return;

  class = TREE_CODE_CLASS (TREE_CODE (node));

  /* ʹ   ʽϴ.   ڰ   ⸦ Ѵٸ,
     ٸ ȣ⿡ debug_tree  ڷν -- node  ּҸ 
     ϸ մϴ.  */

#if 0
  if (indent > 24)
    {
      print_node_brief (file, prefix, node, indent);
      return;
    }
#endif

#if 0
  if (indent > 8 && (class == 't' || class == 'e'))
    {
      print_node_brief (file, prefix, node, indent);
      return;
    }
#endif

  /* It is unsafe to look at any other filds of an ERROR_MARK node.  */
  if (TREE_CODE (node) == ERROR_MARK)
    {
      print_node_brief (file, prefix, node, indent);
      return;
    }

  hash = ((unsigned long) node) % HASH_SIZE;

  /*  node  table  ִٸ,  װ ּҸ Ѵ.  */
  for (b = table[hash]; b; b = b->next)
    if (b->node == node)
      {
	print_node_brief (file, prefix, node, indent);
	return;
      }

  /*  node  table  ߰Ѵ.  */
  b = (struct bucket *) malloc (sizeof (struct bucket));
  b->node = node;
  b->next = table[hash];
  table[hash] = b;

  /* Indent to the specified column, since this is the long form.  */
  indent_to (file, indent);

  /* Print the slot this node is in, and its code, and address.  */
  fprintf (file, "%s <%s ", prefix, tree_code_name[(int) TREE_CODE (node)]);
  fprintf (file, "%p", (char *) node);

  /* , ̸ Ѵ.  */
  if (TREE_CODE (node) == IDENTIFIER_NODE)
    fprintf (file, " %s", IDENTIFIER_POINTER (node));

  /* DECL_ node  ߰ attribute   ִ.  */

  switch (TREE_CODE_CLASS (TREE_CODE (node)))
    {
    case 'a':
      print_node (file, "hostname", TREE_EADDRESS_HOSTNAME (node), indent + 4);
      print_node (file, "pathandsha1", TREE_EADDRESS_PATHANDSHA1 (node), indent + 4);
      break;
    case 'e':
      fprintf (stderr, " access 0x%x", TREE_ENVIRONMENT_ACCESS (node));
      print_node (file, "name", TREE_ENVIRONMENT_NAME (node), indent + 4);
      print_node (file, "access host", TREE_ENVIRONMENT_ACCESS_HOST (node), indent + 4);
      break;
    case 'f':
      fprintf (stderr, " access 0x%x", TREE_FIELD_ACCESS (node));
      print_node (file, "name", TREE_FIELD_NAME (node), indent + 4);
      print_node (file, "type", TREE_FIELD_TYPE (node), indent + 4);
      print_node (file, "access host", TREE_FIELD_ACCESS_HOST (node), indent + 4);
      print_node (file, "value", TREE_FIELD_VALUE (node), indent + 4);
      break;
    case 'm':
      fprintf (stderr, " access 0x%x", TREE_METHOD_ACCESS (node));
      print_node (file, "name", TREE_METHOD_NAME (node), indent + 4);
      print_node (file, "type", TREE_METHOD_TYPE (node), indent + 4);
      print_node (file, "access host", TREE_METHOD_ACCESS_HOST (node), indent + 4);
      indent_to (file, indent + 3);
      fprintf (stderr, " max stack %d", TREE_METHOD_MAX_STACK (node));
      fprintf (stderr, " max locals %d", TREE_METHOD_MAX_LOCALS (node));
      fprintf (stderr, " insn length %d", TREE_METHOD_INSN_LENGTH (node));
      fprintf (stderr, " attribute count %d", TREE_METHOD_ATTRIBUTE_COUNT (node));
      print_node (file, "stmt", TREE_METHOD_STATEMENTS (node), indent + 4);
      break;

    case '1':
      len = TREE_CODE_LENGTH (TREE_CODE (node));

      for (i = 0; i < len; i++)
	{
	  char temp[10];
	  
	  sprintf (temp, "arg %d", i);
	  print_node (file, temp, TREE_OPERAND (node, i), indent + 4);
	}
      break;

    case 't':
      /* The transparent-union flag is used for different things in
	 different nodes.  */
      mode = TYPE_MODE (node);
      fprintf (file, " %s", GET_MODE_NAME (mode));

      print_node (file, "size", TYPE_SIZE (node), indent + 4);
      print_node (file, "unit size", TYPE_SIZE_UNIT (node), indent + 4);
      indent_to (file, indent + 3);

      if (TYPE_USER_ALIGN (node))
	fprintf (file, " user");

      fprintf (file, " align %d", TYPE_ALIGN (node));
      fprintf (file, " alias set ");
      fprintf (file, HOST_WIDE_INT_PRINT_DEC, TYPE_ALIAS_SET (node));

      if (INTEGRAL_TYPE_P (node) || TREE_CODE (node) == REAL_TYPE)
	{
	  fprintf (file, " precision %d", TYPE_PRECISION (node));
	  print_node (file, "min", TYPE_MIN_VALUE (node), indent + 4);
	  print_node (file, "max", TYPE_MAX_VALUE (node), indent + 4);
	}

      print_node (file, "chain", TREE_CHAIN (node), indent + 4);
      break;

    case 'c':
    case 'x':
      switch (TREE_CODE (node))
	{
	case INTEGER_CST:
	  fprintf (file, " ");
	  if (TREE_INT_CST_HIGH (node) == 0)
	    fprintf (file, HOST_WIDE_INT_PRINT_UNSIGNED,
		     TREE_INT_CST_LOW (node));
	  else if (TREE_INT_CST_HIGH (node) == -1
		   && TREE_INT_CST_LOW (node) != 0)
	    {
	      fprintf (file, "-");
	      fprintf (file, HOST_WIDE_INT_PRINT_UNSIGNED,
		       -TREE_INT_CST_LOW (node));
	    }
	  else
	    fprintf (file, HOST_WIDE_INT_PRINT_DOUBLE_HEX,
		     TREE_INT_CST_HIGH (node), TREE_INT_CST_LOW (node));
	  break;

	case REAL_CST:
	  {
	    REAL_VALUE_TYPE d;

	    if (TREE_OVERFLOW (node))
	      fprintf (file, " overflow");

#if !defined(REAL_IS_NOT_DOUBLE) || defined(REAL_ARITHMETIC)
	    d = TREE_REAL_CST (node);
	    if (REAL_VALUE_ISINF (d))
	      fprintf (file, " Inf");
	    else if (REAL_VALUE_ISNAN (d))
	      fprintf (file, " Nan");
	    else
	      {
		char string[100];

		REAL_VALUE_TO_DECIMAL (d, "%e", string);
		fprintf (file, " %s", string);
	      }
#else
	    {
	      int i;
	      unsigned char *p = (unsigned char *) &TREE_REAL_CST (node);
	      fprintf (file, " 0x");
	      for (i = 0; i < sizeof TREE_REAL_CST (node); i++)
		fprintf (file, "%02x", *p++);
	      fprintf (file, "");
	    }
#endif
	  }
	  break;

	case STRING_CST:
	  fprintf (file, " \"%s\"", TREE_STRING_POINTER (node));
	  /* Print the chain at second level.  */
	  if (indent == 4)
	    print_node (file, "chain", TREE_CHAIN (node), indent + 4);
	  else
	    print_node_brief (file, "chain", TREE_CHAIN (node), indent + 4);
	  break;
	case IDENTIFIER_NODE:
	  break;
	case TREE_LIST:
	  print_node (file, "purpose", TREE_PURPOSE (node), indent + 4);
	  print_node (file, "value", TREE_VALUE (node), indent + 4);
	  break;
	case PROGRAM:
	  print_node (file, "info", TREE_PROGRAM_INFO (node), indent + 4);
	  print_node (file, "match", TREE_PROGRAM_MATCH (node), indent + 4);
	  print_node (file, "header", TREE_PROGRAM_HEADER (node), indent + 4);
	  print_node (file, "body", TREE_PROGRAM_BODY (node), indent + 4);
	  break;	  
	case INFO:
	  print_node (file, "writer", TREE_INFO_WRITER (node), indent + 4);
	  print_node (file, "license", TREE_INFO_LICENSE (node), indent + 4);
	  print_node (file, "filename", TREE_INFO_FILENAME (node), indent + 4);
	  break;
	case HEADER:
	  print_node (file, "environement", TREE_HEADER_ENVIRONMENT (node), indent + 4);
	  print_node (file, "super", TREE_HEADER_SUPER (node), indent + 4);
	  print_node (file, "package", TREE_HEADER_PACKAGE (node), indent + 4);
	  break;
	case BODY:
	  print_node (file, "implements", TREE_BODY_IMPLEMENTS (node), indent + 4);
	  indent_to (file, indent + 3);
	  fprintf (stderr, " field count %d", TREE_BODY_FIELDS_COUNT (node));
	  print_node (file, "fields", TREE_BODY_FIELDS (node), indent + 4);
	  indent_to (file, indent + 3);
	  fprintf (stderr, " method count %d", TREE_BODY_METHODS_COUNT (node));
	  print_node (file, "methods", TREE_BODY_METHODS (node), indent + 4);
	  break;
	case LABEL:
	  print_node (file, "name", TREE_LABEL_NAME (node), indent + 4);	  
	  fprintf (stderr, " code offset %d", TREE_LABEL_CODE_OFFSET (node));
	  break;
	case INSN:
	  print_node (file, "code", TREE_INSN_CODE (node), indent + 4);
	  print_node (file, "op1", TREE_INSN_OP1 (node), indent + 4);
	  print_node (file, "op2", TREE_INSN_OP2 (node), indent + 4);
	  break;
	case LIMIT:
	  print_node (file, "name", TREE_LIMIT_NAME (node), indent + 4);
	  print_node (file, "value", TREE_LIMIT_VALUE (node), indent + 4);
	  break;
	case MATCH:
	  print_node (file, "string", TREE_MATCH_STRING (node), indent + 4);
	  print_node (file, "identifier", TREE_MATCH_IDENTIFIER (node), indent + 4);
	  break;
	case IPPORTADDR:
	  print_node (file, "ip", TREE_ADDRESS_IP (node), indent + 4);
	  print_node (file, "port", TREE_ADDRESS_PORT (node), indent + 4);
	  break;
	case ENVIRONMENT_SEGMENT:
	  print_node (file, "absolute_name", TREE_SEGMENT_ABSOLUTE_NAME (node), indent + 4);
	  print_node (file, "name", TREE_SEGMENT_NAME (node), indent + 4);
	  print_node (file, "sha1", TREE_SEGMENT_SHA1 (node), indent + 4);
	  break;
	case FIELDREF:
	  print_node (file, "url", TREE_FIELDREF_URL (node), indent + 4);
	  print_node (file, "type", TREE_FIELDREF_TYPE (node), indent + 4);
	  break;
	case METHODREF:
	  print_node (file, "url", TREE_METHODREF_URL (node), indent + 4);
	  print_node (file, "type", TREE_METHODREF_TYPE (node), indent + 4);
	  break;
	case ENVIRONMENT_URL:
	  print_node (file, "ename", TREE_ENVIRONMENT_URL_NAME (node), indent + 4);
	  print_node (file, "mfname", TREE_ENVIRONMENT_URL_FIELDORMETHOD_NAME (node), indent + 4);
	  break;
	case CASE:
	  print_node (file, "value", TREE_CASE_VALUE (node), indent + 4);
	  print_node (file, "name", TREE_CASE_NAME (node), indent + 4);
	  break;
	case SWITCH_ARGS:
	  print_node (file, "arg1", TREE_SWITCH_ARG1 (node), indent + 4);
	  print_node (file, "arg2", TREE_SWITCH_ARG2 (node), indent + 4);	  
	  break;
	case LOCALS:
	  print_node (file, "locals", TREE_LOCAL_CONSTANT (node), indent + 4);
	  break;
	case SWITCH_TABLE:
	case SWITCH_LOOKUP:
	  print_node (file, "args", TREE_SWITCH_LOOKUP_ARGS (node), indent + 4);
	  print_node (file, "list", TREE_SWITCH_LOOKUP_LIST (node), indent + 4);
	  print_node (file, "default", TREE_SWITCH_LOOKUP_DEFAULT (node), indent + 4);
	  break;
	case ACCESS_EXPR:
	  print_node (file, "", TREE_ACCESS_EXPR_FLAG (node), indent + 4);
	  print_node (file, "ּ", TREE_ACCESS_EXPR_ADDRESS (node), indent + 4);
	  break;
	default:
	  fprintf (stderr, " óϴ  = %s.\n", tree_code_name[TREE_CODE (node)]);
	  break;
	}
      break;
    default:
      fprintf (stderr, "˼  Class = '%c'\n", TREE_CODE_CLASS (TREE_CODE (node)));
      break;
    }

  /* CHAIN  Ǿ  ,    ¢, ̰ Ѵ.  */
  print_node (file, "chain", TREE_CHAIN (node), indent + 4);
  fprintf (file, ">");
}

/*   ǥ  node NODE  Ѵ.
   ̰Ϳ  Ǵ κ node  6 ܰ ̱ 
    µȴ.  */

void
debug_tree (node)
     tree node;
{
  table = (struct bucket **) malloc (HASH_SIZE * sizeof (struct bucket *));
  memset ((char *) table, 0, HASH_SIZE * sizeof (struct bucket *));
  print_node (stderr, "", node, 0);
  table = 0;
  fprintf (stderr, "\n");
}
