#include "eas.h"
#include "machmode.h"
#include "tree.h"
#include "stringpool.h"
#include "errors.h"
#include "sha1.h"
#include "stor-layout.h"
#include "fold-const.h"

tree integer_type_node;
tree unsigned_type_node;
tree long_integer_type_node;
tree long_unsigned_type_node;
tree float_type_node;
tree double_type_node;
tree long_long_unsigned_type_node;
tree long_long_integer_type_node;
tree long_double_type_node;

#define DEFTREECODE(SYM, NAME, TYPE, LENGTH)	TYPE,

const char tree_code_type[MAX_TREE_CODES] = {
#include "tree.def"
};
#undef DEFTREECODE

/* tree code   expression operand    ̺.
    operand  node structure   κԴϴ.
    type  decl    ʽϴ. */

#define DEFTREECODE(SYM, NAME, TYPE, LENGTH)	LENGTH,

int tree_code_length[MAX_TREE_CODES] = {
#include "tree.def"
};
#undef DEFTREECODE

/* tree component  ̸.
   tree  ϰų  ޼   ˴ϴ.  */
#define DEFTREECODE(SYM, NAME, TYPE, LEN)	NAME,

const char *tree_code_name[MAX_TREE_CODES] = {
#include "tree.def"
};
#undef DEFTREECODE

size_t
tree_size (node)
     tree node;
{
  enum tree_code code = TREE_CODE (node);
  
  switch (TREE_CODE_CLASS (code))
    {
    case 'f':
      return (sizeof (struct tree_field));
    case 'h':
      return (sizeof (struct tree_header));
    case 'e':
      return (sizeof (struct tree_env));
    case 'a':
      return (sizeof (struct tree_address));
    case 'i':
      return (sizeof (struct tree_info));
    case 'm':
      return (sizeof (struct tree_method));

    case 't':  /* type node  */
      return sizeof (struct tree_type);

    case '1':	/* a unary arithmetic expression */
      return (sizeof (struct tree_exp)
	      + (TREE_CODE_LENGTH (code) - 1) * sizeof (char *));

    case 'c':  /*  */
      /* 츮 INTEGER_CST  ؼ TREE_CODE_LENGTH    µ,
	 ̰ word   machine-dependent ϱ ε, ͺ
	    ִ ( , long long) HOST_WIDE_INT  
	 ȭ ؼ.  ϰ REAL_CST  ؼ, `double'  ũ
	 alignment  ϱ  word    machine-dependent
	 ϴ.  */
      if (code == INTEGER_CST)
	return sizeof (struct tree_int_cst);
      else if (code == REAL_CST)
	return sizeof (struct tree_real_cst);
      else
	return (sizeof (struct tree_common)
		+ TREE_CODE_LENGTH (code) * sizeof (char *));

    case 'x':
      {
	size_t length;
	length = (sizeof (struct tree_common)
		  + TREE_CODE_LENGTH (code) * sizeof (char *));
	return length;
      }
    default:
      fprintf (stderr, "tree_size: óؾ ϴ . '%c'\n", TREE_CODE_CLASS (code));
      abort ();
    }
}

/* Chain T  ҵ  ݴ մϴ. ׸
   ο chain  ó  (  )  ȯմϴ.  */

tree
nreverse (t)
     tree t;
{
  tree prev = 0, decl, next;
  for (decl = t; decl; decl = next)
    {
      next = TREE_CHAIN (decl);
      TREE_CHAIN (decl) = prev;
      prev = decl;
    }
  return prev;
}

/* CODE  TREE node  .  */
tree
make_node (code)
     enum tree_code code;
{
  tree t;
  int type = TREE_CODE_CLASS (code);
  size_t length;
  struct tree_common ttmp;
  
  TREE_SET_CODE ((tree) &ttmp, code);
  length = tree_size ((tree) &ttmp);

  t = (tree) malloc (length);
  memset ((PTR) t, 0, length);
  TREE_SET_CODE (t, code);
  
  switch (type)
    {
    default:
      break;
    }

  switch (code)
    {
    default:
      break;
    }

  return t;
}

tree
build_info (writer, license, source)
     tree writer, license, source;
{
  tree t = make_node (INFO);

  TREE_INFO_WRITER (t) = writer;
  TREE_INFO_LICENSE (t) = license;
  TREE_INFO_FILENAME (t) = source;

  return t;
}

void
process_package_directive (package)
     tree package;
{
  int i;
  const char *path = TREE_STRING_POINTER (package);
  
  package_path = malloc (TREE_STRING_LENGTH (package) + 3);
  for (i=0; i<TREE_STRING_LENGTH (package); i++)
    {
      if (path[i] == '.')
	package_path[i] = '/';
      else
	package_path[i] = path[i];
    }
  package_path[i++] = '/';
  package_path[i++] = '\0';
}

tree
build_header (package, env, super)
     tree package, env, super;
{
  tree t = make_node (HEADER);

  TREE_HEADER_PACKAGE (t) = package;
  TREE_HEADER_ENVIRONMENT (t) = env;
  TREE_HEADER_SUPER (t) = super;

  return t;
}

tree
build_segment (name, sha1)
     tree name, sha1;
{
  tree t = make_node (ENVIRONMENT_SEGMENT);

  TREE_SEGMENT_NAME (t) = name;
  TREE_SEGMENT_SHA1 (t) = sha1;

  return t;
}

tree
build_ipport (ip, port)
     tree ip, port;
{
  tree t = make_node (IPPORTADDR);

  TREE_ADDRESS_IP (t) = ip;
  TREE_ADDRESS_PORT (t) = port;

  return t;
}

tree
build_address (hostname, pathandsha1)
     tree hostname, pathandsha1;
{
  tree t = make_node (ADDR);

  TREE_EADDRESS_HOSTNAME (t) = hostname;
  TREE_EADDRESS_PATHANDSHA1 (t) = pathandsha1;

  return t;
}

tree
build_env (access, ename, acchost)
     enum esm_access access;
     tree ename, acchost;
{
  tree t= make_node (ENVIRONMENT);

  TREE_ENVIRONMENT_ACCESS (t) = access;
  TREE_ENVIRONMENT_NAME (t) = ename;
  TREE_ENVIRONMENT_ACCESS_HOST (t) = acchost;

  return t;
}

tree
build_interface (access, ename, acchost)
     enum esm_access access;
     tree ename, acchost;
{
  tree t= make_node (INTERFACE);

  TREE_INTERFACE_ACCESS (t) = access;
  TREE_INTERFACE_NAME (t) = ename;
  TREE_INTERFACE_ACCESS_HOST (t) = acchost;

  return t;
}

/* chain 1   node  chain 2  Ű ν  chain
   node  մϴ. (TREE_CHAIN   chain )
   ̰ Lisp ô `nconc' ̴.  */

tree
chainon (op1, op2)
     tree op1, op2;
{
  if (op1)
    {
      tree t1;
      tree t2;
      
      for (t1 = op1; TREE_CHAIN (t1); t1 = TREE_CHAIN (t1))
	;
      TREE_CHAIN (t1) = op2;
      for (t2 = op2; t2; t2 = TREE_CHAIN (t2))
        if (t2 == t1)
          abort ();  /* Circularity created.  */
      return op1;
    }
  else
    return op2;
}

tree
build_field (access, name, type, access_host, val)
          enum esm_access access;
	  tree name, type, access_host, val;
{
  tree t = make_node (FIELD);

  TREE_FIELD_ACCESS (t) = access;
  TREE_FIELD_NAME (t) = name;
  TREE_FIELD_TYPE (t) = type;
  TREE_FIELD_ACCESS_HOST (t) = access_host;

  if (val)
    {
      cp_add (CONSTANT_Utf8, strdup ("ConstantValue"));
      TREE_FIELD_VALUE (t) = val;

      if (!strcmp (IDENTIFIER_POINTER (type), "I"))
	{
	  switch (TREE_CODE (val))
	    {
	    case INTEGER_CST:
	      TREE_CONSTANT_POOL_NUMBER (val) = 
		cp_add (CONSTANT_Integer, (((TREE_INT_CST_HIGH (val) << 16) | TREE_INT_CST_LOW (val))));
	      break;
	    default:
	     warning ("build_field: óؾ ϴ ConstantValue Switch\n"); 
	    }
	}
      else
	warning ("build_field: óؾ ϴ ConstantValue\n");
    }

  return t;
}

tree
build_method_header (access, name, type, access_host)
          enum esm_access access;
	  tree name, type, access_host;
{
  tree t = make_node (METHODS);

  TREE_METHOD_ACCESS (t) = access;
  TREE_METHOD_NAME (t) = name;
  TREE_METHOD_TYPE (t) = type;
  TREE_METHOD_ACCESS_HOST (t) = access_host;

  return t;
}

tree
build_method (method, stmt)
     tree method, stmt;
{
  TREE_METHOD_STATEMENTS (method) = stmt;

  return method;
}

/* Purpose  value ʵ   PARM  VALUE   
   TREE_LIST node  ȯմϴ.  */

tree
build_tree_list (parm, value)
     tree parm, value;
{
  tree t = make_node (TREE_LIST);

  TREE_PURPOSE (t) = parm;
  TREE_VALUE (t) = value;

  return t;
}

tree
build_insn (insn, op1, op2)
     tree insn, op1, op2;
{
  tree t = make_node (INSN);

  TREE_INSN_CODE (t) = insn;
  TREE_INSN_OP1 (t) = op1;
  TREE_INSN_OP2 (t) = op2;

  return t;
}

tree
build_minus (op)
     tree op;
{
  tree t = make_node (NEGATE_EXPR);

  if (TREE_CODE (op) != INTEGER_CST)
    error ("build_minus: Node Code  INTEGER_CST  ƴմϴ. (Ʈ ʿ   )\n");

  TREE_OPERAND (t, 0) = op;

  return t;
}

tree
build_case (value, name)
     tree value, name;
{
  tree t = make_node (CASE);

  TREE_CASE_VALUE (t) = value;
  TREE_CASE_NAME (t) = name;

  return t;
}

tree
build_switch_args (arg1, arg2)
     tree arg1, arg2;
{
  tree t = make_node (SWITCH_ARGS);

  TREE_SWITCH_ARG1 (t) = arg1;
  TREE_SWITCH_ARG2 (t) = arg2;

  return t;
}

tree 
build_switch_lookup (args, list, ddefault)
     tree args, list, ddefault;
{
  tree t = make_node (SWITCH_LOOKUP);

  TREE_SWITCH_LOOKUP_ARGS (t) = args;
  TREE_SWITCH_LOOKUP_LIST (t) = list;
  TREE_SWITCH_LOOKUP_DEFAULT (t) = ddefault;

  return t;
}

tree 
build_switch_table (args, list, ddefault)
     tree args, list, ddefault;
{
  tree t = make_node (SWITCH_TABLE);

  TREE_SWITCH_TABLE_ARGS (t) = args;
  TREE_SWITCH_TABLE_LIST (t) = list;
  TREE_SWITCH_TABLE_DEFAULT (t) = ddefault;

  return t;
}

tree
build_var (op1, name1, name2, from, to)
     tree op1, name1, name2, from, to;
{
  tree t = make_node (VAR);

  return t;
}

tree
build_limit (var, constant)
     tree var, constant;
{
  tree t = make_node (LIMIT);

  TREE_LIMIT_NAME (t) = var;
  TREE_LIMIT_VALUE (t) = constant;

  return t;
}

tree
build_match (str, ident)
     tree str, ident;
{
  tree t = make_node (MATCH);

  TREE_MATCH_STRING (t) = str;
  TREE_MATCH_IDENTIFIER (t) = ident;

  return t;
}

tree
build_line (line)
     tree line;
{
  tree t = make_node (LINE);

  return t;
}

tree
build_throws (env)
     tree env;
{
  tree t = make_node (THROWS);

  return t;
}

tree
build_catch (env, from, to, using)
     tree env, from, to, using;
{
  tree t = make_node (CATCH);

  return t;
}

tree
build_set (op1, op2)
     tree op1, op2;
{
  tree t = make_node (SET);

  return t;
}

tree
build_label (name, code_offset)
     tree name;
     int code_offset;
{
  tree t = make_node (LABEL);

  TREE_LABEL_NAME (t) = name;
  TREE_LABEL_CODE_OFFSET (t) = code_offset;

  return t;
}

tree
build_url (ename, mfname)
     tree ename, mfname;
{
  tree t = make_node (ENVIRONMENT_URL);
  
  TREE_ENVIRONMENT_URL_NAME (t) = ename;
  TREE_ENVIRONMENT_URL_FIELDORMETHOD_NAME (t) = mfname;

  return t;
}

tree
build_fieldref (url, type)
     tree url, type;
{
  tree t = make_node (FIELDREF);
  
  TREE_FIELDREF_URL (t) = url;
  TREE_FIELDREF_TYPE (t) = type;

  return t;
}

tree
build_methodref (url, type)
     tree url, type;
{
  tree t = make_node (METHODREF);
  
  TREE_METHODREF_URL (t) = url;
  TREE_METHODREF_TYPE (t) = type;

  return t;
}

tree
build_body (implements, fields, methods)
     tree implements, fields, methods;
{
  tree t = make_node (BODY);
  
  TREE_BODY_IMPLEMENTS (t) = implements;
  TREE_BODY_FIELDS (t) = fields;
  TREE_BODY_METHODS (t) = methods;

  return t;
}

tree
build_program (info, match, header, body)
     tree info, match, header, body;
{
  tree t = make_node (PROGRAM);

  TREE_PROGRAM_INFO (t) = info;
  TREE_PROGRAM_HEADER (t) = header;
  TREE_PROGRAM_MATCH (t) = match;
  TREE_PROGRAM_BODY (t) = body;

  return t;
}

/* Ŭ Ȥ field, method  ACL  ǥϱ  node  Ѵ.  ACCESS 
    Ÿ, ADDRESS  ش class  ̴ּ.  */
tree
build_access_expr (access, address)
     tree access, address;
{
  tree t = make_node (ACCESS_EXPR);

  TREE_ACCESS_EXPR_FLAG (t) = access;
  TREE_ACCESS_EXPR_ADDRESS (t) = address;

  return t;
}

/*  LEN  ڿ STR     STRING_CST node 
   ȯѴ.
   TREE_TYPE  ʱȭ ʴ´.  */

tree
build_string (len, str)
     int len;
     const char *str;
{
  tree s = make_node (STRING_CST);

  TREE_STRING_LENGTH (s) = len;
  TREE_STRING_POINTER (s) = ggc_alloc_string (str, len);

  return s;
}

/*   ΰ  LOW  HI    
   INTERGET_CST node  ȯѴ.
   TREE_TYPE  `int'  Ǿ.

    Լ `build_int_2' ũθ ؼ ݵ ȣǾ Ѵ.  */

tree
build_int_2_wide (low, hi)
     unsigned int low;
     int hi;
{
  tree t = make_node (INTEGER_CST);

  TREE_INT_CST_LOW (t) = low;
  TREE_INT_CST_HIGH (t) = hi;
  TREE_TYPE (t) = integer_type_node;

  return t;
}

void
process_esegment (node, path, buf, sha1, name, len)
     tree node;
     char *path, *buf;
     tree *sha1, *name;
     int *len;
{
  if (!node)
    return;

  switch (TREE_CODE (node))
    {
    case TREE_LIST:
      process_esegment (TREE_VALUE (node), path, buf, sha1, name, len);
      break;
    case ENVIRONMENT_SEGMENT:
      {
	process_esegment (TREE_SEGMENT_NAME (node), path, buf, sha1, name, len);
	if (!TREE_CHAIN (node))
	  {
	    int size;
	    if (TREE_SEGMENT_SHA1 (node))
	      {
		size = ap_base64decode_binary (buf, TREE_STRING_POINTER (TREE_SEGMENT_SHA1 (node)));
		if (size != SHA_DIGESTSIZE)
		  {
		    error ("process_esegment: BASE64  ؼ Ű ̰ SHA_DIGESTSIZE   ʽϴ.\n");
		    return;
		  }
	      }
	    *name = TREE_SEGMENT_NAME (node);
	    *sha1 = TREE_SEGMENT_SHA1 (node);
	  }
	break;
      }
    case IDENTIFIER_NODE:
      {
	memcpy (path+*len, IDENTIFIER_POINTER (node), IDENTIFIER_LENGTH (node));
	*len += IDENTIFIER_LENGTH (node);
	path[(*len)] = '/';
	*len += 1;
	break;
      }
    default:
      error ("process_esegment: error.\n");
      break;
    }

  process_esegment (TREE_CHAIN (node), path, buf, sha1, name, len);
}

tree
simplify_environment_name (t)
     tree t;
{
  char path[512], *p = path;
  char *s;
  int len = 0;
  tree sha1 = NULL_TREE, name = NULL_TREE;
  tree result = NULL_TREE, ps;

  ps = TREE_EADDRESS_PATHANDSHA1 (t);
  s = TREE_SEGMENT_SHA1_BINARY (ps);
  process_esegment (ps, p, s, &sha1, &name, &len);
  p[len-1] = '\0';

  TREE_SEGMENT_ABSOLUTE_NAME (ps) = get_identifier (p);
  TREE_SEGMENT_NAME (ps) = TREE_VALUE (name);
  TREE_SEGMENT_SHA1 (ps) = sha1;
  TREE_CHAIN (ps) = NULL_TREE;

  /*  SHA1  ٸ,  environment  ɼ ,  
     environment  ƴ ,  ߻Ų.  */
  if (!TREE_SEGMENT_SHA1 (ps) 
      && strcmp (IDENTIFIER_POINTER (current_envname), IDENTIFIER_POINTER (TREE_SEGMENT_NAME (ps))))
    {
      warning ("%d: Environment  SHA1  Ǿ.\n", line);
    }

#if 0
  fprintf (stderr, "simplify_environment_name: %s %p\n", p, sha1);
  debug_tree (t);
#endif

  return result == NULL_TREE ? t : result;
};

/* src/esm.y  ؼ method ̸ , '<' identifier '>'  ,
   method ̸ "<....>"   ־ Ѵ.  */
tree
process_method_name (t)
     tree t;
{
  char buf[256];
  const char *tmp;
  int len;

  tmp = IDENTIFIER_POINTER (TREE_VALUE (t));
  len = IDENTIFIER_LENGTH (TREE_VALUE (t));

  len = sprintf (buf, "<%.*s>", len, tmp);
  buf[len] = '\0';

  TREE_VALUE (t) = get_identifier (buf);

  return t;
}

/* CODE   ù° non-tree operand  index  ȯϰų,  
     tree   operand   ȯѴ.  */

int
first_rtl_op (code)
     enum tree_code code;
{
  switch (code)
    {
    default:
      return TREE_CODE_LENGTH (code);
    }
}

/* Type  TYPE ̰  D  ο REAL_CST node  ȯѴ.  */

tree
build_real (type, d)
     tree type;
     REAL_VALUE_TYPE d;
{
  tree v;
  int overflow = 0;

  v = make_node (REAL_CST);
  TREE_TYPE (v) = type;
  TREE_REAL_CST (v) = d;
  TREE_OVERFLOW (v) = TREE_CONSTANT_OVERFLOW (v) = overflow;

  return v;
}

tree
build_local (t)
     tree t;
{
  tree v = make_node (LOCALS);

  TREE_LOCAL_CONSTANT (v) = t;

  return v;
}

/*   T  ȣ  ȯѴ.
   ȯ T < 0 ̸ -1, T == 0 ̸ 0, T > 0 ̸ 1 ̴.
   T  type  unsign    -1  ȯ  ϶.  */

int
tree_int_cst_sgn (t)
     tree t;
{
  if (TREE_INT_CST_LOW (t) == 0 && TREE_INT_CST_HIGH (t) == 0)
    return 0;
  else if (TREE_UNSIGNED (TREE_TYPE (t)))
    return 1;
  else if (TREE_INT_CST_HIGH (t) < 0)
    return -1;
  else
    return 1;
}

/* NODE  TREE_CHAIN  0 ̰ ̰ ο uid  ٴ 
   ϰ    ο node  ȯѴ.  */

tree
copy_node (node)
     tree node;
{
  tree t;
  size_t length;

  length = tree_size (node);
  t = (tree) malloc (length);
  memcpy (t, node, length);

  TREE_CHAIN (t) = 0;

  return t;
}

/*    C  type TYPE (INTEGER_TYPE )     ִ
     ִٸ 0  ƴ  ȯѴ.  */

int
int_fits_type_p (c, type)
     tree c, type;
{
  /*  type  bound   , 츮 츮 ڽ ˻  ִ.
      ƴ ,  type  subtype ̸ ش ׿  ˻縦 õѴ.
     ׷   precision   ˻縦 , force\_fit\_type  Ѵ.  */
  if (TYPE_MAX_VALUE (type) != NULL_TREE
      && TYPE_MIN_VALUE (type) != NULL_TREE
      && TREE_CODE (TYPE_MAX_VALUE (type)) == INTEGER_CST
      && TREE_CODE (TYPE_MIN_VALUE (type)) == INTEGER_CST)
    {
      if (TREE_UNSIGNED (type))
	return (! INT_CST_LT_UNSIGNED (TYPE_MAX_VALUE (type), c)
		&& ! INT_CST_LT_UNSIGNED (c, TYPE_MIN_VALUE (type))
		/*  int   unsigned type   ʴ´.  */
		&& ! (TREE_INT_CST_HIGH (c) < 0
		      && ! TREE_UNSIGNED (TREE_TYPE (c))));
      else
	return (! INT_CST_LT (TYPE_MAX_VALUE (type), c)
		&& ! INT_CST_LT (c, TYPE_MIN_VALUE (type))
		/* Unsigned ints with top bit set never fit signed types.  */
		&& ! (TREE_INT_CST_HIGH (c) < 0
		      && TREE_UNSIGNED (TREE_TYPE (c))));
    }
  else if (TREE_CODE (type) == INTEGER_TYPE && TREE_TYPE (type) != 0)
    return int_fits_type_p (c, TREE_TYPE (type));
  else
    {
      c = copy_node (c);
      TREE_TYPE (c) = type;
      return !force_fit_type (c, 0);
    }
}

void
build_common_tree_nodes (void)
{
  integer_type_node = make_signed_type (INT_TYPE_SIZE);
  unsigned_type_node = make_unsigned_type (INT_TYPE_SIZE);
  long_integer_type_node = make_signed_type (LONG_TYPE_SIZE);
  long_unsigned_type_node = make_unsigned_type (LONG_TYPE_SIZE);
  long_long_integer_type_node = make_signed_type (LONG_LONG_TYPE_SIZE);
  long_long_unsigned_type_node = make_unsigned_type (LONG_LONG_TYPE_SIZE);

  float_type_node = make_node (REAL_TYPE);
  TYPE_PRECISION (float_type_node) = FLOAT_TYPE_SIZE;
  layout_type (float_type_node);

  double_type_node = make_node (REAL_TYPE);
  TYPE_PRECISION (double_type_node) = DOUBLE_TYPE_SIZE;
  layout_type (double_type_node);

  long_double_type_node = make_node (REAL_TYPE);
  TYPE_PRECISION (long_double_type_node) = LONG_DOUBLE_TYPE_SIZE;
  layout_type (long_double_type_node);
}
