#ifndef ESERVER_TREE_H
#define ESERVER_TREE_H

typedef union tree_node *tree;

#include "sha1.h"
#include "esm.h"

#define MAX_TREE_CODES				256

extern const char tree_code_type[MAX_TREE_CODES];
#define TREE_CODE_CLASS(CODE)			tree_code_type[(int) (CODE)]

/* CLASS  expression   tree-code class   0  ƴ 
   ǵݴϴ.  */

#define IS_EXPR_CODE_CLASS(CLASS) \
  ((CLASS) == '<' || (CLASS) == '1' || (CLASS) == '2' || (CLASS) == 'e')

/*  tree-node  argument-word  (number).  */

extern int tree_code_length[MAX_TREE_CODES];
#define TREE_CODE_LENGTH(CODE)			tree_code_length[(int) (CODE)]

/* tree node   code  */

#define DEFTREECODE(SYM, STRING, TYPE, NARGS)   SYM,

extern const char *tree_code_name[MAX_TREE_CODES];

enum tree_code 
  {
#include "tree.def"
    
    LAST_AND_UNUSED_TREE_CODE	/* NUM_TREE_CODE    
				     */
  };

#undef DEFTREECODE

/* Language-independent tree code  .  */
#define NUM_TREE_CODES				((int) LAST_AND_UNUSED_TREE_CODE)

#define TREE_CHECK(t, code)		(t)
#define TREE_CLASS_CHECK(t, code)	(t)
#define CST_OR_CONSTRUCTOR_CHECK(t)	(t)
#define EXPR_CHECK(t)			(t)

#define TYPE_CHECK(tree)	TREE_CLASS_CHECK  (tree, 't')
#define DECL_CHECK(tree)	TREE_CLASS_CHECK  (tree, 'd')
#define CST_CHECK(tree)		TREE_CLASS_CHECK  (tree, 'c')

struct tree_common
{
  tree chain;
  tree type;

  enum tree_code code : 8;

  int constant_pool_num : 16;

  int is_reserved_word : 1;
  int is_reserved_insn : 1;
  /* TODO:  ʴ 24 bit  ӻ ؾ Ѵ.  */
  int is_labelname : 1;
  int public_flag : 1;
  int static_flag : 1;
  int unsigned_flag : 1;
  int unused : 2;

  int oplen;
  enum constant constant_pool_type;

  int bytecode;
  int insncode;
};

/* expression   node  ̰ expression  date type 
   Ÿϴ.
   POINTER_TYPE node  ̰ ͸ Ű type Դϴ.
   ARRAY_TYPE node  ̰ element  type Դϴ.
   VECTOR_TYPE node  ̰ element  type Դϴ.  */
#define TREE_TYPE(NODE) ((NODE)->common.type)

#define TREE_CODE(NODE)				((enum tree_code) (NODE)->common.code)
#define TREE_SET_CODE(NODE, VALUE)		((NODE)->common.code = (enum tree_code) (VALUE))

/* INTEGER_CST Ȥ REAL_CST, COMPLEX_CST, VECTOR_CST  ̰ folding 
    overflow  ־ ǹѴ.  ̰ TREE_OVERFLOW  ѷϰ
    Ǵµ, ANSI C   ǥĿ overflow  ߻Ͽ , 
    (diagnostic)  䱸ϱ ̴.  */
#define TREE_CONSTANT_OVERFLOW(NODE) ((NODE)->common.static_flag)

/* INTEGER_CST Ȥ REAL_CST, COMPLEX_CST, VECTOR_CST  ̰
   folding  overflow  ־  ǥĿ   
   ߻ ʾ ǹѴ.  TREE_OVERFLOW  TREE_CONSTANT_OVERFLOW 
   , ݴδ ƴϴ.  */
#define TREE_OVERFLOW(NODE) ((NODE)->common.public_flag)

/* INTEGER_TYPE Ȥ ENUMERAL_TYPE node 鿡 unsigned type  ǹѴ.
   FIELD_DECL node 鿡 unsigned bit field  ǹѴ.
    bit  DECL_BUILT_IN_NONANSI ó Լ鳻 ȴ.  */
#define TREE_UNSIGNED(NODE) ((NODE)->common.unsigned_flag)

#define TREE_CHAIN(NODE)			((NODE)->common.chain)

#define IDENTIFIER_LENGTH(NODE)			((NODE)->identifier.id.len)
#define IDENTIFIER_POINTER(NODE)		((const char *) (NODE)->identifier.id.str)

/* Hash table ĺ ͸ tree_identifier ͷ մϴ.
   ݴε մϴ.  */

#define HT_IDENT_TO_GCC_IDENT(NODE) \
  ((tree) ((char *) (NODE) - sizeof (struct tree_common)))
#define GCC_IDENT_TO_HT_IDENT(NODE)		(&((struct tree_identifier *) (NODE))->id)

#define TREE_IS_LABEL_NAME(NODE)		((NODE)->common.is_labelname)
#define TREE_IS_RESERVED_WORD(NODE)		((NODE)->common.is_reserved_word)
#define TREE_IS_RESERVED_INSN(NODE)		((NODE)->common.is_reserved_insn)
#define TREE_RESERVED_WORD_CODE(NODE)		((NODE)->common.bytecode)
#define TREE_BYTE_CODE(NODE)			((NODE)->common.insncode)
#define TREE_INSN_LENGTH(NODE)			((NODE)->common.oplen)
#define TREE_CONSTANT_POOL_NUMBER(NODE)		((NODE)->common.constant_pool_num)
#define TREE_CONSTANT_POOL_TYPE(NODE)		((NODE)->common.constant_pool_type)

struct tree_identifier
{
  struct tree_common common;
  struct ht_identifier id;
};

#define TREE_INFO_WRITER(NODE)			((NODE)->info.writer)
#define TREE_INFO_LICENSE(NODE)			((NODE)->info.license)
#define TREE_INFO_FILENAME(NODE)		((NODE)->info.filename)

struct tree_info
{
  struct tree_common common;

  tree writer;
  tree license;
  tree filename;
};

#define TREE_HEADER_PACKAGE(NODE)		((NODE)->header.package)
#define TREE_HEADER_ENVIRONMENT(NODE)		((NODE)->header.env)
#define TREE_HEADER_SUPER(NODE)			((NODE)->header.super)

struct tree_header
{
  struct tree_common common;

  tree package;
  tree env;
  tree super;
};

#define TREE_SEGMENT_ABSOLUTE_NAME(NODE)	((NODE)->tes.absolute_name)
#define TREE_SEGMENT_NAME(NODE)			((NODE)->tes.name)
#define TREE_SEGMENT_SHA1(NODE)			((NODE)->tes.sha1)
#define TREE_SEGMENT_SHA1_BINARY(NODE)		((NODE)->tes.sha1_binary)

struct tree_environment_segment
{
  struct tree_common common;

  tree absolute_name;
  tree name;
  tree sha1;
  char sha1_binary[SHA_DIGESTSIZE];
};

#define TREE_ADDRESS_IP(NODE)			((NODE)->ipport.ip)
#define TREE_ADDRESS_PORT(NODE)			((NODE)->ipport.port)

struct tree_ipport_address
{
  struct tree_common common;

  tree ip;
  tree port;
};

#define TREE_EADDRESS_HOSTNAME(NODE)		((NODE)->address.hostname)
#define TREE_EADDRESS_PATHANDSHA1(NODE)		((NODE)->address.pathandsha1)

#define TREE_EADDRESS_HOSTNAME_IP(NODE)		(TREE_ADDRESS_IP (TREE_EADDRESS_HOSTNAME ((NODE))))
#define TREE_EADDRESS_PATHANDSHA1_ABSOLUTE_NAME(NODE) \
						(TREE_SEGMENT_ABSOLUTE_NAME (TREE_EADDRESS_PATHANDSHA1 ((NODE))))
#define TREE_EADDRESS_PATHANDSHA1_SHA1_BINARY(NODE) \
						(TREE_SEGMENT_SHA1_BINARY (TREE_EADDRESS_PATHANDSHA1 ((NODE))))

struct tree_address
{
  struct tree_common common;

  tree hostname;
  tree pathandsha1;
};

#define TREE_ENVIRONMENT_ACCESS(NODE)		((NODE)->env.access)
#define TREE_ENVIRONMENT_NAME(NODE)		((NODE)->env.name)
#define TREE_ENVIRONMENT_ACCESS_HOST(NODE)	((NODE)->env.acchost)
#define TREE_ENVIRONMENT_ACCESS_HOST_COUNT(NODE)	((NODE)->env.acchost_count)
#define TREE_INTERFACE_ACCESS(NODE)		((NODE)->env.access)
#define TREE_INTERFACE_NAME(NODE)		((NODE)->env.name)
#define TREE_INTERFACE_ACCESS_HOST(NODE)	((NODE)->env.acchost)
#define TREE_INTERFACE_ACCESS_HOST_COUNT(NODE)	((NODE)->env.acchost_count)

struct tree_env
{
  struct tree_common common;

  enum esm_access access;
  tree name;
  int acchost_count;
  tree acchost;
};

#define TREE_FIELD_ACCESS(NODE)			((NODE)->field.access)
#define TREE_FIELD_NAME(NODE)			((NODE)->field.name)
#define TREE_FIELD_TYPE(NODE)			((NODE)->field.type)
#define TREE_FIELD_ACCESS_HOST(NODE)		((NODE)->field.acchost)
#define TREE_FIELD_ACCESS_HOST_COUNT(NODE)	((NODE)->field.acchost_count)
#define TREE_FIELD_VALUE(NODE)			((NODE)->field.value)

struct tree_field
{
  struct tree_common common;

  enum esm_access access;
  tree name;
  tree type;
  int acchost_count;
  tree acchost;
  tree value;
};

#define TREE_METHOD_ACCESS(NODE)		((NODE)->method.access)
#define TREE_METHOD_NAME(NODE)			((NODE)->method.name)
#define TREE_METHOD_TYPE(NODE)			((NODE)->method.type)
#define TREE_METHOD_ACCESS_HOST(NODE)		((NODE)->method.acchost)
#define TREE_METHOD_ACCESS_HOST_COUNT(NODE)	((NODE)->method.acchost_count)
#define TREE_METHOD_STATEMENTS(NODE)		((NODE)->method.statements)
#define TREE_METHOD_MAX_STACK(NODE)		((NODE)->method.max_stack)
#define TREE_METHOD_MAX_LOCALS(NODE)		((NODE)->method.max_locals)
#define TREE_METHOD_INSN_LENGTH(NODE)		((NODE)->method.insn_len)
#define TREE_METHOD_ATTRIBUTE_COUNT(NODE)	((NODE)->method.attr_count)
#define TREE_METHOD_LABEL_LIST(NODE)		((NODE)->method.label)

struct tree_method
{
  struct tree_common common;

  enum esm_access access;
  tree name;
  tree type;
  int acchost_count;
  tree acchost;

  int max_stack;
  int max_locals;
  int insn_len;
  tree statements;
  
  int attr_count;

  /*  Method  ϴ label .  */
  struct labels *label;
};

#define TREE_LIMIT_NAME(NODE)			((NODE)->limit.name)
#define TREE_LIMIT_VALUE(NODE)			((NODE)->limit.value)

struct tree_limit
{
  struct tree_common common;

  tree name;
  tree value;
};

/* TREE_LIST node  Ǵ  ũ.  */
#define TREE_PURPOSE(NODE)			((NODE)->list.purpose)
#define TREE_VALUE(NODE)			((NODE)->list.value)

struct tree_list
{
  struct tree_common common;

  tree purpose;
  tree value;
};

enum tree_insn_arg_type
  {
    INSN_ARGUMENT_TYPE_UNKNOWN,
    INSN_ARGUMENT_TYPE_STRING
  };

/* TREE_LIST node  Ǵ  ũ.  */
#define TREE_INSN_CODE(NODE)			((NODE)->insn.code)
#define TREE_INSN_CODE_OFFSET(NODE)		((NODE)->insn.offset)
#define TREE_INSN_OP1(NODE)			((NODE)->insn.op1)
#define TREE_INSN_OP2(NODE)			((NODE)->insn.op2)
#define TREE_INSN_TYPE(NODE)			((NODE)->insn.type)

struct tree_insn
{
  struct tree_common common;

  int offset;
  tree code;
  tree op1;
  tree op2;
  enum tree_insn_arg_type type;
};

#define TREE_CASE_VALUE(NODE)			((NODE)->ccase.value)
#define TREE_CASE_NAME(NODE)			((NODE)->ccase.name)

struct tree_case
{
  struct tree_common common;

  tree value;
  tree name;
};

#define TREE_SWITCH_ARG1(NODE)			((NODE)->switch_args.arg1)
#define TREE_SWITCH_ARG2(NODE)			((NODE)->switch_args.arg2)

struct tree_switch_args
{
  struct tree_common common;

  tree arg1;
  tree arg2;
};

#define TREE_SWITCH_LOOKUP_CODE_OFFSET(NODE)		((NODE)->switch_lookup.offset)
#define TREE_SWITCH_LOOKUP_CODE_LENGTH(NODE)		((NODE)->switch_lookup.oplen)
#define TREE_SWITCH_LOOKUP_CASE_NUMBER(NODE)		((NODE)->switch_lookup.num_case)
#define TREE_SWITCH_LOOKUP_ARGS(NODE)			((NODE)->switch_lookup.args)
#define TREE_SWITCH_LOOKUP_LIST(NODE)			((NODE)->switch_lookup.list)
#define TREE_SWITCH_LOOKUP_DEFAULT(NODE)		((NODE)->switch_lookup.ddefault)

struct tree_switch_lookup
{
  struct tree_common common;

  int offset;
  int oplen;
  int num_case;
  tree args;
  tree list;
  tree ddefault;
};

#define TREE_SWITCH_TABLE_CODE_OFFSET(NODE)		((NODE)->switch_table.offset)
#define TREE_SWITCH_TABLE_ARGS(NODE)			((NODE)->switch_table.args)
#define TREE_SWITCH_TABLE_LIST(NODE)			((NODE)->switch_table.list)
#define TREE_SWITCH_TABLE_DEFAULT(NODE)			((NODE)->switch_table.ddefault)

struct tree_switch_table
{
  struct tree_common common;

  int offset;
  tree args;
  tree list;
  tree ddefault;
};

#define TREE_MATCH_STRING(NODE)				((NODE)->match.string)
#define TREE_MATCH_IDENTIFIER(NODE)			((NODE)->match.identifier)

struct tree_match
{
  struct tree_common common;

  tree string;
  tree identifier;
};

#define TREE_BODY_IMPLEMENTS(NODE)			((NODE)->body.implements)
#define TREE_BODY_FIELDS(NODE)				((NODE)->body.fields)
#define TREE_BODY_FIELDS_COUNT(NODE)			((NODE)->body.fields_count)
#define TREE_BODY_METHODS(NODE)				((NODE)->body.methods)
#define TREE_BODY_METHODS_COUNT(NODE)			((NODE)->body.methods_count)

struct tree_body
{
  struct tree_common common;

  tree implements;
  tree fields;
  int fields_count;
  tree methods;
  int methods_count;
};

#define TREE_PROGRAM_INFO(NODE)				((NODE)->program.info)
#define TREE_PROGRAM_HEADER(NODE)			((NODE)->program.header)
#define TREE_PROGRAM_BODY(NODE)				((NODE)->program.body)
#define TREE_PROGRAM_MATCH(NODE)			((NODE)->program.match)

struct tree_program
{
  struct tree_common common;

  tree info;
  tree match;
  tree header;
  tree body;
};

/* STRING_CST   */
#define TREE_STRING_LENGTH(NODE)			((NODE)->string.length)
#define TREE_STRING_POINTER(NODE)			((NODE)->string.pointer)

struct tree_string
{
  struct tree_common common;

  int length;
  const char *pointer;
};

/*  ǥϴ node   ߰ field  accessor  Ѵ.  */

#define build_int_2(LO, HI)  \
  build_int_2_wide ((unsigned int) (LO), (int) (HI))

/* INTEGER_CST node .   ΰ ϳ 2-word  .
    data type  sign  Ǿ ,  װ͵  
   ʰ ִ ϴ sign-extended  2 word ̴.
   2 word   unsign , extra bit  0 ̴.  */
#define TREE_INT_CST(NODE) ((NODE)->int_cst.int_cst)
#define TREE_INT_CST_LOW(NODE) (TREE_INT_CST (NODE).low)
#define TREE_INT_CST_HIGH(NODE) (TREE_INT_CST (NODE).high)

#define INT_CST_LT(A, B)				\
  (TREE_INT_CST_HIGH (A) < TREE_INT_CST_HIGH (B)	\
   || (TREE_INT_CST_HIGH (A) == TREE_INT_CST_HIGH (B)	\
       && TREE_INT_CST_LOW (A) < TREE_INT_CST_LOW (B)))

#define INT_CST_LT_UNSIGNED(A, B)				\
  (((unsigned int) TREE_INT_CST_HIGH (A)		\
    < (unsigned int) TREE_INT_CST_HIGH (B))		\
   || (((unsigned int) TREE_INT_CST_HIGH (A)		\
        == (unsigned int) TREE_INT_CST_HIGH (B))	\
       && TREE_INT_CST_LOW (A) < TREE_INT_CST_LOW (B)))
 
struct tree_int_cst
{
  struct tree_common common;

  /* ⿡ -ü ʿ  Լ `const_hash'  ϳ unit
       word  ȱ⸦ ϰ -ü ּҸ   ν
       bounded pointer    ִ.  */
  struct {
    unsigned HOST_WIDE_INT low;
    HOST_WIDE_INT high;
  } int_cst;
};

#define TREE_ENVIRONMENT_URL_NAME(NODE)				((NODE)->teu.ename)
#define TREE_ENVIRONMENT_URL_FIELDORMETHOD_NAME(NODE)		((NODE)->teu.mfname)

struct tree_environment_url
{
  struct tree_common common;

  tree ename;
  tree mfname;
};

#define TREE_METHODREF_URL(NODE)		((NODE)->tmr.url)
#define TREE_METHODREF_TYPE(NODE)		((NODE)->tmr.type)

struct tree_method_ref
{
  struct tree_common common;

  tree url;
  tree type;
};

#define TREE_FIELDREF_URL(NODE)			((NODE)->tfr.url)
#define TREE_FIELDREF_TYPE(NODE)		((NODE)->tfr.type)

struct tree_field_ref
{
  struct tree_common common;

  tree url;
  tree type;
};

#define TREE_LABEL_NAME(NODE)			((NODE)->label.name)
#define TREE_LABEL_CODE_OFFSET(NODE)		((NODE)->label.code_offset)

struct tree_label
{
  struct tree_common common;

  tree name;
  int code_offset;
};

#define TREE_LOCAL_CONSTANT(NODE)		((NODE)->locals.con)

struct tree_locals
{
  struct tree_common common;

  tree con;
};

#define TREE_ACCESS_EXPR_FLAG(NODE)		((NODE)->access_expr.flag)
#define TREE_ACCESS_EXPR_ADDRESS(NODE)		((NODE)->access_expr.address)

struct tree_access_expr
{
  struct tree_common common;

  tree flag;
  tree address;
};

#define TREE_OPERAND(NODE, I)			((NODE)->exp.operands[I])

struct tree_exp
{
  struct tree_common common;
  int complexity;
  tree operands[1];
};

/* 0  ƴ  TYPE  integral type  Ÿϴ.  츮 ⼭
   COMPLEX type   ʴ´ٴ  ָϽʽÿ.  */

#define INTEGRAL_TYPE_P(TYPE)  \
  (TREE_CODE (TYPE) == INTEGER_TYPE || TREE_CODE (TYPE) == ENUMERAL_TYPE  \
   || TREE_CODE (TYPE) == BOOLEAN_TYPE || TREE_CODE (TYPE) == CHAR_TYPE)

/* 0  ƴ  TYPE  complex floating-point type  ϴ 
   floating-point type  Ÿϴ.  */

#define FLOAT_TYPE_P(TYPE)		\
  (TREE_CODE (TYPE) == REAL_TYPE	\
   || (TREE_CODE (TYPE) == COMPLEX_TYPE \
       && TREE_CODE (TREE_TYPE (TYPE)) == REAL_TYPE))

#define TREE_REAL_CST(NODE) ((NODE)->real_cst.real_cst)

#include "real.h"

struct tree_real_cst
{
  struct tree_common common;
  REAL_VALUE_TYPE real_cst;
};

struct tree_type
{
  struct tree_common common;
  tree values;
  tree size;
  tree size_unit;
  tree attributes;
  unsigned int uid;

  unsigned int precision : 9;
  enum machine_mode mode : 7;

  unsigned string_flag : 1;
  unsigned no_force_blk_flag : 1;
  unsigned needs_constructing_flag : 1;
  unsigned transparent_union_flag : 1;
  unsigned packed_flag : 1;
  unsigned restrict_flag : 1;
  unsigned pointer_depth : 2;

  unsigned lang_flag_0 : 1;
  unsigned lang_flag_1 : 1;
  unsigned lang_flag_2 : 1;
  unsigned lang_flag_3 : 1;
  unsigned lang_flag_4 : 1;
  unsigned lang_flag_5 : 1;
  unsigned lang_flag_6 : 1;
  unsigned user_align : 1;

  unsigned int align;
  tree pointer_to;
  tree reference_to;
  union {int address; char *pointer; } symtab;
  tree name;
  tree minval;
  tree maxval;
  tree next_variant;
  tree main_variant;
  tree binfo;
  tree context;
  HOST_WIDE_INT alias_set;
};

/*  TYPE  aggregate (-component) type  Ÿٸ 0  ƴ .  */

#define AGGREGATE_TYPE_P(TYPE) \
  (TREE_CODE (TYPE) == ARRAY_TYPE || TREE_CODE (TYPE) == RECORD_TYPE \
   || TREE_CODE (TYPE) == UNION_TYPE || TREE_CODE (TYPE) == QUAL_UNION_TYPE \
   || TREE_CODE (TYPE) == SET_TYPE)

/*  type   (-) typed-based alias .
   TYPE_ALIAS_SET  ٸ object   alias   .
    TYPE_ALIAS_SET  -1 ̸,  ƹ alias set   type
    Ҵ  ̴.   TYPE_ALIAS_SET  0 ̸,  type
    object  ٸ type  object 鿡 alias   ִ.  */
#define TYPE_ALIAS_SET(NODE) (TYPE_CHECK (NODE)->type.alias_set)

/*   type   typed-based alias set  Ǿٸ 0  ƴ
    .  */
#define TYPE_ALIAS_SET_KNOWN_P(NODE) (TYPE_CHECK (NODE)->type.alias_set != -1)

/* data type  Ÿ node   field   ũε
   մϴ.  */

/* ̷ field  뿡   tree.def Ͽ  ֽϴ.
    ..._TYPE tree code   ʽÿ.  */

#define TYPE_MODE(NODE) (TYPE_CHECK (NODE)->type.mode)
#define TYPE_SIZE(NODE) (TYPE_CHECK (NODE)->type.size)
#define TYPE_SIZE_UNIT(NODE) (TYPE_CHECK (NODE)->type.size_unit)
#define TYPE_NEXT_VARIANT(NODE) (TYPE_CHECK (NODE)->type.next_variant)
#define TYPE_MAIN_VARIANT(NODE) (TYPE_CHECK (NODE)->type.main_variant)
#define TYPE_MIN_VALUE(NODE) (TYPE_CHECK (NODE)->type.minval)
#define TYPE_MAX_VALUE(NODE) (TYPE_CHECK (NODE)->type.maxval)
#define TYPE_PRECISION(NODE) (TYPE_CHECK (NODE)->type.precision)

/*  type  object   alignment ʿ伺.
    bit  , int ̴.  */
#define TYPE_ALIGN(NODE) (TYPE_CHECK (NODE)->type.align)

/*   type   alignment  "aligned" attribute   ö  1,
    װ  type  ⺻  0.  */
#define TYPE_USER_ALIGN(NODE) (TYPE_CHECK (NODE)->type.user_align)

union tree_node
{
  struct tree_common common;
  struct tree_info info;
  struct tree_address address;
  struct tree_identifier identifier;
  struct tree_env env;
  struct tree_header header;
  struct tree_field field;
  struct tree_method method;
  struct tree_list list;
  struct tree_insn insn;
  struct tree_case ccase;
  struct tree_switch_args switch_args;
  struct tree_switch_lookup switch_lookup;
  struct tree_switch_table switch_table;
  struct tree_body body;
  struct tree_match match;
  struct tree_limit limit;
  struct tree_string string;
  struct tree_int_cst int_cst;
  struct tree_ipport_address ipport;
  struct tree_program program;
  struct tree_environment_segment tes;
  struct tree_method_ref tmr;
  struct tree_field_ref tfr;
  struct tree_label label;
  struct tree_environment_url teu;
  struct tree_exp exp;
  struct tree_real_cst real_cst;
  struct tree_type type;
  struct tree_locals locals;
  struct tree_access_expr access_expr;
};

#define NULL_TREE				((tree) NULL)

/* Data ü ũε, size  ٷ Լ, size  Ÿµ
   Ǵ  type  Ѵ.  */

enum size_type_kind
{
  SIZETYPE,		/* Ʈ size  Ϲ ǥ.  */
  SSIZETYPE,		/* Ʈ size  Sign  ǥ.  */
  USIZETYPE,		/* Ʈ size  Unsign  ǥ.  */
  BITSIZETYPE,		/* Ʈ size  Ϲ ǥ.  */
  SBITSIZETYPE,		/* Ʈ size  Sign  ǥ.  */
  UBITSIZETYPE,	        /* Ʈ size  Unsign  ǥ.  */
  TYPE_KIND_LAST};

extern tree sizetype_tab[(int) TYPE_KIND_LAST];

#define sizetype sizetype_tab[(int) SIZETYPE]
#define bitsizetype sizetype_tab[(int) BITSIZETYPE]
#define ssizetype sizetype_tab[(int) SSIZETYPE]
#define usizetype sizetype_tab[(int) USIZETYPE]
#define sbitsizetype sizetype_tab[(int) SBITSIZETYPE]
#define ubitsizetype sizetype_tab[(int) UBITSIZETYPE]

#define size_int(L) size_int_wide ((HOST_WIDE_INT) (L), SIZETYPE)
#define bitsize_int(L) size_int_wide ((HOST_WIDE_INT) (L), BITSIZETYPE)

extern tree integer_type_node;
extern tree unsigned_type_node;
extern tree long_integer_type_node;
extern tree long_unsigned_type_node;
extern tree float_type_node;
extern tree double_type_node;
extern tree long_long_unsigned_type_node;
extern tree long_long_integer_type_node;
extern tree long_double_type_node;

/* Compute the number of bytes occupied by 'node'.  This routine only
   looks at TREE_CODE and, if the code is TREE_VEC, TREE_VEC_LENGTH.  */

extern size_t tree_size			PARAMS ((tree));

/* Lowest level primitive for allocating a node.
   The TREE_CODE is the only argument.  Contents are initialized
   to zero except for a few of the common fields.  */

extern tree make_node			PARAMS ((enum tree_code));

extern void debug_tree			PARAMS ((tree));

extern tree build_source (tree);
extern tree build_info (tree, tree, tree);
extern tree build_header (tree, tree, tree);
extern tree build_address (tree, tree);
extern tree build_env (enum esm_access, tree, tree);
extern tree build_interface (enum esm_access, tree, tree);
extern tree chainon (tree, tree);
extern tree build_field (enum esm_access, tree, tree, tree, tree);
extern tree build_method_header (enum esm_access, tree, tree, tree);
extern tree build_tree_list (tree, tree);
extern tree build_insn (tree, tree, tree);
extern tree build_minus (tree);
extern tree build_case (tree, tree);
extern tree build_switch_args (tree, tree);
extern tree build_switch_table (tree, tree, tree);
extern tree build_switch_lookup (tree, tree, tree);
extern tree build_var (tree, tree, tree, tree, tree);
extern tree build_limit (tree, tree);
extern tree build_line (tree);
extern tree build_throws (tree);
extern tree build_catch (tree, tree, tree, tree);
extern tree build_set (tree, tree);
extern tree build_label (tree, int);
extern tree build_match (tree, tree);
extern tree build_program (tree, tree, tree, tree);
extern tree build_body (tree, tree, tree);
extern tree build_method (tree, tree);
extern tree build_string (int, const char *);
extern tree build_int_2_wide (unsigned int, int);
extern tree build_ipport (tree, tree);
extern tree build_segment (tree, tree);
extern tree build_fieldref (tree, tree);
extern tree build_methodref (tree, tree);
extern tree build_url (tree, tree);
extern tree simplify_environment_name (tree);
extern tree process_method_name (tree);
extern void process_package_directive (tree);
extern int first_rtl_op (enum tree_code);
extern void build_common_tree_nodes (void);
extern tree build_real (tree, REAL_VALUE_TYPE);
extern int int_fits_type_p (tree, tree);
extern int tree_int_cst_sgn (tree);
extern tree build_local (tree);
extern tree build_access_expr (tree, tree);

#endif
