#ifndef ECONNECT_H
#define ECONNECT_H

/*
   HTTP Status Codes
 */

#define HTTP_CONTINUE                      100
#define HTTP_SWITCHING_PROTOCOLS           101
#define HTTP_PROCESSING                    102
#define HTTP_OK                            200
#define HTTP_CREATED                       201
#define HTTP_ACCEPTED                      202
#define HTTP_NON_AUTHORITATIVE             203
#define HTTP_NO_CONTENT                    204
#define HTTP_RESET_CONTENT                 205
#define HTTP_PARTIAL_CONTENT               206
#define HTTP_MULTI_STATUS                  207
#define HTTP_MULTIPLE_CHOICES              300
#define HTTP_MOVED_PERMANENTLY             301
#define HTTP_MOVED_TEMPORARILY             302
#define HTTP_SEE_OTHER                     303
#define HTTP_NOT_MODIFIED                  304
#define HTTP_USE_PROXY                     305
#define HTTP_TEMPORARY_REDIRECT            307
#define HTTP_BAD_REQUEST                   400
#define HTTP_UNAUTHORIZED                  401
#define HTTP_PAYMENT_REQUIRED              402
#define HTTP_FORBIDDEN                     403
#define HTTP_NOT_FOUND                     404
#define HTTP_METHOD_NOT_ALLOWED            405
#define HTTP_NOT_ACCEPTABLE                406
#define HTTP_PROXY_AUTHENTICATION_REQUIRED 407
#define HTTP_REQUEST_TIME_OUT              408
#define HTTP_CONFLICT                      409
#define HTTP_GONE                          410
#define HTTP_LENGTH_REQUIRED               411
#define HTTP_PRECONDITION_FAILED           412
#define HTTP_REQUEST_ENTITY_TOO_LARGE      413
#define HTTP_REQUEST_URI_TOO_LARGE         414
#define HTTP_UNSUPPORTED_MEDIA_TYPE        415
#define HTTP_RANGE_NOT_SATISFIABLE         416
#define HTTP_EXPECTATION_FAILED            417
#define HTTP_UNPROCESSABLE_ENTITY          422
#define HTTP_LOCKED                        423
#define HTTP_FAILED_DEPENDENCY             424
#define HTTP_UPGRADE_REQUIRED              426
#define HTTP_INTERNAL_SERVER_ERROR         500
#define HTTP_NOT_IMPLEMENTED               501
#define HTTP_BAD_GATEWAY                   502
#define HTTP_SERVICE_UNAVAILABLE           503
#define HTTP_GATEWAY_TIME_OUT              504
#define HTTP_VERSION_NOT_SUPPORTED         505
#define HTTP_VARIANT_ALSO_VARIES           506
#define HTTP_INSUFFICIENT_STORAGE          507
#define HTTP_NOT_EXTENDED                  510

typedef struct ec_request ec_request_t;
typedef struct ec_response ec_response_t;
typedef struct ec_client ec_client_t;
typedef struct ec_file ec_file_t;

#define EC_REQUEST_POOL(NODE)		((NODE)->pool)
#define EC_REQUEST_CLIENT(NODE)		((NODE)->conn)
#define EC_REQUEST_METHOD(NODE)		((NODE)->method)
#define EC_REQUEST_URI(NODE)		((NODE)->uri)
#define EC_REQUEST_HEADERS_OUT(NODE)	((NODE)->headers_out)
#define EC_REQUEST_ARGUMENT(NODE)	((NODE)->args)

struct ec_request
{
  ec_client_t *conn;

  apr_pool_t *pool;

  const char *method;
  const char *uri;
  apr_table_t *headers_out;
  apr_table_t *args;

  apr_bucket_alloc_t *bb_alloc;
  apr_bucket_brigade *bb;

  /* Request  ó , Ǵ response   header  body
      ٷ handler  մϴ.  ̰ ec_request_send () Լ
     ȣ   ˴ϴ.  */
  apr_status_t (*handler) (ec_response_t *, void *);
  void *handler_arg;
};

struct ec_response
{
  ec_client_t *conn;
  apr_pool_t *pool;

  int remaining;
  int read_chunked;

  char *the_request;

  /** Time when the response started */
  apr_time_t response_time;

  char *protocol;
  /* Protocol version number of protocol; 1.1 = 1001 */
  int proto_num;
  /* Response Status Number.  */
  int status;
  char *status_msg;
  
  /* Response Body.  */
  char *body;
  /* Response Body Length.  */
  apr_ssize_t bodylen;

  /* Response Headers table structure.  */
  apr_table_t *headers_in;
  apr_table_t *notes;

  int read_length;

  apr_bucket_alloc_t *bb_alloc;
  apr_bucket_brigade *bb;
};

#define EC_CLIENT_POOL(NODE)		((NODE)->pool)
#define EC_CLIENT_PROXY_HOST(NODE)	((NODE)->proxy_host)
#define EC_CLIENT_PROXY_PORT(NODE)	((NODE)->proxy_port)
#define EC_CLIENT_HOST(NODE)		((NODE)->host)
#define EC_CLIENT_PORT(NODE)		((NODE)->port)
#define EC_CLIENT_USE_INTERFACE(NODE)	((NODE)->use_interface)
#define EC_CLIENT_SOCKET(NODE)		((NODE)->socket)
#define EC_CLIENT_SOCKADDR(NODE)	((NODE)->sockaddr)
#define EC_CLIENT_INTFADDR(NODE)	((NODE)->intfaddr)
#define EC_CLIENT_CONNECTED(NODE)	((NODE)->connected)
#define EC_CLIENT_KEEP_ALIVE(NODE)	((NODE)->keep_alive)
#define EC_CLIENT_TIMEOUT(NODE)		((NODE)->timeout)
					 
struct ec_client
{
  ec_request_t *req;
  ec_response_t *res;
  apr_pool_t *pool;

  const char *proxy_host;
  apr_port_t proxy_port;
  const char *host;
  apr_port_t port;
  const char *use_interface;
  
  apr_socket_t *socket;
  apr_sockaddr_t *sockaddr;
  apr_sockaddr_t *intfaddr;
  
  int connected;
  int keep_alive;
  int timeout;
};

struct ec_file
{
  unsigned char *base;
  unsigned char *cur;
  unsigned char *back_to;

  int size;

  char *filename;
};

enum ec_input_mode_t
  {
    EC_MODE_GETLINE,
    EC_MODE_SPECULATIVE,
    EC_MODE_READBYTES
  };

ec_client_t *ec_client_new (const char *, const int);
apr_status_t ec_client_connect (ec_client_t *);
ec_request_t *ec_request_new (ec_client_t *, const char *, const char *);
apr_status_t ec_request_set_handler (ec_request_t *, apr_status_t (*handler) (ec_response_t *, void *), void *);
apr_status_t ec_request_add_arg (ec_request_t *, const char *, const char *);		       
ec_response_t * ec_request_send (ec_request_t *);
void ec_client_close (ec_client_t *);

struct etx_classobj;
void ec_file_init (ec_file_t *, unsigned char *, int);
extern int ecd_init_reader_args (struct etx_classobj **, ec_file_t *);
extern etx ecd_read_etx (struct etx_classobj *, int *, int *);
extern void ecd_finish_reader (struct etx_classobj *);

#endif
