/* Register Transfer Language (ETL) definitions for GCC
   Copyright (C) 1987, 1991, 1992, 1993, 1994, 1995, 1996, 1997, 1998,
   1999, 2000, 2001, 2002, 2003, 2004 Free Software Foundation, Inc.

This file is part of GCC.

GCC is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

GCC is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#ifndef ECLIB_ETL_H
#define ECLIB_ETL_H

/* Value used by some passes to "recognize" noop moves as valid
 instructions.  */
#define NOOP_MOVE_INSN_CODE	INT_MAX

/* Register Transfer Language EXPRESSIONS CODES */

#define ETX_CODE	enum etx_code
enum etx_code  {

#define DEF_ETL_EXPR(ENUM, NAME, FORMAT, CLASS)   ENUM ,
#include "etl.def"		/* etl expressions are documented here */
#undef DEF_ETL_EXPR

  LAST_AND_UNUSED_ETX_CODE};	/* A convenient way to get a value for
				   NUM_ETX_CODE.
				   Assumes default enum value assignment.  */

#define NUM_ETX_CODE ((int) LAST_AND_UNUSED_ETX_CODE)
				/* The cast here, saves many elsewhere.  */

extern const unsigned char etx_length[NUM_ETX_CODE];
#define GET_ETX_LENGTH(CODE)		(etx_length[(int) (CODE)])

extern const char * const etx_name[NUM_ETX_CODE];
#define GET_ETX_NAME(CODE)		(etx_name[(int) (CODE)])

extern const char * const etx_format[NUM_ETX_CODE];
#define GET_ETX_FORMAT(CODE)		(etx_format[(int) (CODE)])

extern const char etx_class[NUM_ETX_CODE];
#define GET_ETX_CLASS(CODE)		(etx_class[(int) (CODE)])

extern const unsigned char etx_size[NUM_ETX_CODE];
extern const unsigned char etx_next[NUM_ETX_CODE];

/* Common union for an element of an etx.  */

union etunion_def
{
  int rtint;
  unsigned int rtuint;
  const char *rtstr;
  etx etx;
  etvec etvec;
  etree rttree;
};
typedef union etunion_def etunion;

/* ETL expression ("etx").  */

struct etx_def GTY((chain_next ("ETX_NEXT (&%h)"),
		    chain_prev ("ETX_PREV (&%h)")))
{
  /* The kind of expression this is.  */
  enum etx_code code: 16;

  /* The kind of value the expression has.  */
  /* ENUM_BITFIELD(machine_mode) mode : 8; */

  /* 1 in a MEM if we should keep the alias set for this mem unchanged
     when we access a component.
     1 in a CALL_INSN if it is a sibling call.
     1 in a SET that is for a return.
     In a CODE_LABEL, part of the two-bit alternate entry field.  */
  unsigned int jump : 1;
  /* In a CODE_LABEL, part of the two-bit alternate entry field.
     1 in a MEM if it cannot trap.  */
  unsigned int call : 1;
  /* 1 in a REG, MEM, or CONCAT if the value is set at most once, anywhere.
     1 in a SUBREG if it references an unsigned object whose mode has been
     from a promoted to a wider mode.
     1 in a SYMBOL_REF if it addresses something in the per-function
     constants pool.
     1 in a CALL_INSN, NOTE, or EXPR_LIST for a const or pure call.
     1 in a JUMP_INSN, CALL_INSN, or INSN of an annulling branch.  */
  unsigned int unchanging : 1;
  /* 1 in a MEM or ASM_OPERANDS expression if the memory reference is volatile.
     1 in an INSN, CALL_INSN, JUMP_INSN, CODE_LABEL, BARRIER, or NOTE
     if it has been deleted.
     1 in a REG expression if corresponds to a variable declared by the user,
     0 for an internally generated temporary.
     1 in a SUBREG with a negative value.
     1 in a LABEL_REF or in a REG_LABEL note for a non-local label.
     In a SYMBOL_REF, this flag is used for machine-specific purposes.  */
  unsigned int volatil : 1;
  /* 1 in a MEM referring to a field of an aggregate.
     0 if the MEM was a variable or the result of a * operator in C;
     1 if it was the result of a . or -> operator (on a struct) in C.
     1 in a REG if the register is used only in exit code a loop.
     1 in a SUBREG expression if was generated from a variable with a
     promoted mode.
     1 in a CODE_LABEL if the label is used for nonlocal gotos
     and must not be deleted even if its count is zero.
     1 in a LABEL_REF if this is a reference to a label outside the
     current loop.
     1 in an INSN, JUMP_INSN or CALL_INSN if this insn must be scheduled
     together with the preceding insn.  Valid only within sched.
     1 in an INSN, JUMP_INSN, or CALL_INSN if insn is in a delay slot and
     from the target of a branch.  Valid from reorg until end of compilation;
     cleared before used.
     1 in an INSN, JUMP_INSN or CALL_INSN or related etx if this insn is
     dead code.  Valid only during dead-code elimination phase; cleared
     before use.  */
  unsigned int in_struct : 1;
  /* At the end of ETL generation, 1 if this etx is used.  This is used for
     copying shared structure.  See `unshare_all_etl'.
     In a REG, this is not needed for that purpose, and used instead
     in `leaf_renumber_regs_insn'.
     1 in a SYMBOL_REF, means that emit_library_call
     has used it as the function.  */
  unsigned int used : 1;
  /* Nonzero if this etx came from procedure integration.
     1 in a REG or PARALLEL means this etx refers to the return value
     of the current function.
     1 in a SYMBOL_REF if the symbol is weak.  */
  unsigned integrated : 1;
  /* 1 in an INSN or a SET if this etx is related to the call frame,
     either changing how we compute the frame address or saving and
     restoring registers in the prologue and epilogue.
     1 in a MEM if the MEM refers to a scalar, rather than a member of
     an aggregate.
     1 in a REG if the register is a pointer.
     1 in a SYMBOL_REF if it addresses something in the per-function
     constant string pool.  */
  unsigned frame_related : 1;

  /* The first element of the operands of this etx.
     The number of operands and their types are controlled
     by the `code' field, according to etl.def.  */
  union {
    etunion fld[1];
    HOST_WIDE_INT hwint[1];
  } u;
};

/* The size in bytes of an etx header (code, mode and flags).  */
#define ETX_HDR_SIZE offsetof (struct etx_def, u)

/* The size in bytes of an etx with code CODE.  */
#define ETX_SIZE(CODE) etx_size[CODE]

#define NULL_ETX (etx) 0

/* The "next" and "previous" ETX, relative to this one.  */

#define ETX_NEXT(X) (etx_next[ETX_GET_CODE (X)] == 0 ? NULL			\
		     : *(etx *)(((char *)X) + etx_next[ETX_GET_CODE (X)]))

/* FIXME: the "NEXT_INSN (PREV_INSN (X)) == X" condition shouldn't be needed.
 */
#define ETX_PREV(X) ((ETX_GET_CODE (X) == INSN              \
                      || ETX_GET_CODE (X) == CALL_INSN      \
                      || ETX_GET_CODE (X) == JUMP_INSN      \
                      || ETX_GET_CODE (X) == NOTE           \
                      || ETX_GET_CODE (X) == BARRIER        \
                      || ETX_GET_CODE (X) == CODE_LABEL)    \
                     && PREV_INSN (X) != NULL           \
                     && NEXT_INSN (PREV_INSN (X)) == X  \
                     ? PREV_INSN (X) : NULL)

/* Define macros to access the `code' field of the etx.  */

#define ETX_GET_CODE(ETX)	    ((enum etx_code) (ETX)->code)
#define ETX_PUT_CODE(ETX, CODE) ((ETX)->code = (CODE))

/* ETL vector.  These appear inside ETX's when there is a need
   for a variable number of things.  The principle use is inside
   PARALLEL expressions.  */

struct etvec_def GTY(()) {
  int num_elem;		/* number of elements */
  etx GTY ((length ("%h.num_elem"))) elem[1];
};

#define NULL_ETVEC (etvec) 0

#define ETX_GET_NUM_ELEM(ETVEC)		((ETVEC)->num_elem)
#define ETX_PUT_NUM_ELEM(ETVEC, NUM)	((ETVEC)->num_elem = (NUM))

/* 1 if X is a constant value that is an integer.  */
#if 0
#define CONSTANT_P(X)   \
  (ETX_GET_CODE (X) == LABEL_REF || ETX_GET_CODE (X) == SYMBOL_REF		\
   || ETX_GET_CODE (X) == CONST_INT || ETX_GET_CODE (X) == CONST_DOUBLE		\
   || ETX_GET_CODE (X) == CONST || ETX_GET_CODE (X) == HIGH			\
   || ETX_GET_CODE (X) == CONST_VECTOR	                                \
   || ETX_GET_CODE (X) == CONSTANT_P_ETX)
#endif

/* General accessor macros for accessing the fields of an etx.  */

#if defined ENABLE_ETL_CHECKING && (ECLIB_VERSION >= 2007)
/* The bit with a star outside the statement expr and an & inside is
   so that N can be evaluated only once.  */
#define ETL_CHECK1(ETX, N, C1) __extension__				\
(*({ etx const _etx = (ETX); const int _n = (N);			\
     const enum etx_code _code = ETX_GET_CODE (_etx);			\
     if (_n < 0 || _n >= GET_ETX_LENGTH (_code))			\
       etl_check_failed_bounds (_etx, _n, __FILE__, __LINE__,		\
				__FUNCTION__);				\
     if (GET_ETX_FORMAT(_code)[_n] != C1)				\
       etl_check_failed_type1 (_etx, _n, C1, __FILE__, __LINE__,	\
			       __FUNCTION__);				\
     &_etx->u.fld[_n]; }))

#define ETL_CHECK2(ETX, N, C1, C2) __extension__			\
(*({ etx const _etx = (ETX); const int _n = (N);			\
     const enum etx_code _code = ETX_GET_CODE (_etx);			\
     if (_n < 0 || _n >= GET_ETX_LENGTH (_code))			\
       etl_check_failed_bounds (_etx, _n, __FILE__, __LINE__,		\
				__FUNCTION__);				\
     if (GET_ETX_FORMAT(_code)[_n] != C1				\
	 && GET_ETX_FORMAT(_code)[_n] != C2)				\
       etl_check_failed_type2 (_etx, _n, C1, C2, __FILE__, __LINE__,	\
			       __FUNCTION__);				\
     &_etx->u.fld[_n]; }))

#define ETL_CHECKC1(ETX, N, C) __extension__				\
(*({ etx const _etx = (ETX); const int _n = (N);			\
     if (ETX_GET_CODE (_etx) != (C))					\
       etl_check_failed_code1 (_etx, (C), __FILE__, __LINE__,		\
			       __FUNCTION__);				\
     &_etx->u.fld[_n]; }))

#define ETL_CHECKC2(ETX, N, C1, C2) __extension__			\
(*({ etx const _etx = (ETX); const int _n = (N);			\
     const enum etx_code _code = ETX_GET_CODE (_etx);			\
     if (_code != (C1) && _code != (C2))				\
       etl_check_failed_code2 (_etx, (C1), (C2), __FILE__, __LINE__,	\
			       __FUNCTION__); \
     &_etx->u.fld[_n]; }))

#define ETVEC_ELT(ETVEC, I) __extension__				\
(*({ etvec const _etvec = (ETVEC); const int _i = (I);			\
     if (_i < 0 || _i >= ETX_GET_NUM_ELEM (_etvec))				\
       etvec_check_failed_bounds (_etvec, _i, __FILE__, __LINE__,	\
				  __FUNCTION__);			\
     &_etvec->elem[_i]; }))

#define ETX_XWINT(ETX, N) __extension__					\
(*({ etx const _etx = (ETX); const int _n = (N);			\
     const enum etx_code _code = ETX_GET_CODE (_etx);			\
     if (_n < 0 || _n >= GET_ETX_LENGTH (_code))			\
       etl_check_failed_bounds (_etx, _n, __FILE__, __LINE__,		\
				__FUNCTION__);				\
     if (GET_ETX_FORMAT(_code)[_n] != 'w')				\
       etl_check_failed_type1 (_etx, _n, 'w', __FILE__, __LINE__,	\
			       __FUNCTION__);				\
     &_etx->u.hwint[_n]; }))

#define ETX_XCWINT(ETX, N, C) __extension__					\
(*({ etx const _etx = (ETX);						\
     if (ETX_GET_CODE (_etx) != (C))					\
       etl_check_failed_code1 (_etx, (C), __FILE__, __LINE__,		\
			       __FUNCTION__);				\
     &_etx->u.hwint[N]; }))

extern void etl_check_failed_bounds (etx, int, const char *, int,
				     const char *)
    ATTRIBUTE_NORETURN;
extern void etl_check_failed_type1 (etx, int, int, const char *, int,
				    const char *)
    ATTRIBUTE_NORETURN;
extern void etl_check_failed_type2 (etx, int, int, int, const char *,
				    int, const char *)
    ATTRIBUTE_NORETURN;
extern void etl_check_failed_code1 (etx, enum etx_code, const char *,
				    int, const char *)
    ATTRIBUTE_NORETURN;
extern void etl_check_failed_code2 (etx, enum etx_code, enum etx_code,
				    const char *, int, const char *)
    ATTRIBUTE_NORETURN;
extern void etvec_check_failed_bounds (etvec, int, const char *, int,
				       const char *)
    ATTRIBUTE_NORETURN;

#else   /* not ENABLE_ETL_CHECKING */

#define ETL_CHECK1(ETX, N, C1)      ((ETX)->u.fld[N])
#define ETL_CHECK2(ETX, N, C1, C2)  ((ETX)->u.fld[N])
#define ETL_CHECKC1(ETX, N, C)	    ((ETX)->u.fld[N])
#define ETL_CHECKC2(ETX, N, C1, C2) ((ETX)->u.fld[N])
#define ETVEC_ELT(ETVEC, I)	    ((ETVEC)->elem[I])
#define ETX_XWINT(ETX, N)		    ((ETX)->u.hwint[N])
#define ETX_XCWINT(ETX, N, C)	    ((ETX)->u.hwint[N])

#endif

/* General accessor macros for accessing the flags of an etx.  */

/* Access an individual etx flag, with no checking of any kind.  */
#define ETX_FLAG(ETX, FLAG)	((ETX)->FLAG)

#if defined ENABLE_ETL_FLAG_CHECKING && (ECLIB_VERSION >= 2007)
#define ETL_FLAG_CHECK1(NAME, ETX, C1) __extension__			\
({ etx const _etx = (ETX);						\
   if (ETX_GET_CODE(_etx) != C1)						\
     etl_check_failed_flag  (NAME, _etx, __FILE__, __LINE__,		\
			     __FUNCTION__);				\
   _etx; })

#define ETL_FLAG_CHECK2(NAME, ETX, C1, C2) __extension__		\
({ etx const _etx = (ETX);						\
   if (ETX_GET_CODE(_etx) != C1 && ETX_GET_CODE(_etx) != C2)			\
     etl_check_failed_flag  (NAME,_etx, __FILE__, __LINE__,		\
			      __FUNCTION__);				\
   _etx; })

#define ETL_FLAG_CHECK3(NAME, ETX, C1, C2, C3) __extension__		\
({ etx const _etx = (ETX);						\
   if (ETX_GET_CODE(_etx) != C1 && ETX_GET_CODE(_etx) != C2			\
       && ETX_GET_CODE(_etx) != C3)						\
     etl_check_failed_flag  (NAME, _etx, __FILE__, __LINE__,		\
			     __FUNCTION__);				\
   _etx; })

#define ETL_FLAG_CHECK4(NAME, ETX, C1, C2, C3, C4) __extension__	\
({ etx const _etx = (ETX);						\
   if (ETX_GET_CODE(_etx) != C1 && ETX_GET_CODE(_etx) != C2			\
       && ETX_GET_CODE(_etx) != C3 && ETX_GET_CODE(_etx) != C4)			\
     etl_check_failed_flag  (NAME, _etx, __FILE__, __LINE__,		\
			      __FUNCTION__);				\
   _etx; })

#define ETL_FLAG_CHECK5(NAME, ETX, C1, C2, C3, C4, C5) __extension__	\
({ etx const _etx = (ETX);						\
   if (ETX_GET_CODE(_etx) != C1 && ETX_GET_CODE(_etx) != C2			\
       && ETX_GET_CODE(_etx) != C3 && ETX_GET_CODE(_etx) != C4			\
       && ETX_GET_CODE(_etx) != C5)						\
     etl_check_failed_flag  (NAME, _etx, __FILE__, __LINE__,		\
			     __FUNCTION__);				\
   _etx; })

#define ETL_FLAG_CHECK6(NAME, ETX, C1, C2, C3, C4, C5, C6)		\
  __extension__								\
({ etx const _etx = (ETX);						\
   if (ETX_GET_CODE(_etx) != C1 && ETX_GET_CODE(_etx) != C2			\
       && ETX_GET_CODE(_etx) != C3 && ETX_GET_CODE(_etx) != C4			\
       && ETX_GET_CODE(_etx) != C5 && ETX_GET_CODE(_etx) != C6)			\
     etl_check_failed_flag  (NAME,_etx, __FILE__, __LINE__,		\
			     __FUNCTION__);				\
   _etx; })

#define ETL_FLAG_CHECK7(NAME, ETX, C1, C2, C3, C4, C5, C6, C7)		\
  __extension__								\
({ etx const _etx = (ETX);						\
   if (ETX_GET_CODE(_etx) != C1 && ETX_GET_CODE(_etx) != C2			\
       && ETX_GET_CODE(_etx) != C3 && ETX_GET_CODE(_etx) != C4			\
       && ETX_GET_CODE(_etx) != C5 && ETX_GET_CODE(_etx) != C6			\
       && ETX_GET_CODE(_etx) != C7)						\
     etl_check_failed_flag  (NAME, _etx, __FILE__, __LINE__,		\
			     __FUNCTION__);				\
   _etx; })

#define ETL_FLAG_CHECK8(NAME, ETX, C1, C2, C3, C4, C5, C6, C7, C8)	\
  __extension__								\
({ etx const _etx = (ETX);						\
   if (ETX_GET_CODE(_etx) != C1 && ETX_GET_CODE(_etx) != C2			\
       && ETX_GET_CODE(_etx) != C3 && ETX_GET_CODE(_etx) != C4			\
       && ETX_GET_CODE(_etx) != C5 && ETX_GET_CODE(_etx) != C6			\
       && ETX_GET_CODE(_etx) != C7 && ETX_GET_CODE(_etx) != C8)			\
     etl_check_failed_flag  (NAME, _etx, __FILE__, __LINE__,		\
			     __FUNCTION__);				\
   _etx; })

extern void etl_check_failed_flag (const char *, etx, const char *,
				   int, const char *)
    ATTRIBUTE_NORETURN
    ;

#else	/* not ENABLE_ETL_FLAG_CHECKING */

#define ETL_FLAG_CHECK1(NAME, ETX, C1)					(ETX)
#define ETL_FLAG_CHECK2(NAME, ETX, C1, C2)				(ETX)
#define ETL_FLAG_CHECK3(NAME, ETX, C1, C2, C3)				(ETX)
#define ETL_FLAG_CHECK4(NAME, ETX, C1, C2, C3, C4)			(ETX)
#define ETL_FLAG_CHECK5(NAME, ETX, C1, C2, C3, C4, C5)		(ETX)
#define ETL_FLAG_CHECK6(NAME, ETX, C1, C2, C3, C4, C5, C6)		(ETX)
#define ETL_FLAG_CHECK7(NAME, ETX, C1, C2, C3, C4, C5, C6, C7)		(ETX)
#define ETL_FLAG_CHECK8(NAME, ETX, C1, C2, C3, C4, C5, C6, C7, C8)	(ETX)
#endif

#define CLEAR_ETX_FLAGS(ETX)	\
do {				\
  etx const _etx = (ETX);	\
  _etx->call = 0;		\
  _etx->frame_related = 0;	\
  _etx->in_struct = 0;		\
  _etx->integrated = 0;		\
  _etx->jump = 0;		\
  _etx->unchanging = 0;		\
  _etx->used = 0;		\
  _etx->volatil = 0;		\
} while (0)

#define ETX_XINT(ETX, N)	(ETL_CHECK2 (ETX, N, 'i', 'n').rtint)
#define ETX_XSTR(ETX, N)	(ETL_CHECK2 (ETX, N, 's', 'S').rtstr)
#define ETX_XEXP(ETX, N)	(ETL_CHECK2 (ETX, N, 'e', 'u').etx)
#define ETX_XVEC(ETX, N)	(ETL_CHECK2 (ETX, N, 'E', 'V').etvec)
#define ETX_XMODE(ETX, N)	(ETL_CHECK1 (ETX, N, 'M').rttype)
#define ETX_XBITMAP(ETX, N) (ETL_CHECK1 (ETX, N, 'b').rtbit)
#define ETX_XTREE(ETX, N)   (ETL_CHECK1 (ETX, N, 't').rttree)
#define ETX_XBBDEF(ETX, N)	(ETL_CHECK1 (ETX, N, 'B').bb)
#define ETX_XTMPL(ETX, N)	(ETL_CHECK1 (ETX, N, 'T').rtstr)

#define ETX_XVECEXP(ETX, N, M)	ETVEC_ELT (ETX_XVEC (ETX, N), M)
#define ETX_XVECLEN(ETX, N)		ETX_GET_NUM_ELEM (ETX_XVEC (ETX, N))

/* These are like XINT, etc. except that they expect a '0' field instead
   of the normal type code.  */

#define ETX_X0INT(ETX, N)	   (ETL_CHECK1 (ETX, N, '0').rtint)
#define ETX_X0UINT(ETX, N)	   (ETL_CHECK1 (ETX, N, '0').rtuint)
#define ETX_X0STR(ETX, N)	   (ETL_CHECK1 (ETX, N, '0').rtstr)
#define ETX_X0EXP(ETX, N)	   (ETL_CHECK1 (ETX, N, '0').etx)
#define ETX_X0VEC(ETX, N)	   (ETL_CHECK1 (ETX, N, '0').etvec)
#define ETX_X0MODE(ETX, N)	   (ETL_CHECK1 (ETX, N, '0').rttype)
#define ETX_X0BITMAP(ETX, N)   (ETL_CHECK1 (ETX, N, '0').rtbit)
#define ETX_X0TREE(ETX, N)	   (ETL_CHECK1 (ETX, N, '0').rttree)
#define ETX_X0BBDEF(ETX, N)	   (ETL_CHECK1 (ETX, N, '0').bb)
#define ETX_X0ADVFLAGS(ETX, N) (ETL_CHECK1 (ETX, N, '0').rt_addr_diff_vec_flags)
#define ETX_X0CSELIB(ETX, N)   (ETL_CHECK1 (ETX, N, '0').rt_cselib)
#define ETX_X0MEMATTR(ETX, N)  (ETL_CHECKC1 (ETX, N, MEM).rtmem)
#define ETX_X0REGATTR(ETX, N)  (ETL_CHECKC1 (ETX, N, REG).rtreg)

/* Access a '0' field with any type.  */
#define ETX_X0ANY(ETX, N)	   ETL_CHECK1 (ETX, N, '0')

#define ETX_XCINT(ETX, N, C)      (ETL_CHECKC1 (ETX, N, C).rtint)
#define ETX_XCUINT(ETX, N, C)     (ETL_CHECKC1 (ETX, N, C).rtuint)
#define ETX_XCSTR(ETX, N, C)      (ETL_CHECKC1 (ETX, N, C).rtstr)
#define ETX_XCEXP(ETX, N, C)      (ETL_CHECKC1 (ETX, N, C).etx)
#define ETX_XCVEC(ETX, N, C)      (ETL_CHECKC1 (ETX, N, C).etvec)
#define ETX_XCMODE(ETX, N, C)     (ETL_CHECKC1 (ETX, N, C).rttype)
#define ETX_XCBITMAP(ETX, N, C)   (ETL_CHECKC1 (ETX, N, C).rtbit)
#define ETX_XCTREE(RTX, N, C)     (ETL_CHECKC1 (ETX, N, C).rttree)
#define ETX_XCBBDEF(ETX, N, C)    (ETL_CHECKC1 (ETX, N, C).bb)
#define ETX_XCADVFLAGS(ETX, N, C) (ETL_CHECKC1 (ETX, N, C).rt_addr_diff_vec_flags)
#define ETX_XCCSELIB(ETX, N, C)   (ETL_CHECKC1 (ETX, N, C).rt_cselib)

#define ETX_XCVECEXP(ETX, N, M, C)	ETVEC_ELT (XCVEC (ETX, N, C), M)
#define ETX_XCVECLEN(ETX, N, C)	ETX_GET_NUM_ELEM (XCVEC (ETX, N, C))

#define ETX_XC2EXP(ETX, N, C1, C2)      (ETL_CHECKC2 (ETX, N, C1, C2).etx)

/* ACCESS MACROS for particular fields of insns.  */

/* Determines whether X is an insn.  */
#define ETX_INSN_P(X)       (GET_ETX_CLASS (ETX_GET_CODE(X)) == 'i')

/* Holds a unique number for each insn.
   These are not necessarily sequentially increasing.  */
#define ETX_INSN_UID(INSN)  ETX_XINT (INSN, 0)

/* Chain insns together in sequence.  */
#define ETX_PREV_INSN(INSN)	ETX_XEXP (INSN, 1)
#define ETX_NEXT_INSN(INSN)	ETX_XEXP (INSN, 2)

/* Code number of instruction, from when it was recognized.
   -1 means this instruction has not been recognized yet.  */
#define ETX_INSN_CODE(INSN) ETX_XINT (INSN, 6)

/* The base value for branch probability notes.  */
#define REG_BR_PROB_BASE  10000

/* Define macros to extract and insert the reg-note kind in an EXPR_LIST.  */
#define REG_NOTE_KIND(LINK) ((enum reg_note) GET_MODE (LINK))
#define PUT_REG_NOTE_KIND(LINK, KIND) \
  PUT_MODE (LINK, (enum machine_mode) (KIND))

/* Names for REG_NOTE's in EXPR_LIST insn's.  */

extern const char * const reg_note_name[];
#define GET_REG_NOTE_NAME(MODE) (reg_note_name[(int) (MODE)])

#if 0
/* This field is only present on CALL_INSNs.  It holds a chain of EXPR_LIST of
   USE and CLOBBER expressions.
     USE expressions list the registers filled with arguments that
   are passed to the function.
     CLOBBER expressions document the registers explicitly clobbered
   by this CALL_INSN.
     Pseudo registers can not be mentioned in this list.  */
#define CALL_INSN_FUNCTION_USAGE(INSN)	ETX_XEXP(INSN, 9)

/* The label-number of a code-label.  The assembler label
   is made from `L' and the label-number printed in decimal.
   Label numbers are unique in a compilation.  */
#define CODE_LABEL_NUMBER(INSN)	ETX_XINT (INSN, 6)
#endif

#define LINE_NUMBER NOTE

/* In a NOTE that is a line number, this is a string for the file name that the
   line is in.  We use the same field to record block numbers temporarily in
   NOTE_INSN_BLOCK_BEG and NOTE_INSN_BLOCK_END notes.  (We avoid lots of casts
   between ints and pointers if we use a different macro for the block number.)
   */

#if 0
/* Opaque data.  */
#define NOTE_DATA(INSN)	        ETL_CHECKC1 (INSN, 4, NOTE)
#define NOTE_SOURCE_FILE(INSN)	ETX_XCSTR (INSN, 4, NOTE)
#define NOTE_EH_HANDLER(INSN)	ETX_XCINT (INSN, 4, NOTE)
#define NOTE_BASIC_BLOCK(INSN)	ETX_XCBBDEF (INSN, 4, NOTE)
#define NOTE_EXPECTED_VALUE(INSN) ETX_XCEXP (INSN, 4, NOTE)
#define NOTE_PREDICTION(INSN)   ETX_XCINT (INSN, 4, NOTE)
#define NOTE_PRECONDITIONED(INSN)   ETX_XCINT (INSN, 4, NOTE)

/* In a NOTE that is a line number, this is the line number.
   Other kinds of NOTEs are identified by negative numbers here.  */
#define NOTE_LINE_NUMBER(INSN) ETX_XCINT (INSN, 5, NOTE)

/* Nonzero if INSN is a note marking the beginning of a basic block.  */
#define NOTE_INSN_BASIC_BLOCK_P(INSN)			\
  (ETX_GET_CODE (INSN) == NOTE				\
   && NOTE_LINE_NUMBER (INSN) == NOTE_INSN_BASIC_BLOCK)

/* Algorithm and flags for prediction.  */
#define NOTE_PREDICTION_ALG(INSN)   (ETX_XCINT(INSN, 4, NOTE)>>8)
#define NOTE_PREDICTION_FLAGS(INSN) (ETX_XCINT(INSN, 4, NOTE)&0xff)
#define NOTE_PREDICT(ALG,FLAGS)     ((ALG<<8)+(FLAGS))
#endif

#if 0
/* Codes that appear in the NOTE_LINE_NUMBER field
   for kinds of notes that are not line numbers.

   Notice that we do not try to use zero here for any of
   the special note codes because sometimes the source line
   actually can be zero!  This happens (for example) when we
   are generating code for the per-translation-unit constructor
   and destructor routines for some C++ translation unit.

   If you should change any of the following values, or if you
   should add a new value here, don't forget to change the
   note_insn_name array in etl.c.  */

enum insn_note
{
  /* Keep all of these numbers negative.  Adjust as needed.  */
  NOTE_INSN_BIAS = -100,

  /* This note is used to get rid of an insn
     when it isn't safe to patch the insn out of the chain.  */
  NOTE_INSN_DELETED,

  /* These are used to mark the beginning and end of a lexical block.
     See NOTE_BLOCK, identify_blocks and reorder_blocks.  */
  NOTE_INSN_BLOCK_BEG,
  NOTE_INSN_BLOCK_END,

  /* These mark the extremes of a loop.  */
  NOTE_INSN_LOOP_BEG,
  NOTE_INSN_LOOP_END,

  /* Generated at the place in a loop that `continue' jumps to.  */
  NOTE_INSN_LOOP_CONT,
  /* Generated at the start of a duplicated exit test.  */
  NOTE_INSN_LOOP_VTOP,

  /* Generated at the end of a conditional at the top of the loop.
     This is used to perform a lame form of loop rotation in lieu
     of actually understanding the loop structure.  The note is
     discarded after rotation is complete.  */
  NOTE_INSN_LOOP_END_TOP_COND,

  /* This kind of note is generated at the end of the function body,
     just before the return insn or return label.  In an optimizing
     compilation it is deleted by the first jump optimization, after
     enabling that optimizer to determine whether control can fall
     off the end of the function body without a return statement.  */
  NOTE_INSN_FUNCTION_END,

  /* This marks the point immediately after the last prologue insn.  */
  NOTE_INSN_PROLOGUE_END,

  /* This marks the point immediately prior to the first epilogue insn.  */
  NOTE_INSN_EPILOGUE_BEG,

  /* Generated in place of user-declared labels when they are deleted.  */
  NOTE_INSN_DELETED_LABEL,

  /* This note indicates the start of the real body of the function,
     i.e. the point just after all of the parms have been moved into
     their homes, etc.  */
  NOTE_INSN_FUNCTION_BEG,

  /* These note where exception handling regions begin and end.
     Uses NOTE_EH_HANDLER to identify the region in question.  */
  NOTE_INSN_EH_REGION_BEG,
  NOTE_INSN_EH_REGION_END,

  /* Generated whenever a duplicate line number note is output.  For example,
     one is output after the end of an inline function, in order to prevent
     the line containing the inline call from being counted twice in gcov.  */
  NOTE_INSN_REPEATED_LINE_NUMBER,

  /* Record the struct for the following basic block.  Uses NOTE_BASIC_BLOCK.  */
  NOTE_INSN_BASIC_BLOCK,

  /* Record the expected value of a register at a location.  Uses
     NOTE_EXPECTED_VALUE; stored as (eq (reg) (const_int)).  */
  NOTE_INSN_EXPECTED_VALUE,

  /* Record a prediction.  Uses NOTE_PREDICTION.  */
  NOTE_INSN_PREDICTION,

  NOTE_INSN_MAX
};
#endif

/* Names for NOTE insn's other than line numbers.  */

#if 0
extern const char * const note_insn_name[NOTE_INSN_MAX - NOTE_INSN_BIAS];
#define GET_NOTE_INSN_NAME(NOTE_CODE) \
  (note_insn_name[(NOTE_CODE) - (int) NOTE_INSN_BIAS])
#endif

/* In etl.c */
extern etx etx_alloc (apr_pool_t *, ETX_CODE);
extern etvec etvec_alloc (apr_pool_t *, int);
extern etx copy_etx (apr_pool_t *, etx);
extern void dump_etx_statistics (void);

/* In emit-etl.c */
extern etx copy_etx_if_shared (etx);

/* In etl.c */
extern etx copy_most_etx (etx, etx);
extern etx shallow_copy_etx (apr_pool_t *, etx);
extern int etx_equal_p (etx, etx);

/* In emit-etl.c */
extern const char *etx_get_insn_name (int);

typedef struct etx_classobj etx_classobj_t;
struct etx_classobj
{
  apr_pool_t *pool;

  /* Current maximum length of directory names in the search path
     for include files.  (Altered as we get more of them.)  */
  size_t max_include_len;

  struct file_name_list *first_dir_md_include;  /* First dir to search */
  /* First dir to search for <file> */
  struct file_name_list *first_bracket_include;
  struct file_name_list *last_dir_md_include;        /* Last in chain */

  /* The filename for aborting with file and line.  */
  const char *read_etx_filename;
  /* The current line number for the file.  */
  int read_etx_lineno;

  htab_t condition_table;
  size_t n_insn_conditions;

  /* Obstack used for allocating ETL objects.  */
  struct obstack obstack;
  struct obstack *etl_obstack;
  int initialized;

  int sequence_num;
  int errors;

  struct queue_elem *define_attr_queue;
  struct queue_elem **define_attr_tail;
  struct queue_elem *define_insn_queue;
  struct queue_elem **define_insn_tail;
  struct queue_elem *define_cond_exec_queue;
  struct queue_elem **define_cond_exec_tail;
  struct queue_elem *other_queue;
  struct queue_elem **other_tail;

/* Set this to 0 to disable automatic elision of insn patterns which
   can never be used in this configuration.  See genconditions.c.
   Must be set before calling init_md_reader.  */
  int insn_elision;
/* If this is 1, the insn elision table doesn't even exist yet;
   maybe_eval_c_test will always return -1.  This is distinct from
   insn_elision because genflags and gencodes need to see all the
   patterns, but treat elided patterns differently.  */
  int insn_elision_unavailable;

  int predicable_default;
  const char *predicable_true;
  const char *predicable_false;
  char *base_dir;

  htab_t md_constants;

  /* Obstack to allocate bitmap elements from.  */
  struct obstack bitmap_obstack;
  int bitmap_obstack_init;
};

/* In print-etl.c */
extern const char *print_etx_head;
extern void debug_etx (etx);
extern void debug_etx_list (etx, int);
extern void debug_etx_range (etx, etx);
extern etx debug_etx_find (etx, int);
#ifdef BUFSIZ
extern void print_etl (FILE *, etx);
extern void print_simple_etl (FILE *, etx);
extern int print_etl_single (FILE *, etx);
extern void print_inline_etx (FILE *, etx, int);
#endif

/* In etl.c */
extern void init_etl (void);
extern void traverse_ed_constants (struct etx_classobj *, htab_trav, void *);
	
struct ed_constant { char *name, *value; };

#ifdef BUFSIZ
extern int read_etx_skip_spaces (struct etx_classobj *, ec_file_t *);
extern etx read_etx (struct etx_classobj *, ec_file_t *);
#endif

#endif /* ! ECLIB_ETL_H */
