#ifndef ESERVER_ETREE_H
#define ESERVER_ETREE_H

typedef union etree_node *etree;

#define MAX_ETREE_CODES                          256

extern const char etree_code_type[MAX_ETREE_CODES];
#define ETREE_CODE_CLASS(CODE)                   etree_code_type[(int) (CODE)]

/* CLASS  expression   etree-code class   0  ƴ 
   ǵݴϴ.  */

#define ETREE_IS_EXPR_CODE_CLASS(CLASS) \
  ((CLASS) == '<' || (CLASS) == '1' || (CLASS) == '2' || (CLASS) == 'e')

/*  etree-node  argument-word  (number).  */

extern int etree_code_length[MAX_ETREE_CODES];
#define ETREE_CODE_LENGTH(CODE)                  etree_code_length[(int) (CODE)]

/* etree node   code  */

#define DEFTREECODE(SYM, STRING, TYPE, NARGS)   SYM,

extern const char *etree_code_name[MAX_ETREE_CODES];

enum etree_code
  {
#include "etree.def"

    LAST_AND_UNUSED_ETREE_CODE   /* NUM_ETREE_CODE    
                                     */
  };

#undef DEFTREECODE

/* Language-independent etree code  .  */
#define NUM_ETREE_CODES                          ((int) LAST_AND_UNUSED_ETREE_CODE)

#define ETREE_CHECK(t, code)             (t)
#define ETREE_CLASS_CHECK(t, code)       (t)
#define CST_OR_CONSTRUCTOR_CHECK(t)     (t)
#define ETREE_EXPR_CHECK(t)                   (t)

#define ETREE_TYPE_CHECK(etree)        ETREE_CLASS_CHECK  (etree, 't')
#define ETREE_DECL_CHECK(etree)        ETREE_CLASS_CHECK  (etree, 'd')
#define ETREE_CST_CHECK(etree)         ETREE_CLASS_CHECK  (etree, 'c')

struct etree_common
{
  etree chain;
  etree type;

  enum etree_code code : 8;

  int constant_pool_num : 16;

  int is_reserved_word : 1;
  int is_reserved_insn : 1;
  /* TODO:  ʴ 24 bit  ӻ ؾ Ѵ.  */
  int is_labelname : 1;
  int public_flag : 1;
  int static_flag : 1;
  int unsigned_flag : 1;
  int unused : 2;
};

/* expression   node  ̰ expression  date type 
   Ÿϴ.
   POINTER_TYPE node  ̰ ͸ Ű type Դϴ.
   ARRAY_TYPE node  ̰ element  type Դϴ.
   VECTOR_TYPE node  ̰ element  type Դϴ.  */
#define ETREE_TYPE(NODE) ((NODE)->common.type)

#define ETREE_CODE(NODE)                         ((enum etree_code) (NODE)->common.code)
#define ETREE_SET_CODE(NODE, VALUE)              ((NODE)->common.code = (enum etree_code) (VALUE))

/* INTEGER_CST Ȥ REAL_CST, COMPLEX_CST, VECTOR_CST  ̰ folding
    overflow  ־ ǹѴ.  ̰ ETREE_OVERFLOW  ѷϰ
    Ǵµ, ANSI C   ǥĿ overflow  ߻Ͽ ,
    (diagnostic)  䱸ϱ ̴.  */
#define ETREE_CONSTANT_OVERFLOW(NODE) ((NODE)->common.static_flag)

/* INTEGER_CST Ȥ REAL_CST, COMPLEX_CST, VECTOR_CST  ̰
   folding  overflow  ־  ǥĿ   
   ߻ ʾ ǹѴ.  ETREE_OVERFLOW  ETREE_CONSTANT_OVERFLOW 
   , ݴδ ƴϴ.  */
#define ETREE_OVERFLOW(NODE) ((NODE)->common.public_flag)

/* INTEGER_TYPE Ȥ ENUMERAL_TYPE node 鿡 unsigned type  ǹѴ.
   FIELD_DECL node 鿡 unsigned bit field  ǹѴ.
    bit  DECL_BUILT_IN_NONANSI ó Լ鳻 ȴ.  */
#define ETREE_UNSIGNED(NODE) ((NODE)->common.unsigned_flag)

#define ETREE_CHAIN(NODE)                        ((NODE)->common.chain)

/*  ǥϴ node   ߰ field  accessor  Ѵ.  */

#define etree_build_int_2(POOL, LO, HI)  \
  etree_build_int_2_wide ((POOL), (unsigned int) (LO), (int) (HI))

/* INTEGER_CST node .   ΰ ϳ 2-word  .
    data type  sign  Ǿ ,  װ͵  
   ʰ ִ ϴ sign-extended  2 word ̴.
   2 word   unsign , extra bit  0 ̴.  */
#define ETREE_INT_CST(NODE) ((NODE)->int_cst.int_cst)
#define ETREE_INT_CST_LOW(NODE) (ETREE_INT_CST (NODE).low)
#define ETREE_INT_CST_HIGH(NODE) (ETREE_INT_CST (NODE).high)

#define ETREE_INT_CST_LT(A, B)                                \
  (ETREE_INT_CST_HIGH (A) < ETREE_INT_CST_HIGH (B)        \
   || (ETREE_INT_CST_HIGH (A) == ETREE_INT_CST_HIGH (B)   \
       && ETREE_INT_CST_LOW (A) < ETREE_INT_CST_LOW (B)))

#define ETREE_INT_CST_LT_UNSIGNED(A, B)                               \
  (((unsigned int) ETREE_INT_CST_HIGH (A)                \
    < (unsigned int) ETREE_INT_CST_HIGH (B))             \
   || (((unsigned int) ETREE_INT_CST_HIGH (A)            \
        == (unsigned int) ETREE_INT_CST_HIGH (B))        \
       && ETREE_INT_CST_LOW (A) < ETREE_INT_CST_LOW (B)))

struct etree_int_cst
{
  struct etree_common common;

  /* ⿡ -ü ʿ  Լ `const_hash'  ϳ unit
       word  ȱ⸦ ϰ -ü ּҸ   ν
       bounded pointer    ִ.  */
  struct {
    unsigned HOST_WIDE_INT low;
    HOST_WIDE_INT high;
  } int_cst;
};

#define ETREE_OPERAND(NODE, I)                   ((NODE)->exp.operands[I])

struct etree_exp
{
  struct etree_common common;
  int complexity;
  etree operands[1];
};

/* 0  ƴ  TYPE  integral type  Ÿϴ.  츮 ⼭
   COMPLEX type   ʴ´ٴ  ָϽʽÿ.  */

#define ETREE_INTEGRAL_TYPE_P(TYPE)  \
  (ETREE_CODE (TYPE) == ETREE_INTEGER_TYPE || ETREE_CODE (TYPE) == ETREE_ENUMERAL_TYPE  \
   || ETREE_CODE (TYPE) == ETREE_BOOLEAN_TYPE || ETREE_CODE (TYPE) == ETREE_CHAR_TYPE)

/* 0  ƴ  TYPE  complex floating-point type  ϴ
   floating-point type  Ÿϴ.  */

#define ETREE_FLOAT_TYPE_P(TYPE)              \
  (ETREE_CODE (TYPE) == ETREE_REAL_TYPE        \
   || (ETREE_CODE (TYPE) == ETREE_COMPLEX_TYPE \
       && ETREE_CODE (ETREE_TYPE (TYPE)) == ETREE_REAL_TYPE))

#define ETREE_REAL_CST(NODE) ((NODE)->real_cst.real_cst)

#include "real.h"

struct etree_real_cst
{
  struct etree_common common;
  REAL_VALUE_TYPE real_cst;
};

struct etree_type
{
  struct etree_common common;
  etree values;
  etree size;
  etree size_unit;
  etree attributes;
  unsigned int uid;

  unsigned int precision : 9;
  enum etree_machine_mode mode : 7;

  unsigned string_flag : 1;
  unsigned no_force_blk_flag : 1;
  unsigned needs_constructing_flag : 1;
  unsigned transparent_union_flag : 1;
  unsigned packed_flag : 1;
  unsigned restrict_flag : 1;
  unsigned pointer_depth : 2;

  unsigned lang_flag_0 : 1;
  unsigned lang_flag_1 : 1;
  unsigned lang_flag_2 : 1;
  unsigned lang_flag_3 : 1;
  unsigned lang_flag_4 : 1;
  unsigned lang_flag_5 : 1;
  unsigned lang_flag_6 : 1;
  unsigned user_align : 1;

  unsigned int align;
  etree pointer_to;
  etree reference_to;
  union {int address; char *pointer; } symtab;
  etree name;
  etree minval;
  etree maxval;
  etree next_variant;
  etree main_variant;
  etree binfo;
  etree context;
  HOST_WIDE_INT alias_set;
};

/*  TYPE  aggregate (-component) type  Ÿٸ 0  ƴ .  */

#define ETREE_AGGREGATE_TYPE_P(TYPE) \
  (ETREE_CODE (TYPE) == ARRAY_TYPE || ETREE_CODE (TYPE) == RECORD_TYPE \
   || ETREE_CODE (TYPE) == UNION_TYPE || ETREE_CODE (TYPE) == QUAL_UNION_TYPE \
   || ETREE_CODE (TYPE) == SET_TYPE)

/*  type   (-) typed-based alias .
   ETREE_TYPE_ALIAS_SET  ٸ object   alias   .
    ETREE_TYPE_ALIAS_SET  -1 ̸,  ƹ alias set   type
    Ҵ  ̴.   ETREE_TYPE_ALIAS_SET  0 ̸,  type
    object  ٸ type  object 鿡 alias   ִ.  */
#define ETREE_TYPE_ALIAS_SET(NODE) (ETREE_TYPE_CHECK (NODE)->type.alias_set)

/*   type   typed-based alias set  Ǿٸ 0  ƴ
    .  */
#define ETREE_TYPE_ALIAS_SET_KNOWN_P(NODE) (ETREE_TYPE_CHECK (NODE)->type.alias_set != -1)

/* data type  Ÿ node   field   ũε
   մϴ.  */

/* ̷ field  뿡   etree.def Ͽ  ֽϴ.
    ..._TYPE etree code   ʽÿ.  */

#define ETREE_TYPE_MODE(NODE) (ETREE_TYPE_CHECK (NODE)->type.mode)
#define ETREE_TYPE_SIZE(NODE) (ETREE_TYPE_CHECK (NODE)->type.size)
#define ETREE_TYPE_SIZE_UNIT(NODE) (ETREE_TYPE_CHECK (NODE)->type.size_unit)
#define ETREE_TYPE_NEXT_VARIANT(NODE) (ETREE_TYPE_CHECK (NODE)->type.next_variant)
#define ETREE_TYPE_MAIN_VARIANT(NODE) (ETREE_TYPE_CHECK (NODE)->type.main_variant)
#define ETREE_TYPE_MIN_VALUE(NODE) (ETREE_TYPE_CHECK (NODE)->type.minval)
#define ETREE_TYPE_MAX_VALUE(NODE) (ETREE_TYPE_CHECK (NODE)->type.maxval)
#define ETREE_TYPE_PRECISION(NODE) (ETREE_TYPE_CHECK (NODE)->type.precision)

/*  type  object   alignment ʿ伺.
    bit  , int ̴.  */
#define ETREE_TYPE_ALIGN(NODE) (ETREE_TYPE_CHECK (NODE)->type.align)

/*   type   alignment  "aligned" attribute   ö  1,
    װ  type  ⺻  0.  */
#define ETREE_TYPE_USER_ALIGN(NODE) (ETREE_TYPE_CHECK (NODE)->type.user_align)

/* ECONF_STRING_CST   */
#define ETREE_STRING_LENGTH(NODE)                        ((NODE)->econf_string.length)
#define ETREE_STRING_POINTER(NODE)                       ((NODE)->econf_string.pointer)

struct etree_string
{
  struct etree_common common;

  int length;
  const char *pointer;
};

#define ETREE_IDENTIFIER_LENGTH(NODE)			((NODE)->econf_name.length)
#define ETREE_IDENTIFIER_POINTER(NODE)                   ((NODE)->econf_name.pointer)

struct etree_identifier
{
  struct etree_common common;

  int length;
  const char *pointer;
};

union etree_node
{
  struct etree_common common;
  struct etree_int_cst int_cst;
  struct etree_exp exp;
  struct etree_real_cst real_cst;
  struct etree_type type;

  /* Ʒ etree node  ECONF   ͵Դϴ.  */
  struct etree_string econf_string;
  struct etree_identifier econf_name;
};

#define ETREE_NULL_TREE                               ((etree) NULL)

/* Data ü ũε, size  ٷ Լ, size  Ÿµ
   Ǵ  type  Ѵ.  */

enum etree_size_type_kind
{
  ETREE_SIZETYPE,             /* Ʈ size  Ϲ ǥ.  */
  ETREE_SSIZETYPE,            /* Ʈ size  Sign  ǥ.  */
  ETREE_USIZETYPE,            /* Ʈ size  Unsign  ǥ.  */
  ETREE_BITSIZETYPE,          /* Ʈ size  Ϲ ǥ.  */
  ETREE_SBITSIZETYPE,         /* Ʈ size  Sign  ǥ.  */
  ETREE_UBITSIZETYPE,         /* Ʈ size  Unsign  ǥ.  */
  ETREE_TYPE_KIND_LAST};

extern etree etree_sizetype_tab[(int) ETREE_TYPE_KIND_LAST];

#define etree_sizetype etree_sizetype_tab[(int) ETREE_SIZETYPE]
#define etree_bitsizetype etree_sizetype_tab[(int) ETREE_BITSIZETYPE]
#define etree_ssizetype etree_sizetype_tab[(int) ETREE_SSIZETYPE]
#define etree_usizetype etree_sizetype_tab[(int) ETREE_USIZETYPE]
#define etree_sbitsizetype etree_sizetype_tab[(int) ETREE_SBITSIZETYPE]
#define etree_ubitsizetype etree_sizetype_tab[(int) ETREE_UBITSIZETYPE]

#define etree_size_int(L) size_int_wide ((HOST_WIDE_INT) (L), ETREE_SIZETYPE)
#define etree_bitsize_int(L) size_int_wide ((HOST_WIDE_INT) (L), ETREE_BITSIZETYPE)

extern etree etree_integer_type_node;
extern etree etree_unsigned_type_node;
extern etree etree_long_integer_type_node;
extern etree etree_long_unsigned_type_node;
extern etree etree_float_type_node;
extern etree etree_double_type_node;
extern etree etree_long_long_unsigned_type_node;
extern etree etree_long_long_integer_type_node;
extern etree etree_long_double_type_node;

/* Compute the number of bytes occupied by 'node'.  This routine only
   looks at ETREE_CODE and, if the code is ETREE_VEC, ETREE_VEC_LENGTH.  */

extern size_t etree_size                 PARAMS ((etree));

/* Lowest level primitive for allocating a node.
   The ETREE_CODE is the only argument.  Contents are initialized
   to zero except for a few of the common fields.  */

extern etree etree_make_node                   PARAMS ((apr_pool_t *, enum etree_code));

extern void debug_etree                  PARAMS ((etree));

extern etree etree_build_int_2_wide (apr_pool_t *, unsigned int, int);
extern int first_rtl_op (enum etree_code);
extern int etree_int_fits_type_p (apr_pool_t *, etree, etree);
extern int etree_int_cst_sgn (etree);
extern void build_common_etree_nodes (apr_pool_t *);
extern etree etree_build_real (apr_pool_t *, etree, REAL_VALUE_TYPE);
extern etree etree_build_minus (apr_pool_t *, etree);
extern etree etree_build_string (apr_pool_t *, int, const char *);
extern etree lex_number (apr_pool_t *, const char *, unsigned int);
extern etree parse_number (apr_pool_t *, const char *, unsigned int);

#endif
