#ifndef ESERVER_TREE_H
#define ESERVER_TREE_H

typedef union tree_node *tree;

#define MAX_TREE_CODES                          256

extern const char tree_code_type[MAX_TREE_CODES];
#define TREE_CODE_CLASS(CODE)                   tree_code_type[(int) (CODE)]

/* CLASS  expression   tree-code class   0  ƴ 
   ǵݴϴ.  */

#define IS_EXPR_CODE_CLASS(CLASS) \
  ((CLASS) == '<' || (CLASS) == '1' || (CLASS) == '2' || (CLASS) == 'e')

/*  tree-node  argument-word  (number).  */

extern int tree_code_length[MAX_TREE_CODES];
#define TREE_CODE_LENGTH(CODE)                  tree_code_length[(int) (CODE)]

/* tree node   code  */

#define DEFTREECODE(SYM, STRING, TYPE, NARGS)   SYM,

extern const char *tree_code_name[MAX_TREE_CODES];

enum tree_code
  {
#include "tree.def"

    LAST_AND_UNUSED_TREE_CODE   /* NUM_TREE_CODE    
                                     */
  };

#undef DEFTREECODE

/* Language-independent tree code  .  */
#define NUM_TREE_CODES                          ((int) LAST_AND_UNUSED_TREE_CODE)

#define TREE_CHECK(t, code)             (t)
#define TREE_CLASS_CHECK(t, code)       (t)
#define CST_OR_CONSTRUCTOR_CHECK(t)     (t)
#define EXPR_CHECK(t)                   (t)

#define TYPE_CHECK(tree)        TREE_CLASS_CHECK  (tree, 't')
#define DECL_CHECK(tree)        TREE_CLASS_CHECK  (tree, 'd')
#define CST_CHECK(tree)         TREE_CLASS_CHECK  (tree, 'c')

struct tree_common
{
  tree chain;
  tree type;

  enum tree_code code : 8;

  int constant_pool_num : 16;

  int is_reserved_word : 1;
  int is_reserved_insn : 1;
  /* TODO:  ʴ 24 bit  ӻ ؾ Ѵ.  */
  int is_labelname : 1;
  int public_flag : 1;
  int static_flag : 1;
  int unsigned_flag : 1;
  int unused : 2;
};

/* expression   node  ̰ expression  date type 
   Ÿϴ.
   POINTER_TYPE node  ̰ ͸ Ű type Դϴ.
   ARRAY_TYPE node  ̰ element  type Դϴ.
   VECTOR_TYPE node  ̰ element  type Դϴ.  */
#define TREE_TYPE(NODE) ((NODE)->common.type)

#define TREE_CODE(NODE)                         ((enum tree_code) (NODE)->common.code)
#define TREE_SET_CODE(NODE, VALUE)              ((NODE)->common.code = (enum tree_code) (VALUE))

/* INTEGER_CST Ȥ REAL_CST, COMPLEX_CST, VECTOR_CST  ̰ folding
    overflow  ־ ǹѴ.  ̰ TREE_OVERFLOW  ѷϰ
    Ǵµ, ANSI C   ǥĿ overflow  ߻Ͽ ,
    (diagnostic)  䱸ϱ ̴.  */
#define TREE_CONSTANT_OVERFLOW(NODE) ((NODE)->common.static_flag)

/* INTEGER_CST Ȥ REAL_CST, COMPLEX_CST, VECTOR_CST  ̰
   folding  overflow  ־  ǥĿ   
   ߻ ʾ ǹѴ.  TREE_OVERFLOW  TREE_CONSTANT_OVERFLOW 
   , ݴδ ƴϴ.  */
#define TREE_OVERFLOW(NODE) ((NODE)->common.public_flag)

/* INTEGER_TYPE Ȥ ENUMERAL_TYPE node 鿡 unsigned type  ǹѴ.
   FIELD_DECL node 鿡 unsigned bit field  ǹѴ.
    bit  DECL_BUILT_IN_NONANSI ó Լ鳻 ȴ.  */
#define TREE_UNSIGNED(NODE) ((NODE)->common.unsigned_flag)

#define TREE_CHAIN(NODE)                        ((NODE)->common.chain)

/*  ǥϴ node   ߰ field  accessor  Ѵ.  */

#define build_int_2(POOL, LO, HI)  \
  build_int_2_wide ((POOL), (unsigned int) (LO), (int) (HI))

/* INTEGER_CST node .   ΰ ϳ 2-word  .
    data type  sign  Ǿ ,  װ͵  
   ʰ ִ ϴ sign-extended  2 word ̴.
   2 word   unsign , extra bit  0 ̴.  */
#define TREE_INT_CST(NODE) ((NODE)->int_cst.int_cst)
#define TREE_INT_CST_LOW(NODE) (TREE_INT_CST (NODE).low)
#define TREE_INT_CST_HIGH(NODE) (TREE_INT_CST (NODE).high)

#define INT_CST_LT(A, B)                                \
  (TREE_INT_CST_HIGH (A) < TREE_INT_CST_HIGH (B)        \
   || (TREE_INT_CST_HIGH (A) == TREE_INT_CST_HIGH (B)   \
       && TREE_INT_CST_LOW (A) < TREE_INT_CST_LOW (B)))

#define INT_CST_LT_UNSIGNED(A, B)                               \
  (((unsigned int) TREE_INT_CST_HIGH (A)                \
    < (unsigned int) TREE_INT_CST_HIGH (B))             \
   || (((unsigned int) TREE_INT_CST_HIGH (A)            \
        == (unsigned int) TREE_INT_CST_HIGH (B))        \
       && TREE_INT_CST_LOW (A) < TREE_INT_CST_LOW (B)))

struct tree_int_cst
{
  struct tree_common common;

  /* ⿡ -ü ʿ  Լ `const_hash'  ϳ unit
       word  ȱ⸦ ϰ -ü ּҸ   ν
       bounded pointer    ִ.  */
  struct {
    unsigned HOST_WIDE_INT low;
    HOST_WIDE_INT high;
  } int_cst;
};

#define TREE_OPERAND(NODE, I)                   ((NODE)->exp.operands[I])

struct tree_exp
{
  struct tree_common common;
  int complexity;
  tree operands[1];
};

/* 0  ƴ  TYPE  integral type  Ÿϴ.  츮 ⼭
   COMPLEX type   ʴ´ٴ  ָϽʽÿ.  */

#define INTEGRAL_TYPE_P(TYPE)  \
  (TREE_CODE (TYPE) == INTEGER_TYPE || TREE_CODE (TYPE) == ENUMERAL_TYPE  \
   || TREE_CODE (TYPE) == BOOLEAN_TYPE || TREE_CODE (TYPE) == CHAR_TYPE)

/* 0  ƴ  TYPE  complex floating-point type  ϴ
   floating-point type  Ÿϴ.  */

#define FLOAT_TYPE_P(TYPE)              \
  (TREE_CODE (TYPE) == REAL_TYPE        \
   || (TREE_CODE (TYPE) == COMPLEX_TYPE \
       && TREE_CODE (TREE_TYPE (TYPE)) == REAL_TYPE))

#define TREE_REAL_CST(NODE) ((NODE)->real_cst.real_cst)

#include "real.h"

struct tree_real_cst
{
  struct tree_common common;
  REAL_VALUE_TYPE real_cst;
};

struct tree_type
{
  struct tree_common common;
  tree values;
  tree size;
  tree size_unit;
  tree attributes;
  unsigned int uid;

  unsigned int precision : 9;
  enum machine_mode mode : 7;

  unsigned string_flag : 1;
  unsigned no_force_blk_flag : 1;
  unsigned needs_constructing_flag : 1;
  unsigned transparent_union_flag : 1;
  unsigned packed_flag : 1;
  unsigned restrict_flag : 1;
  unsigned pointer_depth : 2;

  unsigned lang_flag_0 : 1;
  unsigned lang_flag_1 : 1;
  unsigned lang_flag_2 : 1;
  unsigned lang_flag_3 : 1;
  unsigned lang_flag_4 : 1;
  unsigned lang_flag_5 : 1;
  unsigned lang_flag_6 : 1;
  unsigned user_align : 1;

  unsigned int align;
  tree pointer_to;
  tree reference_to;
  union {int address; char *pointer; } symtab;
  tree name;
  tree minval;
  tree maxval;
  tree next_variant;
  tree main_variant;
  tree binfo;
  tree context;
  HOST_WIDE_INT alias_set;
};

/*  TYPE  aggregate (-component) type  Ÿٸ 0  ƴ .  */

#define AGGREGATE_TYPE_P(TYPE) \
  (TREE_CODE (TYPE) == ARRAY_TYPE || TREE_CODE (TYPE) == RECORD_TYPE \
   || TREE_CODE (TYPE) == UNION_TYPE || TREE_CODE (TYPE) == QUAL_UNION_TYPE \
   || TREE_CODE (TYPE) == SET_TYPE)

/*  type   (-) typed-based alias .
   TYPE_ALIAS_SET  ٸ object   alias   .
    TYPE_ALIAS_SET  -1 ̸,  ƹ alias set   type
    Ҵ  ̴.   TYPE_ALIAS_SET  0 ̸,  type
    object  ٸ type  object 鿡 alias   ִ.  */
#define TYPE_ALIAS_SET(NODE) (TYPE_CHECK (NODE)->type.alias_set)

/*   type   typed-based alias set  Ǿٸ 0  ƴ
    .  */
#define TYPE_ALIAS_SET_KNOWN_P(NODE) (TYPE_CHECK (NODE)->type.alias_set != -1)

/* data type  Ÿ node   field   ũε
   մϴ.  */

/* ̷ field  뿡   tree.def Ͽ  ֽϴ.
    ..._TYPE tree code   ʽÿ.  */

#define TYPE_MODE(NODE) (TYPE_CHECK (NODE)->type.mode)
#define TYPE_SIZE(NODE) (TYPE_CHECK (NODE)->type.size)
#define TYPE_SIZE_UNIT(NODE) (TYPE_CHECK (NODE)->type.size_unit)
#define TYPE_NEXT_VARIANT(NODE) (TYPE_CHECK (NODE)->type.next_variant)
#define TYPE_MAIN_VARIANT(NODE) (TYPE_CHECK (NODE)->type.main_variant)
#define TYPE_MIN_VALUE(NODE) (TYPE_CHECK (NODE)->type.minval)
#define TYPE_MAX_VALUE(NODE) (TYPE_CHECK (NODE)->type.maxval)
#define TYPE_PRECISION(NODE) (TYPE_CHECK (NODE)->type.precision)

/*  type  object   alignment ʿ伺.
    bit  , int ̴.  */
#define TYPE_ALIGN(NODE) (TYPE_CHECK (NODE)->type.align)

/*   type   alignment  "aligned" attribute   ö  1,
    װ  type  ⺻  0.  */
#define TYPE_USER_ALIGN(NODE) (TYPE_CHECK (NODE)->type.user_align)

/* ECONF_STRING_CST   */
#define TREE_STRING_LENGTH(NODE)                        ((NODE)->econf_string.length)
#define TREE_STRING_POINTER(NODE)                       ((NODE)->econf_string.pointer)

struct tree_string
{
  struct tree_common common;

  int length;
  const char *pointer;
};

#define TREE_IDENTIFIER_LENGTH(NODE)			((NODE)->econf_name.length)
#define TREE_IDENTIFIER_POINTER(NODE)                   ((NODE)->econf_name.pointer)

struct tree_identifier
{
  struct tree_common common;

  int length;
  const char *pointer;
};

union tree_node
{
  struct tree_common common;
  struct tree_int_cst int_cst;
  struct tree_exp exp;
  struct tree_real_cst real_cst;
  struct tree_type type;

  /* Ʒ tree node  ECONF   ͵Դϴ.  */
  struct tree_string econf_string;
  struct tree_identifier econf_name;
};

#define NULL_TREE                               ((tree) NULL)

/* Data ü ũε, size  ٷ Լ, size  Ÿµ
   Ǵ  type  Ѵ.  */

enum size_type_kind
{
  SIZETYPE,             /* Ʈ size  Ϲ ǥ.  */
  SSIZETYPE,            /* Ʈ size  Sign  ǥ.  */
  USIZETYPE,            /* Ʈ size  Unsign  ǥ.  */
  BITSIZETYPE,          /* Ʈ size  Ϲ ǥ.  */
  SBITSIZETYPE,         /* Ʈ size  Sign  ǥ.  */
  UBITSIZETYPE,         /* Ʈ size  Unsign  ǥ.  */
  TYPE_KIND_LAST};

extern tree sizetype_tab[(int) TYPE_KIND_LAST];

#define sizetype sizetype_tab[(int) SIZETYPE]
#define bitsizetype sizetype_tab[(int) BITSIZETYPE]
#define ssizetype sizetype_tab[(int) SSIZETYPE]
#define usizetype sizetype_tab[(int) USIZETYPE]
#define sbitsizetype sizetype_tab[(int) SBITSIZETYPE]
#define ubitsizetype sizetype_tab[(int) UBITSIZETYPE]

#define size_int(L) size_int_wide ((HOST_WIDE_INT) (L), SIZETYPE)
#define bitsize_int(L) size_int_wide ((HOST_WIDE_INT) (L), BITSIZETYPE)

extern tree integer_type_node;
extern tree unsigned_type_node;
extern tree long_integer_type_node;
extern tree long_unsigned_type_node;
extern tree float_type_node;
extern tree double_type_node;
extern tree long_long_unsigned_type_node;
extern tree long_long_integer_type_node;
extern tree long_double_type_node;

/* Compute the number of bytes occupied by 'node'.  This routine only
   looks at TREE_CODE and, if the code is TREE_VEC, TREE_VEC_LENGTH.  */

extern size_t tree_size                 PARAMS ((tree));

/* Lowest level primitive for allocating a node.
   The TREE_CODE is the only argument.  Contents are initialized
   to zero except for a few of the common fields.  */

extern tree make_node                   PARAMS ((apr_pool_t *, enum tree_code));

extern void debug_tree                  PARAMS ((tree));

extern tree build_int_2_wide (apr_pool_t *, unsigned int, int);
extern int first_rtl_op (enum tree_code);
extern int int_fits_type_p (apr_pool_t *, tree, tree);
extern int tree_int_cst_sgn (tree);
extern void build_common_tree_nodes (apr_pool_t *);
extern tree build_real (apr_pool_t *, tree, REAL_VALUE_TYPE);
extern tree build_minus (apr_pool_t *, tree);
extern tree build_string (apr_pool_t *, int, const char *);

#endif
