/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of eserver.

eserver is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

eserver is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "config.h"
#include "apr-1/apr.h"
#include "apr-1/apr_portable.h"
#include "apr-1/apr_strings.h"
#include "apr-1/apr_thread_proc.h"
#include "apr-1/apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr-1/apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE
#include "apr-1/apr_poll.h"
#include "apr-1/apr_buckets.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>            /* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>
#include <setjmp.h>

#include "system.h"
#include "etldefaults.h"
#include "obstack.h"
#include "hashtab.h"

/* TREE   header .  */
#include "config/i386.h"
#include "hwint.h"
#include "machmode.h"
#include "etree.h"
#include "fold-const.h"
#include "stor-layout.h"

#include "econnect.h"
#include "etl.h"
#include "obstack.h"
#include "hashtab.h"
#include "errors.h"

static void float_signal (int);
static void set_float_handler (jmp_buf);
int do_float_handler (void (*fn) PARAMS ((PTR)), PTR);
static void parse_float (PTR);

struct pf_args
{
  /* Input */
  const char *str;
  int fflag;
  int lflag;
  int base;
  /* Output */
  int conversion_errno;
  REAL_VALUE_TYPE value;
  etree type;
};

static int float_handler_set;
int float_handled;
jmp_buf float_handler;

/* Signal    ´.  */

static void
float_signal (signo)
     /* If this is missing, some compilers complain.  */
     int signo ATTRIBUTE_UNUSED;
{
  if (float_handled == 0)
    etl_error ("internal error: %d", signo);
  float_handled = 0;

  /* On System-V derived systems, we must reinstall the signal handler.
     This is harmless on BSD-derived systems.  */
  signal (SIGFPE, float_signal);
  longjmp (float_handler, 1);
}

/* floating arithmetic error  ߻  longjmp    Ѵ.
    HANDLER  0  ,  ̻  ٷ  ǹѴ.  */

static void
set_float_handler (handler)
     jmp_buf handler;
{
  float_handled = (handler != 0);
  if (handler)
    memcpy (float_handler, handler, sizeof (float_handler));

  if (float_handled && ! float_handler_set)
    {
      signal (SIGFPE, float_signal);
      float_handler_set = 1;
    }
}

/* ̰ arithmetic exception  ų  ִ code   wrapper
   function ̴.  ش code  function pointer FN  ޵Ǹ,
   ϳ argument DATA  ǳٴ.  DATA   function FN  
   real input  output  ϰ ִ.   Լ  longjmp  ȣǾ
   (, exception  ߻Ǹ) 0  (н) ȯѴ.  ׷  
   () 1  ȯѴ.  */

int
do_float_handler (fn, data)
     void (*fn) PARAMS ((PTR));
     PTR data;
{
  jmp_buf buf;

  if (setjmp (buf))
    {
      /* 츮 Լ fn ()  exception  ߻  longjmp ()  
          Ѵ.  */
      set_float_handler (NULL);
      return 0;
    }

  set_float_handler (buf);
  (*fn) (data);
  set_float_handler (NULL);
  return 1;
}

static void
parse_float (data)
  PTR data;
{
  struct pf_args * args = (struct pf_args *) data;
  const char *typename;

  args->conversion_errno = 0;
  args->type = etree_double_type_node;
  typename = "double";

  /* The second argument, machine_mode, of REAL_VALUE_ATOF
     tells the desired precision of the binary result
     of decimal-to-binary conversion.  */

  if (args->fflag)
    {
      if (args->lflag)
        etl_error ("both 'f' and 'l' suffixes on floating constant");

      args->type = etree_float_type_node;
      typename = "float";
    }
  else if (args->lflag)
    {
      args->type = etree_long_double_type_node;
      typename = "long double";
    }
  errno = 0;
  if (args->base == 16)
    args->value = REAL_VALUE_HTOF (args->str, ETREE_TYPE_MODE (args->type));
  else
    args->value = REAL_VALUE_ATOF (args->str, ETREE_TYPE_MODE (args->type));

  args->conversion_errno = errno;
  /* A diagnostic is required here by some ISO C testsuites.
     This is not pedwarn, because some people don't want
     an error for this.  */
  if (REAL_VALUE_ISINF (args->value))
    etl_warning ("floating point number exceeds range of '%s'", typename);
}

/* Node   type  մϴ.  */
etree
lex_number (pool, str, len)
     apr_pool_t *pool;
     const char *str;
     unsigned int len;
{
  int count = 0;
  int base = 10;
  etree type, value;
  int c;
  int largest_digit = 0;
  int numdigits = 0;
  int overflow = 0;
  const char *p;
  enum anon1 { NOT_FLOAT = 0, AFTER_POINT, AFTER_EXPON } floatflag = NOT_FLOAT;

  /* 츮   part   HOST_BITS_PER_CHAR bit 鸸 Ѵ.
     Part 迭 ä Ʒ code  host int  host char  ּ ι̻
     ṵ̃, HOST_BITS_PER_WIDE_INT  HOST_BITS_PER_CHAR  ¦ ̶
     Ѵ.   HOST_WIDE_INT  츮   ִ  ū int literal
     ̴.  Ʒ overflow  ϱ ؼ, Part  (TOTAL_PARTS) 
       HOST_WIDE_INT  bit   ִµ ʿ Ȯ part
      number ߸ Ѵ.  */
#define TOTAL_PARTS ((HOST_BITS_PER_WIDE_INT / HOST_BITS_PER_CHAR) * 2)
  unsigned int parts[TOTAL_PARTS];

  /*     ȭѴ.  */
  if (len == 1)
    {
      if (*str == '0')
        return etree_build_int_2 (pool, *str - '0', 0);
      else if (*str == '1')
        return etree_build_int_2 (pool, *str - '0', 0);
      else
        return etree_build_int_2 (pool, *str - '0', 0);
    }

  for (count = 0; count < TOTAL_PARTS; count++)
    parts[count] = 0;

  /*   len  >1  ˷.  */
  p = str;

  if (len > 2 && str[0] == '0' && (str[1] == 'x' || str[1] == 'X'))
    {
      base = 16;
      p = str + 2;
    }
  /* ISDIGIT check  ν 0   suffix   ȥϴ   ̴.  */
  else if (str[0] == '0' && ISDIGIT (str[1]))
    {
      base = 8;
      p = str + 1;
    }

  do
    {
      c = *p++;

      if (c == '.')
        {
          if (floatflag == AFTER_POINT)
            etl_error ("too many decimal points in floating constant");
          else if (floatflag == AFTER_EXPON)
            etl_error ("decimal point in exponent - impossible!");
          else
            floatflag = AFTER_POINT;

          if (base == 8)
            base = 10;
        }
      else if (c == '_')
        /* ߿ Ȯ ɼ , number   _  Ѵ.
           Cosmetic grouping  ϴµ,  , 0x8000_0000 == 0x80000000
           ̴.  ټ б   ִ.  Ada  ̰ ϳ?  */
        etl_error ("underscore in number");
      else
        {
          int n;
          /* Decial point  ƴϴ.  Digit (Ƹ hex digit)  Ѵ.  */

          if (ISDIGIT (c)
              || (base == 16 && ISXDIGIT (c)))
            {
              n = hex_value (c);
            }
          else if (base <= 10 && (c == 'e' || c == 'E'))
            {
              base = 10;
              floatflag = AFTER_EXPON;
              break;
            }
          else if (base == 16 && (c == 'p' || c == 'P'))
            {
              floatflag = AFTER_EXPON;
              break;   /* Exponent   */
            }
          else
            {
              p--;
              break;  /* start of suffix */
            }

          if (n >= largest_digit)
            largest_digit = n;
          numdigits++;

          for (count = 0; count < TOTAL_PARTS; count++)
            {
              parts[count] *= base;
              if (count)
                {
                  parts[count]
                    += (parts[count-1] >> HOST_BITS_PER_CHAR);
                  parts[count-1]
                    &= (1 << HOST_BITS_PER_CHAR) - 1;
                }
              else
                parts[0] += n;
            }

          /*  -- part  overflow  ߻  (host char
                ִ ͺ ũٸ) ü number  overflow Ǿ.
             ̰ ϰ -- part  truncate Ѵ.  */
          if (parts[TOTAL_PARTS - 1] >> HOST_BITS_PER_CHAR)
            {
              overflow = 1;
              parts[TOTAL_PARTS - 1] &= (1 << HOST_BITS_PER_CHAR) - 1;
            }
        }
    }
  while (p < str + len);

  /* ̰ `int i = 0x;'   Է   ߻Ѵ.  */
  if (numdigits == 0)
    etl_error ("numeric constant with no digits");

  if (largest_digit >= base)
    etl_error ("numeric constant contains digits beyond the radix");
  
  if (floatflag != NOT_FLOAT)
    {
      int imag, fflag, lflag, conversion_errno;
      char *copy;
      REAL_VALUE_TYPE real;
      struct pf_args args;

      /* Read explicit exponent if any, and put it in tokenbuf.  */
      if ((base == 10 && ((c == 'e') || (c == 'E')))
          || (base == 16 && (c == 'p' || c == 'P')))
        {
          if (p < str + len)
            c = *p++;
          if (p < str + len && (c == '+' || c == '-'))
            c = *p++;
          /* Exponent is decimal, even if string is a hex float.  */
          if (! ISDIGIT (c))
            etl_error ("floating constant exponent has no digits");
          while (p < str + len && ISDIGIT (c))
            c = *p++;
          if (! ISDIGIT (c))
            p--;
        }

      /*  float constant  Ѵ; 츮 parse_float  ǳ
         ڿ  suffix  ԵǱ  ʴ´.  */
      copy = (char *) apr_pcalloc (pool, p - str + 1);
      memcpy (copy, str, p - str);
      copy[p - str] = '\0';

      /*  suffix  ؼѴ.  */
      fflag = lflag = imag = 0;
      while (p < str + len)
        switch (*p++)
          {
          case 'f': case 'F':
            if (fflag)
              etl_error ("more than one 'f' suffix on floating constant");

            fflag = 1;
            break;

          case 'l': case 'L':
            if (lflag)
              etl_error ("more than one 'l' suffix on floating constant");

            lflag = 1;
            break;

          case 'i': case 'I':
          case 'j': case 'J':
            if (imag)
              etl_error ("more than one 'i' or 'j' suffix on floating constant");

            imag = 1;
            break;

          default:
            etl_error ("invalid suffix on floating constant");
          }

      /* parse_float()   Է° Ѵ. */
      args.str = copy;
      args.fflag = fflag;
      args.lflag = lflag;
      args.base = base;

      /* ڿ double  ϴµ, overflow  ˻Ѵ.  */
      if (do_float_handler (parse_float, (PTR) &args))
        {
          /* parse_float()  output  ޴´.  */
          real = args.value;
        }
      else
        /* We got an exception from parse_float() */
        etl_error ("floating constant out of range");

      conversion_errno = args.conversion_errno;
      type = args.type;

      /*  type   ̿Ͽ node  Ѵ.  */
      if (imag)
        {
          etl_error (" Ҽ ó Ѵ. TODO: ó ƾ ߰ ʿ.");
          value = NULL;
        }
      else
        value = etree_build_real (pool, type, real);
    }
  else
    {
      etree trad_type, ansi_type;
      int high, low;
      int spec_unsigned = 0;
      int spec_long = 0;
      int spec_long_long = 0;
      int spec_imag = 0;
      int suffix_lu = 0;
      int warn = 0, i;

      trad_type = ansi_type = type = ETREE_NULL_TREE;
      while (p < str + len)
        {
          c = *p++;
          switch (c)
            {
            case 'u': case 'U':
              if (spec_unsigned)
                etl_error ("two 'u' suffixes on integer constant");

              spec_unsigned = 1;
              if (spec_long)
                suffix_lu = 1;
              break;

            case 'l': case 'L':
              if (spec_long)
                {
                  if (spec_long_long)
                    etl_error ("three 'l' suffixes on integer constant");
                  else if (suffix_lu)
                    etl_error ("'lul' is not a valid integer suffix");
                  else if (c != spec_long)
                    etl_error ("'Ll' and 'lL' are not valid integer suffixes");
                  spec_long_long = 1;
                }
              spec_long = c;
              break;

            case 'i': case 'I': case 'j': case 'J':
              if (spec_imag)
                etl_error ("more than one 'i' or 'j' suffix on integer constant");
              spec_imag = 1;
              break;

            default:
              etl_error ("invalid suffix on integer constant");
            }
        }

      /*  literal  overflow Ǿٸ,  ׿  Ѵ.  */
      if (overflow)
        {
          warn = 1;
          etl_warning ("integer constant is too large for this configuration of the compiler - truncated to %d bits",
                   HOST_BITS_PER_WIDE_INT * 2);
        }

      /* ̰ 츮  ׻ ̶ Ƿ ȭȴ.  */

      high = low = 0;

      for (i = 0; i < HOST_BITS_PER_WIDE_INT / HOST_BITS_PER_CHAR; i++)
        {
          high |= ((int) parts[i + (HOST_BITS_PER_WIDE_INT
                                              / HOST_BITS_PER_CHAR)]
                   << (i * HOST_BITS_PER_CHAR));
          low |= (int) parts[i] << (i * HOST_BITS_PER_CHAR);
        }

      value = etree_build_int_2 (pool, low, high);
      ETREE_TYPE (value) = etree_long_long_unsigned_type_node;

      /* Calculate the ISO type.  */
      if (! spec_long && ! spec_unsigned
          && etree_int_fits_type_p (pool, value, etree_integer_type_node))
        ansi_type = etree_integer_type_node;
      else if (! spec_long && (base != 10 || spec_unsigned)
               && etree_int_fits_type_p (pool,value, etree_unsigned_type_node))
        ansi_type = etree_unsigned_type_node;
      else if (! spec_unsigned && !spec_long_long
               && etree_int_fits_type_p (pool, value, etree_long_integer_type_node))
        ansi_type = etree_long_integer_type_node;
      else if (! spec_long_long
               && etree_int_fits_type_p (pool, value, etree_long_unsigned_type_node))
        ansi_type = etree_long_unsigned_type_node;
      else if (! spec_unsigned
               && etree_int_fits_type_p (pool, value, etree_long_long_integer_type_node))
        ansi_type = etree_long_long_integer_type_node;
      else if (etree_int_fits_type_p (pool, value, etree_long_long_unsigned_type_node))
        ansi_type = etree_long_long_unsigned_type_node;

      type = ansi_type;

      if (base == 10 && ! spec_unsigned && ETREE_UNSIGNED (type))
        etl_warning ("decimal constant is so large that it is unsigned");

      if (spec_imag)
        {
          if (ETREE_TYPE_PRECISION (type)
              <= ETREE_TYPE_PRECISION (etree_integer_type_node))
            etl_error (" Ҽ ó Ѵ. TODO: ó ƾ ߰ ʿ.");
          else
            etl_error ("complex integer constant is too wide for 'complex int'");
        }
      else
        ETREE_TYPE (value) = type;
    }

  if (p < str + len)
    etl_error ("missing white space after number '%.*s'", (int) (p - str), str);

  return value;
}

etree
parse_number (p, str, len)
     apr_pool_t *p;
     const char *str;
     unsigned int len;
{
  int is_negate = 0;
  char *r, *tmpvalue = apr_pstrndup (p, str, len);
  etree t;
  
  r = tmpvalue;
  while (ISSPACE (*r))
    r++;
  switch (r[0])
    {
    case '-':
      is_negate = 1;
      r++;
      break;
    default:
      break;
    }
  while (ISSPACE (*r))
    r++;
  
  t = lex_number (p, r, strlen (r));
  if (is_negate)
    t = etree_fold (p, etree_build_minus (p, t));

  return t;
}
