/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of libetl.

libetl is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

libetl is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "config.h"
#include "apr-1/apr.h"
#include "apr-1/apr_portable.h"
#include "apr-1/apr_strings.h"
#include "apr-1/apr_thread_proc.h"
#include "apr-1/apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr-1/apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE
#include "apr-1/apr_poll.h"
#include "apr-1/apr_buckets.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>            /* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#define HTTP_NOT_CONNECTED -1
#include "system.h"
#include "etldefaults.h"
#include "obstack.h"
#include "hashtab.h"

/* TREE   header .  */
#include "config/i386.h"
#include "hwint.h"
#include "machmode.h"
#include "etree.h"
#include "fold-const.h"
#include "stor-layout.h"

#include "econnect.h"
#include "etl.h"
#include "obstack.h"
#include "errors.h"
#include "hashtab.h"
#include "econnect.h"
#include "gensupport.h"

static const char * etl_alloc_string (apr_pool_t *, const char *, int);
		       
etree nreverse (etree);
etree chainon (etree, etree);
etree copy_node (apr_pool_t *, etree);

etree etree_integer_type_node;
etree etree_unsigned_type_node;
etree etree_long_integer_type_node;
etree etree_long_unsigned_type_node;
etree etree_float_type_node;
etree etree_double_type_node;
etree etree_long_long_unsigned_type_node;
etree etree_long_long_integer_type_node;
etree etree_long_double_type_node;

/* build_common_etree_nodes () Լ ü process 󿡼 ѹ ϱ
   ؼ   Ѵ.   0  ƴ ,  ȣǾ ǹѴ.  */
static int build_common_etree_nodes_var;

#define DEFTREECODE(SYM, NAME, TYPE, LENGTH)    TYPE,

const char etree_code_type[MAX_ETREE_CODES] = {
#include "etree.def"
};
#undef DEFTREECODE

/* etree code   expression operand    ̺.
    operand  node structure   κԴϴ.
    type  decl    ʽϴ. */

#define DEFTREECODE(SYM, NAME, TYPE, LENGTH)    LENGTH,

int etree_code_length[MAX_ETREE_CODES] = {
#include "etree.def"
};
#undef DEFTREECODE

/* etree component  ̸.
   etree  ϰų  ޼   ˴ϴ.  */
#define DEFTREECODE(SYM, NAME, TYPE, LEN)       NAME,

const char *etree_code_name[MAX_ETREE_CODES] = {
#include "etree.def"
};
#undef DEFTREECODE

size_t
etree_size (node)
     etree node;
{
  enum etree_code code = ETREE_CODE (node);

  switch (ETREE_CODE_CLASS (code))
    {
    case 't':  /* type node  */
      return sizeof (struct etree_type);

    case '1':   /* a unary arithmetic expression */
      return (sizeof (struct etree_exp)
              + (ETREE_CODE_LENGTH (code) - 1) * sizeof (char *));

    case 'c':  /*  */
      /* 츮 ETREE_INTEGER_CST  ؼ ETREE_CODE_LENGTH    µ,
         ̰ word   machine-dependent ϱ ε, ͺ
            ִ ( , long long) HOST_WIDE_INT  
         ȭ ؼ.  ϰ ETREE_ETREE_REAL_CST  ؼ, `double'  ũ
         alignment  ϱ  word    machine-dependent
         ϴ.  */
      if (code == ETREE_INTEGER_CST)
        return sizeof (struct etree_int_cst);
      else if (code == ETREE_REAL_CST)
        return sizeof (struct etree_real_cst);
      else
        return (sizeof (struct etree_common)
                + ETREE_CODE_LENGTH (code) * sizeof (char *));

    case 'x':
      {
        size_t length;
        length = (sizeof (struct etree_common)
                  + ETREE_CODE_LENGTH (code) * sizeof (char *));
        return length;
      }
    default:
      fprintf (stderr, "etree_size: óؾ ϴ . '%c'\n", ETREE_CODE_CLASS (code));
      abort ();
    }
}

/* Chain T  ҵ  ݴ մϴ. ׸
   ο chain  ó  (  )  ȯմϴ.  */

etree
nreverse (t)
     etree t;
{
  etree prev = 0, decl, next;
  for (decl = t; decl; decl = next)
    {
      next = ETREE_CHAIN (decl);
      ETREE_CHAIN (decl) = prev;
      prev = decl;
    }
  return prev;
}

/* CODE  TREE node  .  */
etree
etree_make_node (p, code)
     apr_pool_t *p;
     enum etree_code code;
{
  etree t;
  int type = ETREE_CODE_CLASS (code);
  size_t length;
  struct etree_common ttmp;

  ETREE_SET_CODE ((etree) &ttmp, code);
  length = etree_size ((etree) &ttmp);

  t = (etree) apr_pcalloc (p, length);
  memset ((PTR) t, 0, length);
  ETREE_SET_CODE (t, code);

  switch (type)
    {
    default:
      break;
    }

  switch (code)
    {
    default:
      break;
    }

  return t;
}

/* chain 1   node  chain 2  Ű ν  chain
   node  մϴ. (ETREE_CHAIN   chain )
   ̰ Lisp ô `nconc' ̴.  */

etree
chainon (op1, op2)
     etree op1, op2;
{
  if (op1)
    {
      etree t1;
      etree t2;

      for (t1 = op1; ETREE_CHAIN (t1); t1 = ETREE_CHAIN (t1))
        ;
      ETREE_CHAIN (t1) = op2;
      for (t2 = op2; t2; t2 = ETREE_CHAIN (t2))
        if (t2 == t1)
          abort ();  /* Circularity created.  */
      return op1;
    }
  else
    return op2;
}

etree
etree_build_minus (p, op)
     apr_pool_t *p;
     etree op;
{
  etree t = etree_make_node (p, ETREE_NEGATE_EXPR);

  if (ETREE_CODE (op) != ETREE_INTEGER_CST)
    etl_error ("build_minus: Node Code  ETREE_INTEGER_CST  ƴմϴ. (Ʈ ʿ   )\n");

  ETREE_OPERAND (t, 0) = op;

  return t;
}

/*   ΰ  LOW  HI    
   INTERGET_CST node  ȯѴ.
   ETREE_TYPE  `int'  Ǿ.

    Լ `build_int_2' ũθ ؼ ݵ ȣǾ Ѵ.  */

etree
etree_build_int_2_wide (p, low, hi)
     apr_pool_t *p;
     unsigned int low;
     int hi;
{
  etree t = etree_make_node (p, ETREE_INTEGER_CST);

  ETREE_INT_CST_LOW (t) = low;
  ETREE_INT_CST_HIGH (t) = hi;
  ETREE_TYPE (t) = etree_integer_type_node;

  return t;
}

/* CODE   ù° non-etree operand  index  ȯϰų, 
     etree   operand   ȯѴ.  */

int
first_rtl_op (code)
     enum etree_code code;
{
  switch (code)
    {
    default:
      return ETREE_CODE_LENGTH (code);
    }
}

/* Type  TYPE ̰  D  ο ETREE_ETREE_REAL_CST node  ȯѴ.  */

etree
etree_build_real (p, type, d)
     apr_pool_t *p;
     etree type;
     REAL_VALUE_TYPE d;
{
  etree v;
  int overflow = 0;

  v = etree_make_node (p, ETREE_REAL_CST);
  ETREE_TYPE (v) = type;
  ETREE_REAL_CST (v) = d;
  ETREE_OVERFLOW (v) = ETREE_CONSTANT_OVERFLOW (v) = overflow;

  return v;
}

/*   T  ȣ  ȯѴ.
   ȯ T < 0 ̸ -1, T == 0 ̸ 0, T > 0 ̸ 1 ̴.
   T  type  unsign    -1  ȯ  ϶.  */

int
etree_int_cst_sgn (t)
     etree t;
{
  if (ETREE_INT_CST_LOW (t) == 0 && ETREE_INT_CST_HIGH (t) == 0)
    return 0;
  else if (ETREE_UNSIGNED (ETREE_TYPE (t)))
    return 1;
  else if (ETREE_INT_CST_HIGH (t) < 0)
    return -1;
  else
    return 1;
}

/* ""   Ȱ ڿ.  */
const char empty_string[] = "";

#define digit_string(d) (digit_vector + ((d) * 2))

/*   ϴ Character ڿ.
 *     ϱ    ϴ.  */
const char digit_vector[] = {
  '0', 0, '1', 0, '2', 0, '3', 0, '4', 0,
  '5', 0, '6', 0, '7', 0, '8', 0, '9', 0
};

/* ̰ LENGTH ̰ CONTENTS  ϴ ڿ constant  ҴϿ
   ȯմϴ.  LENGTH  -1 ̸ CONTENTS  null-terminated
   ڿ մϴ. ̴ strlen  Ͽ ˴ϴ. 
    ڿ constant   ҴǾٸ  纻   
   ȯ˴ϴ.  */

static const char *
etl_alloc_string (p, contents, length)
     apr_pool_t *p;
     const char *contents;
     int length;
{
  if (length == -1)
    length = strlen (contents);

  if (length == 0)
    return empty_string;
  if (length == 1 && ISDIGIT (contents[0]))
    return digit_string (contents[0] - '0');

  return apr_pstrndup (p, contents, length);
}

/*  LEN  ڿ STR     STRING_CST node 
   ȯѴ.
   ETREE_TYPE  ʱȭ ʴ´.  */

etree
etree_build_string (p, len, str)
     apr_pool_t *p;
     int len;
     const char *str;
{
  etree s = etree_make_node (p, ECONF_STRING_CST);

  ETREE_STRING_LENGTH (s) = len;
  ETREE_STRING_POINTER (s) = etl_alloc_string (p, str, len);

  return s;
}

/* NODE  ETREE_CHAIN  0 ̰ ̰ ο uid  ٴ 
   ϰ    ο node  ȯѴ.  */

etree
copy_node (p, node)
     apr_pool_t *p;
     etree node;
{
  etree t;
  size_t length;

  length = etree_size (node);
  t = (etree) apr_pcalloc (p, length);
  memcpy (t, node, length);

  ETREE_CHAIN (t) = 0;

  return t;
}

/*    C  type TYPE (ETREE_INTEGER_TYPE )     ִ
     ִٸ 0  ƴ  ȯѴ.  */

int
etree_int_fits_type_p (p, c, type)
     apr_pool_t *p;
     etree c, type;
{
  /*  type  bound   , 츮 츮 ڽ ˻  ִ.
      ƴ ,  type  subtype ̸ ش ׿  ˻縦 õѴ.
     ׷   precision   ˻縦 , force\_fit\_type  Ѵ.  */
  if (ETREE_TYPE_MAX_VALUE (type) != ETREE_NULL_TREE
      && ETREE_TYPE_MIN_VALUE (type) != ETREE_NULL_TREE
      && ETREE_CODE (ETREE_TYPE_MAX_VALUE (type)) == ETREE_INTEGER_CST
      && ETREE_CODE (ETREE_TYPE_MIN_VALUE (type)) == ETREE_INTEGER_CST)
    {
      if (ETREE_UNSIGNED (type))
        return (! ETREE_INT_CST_LT_UNSIGNED (ETREE_TYPE_MAX_VALUE (type), c)
                && ! ETREE_INT_CST_LT_UNSIGNED (c, ETREE_TYPE_MIN_VALUE (type))
                /*  int   unsigned type   ʴ´.  */
                && ! (ETREE_INT_CST_HIGH (c) < 0
                      && ! ETREE_UNSIGNED (ETREE_TYPE (c))));
      else
        return (! ETREE_INT_CST_LT (ETREE_TYPE_MAX_VALUE (type), c)
                && ! ETREE_INT_CST_LT (c, ETREE_TYPE_MIN_VALUE (type))
                /* Unsigned ints with top bit set never fit signed types.  */
                && ! (ETREE_INT_CST_HIGH (c) < 0
                      && ETREE_UNSIGNED (ETREE_TYPE (c))));
    }
  else if (ETREE_CODE (type) == ETREE_INTEGER_TYPE && ETREE_TYPE (type) != 0)
    return etree_int_fits_type_p (p, c, ETREE_TYPE (type));
  else
    {
      c = copy_node (p, c);
      ETREE_TYPE (c) = type;
      return !etree_force_fit_type (c, 0);
    }
}

void
build_common_etree_nodes (p)
     apr_pool_t *p;
{
  if (build_common_etree_nodes_var++)
    return;
  
  etree_integer_type_node = etree_make_signed_type (p, INT_TYPE_SIZE);
  etree_unsigned_type_node = etree_make_unsigned_type (p, INT_TYPE_SIZE);
  etree_long_integer_type_node = etree_make_signed_type (p, LONG_TYPE_SIZE);
  etree_long_unsigned_type_node = etree_make_unsigned_type (p, LONG_TYPE_SIZE);
  etree_long_long_integer_type_node = etree_make_signed_type (p, LONG_LONG_TYPE_SIZE);
  etree_long_long_unsigned_type_node = etree_make_unsigned_type (p, LONG_LONG_TYPE_SIZE);

  etree_float_type_node = etree_make_node (p, ETREE_REAL_TYPE);
  ETREE_TYPE_PRECISION (etree_float_type_node) = FLOAT_TYPE_SIZE;
  etree_layout_type (etree_float_type_node);

  etree_double_type_node = etree_make_node (p, ETREE_REAL_TYPE);
  ETREE_TYPE_PRECISION (etree_double_type_node) = DOUBLE_TYPE_SIZE;
  etree_layout_type (etree_double_type_node);

  etree_long_double_type_node = etree_make_node (p, ETREE_REAL_TYPE);
  ETREE_TYPE_PRECISION (etree_long_double_type_node) = LONG_DOUBLE_TYPE_SIZE;
  etree_layout_type (etree_long_double_type_node);
}
