/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of libetl.

libetl is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

libetl is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "config.h"
#include "apr-1/apr.h"
#include "apr-1/apr_portable.h"
#include "apr-1/apr_strings.h"
#include "apr-1/apr_thread_proc.h"
#include "apr-1/apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr-1/apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE
#include "apr-1/apr_poll.h"
#include "apr-1/apr_buckets.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>            /* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#define HTTP_NOT_CONNECTED -1
#include "system.h"
#include "etldefaults.h"
#include "obstack.h"
#include "hashtab.h"

/* TREE   header .  */
#include "config/i386.h"
#include "hwint.h"
#include "machmode.h"
#include "etree.h"
#include "fold-const.h"
#include "stor-layout.h"

#include "econnect.h"
#include "etl.h"
#include "obstack.h"
#include "errors.h"
#include "hashtab.h"
#include "econnect.h"
#include "gensupport.h"

static int neg_double (unsigned HOST_WIDE_INT, HOST_WIDE_INT, unsigned HOST_WIDE_INT *, HOST_WIDE_INT *);

/*  doubleword  doubleword  negate Ѵ.
   װ sign     operation  overflow ̸, 0  ƴ
    ȯѴ.
    argument  L1  H1   `HOST_WIDE_INT'  ־.
    *LV  *HV   `HOST_WIDE_INT'  ȴ.  */

static int
neg_double (l1, h1, lv, hv)
     unsigned HOST_WIDE_INT l1;
     HOST_WIDE_INT h1;
     unsigned HOST_WIDE_INT *lv;
     HOST_WIDE_INT *hv;
{
  if (l1 == 0)
    {
      *lv = 0;
      *hv = - h1;
      return (*hv & h1) < 0;
    }
  else
    {
      *lv = -l1;
      *hv = ~h1;
      return 0;
    }
}

/* Type   ʴ constant   bit  0 Ȥ 1  ν 
   constant T  type   T  ȿϵ .

    signed overflow  ߻ϸ 1  ׷  0  ȯѴ.  
   OVERFLOW  0  ƴϸ, signed overfow  T  ϴ  ̹ ߻
   ̴. ׷ װ ܼ ø.

   ( ) CHECK_FLOAT_VALUE  ȣν type  real constant T 
    ȿϵ .  */

int
etree_force_fit_type (t, overflow)
     etree t;
     int overflow;
{
  unsigned HOST_WIDE_INT low;
  HOST_WIDE_INT high;
  unsigned int prec;

  if (ETREE_CODE (t) == ETREE_REAL_CST)
    {
      return overflow;
    }
  else if (ETREE_CODE (t) != ETREE_INTEGER_CST)
    return overflow;

  low = ETREE_INT_CST_LOW (t);
  high = ETREE_INT_CST_HIGH (t);

  if (ETREE_CODE (t) == ETREE_INTEGER_CST)
    prec = HOST_BITS_PER_WIDE_INT;
  else
    prec = 32;

  /*  type  precision   bit   Ѵ.  */

  if (prec == 2 * HOST_BITS_PER_WIDE_INT)
    ;
  else if (prec > HOST_BITS_PER_WIDE_INT)
    ETREE_INT_CST_HIGH (t)
      &= ~((HOST_WIDE_INT) (-1) << (prec - HOST_BITS_PER_WIDE_INT));
  else
    {
      ETREE_INT_CST_HIGH (t) = 0;
      if (prec < HOST_BITS_PER_WIDE_INT)
	ETREE_INT_CST_LOW (t) &= ~((unsigned HOST_WIDE_INT) (-1) << prec);
    }

  /*   sign bit  Ǿ ִٸ, sign  ȮѴ.  */
  if (prec != 2 * HOST_BITS_PER_WIDE_INT
      && (prec > HOST_BITS_PER_WIDE_INT
	  ? 0 != (ETREE_INT_CST_HIGH (t)
		  & ((HOST_WIDE_INT) 1
		     << (prec - HOST_BITS_PER_WIDE_INT - 1)))
	  : 0 != (ETREE_INT_CST_LOW (t)
		  & ((unsigned HOST_WIDE_INT) 1 << (prec - 1)))))
    {
      /* Value is negative:
	 set to 1 all the bits that are outside this type's precision.  */
      if (prec > HOST_BITS_PER_WIDE_INT)
	ETREE_INT_CST_HIGH (t)
	  |= ((HOST_WIDE_INT) (-1) << (prec - HOST_BITS_PER_WIDE_INT));
      else
	{
	  ETREE_INT_CST_HIGH (t) = -1;
	  if (prec < HOST_BITS_PER_WIDE_INT)
	    ETREE_INT_CST_LOW (t) |= ((unsigned HOST_WIDE_INT) (-1) << prec);
	}
    }

  /*  signed overflow  ߻Ͽٸ 0  ƴ  ȯѴ.  */
  return
    ((overflow | (low ^ ETREE_INT_CST_LOW (t)) | (high ^ ETREE_INT_CST_HIGH (t)))
     != 0);
}

etree
etree_fold (p, expr)
     apr_pool_t *p;
     etree expr;
{
  etree t = expr;
  etree arg0 = ETREE_NULL_TREE, arg1 = ETREE_NULL_TREE;
  enum etree_code code = ETREE_CODE (t);
  int kind = ETREE_CODE_CLASS (code);
  /* WINS   operand   switch   
     0  ƴ   ̴.  */
  int wins = 1;

  if (ETREE_IS_EXPR_CODE_CLASS (kind) || kind == 'r')
    {
      int len = first_rtl_op (code);
      int i;
      for (i = 0; i < len; i++)
	{
	  etree op = ETREE_OPERAND (t, i);
	  etree subop;

	  if (op == 0)
	    continue;		/* Valid for CALL_EXPR, at least.  */

	  subop = op;

	  if (ETREE_CODE (subop) != ETREE_INTEGER_CST
	      && ETREE_CODE (subop) != ETREE_REAL_CST)
	    /* Note that ETREE_CONSTANT isn't enough:
	       static var addresses are constant but we can't
	       do arithmetic on them.  */
	    wins = 0;

	  if (i == 0)
	    arg0 = op;
	  else if (i == 1)
	    arg1 = op;
	}
    }

  switch (code)
    {
    case ETREE_INTEGER_CST:
      return t;
    case ETREE_NEGATE_EXPR:
      if (wins)
	{
	  if (ETREE_CODE (arg0) == ETREE_INTEGER_CST)
	    {
	      unsigned HOST_WIDE_INT low;
	      HOST_WIDE_INT high;
	      int overflow = neg_double (ETREE_INT_CST_LOW (arg0),
					 ETREE_INT_CST_HIGH (arg0),
					 &low, &high);
	      t = etree_build_int_2 (p, low, high);
	      ETREE_OVERFLOW (t)	= (ETREE_OVERFLOW (arg0) | etree_force_fit_type (t, overflow));
	      ETREE_CONSTANT_OVERFLOW (t) = ETREE_OVERFLOW (t) | ETREE_CONSTANT_OVERFLOW (arg0);
	    }
	}
      else if (ETREE_CODE (arg0) == ETREE_NEGATE_EXPR)
	return ETREE_OPERAND (arg0, 0);

      return t;
    default:
      etl_error ("fold: óؾ ϴ etree node.\n");
      debug_etree (t);
      return t;
    }
}
