/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of libetl.

libetl is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

libetl is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "config.h"
#include "apr-1/apr.h"
#include "apr-1/apr_portable.h"
#include "apr-1/apr_strings.h"
#include "apr-1/apr_thread_proc.h"
#include "apr-1/apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr-1/apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE
#include "apr-1/apr_poll.h"
#include "apr-1/apr_buckets.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>            /* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#define HTTP_NOT_CONNECTED -1
#include "system.h"
#include "etldefaults.h"
#include "obstack.h"
#include "hashtab.h"

/* TREE   header .  */
#include "config/i386.h"
#include "hwint.h"
#include "machmode.h"
#include "etree.h"
#include "fold-const.h"
#include "stor-layout.h"

#include "econnect.h"
#include "etl.h"
#include "obstack.h"
#include "errors.h"
#include "hashtab.h"
#include "econnect.h"
#include "gensupport.h"

/**
 *    machine mode  Ǿ  machine mode  ̸
 *       մϴ.  ̸ "mode" ڸ  ʽϴ.
 *       */

#define DEF_MACHMODE(SYM, NAME, CLASS, BITSIZE, SIZE, UNIT, WIDER, INNER)  NAME,

const char * const etree_mode_name[ETREE_NUM_MACHINE_MODES] = {
#include "machmode.def"
};

#undef DEF_MACHMODE

/** machine mode  Ǿ  GET_MODE_CLASS   class
 *    mode  մϴ.  */

#define DEF_MACHMODE(SYM, NAME, CLASS, BITSIZE, SIZE, UNIT, WIDER, INNER)  CLASS,

const enum etree_mode_class etree_mode_class[ETREE_NUM_MACHINE_MODES] = {
#include "machmode.def"
};

#undef DEF_MACHMODE

/** machine mode  Ǿ  mode  ̸ Ʈ ũ
 *    մϴ. GET_MODE_BITSIZE  ̰ մϴ.  */

#define DEF_MACHMODE(SYM, NAME, CLASS, BITSIZE, SIZE, UNIT, WIDER, INNER)  BITSIZE,

const unsigned short etree_mode_bitsize[ETREE_NUM_MACHINE_MODES] = {
#include "machmode.def"
};

#undef DEF_MACHMODE

/** machine mode  Ǿ  mode  ̸ Ʈ ũ
 *    մϴ. GET_MODE_SIZE  ̰ մϴ.  */

#define DEF_MACHMODE(SYM, NAME, CLASS, BITSIZE, SIZE, UNIT, WIDER, INNER)  SIZE,

const unsigned char etree_mode_size[ETREE_NUM_MACHINE_MODES] = {
#include "machmode.def"
};

#undef DEF_MACHMODE

/** machine mode  Ǿ  mode  subunit  ̸
 *    մϴ. GET_MODE_UNIT_SIZE  ̰ մϴ.  */

#define DEF_MACHMODE(SYM, NAME, CLASS, BITSIZE, SIZE, UNIT, WIDER, INNER)  UNIT,

const unsigned char etree_mode_unit_size[ETREE_NUM_MACHINE_MODES] = {
#include "machmode.def"         /* machine mode   ȭǾ  */
};

#undef DEF_MACHMODE
/** Machine mode  Ǿ      mode 
 *    ݴϴ.  (QI -> HI -> SI -> DI, Ÿ )  Widening multiply
 *       instruction  ̰ մϴ.  */

#define DEF_MACHMODE(SYM, NAME, CLASS, BITSIZE, SIZE, UNIT, WIDER, INNER)  \
  (unsigned char) WIDER,

const unsigned char etree_mode_wider_mode[ETREE_NUM_MACHINE_MODES] = {
#include "machmode.def"         /* machine mode   ȭǾ  */
};

#undef DEF_MACHMODE

#define DEF_MACHMODE(SYM, NAME, CLASS, BITSIZE, SIZE, UNIT, WIDER, INNER)  \
  ((BITSIZE) >= HOST_BITS_PER_WIDE_INT) ? ~(unsigned HOST_WIDE_INT) 0 : ((unsigned HOST_WIDE_INT) 1 << (BITSIZE)) - 1,

/** Machine mode  ĵǾ, mode  significant bit  mask 
 *    Ѵ.
 *
 *       Indexed by machine mode, gives mask of significant bits in mode.  */

const unsigned HOST_WIDE_INT etree_mode_mask_array[ETREE_NUM_MACHINE_MODES] = {
#include "machmode.def"
};

#undef DEF_MACHMODE

#define DEF_MACHMODE(SYM, NAME, CLASS, BITSIZE, SIZE, UNIT, WIDER, INNER) INNER,

/** Machine mode  Ǿ  Vector type  inner element
 *     mode  մϴ.  */

const enum etree_machine_mode etree_inner_mode_array[ETREE_NUM_MACHINE_MODES] = {
#include "machmode.def"
};
/**
 * Mode class  зǾ   class     mode 
 * ˷ݴϴ.  Q mode  ׻ ʺ 1 (Ҽ (complex)  ؼ
 * 2) ̱  ٸ mode   ﷡ ׷  ϴ.
 *
 * 츮 MODE_INT  BImode ſ QImode  ϴ  ˰
 * ʽÿ.  ̷ ϴ  middle end  츮  
 * ʴ ü ׿  ͵鿡 bitfields   õ
 *  ֱ Դϴ.   target md file  BImode widget 
 * ؾ մϴ.
 *                         */
const enum etree_machine_mode etree_class_narrowest_mode[(int) ETREE_MAX_MODE_CLASS] = {
    /* MODE_RANDOM */           ETREE_VOIDmode,
    /* MODE_INT */              ETREE_QImode,
    /* MODE_FLOAT */            ETREE_QFmode,
    /* MODE_PARTIAL_INT */      ETREE_PQImode,
    /* MODE_CC */               ETREE_CCmode,
    /* MODE_COMPLEX_INT */      ETREE_CQImode,
    /* MODE_COMPLEX_FLOAT */    ETREE_QCmode,
    /* MODE_VECTOR_INT */       ETREE_V2QImode,
    /* MODE_VECTOR_FLOAT */     ETREE_V2SFmode
};
