/*  -
   Copyright (C) 2005 Weongyo Jeong (weongyo@gmail.com)

This file is part of libetl.

libetl is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

libetl is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "config.h"
#include "apr-1/apr.h"
#include "apr-1/apr_portable.h"
#include "apr-1/apr_strings.h"
#include "apr-1/apr_thread_proc.h"
#include "apr-1/apr_signal.h"
#define APR_WANT_STDIO
#define APR_WANT_STRFUNC
#include "apr-1/apr_want.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if APR_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define CORE_PRIVATE
#include "apr-1/apr_poll.h"
#include "apr-1/apr_buckets.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>            /* for IRIX, FD_SET calls bzero () */
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_PROCESSOR_H
#include <sys/processor.h> /* for bindprocessor () */
#endif

#include <signal.h>
#include <sys/times.h>

#define HTTP_NOT_CONNECTED -1
#include "system.h"
#include "etldefaults.h"
#include "obstack.h"
#include "hashtab.h"

/* TREE   header .  */
#include "config/i386.h"
#include "hwint.h"
#include "machmode.h"
#include "etree.h"
#include "fold-const.h"
#include "stor-layout.h"

#include "econnect.h"
#include "etl.h"
#include "obstack.h"
#include "errors.h"
#include "hashtab.h"
#include "econnect.h"
#include "gensupport.h"

#ifndef MAX_FIXED_MODE_SIZE
#define MAX_FIXED_MODE_SIZE ETREE_GET_MODE_BITSIZE (ETREE_DImode)
#endif

void fixup_signed_type (apr_pool_t *, etree);
void fixup_unsigned_type (apr_pool_t *, etree);

/* bit ũ precision  Ͽ TYPE    ϰ,
   ׷  װ ̾ƿѴ.  Tree code  INTEGER_TYPE  ƴϱ
    make_signed_type      ȴ.
    , Pascal  -fsigned-char option  ־ 
   ȴ.  */

void
fixup_signed_type (p, type)
     apr_pool_t *p;
     etree type;
{
  int precision = ETREE_TYPE_PRECISION (type);

  /* i386 vector extension   ģ鿡  ǰ 
     type  ʿ Ҷ 2 * HOST_BITS_PER_WIDE_INT 
     ū   ǥ  .  */
  if (precision > HOST_BITS_PER_WIDE_INT * 2)
    precision = HOST_BITS_PER_WIDE_INT * 2;

  ETREE_TYPE_MIN_VALUE (type)
    = etree_build_int_2 (p, (precision - HOST_BITS_PER_WIDE_INT > 0
		       ? 0 : (HOST_WIDE_INT) (-1) << (precision - 1)),
		   (((HOST_WIDE_INT) (-1)
		     << (precision - HOST_BITS_PER_WIDE_INT - 1 > 0
			 ? precision - HOST_BITS_PER_WIDE_INT - 1
			 : 0))));
  ETREE_TYPE_MAX_VALUE (type)
    = etree_build_int_2 (p, (precision - HOST_BITS_PER_WIDE_INT > 0
		       ? -1 : ((HOST_WIDE_INT) 1 << (precision - 1)) - 1),
		   (precision - HOST_BITS_PER_WIDE_INT - 1 > 0
		    ? (((HOST_WIDE_INT) 1
			<< (precision - HOST_BITS_PER_WIDE_INT - 1))) - 1
		    : 0));
  
  ETREE_TYPE (ETREE_TYPE_MIN_VALUE (type)) = type;
  ETREE_TYPE (ETREE_TYPE_MAX_VALUE (type)) = type;

  /* type  ̾ƿ Ѵ: ̰ alignment, size, Ÿ Ѵ.  */
  etree_layout_type (type);
}

/* bit ũ precision  Ͽ TYPE    ϰ,
   ׷  װ ̾ƿѴ.  ̰ `make_unsigned_type' 
    enumeral type   ȴ.  */

void
fixup_unsigned_type (p, type)
     apr_pool_t *p;
     etree type;
{
  int precision = ETREE_TYPE_PRECISION (type);

  /* i386 vector extension   ģ鿡  ǰ 
     type  ʿ Ҷ 2 * HOST_BITS_PER_WIDE_INT 
     ū   ǥ  .  */
  if (precision > HOST_BITS_PER_WIDE_INT * 2)
    precision = HOST_BITS_PER_WIDE_INT * 2;

  ETREE_TYPE_MIN_VALUE (type) = etree_build_int_2 (p, 0, 0);
  ETREE_TYPE_MAX_VALUE (type)
    = etree_build_int_2 (p, precision - HOST_BITS_PER_WIDE_INT >= 0
		   ? -1 : ((HOST_WIDE_INT) 1 << precision) - 1,
		   precision - HOST_BITS_PER_WIDE_INT > 0
		   ? ((unsigned HOST_WIDE_INT) ~0
		      >> (HOST_BITS_PER_WIDE_INT
			  - (precision - HOST_BITS_PER_WIDE_INT)))
		   : 0);
  ETREE_TYPE (ETREE_TYPE_MIN_VALUE (type)) = type;
  ETREE_TYPE (ETREE_TYPE_MAX_VALUE (type)) = type;

  /* type  ̾ƿ Ѵ: ̰ alignment, size, Ÿ Ѵ.  */
  etree_layout_type (type);
}

/* PRECISION bit  signed   type  Ͽ ȯѴ.  */

etree
etree_make_signed_type (p, precision)
     apr_pool_t *p;
     int precision;
{
  etree type = etree_make_node (p, ETREE_INTEGER_TYPE);

  ETREE_TYPE_PRECISION (type) = precision;

  fixup_signed_type (p, type);

  return type;
}

/* PRECISION bit  unsigned   type  Ͽ ȯѴ.  */

etree
etree_make_unsigned_type (p, precision)
     apr_pool_t *p;
     int precision;
{
  etree type = etree_make_node (p, ETREE_INTEGER_TYPE);

  ETREE_TYPE_PRECISION (type) = precision;

  fixup_unsigned_type (p, type);

  return type;
}

/* SIZE bit  nonscalar   machine mode  ȯ.
    mode  ݵ class CLASS   ̾ ϸ Ȯ 
   bit   Ǿ ־ Ѵ.
    LIMIT  0  ƴ϶ MAX_FIXED_MODE_SIZE   mode 
     ̴.  */

enum etree_machine_mode
mode_for_size (size, class, limit)
     unsigned int size;
     enum etree_mode_class class;
     int limit;
{
  enum etree_machine_mode mode;

  if (limit && size > MAX_FIXED_MODE_SIZE)
    return ETREE_BLKmode;

  /*  class   size  ó mode  ´.  */
  for (mode = ETREE_GET_CLASS_NARROWEST_MODE (class); mode != ETREE_VOIDmode;
       mode = ETREE_GET_MODE_WIDER_MODE (mode))
    if (ETREE_GET_MODE_BITSIZE (mode) == size)
      return mode;

  return ETREE_BLKmode;
}

/* ,  BLKmode  ȯ ʴ´;  û bit  
     ִ   mode  ȯѴ.  */

enum etree_machine_mode
smallest_mode_for_size (size, class)
     unsigned int size;
     enum etree_mode_class class;
{
  enum etree_machine_mode mode;

  /* Ư з  size  ´ ó mode   .  */
  for (mode = ETREE_GET_CLASS_NARROWEST_MODE (class); mode != ETREE_VOIDmode;
       mode = ETREE_GET_MODE_WIDER_MODE (mode))
    if (ETREE_GET_MODE_BITSIZE (mode) >= size)
      return mode;

  abort ();
}

/* TYPE   mode  size, alignment  Ѵ.
   迭 type  ؼ element   Ѵ.
   TYPE   Ȥ ӽ type     ϴµ,
   ̷ dbxout  ̰ ã ̴.

   type  ETREE_TYPE_SIZE  ̹ type  ̾ƿ Ǿ   0 
   ƴ ̴. ׷  layout_type  type   ƹ͵ 
   ʴ´.

    type    ʴٸ, װ ETREE_TYPE_SIZE   0
    Ǿ ִ.  */

void
etree_layout_type (type)
     etree type;
{
  if (type == 0)
    abort ();

  /*  type   ̹ ̾ƿǾٸ ƹ͵  ʴ´.  */
  if (ETREE_TYPE_SIZE (type))
    return;

  switch (ETREE_CODE (type))
    {
    case ETREE_INTEGER_TYPE:
      if (ETREE_CODE (ETREE_TYPE_MIN_VALUE (type)) == ETREE_INTEGER_CST
	  && etree_int_cst_sgn (ETREE_TYPE_MIN_VALUE (type)) >= 0)
	ETREE_UNSIGNED (type) = 1;

      ETREE_TYPE_MODE (type) = smallest_mode_for_size (ETREE_TYPE_PRECISION (type),
						 ETREE_MODE_INT);
      break;
    case ETREE_REAL_TYPE:
      ETREE_TYPE_MODE (type) = mode_for_size (ETREE_TYPE_PRECISION (type), ETREE_MODE_FLOAT, 0);
      break;

    default:
      etl_warning ("layout_type: óؾ  type [%s]\n", etree_code_name[ETREE_CODE (type)]);
      abort ();
    }
}
